@@ -46,6 +46,16 @@ F2x_degree(GEN x)
   return F2x_degree_lg(x, lg(x));
 }
 
+GEN
+monomial_F2x(long d, long vs)
+{
+  long l=nbits2lg(d+1);
+  GEN z = zero_zv(l-1);
+  z[1] = vs;
+  F2x_set(z,d);
+  return z;
+}
+
 GEN
 F2x_to_ZX(GEN x)
 {
@@ -59,15 +69,6 @@ F2x_to_ZX(GEN x)
   return z;
 }
 
-GEN
-F2xC_to_ZXC(GEN v)
-{
-  long j, N = lg(v);
-  GEN y = cgetg(N, t_COL);
-  for (j=1; j<N; j++) gel(y,j) = F2x_to_ZX(gel(v,j));
-  return y;
-}
-
 GEN
 F2x_to_Flx(GEN x)
 {
@@ -176,7 +177,8 @@ Rg_to_F2xq(GEN x, GEN T)
       if (is_const_t(ta)) return Rg_to_F2(a)? pol1_F2x(v): pol0_F2x(v);
       b = RgX_to_F2x(b); if (b[1] != v) break;
       a = RgX_to_F2x(a); if (F2x_equal(b,T)) return a;
-      return F2x_rem(a, T);
+      if (lgpol(F2x_rem(b,T))==0) return F2x_rem(a, T);
+      break;
     case t_POL:
       x = RgX_to_F2x(x);
       if (x[1] != v) break;
@@ -190,6 +192,28 @@ Rg_to_F2xq(GEN x, GEN T)
   return NULL; /* not reached */
 }
 
+ulong
+F2x_eval(GEN P, ulong x)
+{
+  if (odd(x))
+  {
+    long i, lP = lg(P);
+    ulong c = 0;
+    for (i=2; i<lP; i++)
+      c ^= P[i];
+#ifdef LONG_IS_64BIT
+    c ^= c >> 32;
+#endif
+    c ^= c >> 16;
+    c ^= c >>  8;
+    c ^= c >>  4;
+    c ^= c >>  2;
+    c ^= c >>  1;
+    return c & 1;
+  }
+  else return F2x_coeff(P,0);
+}
+
 GEN
 F2x_add(GEN x, GEN y)
 {
@@ -212,7 +236,15 @@ F2x_addspec(GEN x, GEN y, long lx, long ly)
 
   if (ly>lx) swapspec(x,y, lx,ly);
   lz = lx+2; z = cgetg(lz, t_VECSMALL) + 2;
-  for (i=0; i<ly; i++) z[i] = x[i]^y[i];
+  for (i=0; i<ly-3; i+=4)
+  {
+    z[i] = x[i]^y[i];
+    z[i+1] = x[i+1]^y[i+1];
+    z[i+2] = x[i+2]^y[i+2];
+    z[i+3] = x[i+3]^y[i+3];
+  }
+  for (; i<ly; i++)
+    z[i] = x[i]^y[i];
   for (   ; i<lx; i++) z[i] = x[i];
   z -= 2; return F2x_renormalize(z, lz);
 }
@@ -234,6 +266,106 @@ F2x_1_add(GEN y)
   return z;
 }
 
+INLINE void
+F2x_addshiftipspec(GEN x, GEN y, long ny, ulong db)
+{
+  long i;
+  if (db)
+  {
+    ulong dc=BITS_IN_LONG-db;
+    ulong r=0, yi;
+    for(i=0; i<ny-3; i+=4)
+    {
+      yi = uel(y,i);   x[i]   ^= (yi<<db)|r; r = yi>>dc;
+      yi = uel(y,i+1); x[i+1] ^= (yi<<db)|r; r = yi>>dc;
+      yi = uel(y,i+2); x[i+2] ^= (yi<<db)|r; r = yi>>dc;
+      yi = uel(y,i+3); x[i+3] ^= (yi<<db)|r; r = yi>>dc;
+    }
+    for(  ; i<ny; i++)
+    {
+      ulong yi = uel(y,i); x[i] ^= (yi<<db)|r; r = yi>>dc;
+    }
+    if (r) x[i] ^= r;
+  }
+  else
+  {
+    for(i=0; i<ny-3; i+=4)
+    {
+      x[i]   ^= y[i];
+      x[i+1] ^= y[i+1];
+      x[i+2] ^= y[i+2];
+      x[i+3] ^= y[i+3];
+    }
+    for(   ; i<ny; i++)
+      x[i] ^= y[i];
+  }
+}
+
+INLINE void
+F2x_addshiftip(GEN x, GEN y, ulong d)
+{
+  ulong db, dl=dvmduBIL(d, &db);
+  F2x_addshiftipspec(x+2+dl, y+2, lgpol(y), db);
+}
+
+static GEN
+F2x_mul1(ulong x, ulong y)
+{
+  ulong x1=(x&HIGHMASK)>>BITS_IN_HALFULONG;
+  ulong x2=x&LOWMASK;
+  ulong y1=(y&HIGHMASK)>>BITS_IN_HALFULONG;
+  ulong y2=y&LOWMASK;
+  ulong r1,r2,rr;
+  GEN z;
+  ulong i;
+  rr=r1=r2=0UL;
+  if (x2)
+    for(i=0;i<BITS_IN_HALFULONG;i++)
+      if (x2&(1UL<<i))
+      {
+        r2^=y2<<i;
+        rr^=y1<<i;
+      }
+  if (x1)
+    for(i=0;i<BITS_IN_HALFULONG;i++)
+      if (x1&(1UL<<i))
+      {
+        r1^=y1<<i;
+        rr^=y2<<i;
+      }
+  r2^=(rr&LOWMASK)<<BITS_IN_HALFULONG;
+  r1^=(rr&HIGHMASK)>>BITS_IN_HALFULONG;
+  z=cgetg((r1?4:3),t_VECSMALL);
+  z[2]=r2;
+  if (r1) z[3]=r1;
+  return z;
+}
+
+static GEN
+F2x_mulspec_basecase(GEN x, GEN y, long nx, long ny)
+{
+  long l, i, j;
+  GEN z;
+  l = nx + ny;
+  z = zero_Flv(l+1);
+  for(i=0; i < ny-1; i++)
+  {
+    GEN zi = z+2+i;
+    ulong yi = uel(y,i);
+    if (yi)
+      for(j=0; j < BITS_IN_LONG; j++)
+        if (yi&(1UL<<j)) F2x_addshiftipspec(zi,x,nx,j);
+  }
+  {
+    GEN zi = z+2+i;
+    ulong yi = uel(y,i);
+    long c = BITS_IN_LONG-bfffo(yi);
+    for(j=0; j < c; j++)
+      if (yi&(1UL<<j)) F2x_addshiftipspec(zi,x,nx,j);
+  }
+  return F2x_renormalize(z, l+2);
+}
+
 static GEN
 F2x_addshift(GEN x, GEN y, long d)
 {
@@ -279,39 +411,6 @@ F2x_shiftip(pari_sp av, GEN x, long v)
   avma = (pari_sp)y; return y;
 }
 
-static GEN
-F2x_mul1(ulong x, ulong y)
-{
-  ulong x1=(x&HIGHMASK)>>BITS_IN_HALFULONG;
-  ulong x2=x&LOWMASK;
-  ulong y1=(y&HIGHMASK)>>BITS_IN_HALFULONG;
-  ulong y2=y&LOWMASK;
-  ulong r1,r2,rr;
-  GEN z;
-  ulong i;
-  rr=r1=r2=0UL;
-  if (x2)
-    for(i=0;i<BITS_IN_HALFULONG;i++)
-      if (x2&(1UL<<i))
-      {
-        r2^=y2<<i;
-        rr^=y1<<i;
-      }
-  if (x1)
-    for(i=0;i<BITS_IN_HALFULONG;i++)
-      if (x1&(1UL<<i))
-      {
-        r1^=y1<<i;
-        rr^=y2<<i;
-      }
-  r2^=(rr&LOWMASK)<<BITS_IN_HALFULONG;
-  r1^=(rr&HIGHMASK)>>BITS_IN_HALFULONG;
-  z=cgetg((r1?4:3),t_VECSMALL);
-  z[2]=r2;
-  if (r1) z[3]=r1;
-  return z;
-}
-
 /* fast product (Karatsuba) of polynomials a,b. These are not real GENs, a+2,
  * b+2 were sent instead. na, nb = number of terms of a, b.
  * Only c, c0, c1, c2 are genuine GEN.
@@ -329,7 +428,10 @@ F2x_mulspec(GEN a, GEN b, long na, long nb)
   if (!nb) return pol0_F2x(0);
 
   av = avma;
-  if (na <=1) return F2x_shiftip(av,F2x_mul1(*a,*b),v);
+  if (na == 1)
+    return F2x_shiftip(av, F2x_mul1(*a,*b), v);
+  if (na < F2x_MUL_KARATSUBA_LIMIT)
+    return F2x_shiftip(av, F2x_mulspec_basecase(a, b, na, nb), v);
   i=(na>>1); n0=na-i; na=i;
   a0=a+n0; n0a=n0;
   while (n0a && !a[n0a-1]) n0a--;
@@ -445,27 +547,6 @@ F2x_sqrt(GEN x)
   return F2x_renormalize(z, lz);
 }
 
-INLINE void
-F2x_addshiftip(GEN x, GEN y, ulong d)
-{
-  ulong db, dl=dvmduBIL(d, &db);
-  long i, ly = lg(y);
-  if (db)
-  {
-    ulong dc=BITS_IN_LONG-db;
-    ulong r=0;
-    for(i=2; i<ly; i++)
-    {
-      x[i+dl] ^= (((ulong)y[i])<<db)|r;
-      r = ((ulong)y[i])>>dc;
-    }
-    if (r) x[i+dl] ^= r;
-  }
-  else
-    for(i=2; i<ly; i++)
-      x[i+dl] ^= y[i];
-}
-
 static GEN
 F2x_shiftneg(GEN y, ulong d)
 {
@@ -527,7 +608,7 @@ F2x_rem(GEN x, GEN y)
   long lx=lg(x);
   dy = F2x_degree(y); if (!dy) return pol0_F2x(x[1]);
   dx = F2x_degree_lg(x,lx);
-  x  = vecsmall_copy(x);
+  x  = F2x_copy(x);
   while (dx>=dy)
   {
     F2x_addshiftip(x,y,dx-dy);
@@ -548,7 +629,7 @@ F2x_divrem(GEN x, GEN y, GEN *pr)
   if (pr == ONLY_REM) return F2x_rem(x, y);
   if (!dy)
   {
-    z = vecsmall_copy(x);
+    z = F2x_copy(x);
     if (pr && pr != ONLY_DIVIDES) *pr = pol0_F2x(vs);
     return z;
   }
@@ -556,13 +637,13 @@ F2x_divrem(GEN x, GEN y, GEN *pr)
   dz = dx-dy;
   if (dz < 0)
   {
-    if (pr == ONLY_DIVIDES) return dx < 0? vecsmall_copy(x): NULL;
+    if (pr == ONLY_DIVIDES) return dx < 0? F2x_copy(x): NULL;
     z = pol0_F2x(vs);
-    if (pr) *pr = vecsmall_copy(x);
+    if (pr) *pr = F2x_copy(x);
     return z;
   }
   z = zero_zv(lg(x)-lg(y)+2); z[1] = vs;
-  x = vecsmall_copy(x);
+  x = F2x_copy(x);
   while (dx>=dy)
   {
     F2x_set(z,dx-dy);
@@ -585,7 +666,7 @@ F2x_valrem(GEN x, GEN *Z)
 {
   long v, v2, i, l=lg(x);
   GEN y;
-  if (l==2) { *Z = leafcopy(x); return LONG_MAX; }
+  if (l==2) { *Z = F2x_copy(x); return LONG_MAX; }
   for (i=2; i<l && x[i]==0; i++) /*empty*/;
   v = i-2;
   v2 = (i < l)? vals(x[i]): 0;
@@ -616,8 +697,8 @@ F2x_deflate(GEN x, long d)
 {
   GEN y;
   long i,id, dy, dx = F2x_degree(x);
-  if (d <= 1) return Flx_copy(x);
-  if (dx < 0) return leafcopy(x);
+  if (d <= 1) return F2x_copy(x);
+  if (dx < 0) return F2x_copy(x);
   dy = dx/d; /* dy+1 coefficients + 1 extra word for variable */
   y = zero_zv(nbits2lg(dy+1)-1); y[1] = x[1];
   for (i=id=0; i<=dy; i++,id+=d)
@@ -632,7 +713,7 @@ F2x_even_odd(GEN p, GEN *pe, GEN *po)
   long n = F2x_degree(p), n0, n1, i;
   GEN p0, p1;
 
-  if (n <= 0) { *pe = leafcopy(p); *po = pol0_F2x(p[1]); return; }
+  if (n <= 0) { *pe = F2x_copy(p); *po = pol0_F2x(p[1]); return; }
 
   n0 = (n>>1)+1; n1 = n+1 - n0; /* n1 <= n0 <= n1+1 */
   p0 = zero_zv(nbits2lg(n0+1)-1); p0[1] = p[1];
@@ -660,26 +741,26 @@ F2x_deriv(GEN z)
 GEN
 F2x_gcd(GEN a, GEN b)
 {
-  pari_sp av = avma, lim = stack_lim(av,2);
+  pari_sp av = avma;
   if (lg(b) > lg(a)) swap(a, b);
   while (lgpol(b))
   {
     GEN c = F2x_rem(a,b);
     a = b; b = c;
-    if (low_stack(lim,stack_lim(av,2)))
+    if (gc_needed(av,2))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"F2x_gcd (d = %ld)",F2x_degree(c));
       gerepileall(av,2, &a,&b);
     }
   }
-  if (low_stack(lim,stack_lim(av,2))) a = gerepileuptoleaf(av, a);
+  if (gc_needed(av,2)) a = gerepileuptoleaf(av, a);
   return a;
 }
 
 GEN
 F2x_extgcd(GEN a, GEN b, GEN *ptu, GEN *ptv)
 {
-  pari_sp av=avma, lim = stack_lim(av,2);
+  pari_sp av=avma;
   GEN u,v,d,d1,v1;
   long vx = a[1];
   d = a; d1 = b;
@@ -690,7 +771,7 @@ F2x_extgcd(GEN a, GEN b, GEN *ptu, GEN *ptv)
     v = F2x_add(v,F2x_mul(q,v1));
     u=v; v=v1; v1=u;
     u=r; d=d1; d1=u;
-    if (low_stack(lim,stack_lim(av,2)))
+    if (gc_needed(av,2))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"F2x_extgcd (d = %ld)",F2x_degree(d));
       gerepileall(av,5, &d,&d1,&u,&v,&v1);
@@ -698,14 +779,14 @@ F2x_extgcd(GEN a, GEN b, GEN *ptu, GEN *ptv)
   }
   if (ptu) *ptu = F2x_div(F2x_add(d, F2x_mul(b,v)), a);
   *ptv = v;
-  if (low_stack(lim,stack_lim(av,2))) gerepileall(av,ptu?3:2,&d,ptv,ptu);
+  if (gc_needed(av,2)) gerepileall(av,ptu?3:2,&d,ptv,ptu);
   return d;
 }
 
 static GEN
 F2x_halfgcd_i(GEN a, GEN b)
 {
-  pari_sp av=avma, lim = stack_lim(av,2);
+  pari_sp av=avma;
   GEN u,u1,v,v1;
   long vx = a[1];
   long n = (F2x_degree(a)+1)>>1;
@@ -717,7 +798,7 @@ F2x_halfgcd_i(GEN a, GEN b)
     a = b; b = r; swap(u,u1); swap(v,v1);
     u1 = F2x_add(u1, F2x_mul(u, q));
     v1 = F2x_add(v1, F2x_mul(v, q));
-    if (low_stack(lim,stack_lim(av,2)))
+    if (gc_needed(av,2))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"F2x_halfgcd (d = %ld)",F2x_degree(b));
       gerepileall(av,6, &a,&b,&u1,&v1,&u,&v);
@@ -767,7 +848,7 @@ F2xq_inv(GEN x,GEN T)
 {
   pari_sp av=avma;
   GEN U = F2xq_invsafe(x, T);
-  if (!U) pari_err_INV("F2xq_inv",x);
+  if (!U) pari_err_INV("F2xq_inv", F2x_to_ZX(x));
   return gerepileuptoleaf(av, U);
 }
 
@@ -820,7 +901,7 @@ F2xq_pow(GEN x, GEN n, GEN pol)
 
   if (!signe(n)) return pol1_F2x(x[1]);
   if (is_pm1(n)) /* +/- 1 */
-    return (signe(n) < 0)? F2xq_inv(x,pol): vecsmall_copy(x);
+    return (signe(n) < 0)? F2xq_inv(x,pol): F2x_copy(x);
 
   if (signe(n) < 0) x = F2xq_inv(x,pol);
   y = gen_pow(x, n, (void*)pol, &_F2xq_sqr, &_F2xq_mul);
@@ -835,7 +916,7 @@ F2xq_powu(GEN x, ulong n, GEN pol)
   switch(n)
   {
     case 0: return pol1_F2x(x[1]);
-    case 1: return vecsmall_copy(x);
+    case 1: return F2x_copy(x);
     case 2: return F2xq_sqr(x,pol);
   }
   y = gen_powu(x, n, (void*)pol, &_F2xq_sqr, &_F2xq_mul);
@@ -846,7 +927,7 @@ F2xq_powu(GEN x, ulong n, GEN pol)
 GEN
 F2xq_powers(GEN x, long l, GEN T)
 {
-  int use_sqr = (F2x_degree(x)<<1) >= F2x_degree(T);
+  int use_sqr = 2*F2x_degree(x) >= F2x_degree(T);
   return gen_powers(x, l, use_sqr, (void*)T, &_F2xq_sqr, &_F2xq_mul, &_F2xq_one);
 }
 
@@ -856,7 +937,21 @@ F2xq_matrix_pow(GEN y, long n, long m, GEN P)
   return F2xV_to_F2m(F2xq_powers(y,m-1,P),n);
 }
 
-static struct bb_algebra F2xq_algebra = { _F2xq_red,_F2xq_add,_F2xq_mul,_F2xq_sqr,_F2xq_one,_F2xq_zero};
+GEN
+F2x_Frobenius(GEN T)
+{
+  return F2xq_sqr(polx_F2x(T[1]), T);
+}
+
+GEN
+F2x_matFrobenius(GEN T)
+{
+  long n = F2x_degree(T);
+  return F2xq_matrix_pow(F2x_Frobenius(T), n, n, T);
+}
+
+static struct bb_algebra F2xq_algebra = { _F2xq_red, _F2xq_add, _F2xq_add,
+              _F2xq_mul, _F2xq_sqr, _F2xq_one, _F2xq_zero};
 
 GEN
 F2x_F2xqV_eval(GEN Q, GEN x, GEN T)
@@ -869,7 +964,7 @@ GEN
 F2x_F2xq_eval(GEN Q, GEN x, GEN T)
 {
   long d = F2x_degree(Q);
-  int use_sqr = (F2x_degree(x)<<1) >= F2x_degree(T);
+  int use_sqr = 2*F2x_degree(x) >= F2x_degree(T);
   return gen_bkeval(Q, d, x, use_sqr, (void*)T, &F2xq_algebra, _F2xq_cmul);
 }
 
@@ -902,21 +997,11 @@ F2xq_conjvec(GEN x, GEN T)
 {
   long i, l = F2x_degree(T);
   GEN z = cgetg(l,t_COL);
-  gel(z,1) = vecsmall_copy(x);
+  gel(z,1) = F2x_copy(x);
   for (i=2; i<l; i++) gel(z,i) = F2xq_sqr(gel(z,i-1), T);
   return z;
 }
 
-GEN
-random_F2x(long d, long vs)
-{
-  long i, l = nbits2lg(d+1);
-  GEN y = cgetg(l,t_VECSMALL); y[1] = vs;
-  for (i=2; i<l; i++) y[i] = pari_rand();
-  y[l-1] &= (1UL<<remsBIL(d))-1UL;
-  return F2x_renormalize(y,l);
-}
-
 static GEN
 _F2xq_pow(void *data, GEN x, GEN n)
 {
@@ -977,7 +1062,6 @@ F2x_is_smooth(GEN g, long r)
     if (F2x_degree(f)==0) return 1;
     g = F2x_issquare(f) ? F2x_sqrt(f): f;
   }
-  return 0;
 }
 
 static GEN
@@ -1084,7 +1168,7 @@ F2xq_log_Coppersmith_d(GEN W, GEN g, long r, long n, GEN T, GEN mo)
 static GEN
 F2xq_log_find_rel(GEN b, long r, GEN T, GEN *g, ulong *e)
 {
-  pari_sp av = avma, lim = stack_lim(av,2);
+  pari_sp av = avma;
   while (1)
   {
     GEN M;
@@ -1103,7 +1187,7 @@ F2xq_log_find_rel(GEN b, long r, GEN T, GEN *g, ulong *e)
         return rel;
       }
     }
-    if (low_stack(lim, stack_lim(av,2)))
+    if (gc_needed(av,2))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"F2xq_log_find_rel");
       *g = gerepileuptoleaf(av, *g);
@@ -1239,15 +1323,15 @@ check_kernel(long N, GEN M, GEN T, GEN m)
   GEN K = FpMs_leftkernel_elt(M, N, m);
   long i, f=0;
   long l = lg(K), lm = lgefint(m);
-  GEN g = polx_F2x(T[1]);
   GEN idx = diviiexact(subis(int2n(F2x_degree(T)),1),m);
+  GEN g = F2xq_pow(polx_F2x(T[1]), idx, T);
   pari_timer ti;
   if (DEBUGLEVEL) timer_start(&ti);
-  K = FpC_Fp_mul(K, Fp_inv(gel(K,g[2]), m), m);
+  K = FpC_Fp_mul(K, Fp_inv(gel(K,2), m), m);
   for(i=1; i<l; i++)
   {
     GEN k = gel(K,i);
-    if (signe(k)==0 || !F2x_equal(F2xq_pow(g, mulii(k,idx), T),
+    if (signe(k)==0 || !F2x_equal(F2xq_pow(g, k, T),
                                   F2xq_pow(mkF2(i,T[1]), idx, T)))
       gel(K,i) = cgetineg(lm);
     else
@@ -1305,7 +1389,7 @@ F2xq_easylog(void* E, GEN a, GEN g, GEN ord)
 GEN
 F2xq_log(GEN a, GEN g, GEN ord, GEN T)
 {
-  GEN z, v = dlog_get_ordfa(ord);
+  GEN z, v = get_arith_ZZM(ord);
   ord = mkvec2(gel(v,1),ZM_famat_limit(gel(v,2),int2n(28)));
   z = gen_PH_log(a,g,ord,(void*)T,&F2xq_star);
   return z? z: cgetg(1,t_VEC);
@@ -1316,21 +1400,15 @@ F2xq_Artin_Schreier(GEN a, GEN T)
 {
   pari_sp ltop=avma;
   long j,N = F2x_degree(T);
-  GEN Q, XP;
-  pari_timer ti;
-  timer_start(&ti);
-  XP = F2xq_sqr(polx_F2x(T[1]),T);
-  Q  = F2xq_matrix_pow(XP,N,N,T);
+  GEN Q = F2x_matFrobenius(T);
   for (j=1; j<=N; j++)
     F2m_flip(Q,j,j);
-  if(DEBUGLEVEL>=9) timer_printf(&ti,"Berlekamp matrix");
   F2v_add_inplace(gel(Q,1),a);
   Q = F2m_ker_sp(Q,0);
-  if(DEBUGLEVEL>=9) timer_printf(&ti,"kernel");
   if (lg(Q)!=2) return NULL;
   Q = gel(Q,1);
   Q[1] = T[1];
-  return gerepileuptoleaf(ltop, Q);
+  return gerepileuptoleaf(ltop, F2x_renormalize(Q, lg(Q)));
 }
 
 GEN
@@ -1350,7 +1428,7 @@ F2xq_sqrt(GEN a, GEN T)
   pari_sp av = avma;
   long n = F2x_degree(T);
   GEN sqx;
-  if (n==1) return leafcopy(a);
+  if (n==1) return F2x_copy(a);
   if (n==2) return F2xq_sqr(a,T);
   sqx = F2xq_autpow(mkF2(4, T[1]), n-1, T);
   return gerepileuptoleaf(av, F2x_is_x(a)? sqx: F2xq_sqrt_fast(a,sqx,T));
@@ -1385,7 +1463,7 @@ gener_F2xq(GEN T, GEN *po)
   L2 = leafcopy( gel(o, 1) );
   for (i = j = 1; i < lg(L2); i++)
   {
-    if (equaliu(gel(L2,i),2)) continue;
+    if (absequaliu(gel(L2,i),2)) continue;
     gel(L2,j++) = diviiexact(q, gel(L2,i));
   }
   setlg(L2, j);
@@ -1408,29 +1486,9 @@ gener_F2xq(GEN T, GEN *po)
   return g;
 }
 
-GEN
-ZXX_to_F2xX(GEN B, long v)
-{
-  long lb=lg(B);
-  long i;
-  GEN b=cgetg(lb,t_POL);
-  b[1]=evalsigne(1)|(((ulong)B[1])&VARNBITS);
-  for (i=2; i<lb; i++)
-    switch (typ(gel(B,i)))
-    {
-    case t_INT:
-      gel(b,i) = Z_to_F2x(gel(B,i), v);
-      break;
-    case t_POL:
-      gel(b,i) = ZX_to_F2x(gel(B,i));
-      break;
-    }
-  return FlxX_renormalize(b, lb);
-}
-
 static GEN
 _F2xq_neg(void *E, GEN x)
-{ (void) E; return vecsmall_copy(x); }
+{ (void) E; return F2x_copy(x); }
 
 static GEN
 _F2xq_rmul(void *E, GEN x, GEN y)
@@ -1540,7 +1598,7 @@ F2m_to_mod(GEN z)
 }
 
 GEN
-F2c_to_Flc(GEN x)
+F2v_to_Flv(GEN x)
 {
   long l=x[1]+1;
   GEN  z=cgetg(l, t_VECSMALL);
@@ -1556,7 +1614,7 @@ F2m_to_Flm(GEN z)
 {
   long i, l = lg(z);
   GEN x = cgetg(l,t_MAT);
-  for (i=1; i<l; i++) gel(x,i) = F2c_to_Flc(gel(z,i));
+  for (i=1; i<l; i++) gel(x,i) = F2v_to_Flv(gel(z,i));
   return x;
 }
 
@@ -1635,6 +1693,16 @@ Flm_to_F2m(GEN x)
   return y;
 }
 
+GEN
+const_F2v(long m)
+{
+  long i, l = nbits2lg(m);
+  GEN c = cgetg(l, t_VECSMALL);
+  for (i = 2; i <= l; i++) c[i] = -1;
+  if (remsBIL(m)) c[l] = (1UL<<remsBIL(m))-1UL;
+  return c;
+}
+
 /* Allow lg(y)<lg(x) */
 void
 F2v_add_inplace(GEN x, GEN y)
@@ -1662,6 +1730,33 @@ F2v_add_inplace(GEN x, GEN y)
 /***********************************************************************/
 /* F2xV are t_VEC with F2x coefficients. */
 
+GEN
+FlxC_to_F2xC(GEN x)
+{
+  long i, l=lg(x);
+  GEN z = cgetg(l,t_COL);
+  for (i=1; i<l ; i++) gel(z,i) = Flx_to_F2x(gel(x,i));
+  return z;
+}
+
+GEN
+F2xC_to_FlxC(GEN x)
+{
+  long i, l=lg(x);
+  GEN z = cgetg(l,t_COL);
+  for (i=1; i<l ; i++) gel(z,i) = F2x_to_Flx(gel(x,i));
+  return z;
+}
+
+GEN
+F2xC_to_ZXC(GEN v)
+{
+  long j, N = lg(v);
+  GEN y = cgetg(N, t_COL);
+  for (j=1; j<N; j++) gel(y,j) = F2x_to_ZX(gel(v,j));
+  return y;
+}
+
 GEN
 F2xV_to_F2m(GEN v, long n)
 {
@@ -1670,3 +1765,588 @@ F2xV_to_F2m(GEN v, long n)
   for (j=1; j<N; j++) gel(y,j) = F2x_to_F2v(gel(v,j), n);
   return y;
 }
+
+/***********************************************************************/
+/**                                                                   **/
+/**                             F2xX                                  **/
+/**                                                                   **/
+/***********************************************************************/
+
+GEN
+F2xX_renormalize(GEN /*in place*/ x, long lx)
+{ return FlxX_renormalize(x, lx); }
+
+GEN
+pol1_F2xX(long v, long sv) { return pol1_FlxX(v, sv); }
+
+GEN
+polx_F2xX(long v, long sv) { return polx_FlxX(v, sv); }
+
+long
+F2xY_degreex(GEN b)
+{
+  long deg = 0, i;
+  if (!signe(b)) return -1;
+  for (i = 2; i < lg(b); ++i)
+    deg = maxss(deg, F2x_degree(gel(b, i)));
+  return deg;
+}
+
+GEN
+FlxX_to_F2xX(GEN B)
+{
+  long lb=lg(B);
+  long i;
+  GEN b=cgetg(lb,t_POL);
+  b[1]=evalsigne(1)|(((ulong)B[1])&VARNBITS);
+  for (i=2; i<lb; i++)
+    gel(b,i) = Flx_to_F2x(gel(B,i));
+  return F2xX_renormalize(b, lb);
+}
+
+GEN
+ZXX_to_F2xX(GEN B, long v)
+{
+  long lb=lg(B);
+  long i;
+  GEN b=cgetg(lb,t_POL);
+  b[1]=evalsigne(1)|(((ulong)B[1])&VARNBITS);
+  for (i=2; i<lb; i++)
+    switch (typ(gel(B,i)))
+    {
+    case t_INT:
+      gel(b,i) = Z_to_F2x(gel(B,i), v);
+      break;
+    case t_POL:
+      gel(b,i) = ZX_to_F2x(gel(B,i));
+      break;
+    }
+  return F2xX_renormalize(b, lb);
+}
+
+GEN
+F2xX_to_ZXX(GEN B)
+{
+  long i, lb = lg(B);
+  GEN b = cgetg(lb,t_POL);
+  for (i=2; i<lb; i++)
+  {
+    GEN c = gel(B,i);
+    gel(b,i) = lgpol(c) ?  F2x_equal1(c) ? gen_1 : F2x_to_ZX(c) : gen_0;
+  }
+  b[1] = B[1]; return b;
+}
+
+GEN
+F2xX_deriv(GEN z)
+{
+  long i,l = lg(z)-1;
+  GEN x;
+  if (l < 2) l = 2;
+  x = cgetg(l, t_POL); x[1] = z[1];
+  for (i=2; i<l; i++) gel(x,i) = odd(i) ? pol0_F2x(mael(z,i+1,1)): gel(z,i+1);
+  return F2xX_renormalize(x,l);
+}
+
+GEN
+F2xX_add(GEN x, GEN y)
+{
+  long i,lz;
+  GEN z;
+  long lx=lg(x);
+  long ly=lg(y);
+  if (ly>lx) swapspec(x,y, lx,ly);
+  lz = lx; z = cgetg(lz, t_POL); z[1]=x[1];
+  for (i=2; i<ly; i++) gel(z,i) = F2x_add(gel(x,i), gel(y,i));
+  for (   ; i<lx; i++) gel(z,i) = F2x_copy(gel(x,i));
+  return F2xX_renormalize(z, lz);
+}
+
+GEN
+F2xX_F2x_mul(GEN P, GEN U)
+{
+  long i, lP = lg(P);
+  GEN res = cgetg(lP,t_POL);
+  res[1] = P[1];
+  for(i=2; i<lP; i++)
+    gel(res,i) = F2x_mul(U,gel(P,i));
+  return F2xX_renormalize(res, lP);
+}
+
+GEN
+F2xY_F2xqV_evalx(GEN P, GEN x, GEN T)
+{
+  long i, lP = lg(P);
+  GEN res = cgetg(lP,t_POL);
+  res[1] = P[1];
+  for(i=2; i<lP; i++)
+    gel(res,i) = F2x_F2xqV_eval(gel(P,i), x, T);
+  return F2xX_renormalize(res, lP);
+}
+
+GEN
+F2xY_F2xq_evalx(GEN P, GEN x, GEN T)
+{
+  pari_sp av = avma;
+  long n = brent_kung_optpow(F2x_degree(T)-1,lgpol(P),1);
+  GEN xp = F2xq_powers(x, n, T);
+  return gerepileupto(av, F2xY_F2xqV_evalx(P, xp, T));
+}
+
+GEN
+F2xX_to_Kronecker(GEN P, long d)
+{
+  long i, k, N = 2*d + 1;
+  long dP = degpol(P);
+  long l = nbits2nlong(N*dP+d+1);
+  GEN x = zero_zv(l+1);
+  for (k=i=0; i<=dP; i++, k+=N)
+  {
+    GEN c = gel(P,i+2);
+    F2x_addshiftip(x, c, k);
+  }
+  x[1] = P[1]&VARNBITS; return F2x_renormalize(x, l+2);
+}
+
+static GEN
+F2x_slice(GEN x, long n, long d)
+{
+  ulong ib, il=dvmduBIL(n, &ib);
+  ulong db, dl=dvmduBIL(d, &db);
+  long lN = dl+2+(db?1:0);
+  GEN t = cgetg(lN,t_VECSMALL);
+  t[1] = x[1];
+  if (ib)
+  {
+    ulong i, ic = BITS_IN_LONG-ib;
+    ulong r = uel(x,2+il)>>ib;
+    for(i=0; i<dl; i++)
+    {
+      uel(t,2+i) = (uel(x,3+il+i)<<ic)|r;
+      r = uel(x,3+il+i)>>ib;
+    }
+    if (db)
+      uel(t,2+i) = (uel(x,3+il+i)<<ic)|r;
+  }
+  else
+  {
+    long i;
+    for(i=2; i<lN; i++)
+      uel(t,i) = uel(x,il+i);
+  }
+  if (db) uel(t,lN-1) &= (1UL<<db)-1;
+  return F2x_renormalize(t, lN);
+}
+
+GEN
+Kronecker_to_F2xqX(GEN z, GEN T)
+{
+  long lz = F2x_degree(z)+1;
+  long i, j, N = F2x_degree(T)*2 + 1;
+  long lx = lz / (N-2);
+  GEN x = cgetg(lx+3,t_POL);
+  x[1] = z[1];
+  for (i=0, j=2; i<lz; i+=N, j++)
+  {
+    GEN t = F2x_slice(z,i,minss(lz-i,N));
+    t[1] = T[1];
+    gel(x,j) = F2x_rem(t, T);
+  }
+  return F2xX_renormalize(x, j);
+}
+
+/***********************************************************************/
+/**                                                                   **/
+/**                             F2xXV/F2xXC                           **/
+/**                                                                   **/
+/***********************************************************************/
+
+GEN
+FlxXC_to_F2xXC(GEN B)
+{
+  long i, l = lg(B);
+  GEN z = cgetg(l, t_COL);
+  for (i=1; i<l; i++)
+    gel(z,i) = FlxX_to_F2xX(gel(B,i));
+  return z;
+}
+
+GEN
+F2xXC_to_ZXXC(GEN B)
+{
+  long i, l = lg(B);
+  GEN z = cgetg(l, t_COL);
+  for (i=1; i<l; i++)
+    gel(z,i) = F2xX_to_ZXX(gel(B,i));
+  return z;
+}
+
+/***********************************************************************/
+/**                                                                   **/
+/**                             F2xqX                                 **/
+/**                                                                   **/
+/***********************************************************************/
+
+GEN
+random_F2xqX(long d1, long v, GEN T)
+{
+  long dT = F2x_degree(T), vT = T[1];
+  long i, d = d1+2;
+  GEN y = cgetg(d,t_POL); y[1] = evalsigne(1) | evalvarn(v);
+  for (i=2; i<d; i++) gel(y,i) = random_F2x(dT, vT);
+  return FlxX_renormalize(y,d);
+}
+
+GEN
+F2xqX_red(GEN z, GEN T)
+{
+  GEN res;
+  long i, l = lg(z);
+  res = cgetg(l,t_POL); res[1] = z[1];
+  for(i=2;i<l;i++) gel(res,i) = F2x_rem(gel(z,i),T);
+  return F2xX_renormalize(res,l);
+}
+
+GEN
+F2xqX_F2xq_mul(GEN P, GEN U, GEN T)
+{
+  long i, lP = lg(P);
+  GEN res = cgetg(lP,t_POL);
+  res[1] = P[1];
+  for(i=2; i<lP; i++)
+    gel(res,i) = F2xq_mul(U,gel(P,i), T);
+  return F2xX_renormalize(res, lP);
+}
+
+GEN
+F2xqX_F2xq_mul_to_monic(GEN P, GEN U, GEN T)
+{
+  long i, lP = lg(P);
+  GEN res = cgetg(lP,t_POL);
+  res[1] = P[1];
+  for(i=2; i<lP-1; i++) gel(res,i) = F2xq_mul(U,gel(P,i), T);
+  gel(res,lP-1) = pol1_F2x(T[1]);
+  return F2xX_renormalize(res, lP);
+}
+
+GEN
+F2xqX_normalize(GEN z, GEN T)
+{
+  GEN p1 = leading_coeff(z);
+  if (!lgpol(z) || (!degpol(p1) && p1[1] == 1)) return z;
+  return F2xqX_F2xq_mul_to_monic(z, F2xq_inv(p1,T), T);
+}
+
+GEN
+F2xqX_mul(GEN x, GEN y, GEN T)
+{
+  pari_sp ltop=avma;
+  GEN z,kx,ky;
+  kx= F2xX_to_Kronecker(x, F2x_degree(T));
+  ky= F2xX_to_Kronecker(y, F2x_degree(T));
+  z = F2x_mul(ky, kx);
+  z = Kronecker_to_F2xqX(z,T);
+  return gerepileupto(ltop,z);
+}
+
+GEN
+F2xqX_sqr(GEN x, GEN T)
+{
+  long d = degpol(x), l = 2*d+3;
+  GEN z;
+  if (!signe(x)) return pol_0(varn(x));
+  z = cgetg(l,t_POL);
+  z[1] = x[1];
+  if (d > 0)
+  {
+    GEN u = pol0_F2x(T[1]);
+    long i,j;
+    for (i=2,j=2; i<d+2; i++,j+=2)
+    {
+      gel(z, j) = F2xq_sqr(gel(x, i), T);
+      gel(z, j+1) = u;
+    }
+  }
+  gel(z, 2+2*d) = F2xq_sqr(gel(x, 2+d), T);
+  return FlxX_renormalize(z, l);
+}
+
+GEN
+F2xqX_divrem(GEN x, GEN y, GEN T, GEN *pr)
+{
+  long vx, dx, dy, dz, i, j, sx, lr;
+  pari_sp av0, av, tetpil;
+  GEN z,p1,rem,lead;
+
+  if (!signe(y)) pari_err_INV("F2xqX_divrem",y);
+  vx=varn(x); dy=degpol(y); dx=degpol(x);
+  if (dx < dy)
+  {
+    if (pr)
+    {
+      av0 = avma; x = F2xqX_red(x, T);
+      if (pr == ONLY_DIVIDES) { avma=av0; return signe(x)? NULL: pol_0(vx); }
+      if (pr == ONLY_REM) return x;
+      *pr = x;
+    }
+    return pol_0(vx);
+  }
+  lead = leading_term(y);
+  if (!dy) /* y is constant */
+  {
+    if (pr && pr != ONLY_DIVIDES)
+    {
+      if (pr == ONLY_REM) return pol_0(vx);
+      *pr = pol_0(vx);
+    }
+    if (F2x_equal1(lead)) return gcopy(x);
+    av0 = avma; x = F2xqX_F2xq_mul(x,F2xq_inv(lead,T),T);
+    return gerepileupto(av0,x);
+  }
+  av0 = avma; dz = dx-dy;
+  lead = F2x_equal1(lead)? NULL: gclone(F2xq_inv(lead,T));
+  avma = av0;
+  z = cgetg(dz+3,t_POL); z[1] = x[1];
+  x += 2; y += 2; z += 2;
+
+  p1 = gel(x,dx); av = avma;
+  gel(z,dz) = lead? gerepileupto(av, F2xq_mul(p1,lead, T)): leafcopy(p1);
+  for (i=dx-1; i>=dy; i--)
+  {
+    av=avma; p1=gel(x,i);
+    for (j=i-dy+1; j<=i && j<=dz; j++)
+      p1 = F2x_add(p1, F2x_mul(gel(z,j),gel(y,i-j)));
+    if (lead) p1 = F2x_mul(p1, lead);
+    tetpil=avma; gel(z,i-dy) = gerepile(av,tetpil,F2x_rem(p1,T));
+  }
+  if (!pr) { if (lead) gunclone(lead); return z-2; }
+
+  rem = (GEN)avma; av = (pari_sp)new_chunk(dx+3);
+  for (sx=0; ; i--)
+  {
+    p1 = gel(x,i);
+    for (j=0; j<=i && j<=dz; j++)
+      p1 = F2x_add(p1, F2x_mul(gel(z,j),gel(y,i-j)));
+    tetpil=avma; p1 = F2x_rem(p1, T); if (lgpol(p1)) { sx = 1; break; }
+    if (!i) break;
+    avma=av;
+  }
+  if (pr == ONLY_DIVIDES)
+  {
+    if (lead) gunclone(lead);
+    if (sx) { avma=av0; return NULL; }
+    avma = (pari_sp)rem; return z-2;
+  }
+  lr=i+3; rem -= lr;
+  rem[0] = evaltyp(t_POL) | evallg(lr);
+  rem[1] = z[-1];
+  p1 = gerepile((pari_sp)rem,tetpil,p1);
+  rem += 2; gel(rem,i) = p1;
+  for (i--; i>=0; i--)
+  {
+    av=avma; p1 = gel(x,i);
+    for (j=0; j<=i && j<=dz; j++)
+      p1 = F2x_add(p1, F2x_mul(gel(z,j),gel(y,i-j)));
+    tetpil=avma; gel(rem,i) = gerepile(av,tetpil, F2x_rem(p1, T));
+  }
+  rem -= 2;
+  if (lead) gunclone(lead);
+  if (!sx) (void)F2xX_renormalize(rem, lr);
+  if (pr == ONLY_REM) return gerepileupto(av0,rem);
+  *pr = rem; return z-2;
+}
+
+GEN
+F2xqX_rem(GEN x, GEN S, GEN T)
+{
+  GEN y = S;
+  long dy = degpol(y), dx = degpol(x), d = dx-dy;
+  if (d < 0) return F2xqX_red(x, T);
+  return F2xqX_divrem(x,y, T, ONLY_REM);
+}
+
+GEN
+F2xqX_gcd(GEN a, GEN b, GEN T)
+{
+  pari_sp av = avma, av0=avma;
+  while (signe(b))
+  {
+    GEN c;
+    if (gc_needed(av0,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"F2xqX_gcd (d = %ld)",degpol(b));
+      gerepileall(av0,2, &a,&b);
+    }
+    av = avma; c = F2xqX_rem(a, b, T); a=b; b=c;
+  }
+  avma = av; return a;
+}
+
+/***********************************************************************/
+/**                                                                   **/
+/**                             F2xqXQ                                **/
+/**                                                                   **/
+/***********************************************************************/
+
+GEN
+F2xqXQ_mul(GEN x, GEN y, GEN S, GEN T) {
+  return F2xqX_rem(F2xqX_mul(x,y,T),S,T);
+}
+
+GEN
+F2xqXQ_sqr(GEN x, GEN S, GEN T) {
+  return F2xqX_rem(F2xqX_sqr(x,T),S,T);
+}
+
+struct _F2xqXQ {
+  GEN T, S;
+};
+
+static GEN
+_F2xqXQ_add(void *data, GEN x, GEN y) {
+  (void) data;
+  return F2xX_add(x,y);
+}
+static GEN
+_F2xqXQ_cmul(void *data, GEN P, long a, GEN x) {
+  (void) data;
+  return F2xX_F2x_mul(x,gel(P,a+2));
+}
+static GEN
+_F2xqXQ_red(void *data, GEN x) {
+  struct _F2xqXQ *d = (struct _F2xqXQ*) data;
+  return F2xqX_red(x, d->T);
+}
+static GEN
+_F2xqXQ_mul(void *data, GEN x, GEN y) {
+  struct _F2xqXQ *d = (struct _F2xqXQ*) data;
+  return F2xqXQ_mul(x,y, d->S,d->T);
+}
+static GEN
+_F2xqXQ_sqr(void *data, GEN x) {
+  struct _F2xqXQ *d = (struct _F2xqXQ*) data;
+  return F2xqXQ_sqr(x, d->S,d->T);
+}
+static GEN
+_F2xqXQ_zero(void *data) {
+  struct _F2xqXQ *d = (struct _F2xqXQ*) data;
+  return pol_0(varn(d->S));
+}
+static GEN
+_F2xqXQ_one(void *data) {
+  struct _F2xqXQ *d = (struct _F2xqXQ*) data;
+  return pol1_F2xX(varn(d->S),d->T[1]);
+}
+
+static struct bb_algebra F2xqXQ_algebra = { _F2xqXQ_red,
+ _F2xqXQ_add, _F2xqXQ_add, _F2xqXQ_mul,_F2xqXQ_sqr,_F2xqXQ_one,_F2xqXQ_zero };
+
+GEN
+F2xqXQ_pow(GEN x, GEN n, GEN S, GEN T)
+{
+  struct _F2xqXQ D;
+  long s = signe(n);
+  if (!s) return pol1_F2xX(varn(S), T[1]);
+  if (s < 0) pari_err_IMPL("F2xqXQ_inv");
+  if (is_pm1(n)) return s < 0 ? x : gcopy(x);
+  if (degpol(x)>=degpol(S)) x = F2xqX_rem(x,S,T);
+  D.S = S; D.T = T;
+  return gen_pow(x, n, (void*)&D, &_F2xqXQ_sqr, &_F2xqXQ_mul);
+}
+
+GEN
+F2xqXQ_powers(GEN x, long l, GEN S, GEN T)
+{
+  struct _F2xqXQ D;
+  int use_sqr = 2*degpol(x) >= degpol(S);
+  D.S = S; D.T = T;
+  return gen_powers(x, l, use_sqr, (void*)&D, &_F2xqXQ_sqr, &_F2xqXQ_mul,&_F2xqXQ_one);
+}
+
+GEN
+F2xqX_F2xqXQV_eval(GEN P, GEN V, GEN S, GEN T)
+{
+  struct _F2xqXQ D;
+  D.S = S; D.T = T;
+  return gen_bkeval_powers(P, degpol(P), V, (void*)&D, &F2xqXQ_algebra,
+                                                   _F2xqXQ_cmul);
+}
+
+GEN
+F2xqX_F2xqXQ_eval(GEN Q, GEN x, GEN S, GEN T)
+{
+  struct _F2xqXQ D;
+  int use_sqr = 2*degpol(x) >= degpol(S);
+  D.S = S; D.T = T;
+  return gen_bkeval(Q, degpol(Q), x, use_sqr, (void*)&D, &F2xqXQ_algebra,
+                                                    _F2xqXQ_cmul);
+}
+
+static GEN
+F2xqXQ_autpow_sqr(void * E, GEN x)
+{
+  struct _F2xqXQ *D = (struct _F2xqXQ *)E;
+  GEN T = D->T;
+  GEN phi = gel(x,1), S = gel(x,2);
+  long n = brent_kung_optpow(F2x_degree(T)-1,lgpol(S)+1,1);
+  GEN V = F2xq_powers(phi, n, T);
+  GEN phi2 = F2x_F2xqV_eval(phi, V, T);
+  GEN Sphi = F2xY_F2xqV_evalx(S, V, T);
+  GEN S2 = F2xqX_F2xqXQ_eval(Sphi, S, D->S, T);
+  return mkvec2(phi2, S2);
+}
+
+static GEN
+F2xqXQ_autpow_mul(void * E, GEN x, GEN y)
+{
+  struct _F2xqXQ *D = (struct _F2xqXQ *)E;
+  GEN T = D->T;
+  GEN phi1 = gel(x,1), S1 = gel(x,2);
+  GEN phi2 = gel(y,1), S2 = gel(y,2);
+  long n = brent_kung_optpow(F2x_degree(T)-1,lgpol(S1)+1,1);
+  GEN V = F2xq_powers(phi2, n, T);
+  GEN phi3 = F2x_F2xqV_eval(phi1,V,T);
+  GEN Sphi = F2xY_F2xqV_evalx(S1,V,T);
+  GEN S3 = F2xqX_F2xqXQ_eval(Sphi, S2, D->S, T);
+  return mkvec2(phi3, S3);
+}
+
+GEN
+F2xqXQV_autpow(GEN aut, long n, GEN S, GEN T)
+{
+  struct _F2xqXQ D;
+  D.S = S; D.T = T;
+  return gen_powu(aut,n,&D,F2xqXQ_autpow_sqr,F2xqXQ_autpow_mul);
+}
+
+static GEN
+F2xqXQ_auttrace_mul(void *E, GEN x, GEN y)
+{
+  struct _F2xqXQ *D = (struct _F2xqXQ *) E;
+  GEN T = D->T;
+  GEN phi1 = gel(x,1), S1 = gel(x,2), a1 = gel(x,3);
+  GEN phi2 = gel(y,1), S2 = gel(y,2), a2 = gel(y,3);
+  long n2 = brent_kung_optpow(F2x_degree(T)-1, lgpol(S1)+lgpol(a1)+1, 1);
+  GEN V2 = F2xq_powers(phi2, n2, T);
+  GEN phi3 = F2x_F2xqV_eval(phi1, V2, T);
+  GEN Sphi = F2xY_F2xqV_evalx(S1, V2, T);
+  GEN aphi = F2xY_F2xqV_evalx(a1, V2, T);
+  long n = brent_kung_optpow(maxss(degpol(Sphi),degpol(aphi)),2,1);
+  GEN V = F2xqXQ_powers(S2, n, D->S, T);
+  GEN S3 = F2xqX_F2xqXQV_eval(Sphi, V, D->S, T);
+  GEN aS = F2xqX_F2xqXQV_eval(aphi, V, D->S, T);
+  GEN a3 = F2xX_add(aS, a2);
+  return mkvec3(phi3, S3, a3);
+}
+
+static GEN
+F2xqXQ_auttrace_sqr(void *E, GEN x)
+{ return F2xqXQ_auttrace_mul(E, x, x); }
+
+GEN
+F2xqXQV_auttrace(GEN aut, long n, GEN S, GEN T)
+{
+  struct _F2xqXQ D;
+  D.S = S; D.T = T;
+  return gen_powu(aut,n,&D,F2xqXQ_auttrace_sqr,F2xqXQ_auttrace_mul);
+}
