@@ -70,6 +70,39 @@ F2xqE_changepointinv(GEN x, GEN ch, GEN T)
   return z;
 }
 
+static GEN
+nonzerotrace_F2xq(GEN T)
+{
+  pari_sp av = avma;
+  long n = F2x_degree(T), vs = T[1];
+  GEN a;
+  if (odd(n))
+    return pol1_F2x(vs);
+  do
+  {
+    avma = av;
+    a = random_F2x(n, vs);
+  } while (F2xq_trace(a, T)==0);
+  return a;
+}
+
+void
+F2xq_elltwist(GEN a, GEN a6, GEN T, GEN *pt_a, GEN *pt_a6)
+{
+  pari_sp av = avma;
+  GEN n = nonzerotrace_F2xq(T);
+  if (typ(a)==t_VECSMALL)
+  {
+    *pt_a = gerepileuptoleaf(av, F2x_add(n, a));
+    *pt_a6 = vecsmall_copy(a6);
+  } else
+  {
+    GEN a6t = F2x_add(a6, F2xq_mul(n, F2xq_sqr(gel(a,1), T), T));
+    *pt_a6 = gerepileuptoleaf(av, a6t);
+    *pt_a = vecsmall_copy(a);
+  }
+}
+
 static GEN
 F2xqE_dbl_slope(GEN P, GEN a, GEN T, GEN *slope)
 {
@@ -289,9 +322,47 @@ F2xqE_log(GEN a, GEN b, GEN o, GEN a2, GEN T)
 static GEN
 F2xqE_vert(GEN P, GEN Q, GEN T)
 {
+  long vT = T[1];
   if (ell_is_inf(P))
     return pol1_F2x(T[1]);
-  return F2x_add(gel(Q, 1), gel(P, 1));
+  if (!F2x_equal(gel(Q, 1), gel(P, 1)))
+    return F2x_add(gel(Q, 1), gel(P, 1));
+  if (lgpol(gel(Q, 1))) return pol1_F2x(vT);
+  return F2xq_inv(gel(Q,2), T);
+}
+
+static GEN
+F2xqE_Miller_line(GEN R, GEN Q, GEN slope, GEN a, GEN T)
+{
+  long vT = T[1];
+  GEN x = gel(Q, 1), y = gel(Q, 2);
+  GEN tmp1 = F2x_add(x, gel(R, 1));
+  GEN tmp2 = F2x_add(F2xq_mul(tmp1, slope, T), gel(R, 2));
+  GEN s1, s2, ix;
+  if (!F2x_equal(y, tmp2))
+    return F2x_add(y, tmp2);
+  if (lgpol(x) == 0) return pol1_F2x(vT);
+  if (typ(a)==t_VEC)
+  {
+    GEN a4 = gel(a,2), a3i = gel(a,3);
+    s1 = F2xq_mul(F2x_add(a4, F2xq_sqr(x, T)), a3i, T);
+    if (!F2x_equal(s1, slope))
+      return F2x_add(s1, slope);
+    s2 = F2xq_mul(F2x_add(x, F2xq_sqr(s1, T)), a3i, T);
+    if (lgpol(s2)) return s2;
+    return zv_copy(a3i);
+  } else
+  {
+    GEN a2 = a ;
+    ix = F2xq_inv(x, T);
+    s1 = F2x_add(x, F2xq_mul(y, ix, T));
+    if (!F2x_equal(s1, slope))
+      return F2x_add(s1, slope);
+    s2 =F2x_add(a2, F2x_add(F2xq_sqr(s1,T), s1));
+    if (!F2x_equal(s2, x))
+      return  F2x_add(pol1_F2x(vT), F2xq_mul(s2, ix, T));
+    return ix;
+  }
 }
 
 /* Computes the equation of the line tangent to R and returns its
@@ -311,11 +382,9 @@ F2xqE_tangent_update(GEN R, GEN Q, GEN a2, GEN T, GEN *pt_R)
     *pt_R = ellinf();
     return F2xqE_vert(R, Q, T);
   } else {
-    GEN slope, tmp1, tmp2;
+    GEN slope;
     *pt_R = F2xqE_dbl_slope(R, a2, T, &slope);
-    tmp1 = F2x_add(gel(Q, 1), gel(R, 1));
-    tmp2 = F2x_add(F2xq_mul(tmp1, slope, T), gel(R,2));
-    return F2x_add(gel(Q, 2), tmp2);
+    return F2xqE_Miller_line(R, Q, slope, a2, T);
   }
 }
 
@@ -346,11 +415,9 @@ F2xqE_chord_update(GEN R, GEN P, GEN Q, GEN a2, GEN T, GEN *pt_R)
       return F2xqE_vert(R, Q, T);
     }
   } else {
-    GEN slope, tmp1, tmp2;
+    GEN slope;
     *pt_R = F2xqE_add_slope(P, R, a2, T, &slope);
-    tmp1  = F2xq_mul(F2x_add(gel(Q, 1), gel(R, 1)), slope, T);
-    tmp2  = F2x_add(tmp1, gel(R, 2));
-    return F2x_add(gel(Q, 2), tmp2);
+    return F2xqE_Miller_line(R, Q, slope, a2, T);
   }
 }
 
@@ -407,7 +474,6 @@ F2xqE_Miller(GEN Q, GEN P, GEN m, GEN a2, GEN T)
   g1 = pol1_F2x(T[1]);
   v = gen_pow(mkvec3(g1,g1,Q), m, (void*)&d, F2xqE_Miller_dbl, F2xqE_Miller_add);
   num = gel(v,1); denom = gel(v,2);
-  if (!lgpol(num) || !lgpol(denom)) { avma = ltop; return NULL; }
   return gerepileupto(ltop, F2xq_div(num, denom, T));
 }
 
@@ -419,9 +485,7 @@ F2xqE_weilpairing(GEN P, GEN Q, GEN m, GEN a2, GEN T)
   if (ell_is_inf(P) || ell_is_inf(Q) || F2x_equal(P,Q))
     return pol1_F2x(T[1]);
   num    = F2xqE_Miller(P, Q, m, a2, T);
-  if (!num) return pol1_F2x(T[1]);
   denom  = F2xqE_Miller(Q, P, m, a2, T);
-  if (!denom) { avma=ltop; return pol1_F2x(T[1]); }
   result = F2xq_div(num, denom, T);
   return gerepileupto(ltop, result);
 }
@@ -429,11 +493,9 @@ F2xqE_weilpairing(GEN P, GEN Q, GEN m, GEN a2, GEN T)
 GEN
 F2xqE_tatepairing(GEN P, GEN Q, GEN m, GEN a2, GEN T)
 {
-  GEN num;
   if (ell_is_inf(P) || ell_is_inf(Q))
     return pol1_F2x(T[1]);
-  num = F2xqE_Miller(P, Q, m, a2, T);
-  return num? num: pol1_F2x(T[1]);
+  return F2xqE_Miller(P, Q, m, a2, T);
 }
 
 /***********************************************************************/
@@ -719,48 +781,85 @@ F2xq_elltrace_Harley(GEN a6, GEN T2)
   return gerepileuptoint(ltop, t);
 }
 
+static GEN
+get_v(GEN u, GEN a, GEN T)
+{
+  GEN a4 = gel(a,2), a3i = gel(a,3);
+  GEN ui2 = F2xq_mul(u, a3i, T), ui4 = F2xq_sqr(ui2, T);
+  return F2xq_mul(a4, ui4, T);
+}
+
+static GEN
+get_r(GEN w, GEN u, GEN T)
+{
+  return F2xq_sqr(F2xq_mul(F2xq_Artin_Schreier(w, T), u, T), T);
+}
+
+static GEN
+F2xq_elltracej(GEN a, GEN a6, GEN T, GEN q, long n)
+{
+  GEN a3 = gel(a,1), a4 = gel(a,2), a3i = gel(a,3);
+  GEN r, pi, rhs;
+  long t, s, m = n >> 1;
+  if (odd(n))
+  {
+    GEN u, v, w;
+    u = F2xq_pow(a3,diviuexact(subis(shifti(q,1), 1), 3),T);
+    v = get_v(u, a, T);
+    if (F2xq_trace(v, T)==0) return gen_0;
+    w = F2xq_Artin_Schreier(F2x_1_add(v), T);
+    if (F2xq_trace(w, T)) w = F2x_1_add(w);
+    r = get_r(w, u, T);
+    pi = int2n(m+1);
+    s = (((m+3)&3L) <= 1) ? -1: 1;
+  }
+  else
+  {
+    if (F2x_degree(F2xq_pow(a3,diviuexact(subis(q, 1), 3),T))==0)
+    {
+      GEN u, v, w;
+      u = F2xq_sqrtn(a3, utoi(3), T, NULL);
+      v = get_v(u, a, T);
+      if (F2xq_trace(v, T)==1) return gen_0;
+      w = F2xq_Artin_Schreier(v, T);
+      if (F2xq_trace(w, T)==1) return gen_0;
+      r = get_r(w, u, T);
+      pi = int2n(m+1);
+      s = odd(m+1) ? -1: 1;
+    }
+    else
+    {
+      long sv = T[1];
+      GEN P = mkpoln(5, pol1_F2x(sv), pol0_F2x(sv), pol0_F2x(sv), a3, a4);
+      r = F2xq_sqr(gel(F2xqX_roots(P,T), 1), T);
+      pi = int2n(m);
+      s = odd(m) ? -1: 1;
+    }
+  }
+  rhs = F2x_add(F2xq_mul(F2x_add(F2xq_sqr(r, T), a4), r, T), a6);
+  t = F2xq_trace(F2xq_mul(rhs, F2xq_sqr(a3i, T), T), T);
+  return (t==0)^(s==1)? pi: negi(pi);
+}
+
 GEN
 F2xq_ellcard(GEN a, GEN a6, GEN T)
 {
   pari_sp av = avma;
   long n = F2x_degree(T);
-  GEN q = int2u(n), c;
+  GEN c;
   if (typ(a)==t_VECSMALL)
   {
     GEN t = F2xq_elltrace_Harley(a6, T);
-    c = addii(q, F2xq_trace(a,T) ? addui(1,t): subui(1,t));
+    c = addii(int2u(n), F2xq_trace(a,T) ? addui(1,t): subui(1,t));
   } else if (n==1)
   {
     long a4i = lgpol(gel(a,2)), a6i = lgpol(a6);
     return utoi(a4i? (a6i? 1: 5): 3);
   }
-  else if (n==2)
-  {
-    GEN a3 = gel(a,1), a4 = gel(a,2), x = polx_F2x(T[1]), x1 = pol1_F2x(T[1]);
-    GEN a613 = F2xq_mul(F2x_add(x1, a6),a3,T), a43= F2xq_mul(a4,a3,T);
-    long f0= F2xq_trace(F2xq_mul(a6,a3,T),T);
-    long f1= F2xq_trace(F2x_add(a43,a613),T);
-    long f2= F2xq_trace(F2x_add(F2xq_mul(a43,x,T),a613),T);
-    long f3= F2xq_trace(F2x_add(F2xq_mul(a43,F2x_add(x,x1),T),a613),T);
-    c = utoi(9-2*(f0+f1+f2+f3));
-  }
   else
   {
-    struct _F2xqE e;
-    long m = (n+1)>>1;
-    GEN q1 = addis(q, 1);
-    GEN v = n==4 ? mkvec4s(13,17,21,25)
-                 : odd(n) ? mkvec3(subii(q1,int2u(m)),q1,addii(q1,int2u(m))):
-                            mkvec5(subii(q1,int2u(m+1)),subii(q1,int2u(m)),q1,
-                                   addii(q1,int2u(m)),addii(q1,int2u(m+1)));
-    e.a2=a; e.a6=a6; e.T=T;
-    c = gen_select_order(v,(void*)&e, &F2xqE_group);
-    if (n==4 && equaliu(c, 21)) /* Ambiguous case */
-    {
-      GEN d = F2xq_powu(polx_F2x(T[1]),3,T), a3 = gel(a,1);
-      e.a6 = F2x_add(a6,F2xq_mul(d,F2xq_sqr(a3,T),T)); /* twist */
-      c = subui(34, gen_select_order(mkvec2s(13,25),(void*)&e, &F2xqE_group));
-    }
+    GEN q = int2u(n);
+    c = subii(addis(q, 1), F2xq_elltracej(a, a6, T, q, n));
   }
   return gerepileuptoint(av, c);
 }
