@@ -20,6 +20,12 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /**                                                                     **/
 /*************************************************************************/
 
+/*************************************************************************/
+/**                                                                     **/
+/**                   Low-level constructors                            **/
+/**                                                                     **/
+/*************************************************************************/
+
 INLINE void
 _getFF(GEN x, GEN *T, GEN *p, ulong *pp)
 {
@@ -66,6 +72,132 @@ mkFF_i(GEN x, GEN r)
   return _mkFF_i(x,z,r);
 }
 
+/*************************************************************************/
+/**                                                                     **/
+/**                   medium-level constructors                         **/
+/**                                                                     **/
+/*************************************************************************/
+
+static GEN
+Z_to_raw(GEN x, GEN ff)
+{
+  ulong pp;
+  GEN T, p;
+  _getFF(ff,&T,&p,&pp);
+  switch(ff[1])
+  {
+  case t_FF_FpXQ:
+    return scalarpol(x, varn(T));
+  case t_FF_F2xq:
+    return Z_to_F2x(x,varn(T));
+  default:
+    return Z_to_Flx(x,pp,T[1]);
+  }
+}
+
+static GEN
+Rg_to_raw(GEN x, GEN ff)
+{
+  long tx = typ(x);
+  switch(tx)
+  {
+  case t_INT: case t_FRAC: case t_PADIC: case t_INTMOD:
+    return Z_to_raw(Rg_to_Fp(x, FF_p_i(ff)), ff);
+  case t_FFELT:
+    if (!FF_samefield(x,ff))
+      pari_err_MODULUS("Rg_to_raw",x,ff);
+    return gel(x,2);
+  }
+  pari_err_TYPE("Rg_to_raw",x);
+  return NULL; /* not reached */
+}
+
+static GEN
+FFX_to_raw(GEN x, GEN ff)
+{
+  long i, lx;
+  GEN y = cgetg_copy(x,&lx);
+  y[1] = x[1];
+  for(i=2; i<lx; i++)
+    gel(y, i) = Rg_to_raw(gel(x, i), ff);
+  return y;
+}
+
+static GEN
+FFC_to_raw(GEN x)
+{
+  long i, lx;
+  GEN y = cgetg_copy(x,&lx);
+  for(i=1; i<lx; i++)
+    gel(y, i) = gel(gel(x, i), 2);
+  return y;
+}
+
+static GEN
+FFM_to_raw(GEN x)
+{
+  long i, lx;
+  GEN y = cgetg_copy(x,&lx);
+  for(i=1; i<lx; i++)
+    gel(y, i) = FFC_to_raw(gel(x, i));
+  return y;
+}
+
+/* in place */
+static GEN
+raw_to_FFC(GEN x, GEN ff)
+{
+  long i, lx = lg(x);
+  for (i=1; i<lx; i++) gel(x,i) = mkFF_i(ff, gel(x,i));
+  return x;
+}
+
+/* in place */
+static GEN
+raw_to_FFM(GEN x, GEN ff)
+{
+  long i, lx = lg(x);
+  for (i=1; i<lx; i++) gel(x,i) = raw_to_FFC(gel(x,i), ff);
+  return x;
+}
+
+GEN
+Fq_to_FF(GEN x, GEN ff)
+{
+  ulong pp;
+  GEN r, T, p, z=_initFF(ff,&T,&p,&pp);
+  int is_int = typ(x)==t_INT;
+  switch(ff[1])
+  {
+  case t_FF_FpXQ:
+    r= is_int ? scalarpol(x, varn(T)): x;
+    break;
+  case t_FF_F2xq:
+    r= is_int ? Z_to_F2x(x,T[1]): ZX_to_F2x(x);
+    break;
+  default:
+    r= is_int ? Z_to_Flx(x,pp,T[1]): ZX_to_Flx(x,pp);
+  }
+  setvarn(r, varn(T)); /* paranoia */
+  return _mkFF_i(ff,z,r);
+}
+
+/* in place */
+static GEN
+to_FFX(GEN x, GEN ff)
+{
+  long i, lx = lg(x);
+  if (typ(x) != t_POL) pari_err_TYPE("to_FFX",x);
+  for (i=2; i<lx; i++) gel(x,i) = Fq_to_FF(gel(x,i), ff);
+  return x;
+}
+
+/*************************************************************************/
+/**                                                                     **/
+/**                   Public functions                                  **/
+/**                                                                     **/
+/*************************************************************************/
+
 /* Return true if x and y are defined in the same field */
 int
 FF_samefield(GEN x, GEN y)
@@ -116,7 +248,7 @@ FF_equalm1(GEN x)
   case t_FF_FpXQ:
     return (degpol(y) == 0 && Fp_cmp_1(gel(y,2), p));
   default:
-    return (degpol(y) == 0 && (ulong)y[2] == pp-1);
+    return (degpol(y) == 0 && uel(y,2) == pp-1);
   }
 }
 
@@ -980,15 +1112,6 @@ F2xq_ell_to_a4a6(GEN E, GEN T)
   return NULL;
 }
 
-static GEN
-Fq_to_FpXQ(GEN x, GEN T)
-{
-  if (typ(x)==t_INT)
-    return scalarpol(x, varn(T));
-  else
-    return x;
-}
-
 static GEN
 FqV_to_FpXQV(GEN x, GEN T)
 {
@@ -1016,7 +1139,7 @@ FF_ellcard(GEN E)
   switch(fg[1])
   {
   case t_FF_FpXQ:
-    return FpXQ_ellcard(Fq_to_FpXQ(gel(e,1),T), Fq_to_FpXQ(gel(e,2),T),T,p);
+    return FpXQ_ellcard(Fq_to_FpXQ(gel(e,1),T,p), Fq_to_FpXQ(gel(e,2),T,p),T,p);
   case t_FF_F2xq:
     return F2xq_ellcard(gel(e,1),gel(e,2),T);
   default:
@@ -1024,6 +1147,32 @@ FF_ellcard(GEN E)
   }
 }
 
+GEN
+FF_ellcard_SEA(GEN E, ulong smallfact)
+{
+  pari_sp av = avma;
+  GEN T,p;
+  ulong pp;
+  GEN e = ellff_get_a4a6(E), a4, a6, card;
+  GEN fg = ellff_get_field(E);
+  _getFF(fg,&T,&p,&pp);
+  switch(fg[1])
+  {
+  case t_FF_FpXQ:
+    a4 = Fq_to_FpXQ(gel(e,1), T, p);
+    a6 = Fq_to_FpXQ(gel(e,2), T, p);
+    card = Fq_ellcard_SEA(a4, a6, powiu(p,degpol(T)), T,p, smallfact);
+    break;
+  case t_FF_F2xq:
+    pari_err_IMPL("SEA for char 2");
+  default:
+    a4 = Flx_to_ZX(gel(e,1));
+    a6 = Flx_to_ZX(gel(e,2));
+    card = Fq_ellcard_SEA(a4, a6, powuu(pp,degpol(T)), Flx_to_ZX(T), p, smallfact);
+  }
+  return gerepileuptoint(av, card);
+}
+
 /* return [G,m] */
 GEN
 FF_ellgroup(GEN E)
@@ -1038,7 +1187,7 @@ FF_ellgroup(GEN E)
   switch(fg[1])
   {
   case t_FF_FpXQ:
-    G = FpXQ_ellgroup(Fq_to_FpXQ(gel(e,1),T),Fq_to_FpXQ(gel(e,2),T),N,T,p,&m);
+    G = FpXQ_ellgroup(Fq_to_FpXQ(gel(e,1),T,p),Fq_to_FpXQ(gel(e,2),T,p),N,T,p,&m);
     break;
   case t_FF_F2xq:
     G = F2xq_ellgroup(gel(e,1),gel(e,2),N,T,&m); break;
@@ -1062,7 +1211,7 @@ FF_ellgens(GEN E)
   {
   case t_FF_FpXQ:
     e3 = FqV_to_FpXQV(gel(e,3),T);
-    F = FpXQ_ellgens(Fq_to_FpXQ(gel(e,1),T),Fq_to_FpXQ(gel(e,2),T),e3,G,m,T,p);
+    F = FpXQ_ellgens(Fq_to_FpXQ(gel(e,1),T,p),Fq_to_FpXQ(gel(e,2),T,p),e3,G,m,T,p);
     break;
   case t_FF_F2xq:
     F = F2xq_ellgens(gel(e,1),gel(e,2),gel(e,3),G,m,T);
@@ -1130,6 +1279,35 @@ FF_ellinit(GEN E, GEN fg)
   return y;
 }
 
+GEN
+FF_elltwist(GEN E)
+{
+  pari_sp av = avma;
+  GEN fg = ellff_get_field(E), e = ellff_get_a4a6(E);
+  GEN T, p, a, a6, V;
+  ulong pp;
+  _getFF(fg,&T,&p,&pp);
+  switch (fg[1])
+  {
+  case t_FF_FpXQ:
+    FpXQ_elltwist(gel(e,1), gel(e,2), T, p, &a, &a6);
+    V = mkvec5(gen_0, gen_0, gen_0, mkFF_i(fg, a), mkFF_i(fg, a6));
+    break;
+  case t_FF_F2xq:
+    F2xq_elltwist(gel(e,1), gel(e,2), T, &a, &a6);
+    V = typ(a)==t_VECSMALL ?
+          mkvec5(gen_1, mkFF_i(fg, a), gen_0, gen_0, mkFF_i(fg, a6)):
+          mkvec5(gen_0, gen_0, mkFF_i(fg, gel(a,1)), mkFF_i(fg, gel(a,2)), mkFF_i(fg, a6));
+    break;
+  default:
+    Flxq_elltwist(gel(e,1), gel(e,2), T, pp, &a, &a6);
+    V = typ(a)==t_VECSMALL ?
+          mkvec5(gen_0, gen_0, gen_0, mkFF_i(fg, a), mkFF_i(fg, a6)):
+          mkvec5(gen_0, mkFF_i(fg, gel(a,1)), gen_0, gen_0, mkFF_i(fg, a6));
+  }
+  return gerepilecopy(av, V);
+}
+
 static long
 F3x_equalm1(GEN x)
 { return degpol(x)==0 && x[2] == 2; }
@@ -1145,7 +1323,7 @@ FF_ellrandom(GEN E)
   switch (fg[1])
   {
   case t_FF_FpXQ:
-    Q = random_FpXQE(Fq_to_FpXQ(gel(e,1),T), Fq_to_FpXQ(gel(e,2),T), T, p);
+    Q = random_FpXQE(Fq_to_FpXQ(gel(e,1),T,p), Fq_to_FpXQ(gel(e,2),T,p), T, p);
     Q = FpXQE_changepoint(Q, FqV_to_FpXQV(gel(e,3), T) , T, p);
     break;
   case t_FF_F2xq:
@@ -1320,61 +1498,37 @@ FF_elltatepairing(GEN E, GEN P, GEN Q, GEN m)
   return _mkFF(fg,z,r);
 }
 
-static GEN
-to_FF(GEN x, GEN ff)
+GEN
+FFX_roots(GEN Pf, GEN ff)
 {
+  pari_sp av = avma;
+  GEN r,T,p;
   ulong pp;
-  GEN r, T, p, z=_initFF(ff,&T,&p,&pp);
-  int is_int = typ(x)==t_INT;
+  GEN P = FFX_to_raw(Pf, ff);
+  _getFF(ff,&T,&p,&pp);
   switch(ff[1])
   {
   case t_FF_FpXQ:
-    r= is_int ? scalarpol(x, varn(T)): x;
+    r = FpXQX_roots(P, T, p);
     break;
   case t_FF_F2xq:
-    r= is_int ? Z_to_F2x(x,T[1]): ZX_to_F2x(x);
+    r = F2xqX_roots(P, T);
     break;
   default:
-    r= is_int ? Z_to_Flx(x,pp,T[1]): ZX_to_Flx(x,pp);
+    r = FlxqX_roots(P, T, pp);
   }
-  setvarn(r, varn(T)); /* paranoia */
-  return _mkFF_i(ff,z,r);
+  r = raw_to_FFC(r, ff);
+  return gerepilecopy(av, r);
 }
 
-/* in place */
-static GEN
-to_FFX(GEN x, GEN ff)
+GEN
+FqX_to_FFX(GEN x, GEN ff)
 {
-  long i, lx = lg(x);
-  if (typ(x) != t_POL) pari_err_TYPE("to_FFX",x);
-  for (i=2; i<lx; i++) gel(x,i) = to_FF(gel(x,i), ff);
-  return x;
-}
-/* in place */
-static GEN
-to_FFC(GEN x, GEN ff)
-{
-  long i, lx = lg(x);
-  for (i=1; i<lx; i++) gel(x,i) = to_FF(gel(x,i), ff);
-  return x;
-}
-
-/* in place */
-static GEN
-raw_to_FFC(GEN x, GEN ff)
-{
-  long i, lx = lg(x);
-  for (i=1; i<lx; i++) gel(x,i) = mkFF_i(ff, gel(x,i));
-  return x;
-}
-
-/* in place */
-static GEN
-raw_to_FFM(GEN x, GEN ff)
-{
-  long i, lx = lg(x);
-  for (i=1; i<lx; i++) gel(x,i) = raw_to_FFC(gel(x,i), ff);
-  return x;
+  long i, lx;
+  GEN y =  cgetg_copy(x,&lx);
+  y[1] = x[1];
+  for (i=2; i<lx; i++) gel(y,i) = Fq_to_FF(gel(x,i), ff);
+  return y;
 }
 
 /* P vector of t_POL, E t_VECSMALL of exponents, ff a t_FFELT. Update elts of
@@ -1393,7 +1547,7 @@ to_FF_fact(long vP, GEN P, GEN E, GEN ff, pari_sp av)
     GEN Q = simplify_shallow(gel(P,j)); /* may contain pols of degree 0 */
     if (typ(Q) == t_POL) setvarn(Q, vP);
     gel(u,j) = Q;
-    gel(v,j) = utoi((ulong)E[j]);
+    gel(v,j) = utoi(uel(E,j));
   }
   y = gerepilecopy(av, y); u = gel(y,1);
   zf = FF_zero(ff);
@@ -1401,29 +1555,6 @@ to_FF_fact(long vP, GEN P, GEN E, GEN ff, pari_sp av)
   return y;
 }
 
-/*Warning: FFX are polynomials whose coefficients are compatible with FF:
- * t_INT t_INTMOD, t_FFELT. Assume varncmp(varn(T), varn(x)) < 0 */
-static GEN
-FFX_to_FqX(GEN x, GEN T, GEN p)
-{
-  long i, l = lg(x);
-  GEN z = cgetg(l, t_POL); z[1] = x[1];
-
-  for (i = 2; i < l; i++)
-  {
-    GEN y = gel(x,i);
-    if (typ(y) == t_FFELT)
-    {
-      y = FF_to_FpXQ(y);
-      setvarn(y, varn(T)); /* paranoia */
-    }
-    else
-      y = Rg_to_FpXQ(y, T,p);
-    gel(z,i) = simplify_shallow(y);
-  }
-  return normalizepol_lg(z, l);
-}
-
 static GEN
 FFX_init_fix_varn(GEN P, GEN x, GEN *pT, GEN *pp)
 {
@@ -1443,7 +1574,7 @@ FFX_init_fix_varn(GEN P, GEN x, GEN *pT, GEN *pp)
     T=Flx_to_ZX(T);
   }
   setvarn(T, 1);
-  Q = FFX_to_FqX(P, T,p);
+  Q = RgX_to_FqX(P, T,p);
   setvarn(Q, 0);
 
   *pT = T;
@@ -1459,22 +1590,10 @@ FFX_factor(GEN P, GEN x)
   pari_sp av=avma;
 
   P = FFX_init_fix_varn(P, x, &T, &p);
-  r = FqX_factor(P, T,p);
+  r = FpXQX_factor(P, T,p);
   return to_FF_fact(vP, gel(r,1),gel(r,2), x,av);
 }
 
-/* Roots of P over the field of definition of x */
-GEN
-FFX_roots(GEN P, GEN x)
-{
-  GEN r, T, p;
-  pari_sp av=avma;
-
-  P = FFX_init_fix_varn(P, x, &T, &p);
-  r = FqX_roots(P, T,p);
-  return gerepilecopy(av, to_FFC(r, x));
-}
-
 GEN
 ffgen(GEN T, long v)
 {
@@ -1490,11 +1609,21 @@ ffgen(GEN T, long v)
       if (!FpX_is_squarefree(T,p)) pari_err_IRREDPOL("ffgen",T);
       break;
     case t_INT:
-      d = Z_isanypower(T, &p);
-      if (!d) { d = 1; p = T; }
-      if (!BPSW_psp(p)) pari_err_PRIME("ffgen",p);
+      d = ispseudoprimepower(T,&p);
+      if (!d) pari_err_PRIME("ffgen",T);
       T = init_Fq(p, d, v);
       break;
+    case t_VEC: case t_COL:
+      if (lg(T) == 3) {
+        p = gel(T,1);
+        A = gel(T,2);
+        if (typ(p) == t_INT && typ(A) == t_INT)
+        {
+          d = itos(A);
+          T = init_Fq(p, d, v);
+          break;
+        }
+      }
     default:
       pari_err_TYPE("ffgen",T);
       return NULL;
@@ -1566,6 +1695,44 @@ p_to_FF(GEN p, long v)
   gel(ff,3) = T;
   gel(ff,4) = p; return ff;
 }
+GEN
+Tp_to_FF(GEN T, GEN p)
+{
+  GEN A, ff;
+  long v;
+  if (!T) return p_to_FF(p,0);
+  ff = cgetg(5,t_FFELT);
+  v = varn(T);
+  if (lgefint(p)==3)
+  {
+    ulong pp = p[2];
+    long sv = evalvarn(v);
+    if (pp==2)
+    {
+      ff[1] = t_FF_F2xq;
+      T = ZX_to_F2x(T);
+      A = pol1_F2x(sv);
+      p = gen_2;
+    }
+    else
+    {
+      ff[1] = t_FF_Flxq;
+      T = ZX_to_Flx(T, pp);
+      A = pol1_Flx(sv);
+      p = icopy(p);
+    }
+  }
+  else
+  {
+    ff[1] = t_FF_FpXQ;
+    T = ZX_copy(T);
+    A = pol_1(v);
+    p = icopy(p);
+  }
+  gel(ff,2) = A;
+  gel(ff,3) = T;
+  gel(ff,4) = p; return ff;
+}
 
 GEN
 fforder(GEN x, GEN o)
@@ -1642,42 +1809,22 @@ RgM_is_FFM(GEN x, GEN *ff)
 }
 
 static GEN
-FFC_to_raw(GEN x)
+FqC_to_FpXQC(GEN x, GEN T, GEN p)
 {
   long i, lx;
   GEN y = cgetg_copy(x,&lx);
   for(i=1; i<lx; i++)
-    gel(y, i) = gmael(x, i, 2);
+    gel(y, i) = Fq_to_FpXQ(gel(x, i), T, p);
   return y;
 }
 
 static GEN
-FFM_to_raw(GEN x)
+FqM_to_FpXQM(GEN x, GEN T, GEN p)
 {
   long i, lx;
   GEN y = cgetg_copy(x,&lx);
   for(i=1; i<lx; i++)
-    gel(y, i) = FFC_to_raw(gel(x, i));
-  return y;
-}
-
-static GEN
-FqC_to_FpXQC(GEN x, GEN T)
-{
-  long i, lx;
-  GEN y = cgetg_copy(x,&lx);
-  for(i=1; i<lx; i++)
-    gel(y, i) = Fq_to_FpXQ(gel(x, i), T);
-  return y;
-}
-
-static GEN
-FqM_to_FpXQM(GEN x, GEN T)
-{
-  long i, lx;
-  GEN y = cgetg_copy(x,&lx);
-  for(i=1; i<lx; i++)
-    gel(y, i) = FqC_to_FpXQC(gel(x, i), T);
+    gel(y, i) = FqC_to_FpXQC(gel(x, i), T, p);
   return y;
 }
 
@@ -1692,7 +1839,7 @@ FFM_wrap(GEN M, GEN ff, GEN (*Fq)(GEN,GEN,GEN),
   _getFF(ff,&T,&p,&pp); M = FFM_to_raw(M);
   switch(ff[1])
   {
-  case t_FF_FpXQ: M = FqM_to_FpXQM(Fq(M,T,p), T); break;
+  case t_FF_FpXQ: M = FqM_to_FpXQM(Fq(M,T,p), T, p); break;
   case t_FF_F2xq: M = F2xq(M,T); break;
   default: M = Flxq(M,T,pp); break;
   }
@@ -1763,9 +1910,10 @@ FFM_mul(GEN M, GEN N, GEN ff)
   pari_sp av = avma;
   ulong pp;
   GEN P, T, p;
+  int is_sqr = M==N;
   _getFF(ff, &T, &p, &pp);
   M = FFM_to_raw(M);
-  N = FFM_to_raw(N);
+  N = is_sqr? M: FFM_to_raw(N);
   switch (ff[1])
   {
   case t_FF_FpXQ: P = FqM_mul(M, N, T, p); break;
