@@ -68,6 +68,48 @@ FlxqE_changepointinv(GEN x, GEN ch, GEN T, ulong p)
   return z;
 }
 
+static ulong
+nonsquare_Fl(ulong p)
+{
+  ulong a;
+  do
+    a = random_Fl(p);
+  while (krouu(a, p) >= 0);
+  return a;
+}
+
+static GEN
+nonsquare_Flxq(GEN T, ulong p)
+{
+  pari_sp av = avma;
+  long n = degpol(T), vs = T[1];
+  GEN a;
+  if (odd(n))
+    return mkvecsmall2(vs, nonsquare_Fl(p));
+  do
+  {
+    avma = av;
+    a = random_Flx(n, vs, p);
+  } while (Flxq_issquare(a, T, p));
+  return a;
+}
+
+void
+Flxq_elltwist(GEN a, GEN a6, GEN T, ulong p, GEN *pt_a, GEN *pt_a6)
+{
+  GEN d = nonsquare_Flxq(T, p);
+  GEN d2 = Flxq_sqr(d, T, p), d3 = Flxq_mul(d2, d, T, p);
+  if (typ(a)==t_VECSMALL)
+  {
+    *pt_a  = Flxq_mul(a,  d2, T, p);
+    *pt_a6 = Flxq_mul(a6, d3, T, p);
+  } else
+  {
+    *pt_a  = mkvec(Flxq_mul(gel(a,1), d, T, p));
+    *pt_a6 = Flxq_mul(a6, d3, T, p);
+  }
+}
+
 static GEN
 FlxqE_dbl_slope(GEN P, GEN a4, GEN T, ulong p, GEN *slope)
 {
@@ -278,11 +320,46 @@ FlxqE_log(GEN a, GEN b, GEN o, GEN a4, GEN T, ulong p)
 /* Derived from APIP from and by Jerome Milan, 2012 */
 
 static GEN
-FlxqE_vert(GEN P, GEN Q, GEN T, ulong p)
+FlxqE_vert(GEN P, GEN Q, GEN a4, GEN T, ulong p)
 {
+  long vT = get_Flx_var(T);
+  GEN df;
   if (ell_is_inf(P))
-    return pol1_Flx(get_Flx_var(T));
-  return Flx_sub(gel(Q, 1), gel(P, 1), p);
+    return pol1_Flx(vT);
+  if (!Flx_equal(gel(Q, 1), gel(P, 1)))
+    return Flx_sub(gel(Q, 1), gel(P, 1), p);
+  if (lgpol(gel(P,2))!=0) return pol1_Flx(vT);
+  df = typ(a4)==t_VEC ? Flxq_mul(gel(P,1), Flx_mulu(gel(a4, 1), 2, p), T, p)
+                      : a4;
+  return Flxq_inv(Flx_add(Flx_mulu(Flxq_sqr(gel(P,1), T, p), 3, p),
+                          df, p), T, p);
+}
+
+static GEN
+FlxqE_Miller_line(GEN R, GEN Q, GEN slope, GEN a4, GEN T, ulong p)
+{
+  long vT = get_Flx_var(T);
+  GEN x = gel(Q, 1), y = gel(Q, 2);
+  GEN tmp1 = Flx_sub(x, gel(R, 1), p);
+  GEN tmp2 = Flx_add(Flxq_mul(tmp1, slope, T, p), gel(R, 2), p);
+  if (!Flx_equal(y, tmp2))
+    return Flx_sub(y, tmp2, p);
+  if (lgpol(y) == 0)
+    return pol1_Flx(vT);
+  else
+  {
+    GEN s1, s2, a2 = typ(a4)==t_VEC ? gel(a4,1): NULL;
+    GEN y2i = Flxq_inv(Flx_mulu(y, 2, p), T, p);
+    GEN df = a2 ? Flxq_mul(x, Flx_mulu(a2, 2, p), T, p): a4;
+    GEN x3, ddf;
+    s1 = Flxq_mul(Flx_add(Flx_mulu(Flxq_sqr(x, T, p), 3, p), df, p), y2i, T, p);
+    if (!Flx_equal(s1, slope))
+      return Flx_sub(s1, slope, p);
+    x3 = Flx_mulu(x, 3, p);
+    ddf = a2 ? Flx_add(x3, a2, p): x3;
+    s2 = Flxq_mul(Flx_sub(ddf, Flxq_sqr(s1, T, p), p), y2i, T, p);
+    return lgpol(s2)!=0 ? s2: y2i;
+  }
 }
 
 /* Computes the equation of the line tangent to R and returns its
@@ -300,13 +377,11 @@ FlxqE_tangent_update(GEN R, GEN Q, GEN a4, GEN T, ulong p, GEN *pt_R)
   else if (!lgpol(gel(R,2)))
   {
     *pt_R = ellinf();
-    return FlxqE_vert(R, Q, T, p);
+    return FlxqE_vert(R, Q, a4, T, p);
   } else {
-    GEN slope, tmp1, tmp2;
+    GEN slope;
     *pt_R = FlxqE_dbl_slope(R, a4, T, p, &slope);
-    tmp1 = Flx_sub(gel(Q, 1), gel(R, 1), p);
-    tmp2 = Flx_add(Flxq_mul(tmp1, slope, T, p), gel(R,2), p);
-    return Flx_sub(gel(Q, 2), tmp2, p);
+    return FlxqE_Miller_line(R, Q, slope, a4, T, p);
   }
 }
 
@@ -320,12 +395,12 @@ FlxqE_chord_update(GEN R, GEN P, GEN Q, GEN a4, GEN T, ulong p, GEN *pt_R)
   if (ell_is_inf(R))
   {
     *pt_R = gcopy(P);
-    return FlxqE_vert(P, Q, T, p);
+    return FlxqE_vert(P, Q, a4, T, p);
   }
   else if (ell_is_inf(P))
   {
     *pt_R = gcopy(R);
-    return FlxqE_vert(R, Q, T, p);
+    return FlxqE_vert(R, Q, a4, T, p);
   }
   else if (Flx_equal(gel(P, 1), gel(R, 1)))
   {
@@ -334,14 +409,12 @@ FlxqE_chord_update(GEN R, GEN P, GEN Q, GEN a4, GEN T, ulong p, GEN *pt_R)
     else
     {
       *pt_R = ellinf();
-      return FlxqE_vert(R, Q, T, p);
+      return FlxqE_vert(R, Q, a4, T, p);
     }
   } else {
-    GEN slope, tmp1, tmp2;
+    GEN slope;
     *pt_R = FlxqE_add_slope(P, R, a4, T, p, &slope);
-    tmp1  = Flxq_mul(Flx_sub(gel(Q, 1), gel(R, 1), p), slope, T, p);
-    tmp2  = Flx_add(tmp1, gel(R, 2), p);
-    return Flx_sub(gel(Q, 2), tmp2, p);
+    return FlxqE_Miller_line(R, Q, slope, a4, T, p);
   }
 }
 
@@ -367,7 +440,7 @@ FlxqE_Miller_dbl(void* E, GEN d)
   GEN point = gel(d,3);
   line = FlxqE_tangent_update(point, P, a4, T, p, &point);
   num  = Flxq_mul(num, line, T, p);
-  v = FlxqE_vert(point, P, T, p);
+  v = FlxqE_vert(point, P, a4, T, p);
   denom = Flxq_mul(denom, v, T, p);
   return mkvec3(num, denom, point);
 }
@@ -385,7 +458,7 @@ FlxqE_Miller_add(void* E, GEN va, GEN vb)
   GEN denom = Flxq_mul(da, db, T, p);
   line = FlxqE_chord_update(pa, pb, P, a4, T, p, &point);
   num  = Flxq_mul(num, line, T, p);
-  v = FlxqE_vert(point, P, T, p);
+  v = FlxqE_vert(point, P, a4, T, p);
   denom = Flxq_mul(denom, v, T, p);
   return mkvec3(num, denom, point);
 }
@@ -401,7 +474,6 @@ FlxqE_Miller(GEN Q, GEN P, GEN m, GEN a4, GEN T, ulong p)
   g1 = pol1_Flx(get_Flx_var(T));
   v = gen_pow(mkvec3(g1,g1,Q), m, (void*)&d, FlxqE_Miller_dbl, FlxqE_Miller_add);
   num = gel(v,1); denom = gel(v,2);
-  if (!lgpol(num) || !lgpol(denom)) { avma = ltop; return NULL; }
   return gerepileupto(ltop, Flxq_div(num, denom, T, p));
 }
 
@@ -413,9 +485,7 @@ FlxqE_weilpairing(GEN P, GEN Q, GEN m, GEN a4, GEN T, ulong p)
   if (ell_is_inf(P) || ell_is_inf(Q) || Flx_equal(P,Q))
     return pol1_Flx(get_Flx_var(T));
   num    = FlxqE_Miller(P, Q, m, a4, T, p);
-  if (!num) return pol1_Flx(get_Flx_var(T));
   denom  = FlxqE_Miller(Q, P, m, a4, T, p);
-  if (!denom) {avma = ltop; return pol1_Flx(get_Flx_var(T)); }
   result = Flxq_div(num, denom, T, p);
   if (mpodd(m))
     result  = Flx_neg(result, p);
@@ -425,11 +495,9 @@ FlxqE_weilpairing(GEN P, GEN Q, GEN m, GEN a4, GEN T, ulong p)
 GEN
 FlxqE_tatepairing(GEN P, GEN Q, GEN m, GEN a4, GEN T, ulong p)
 {
-  GEN num;
   if (ell_is_inf(P) || ell_is_inf(Q))
     return pol1_Flx(get_Flx_var(T));
-  num = FlxqE_Miller(P, Q, m, a4, T, p);
-  return num? num: pol1_Flx(get_Flx_var(T));
+  return FlxqE_Miller(P, Q, m, a4, T, p);
 }
 
 static GEN
@@ -572,13 +640,13 @@ _can5_iter(void *E, GEN f, GEN q)
   pari_sp av = avma;
   struct _can_mul D;
   ulong p = *(ulong*)E;
-  long i;
+  long i, vT = fetch_var();
   GEN N, P, d, V, fs;
-  D.q = q; D.T = ZX_Z_sub(monomial(gen_1,p,MAXVARN),gen_1);
+  D.q = q; D.T = ZX_Z_sub(pol_xn(p,vT),gen_1);
   D.p = p;
-  fs = mkvec2(_shift(f, 1, p, MAXVARN), gen_1);
+  fs = mkvec2(_shift(f, 1, p, vT), gen_1);
   N = gel(gen_powu(fs,p-1,(void*)&D,_can5_sqr,_can5_mul),1);
-  N = simplify_shallow(FpXQX_red(N,polcyclo(p,MAXVARN),q));
+  N = simplify_shallow(FpXQX_red(N,polcyclo(p,vT),q));
   P = FpX_mul(N,f,q);
   P = RgX_deflate(P, p);
   d = RgX_splitting(N, p);
@@ -586,7 +654,7 @@ _can5_iter(void *E, GEN f, GEN q)
   gel(V,1) = ZX_mulu(gel(d,1), p);
   for(i=2; i<= (long)p; i++)
     gel(V,i) = ZX_mulu(RgX_shift_shallow(gel(d,p+2-i), 1), p);
-  return gerepilecopy(av, mkvec2(ZX_sub(f,P),V));
+  (void)delete_var(); return gerepilecopy(av, mkvec2(ZX_sub(f,P),V));
 }
 
 static GEN
@@ -707,10 +775,11 @@ _lift_iter(void *E, GEN x2, GEN q)
 {
   struct _lift_iso *d = (struct _lift_iso *) E;
   ulong p = d->p;
+  long n = lg(d->phi)-2;
   GEN TN = FpXT_red(d->T, q), XN = FpXV_red(d->Xm, q);
   GEN y2 = ZpXQ_frob(x2, XN, TN, q, p);
-  GEN xp = FpXQ_powers(x2, p, TN, q);
-  GEN yp = FpXQ_powers(y2, p, TN, q);
+  GEN xp = FpXQ_powers(x2, n, TN, q);
+  GEN yp = FpXQ_powers(y2, n, TN, q);
   GEN V  = FpM_FpXQV_bilinear(d->phi,xp,yp,TN,q);
   return mkvec3(V,xp,yp);
 }
@@ -750,7 +819,7 @@ getc2(GEN act, GEN X, GEN T, GEN q, ulong p, long N)
   GEN xp = FpXQ_powers(X,n,T,q);
   GEN P  = FpX_FpXQV_eval(A1, xp, T, q);
   GEN Q  = FpX_FpXQV_eval(A2, xp, T, q);
-  return FpXQ_mul(P,ZpXQ_inv(Q,T,utoi(p),N),T,q);
+  return ZpXQ_div(P, Q, T, q, utoi(p), N);
 }
 
 struct _ZpXQ_norm
@@ -882,7 +951,7 @@ get_norm(GEN a4, GEN a6, GEN T, ulong p, long N)
   else
   {
     GEN P = mkpoln(4, pol1_Flx(sv), pol0_Flx(sv), a4, a6);
-    a = gel(FlxqX_pow(P,p>>1,T,p),2+p-1);
+    a = gel(FlxqX_powu(P,p>>1,T,p),2+p-1);
   }
   return Zp_sqrtnlift(gen_1,subss(p,1),utoi(Flxq_norm(a,T,p)),utoi(p), N);
 }
@@ -966,7 +1035,7 @@ zx_is_pcyc(GEN T)
 }
 
 static GEN
-Flxq_ellcard_Harley(GEN a4, GEN a6, GEN T, ulong p)
+Flxq_ellcard_Kohel(GEN a4, GEN a6, GEN T, ulong p)
 {
   pari_sp av = avma, av2;
   pari_timer ti;
@@ -989,7 +1058,7 @@ Flxq_ellcard_Harley(GEN a4, GEN a6, GEN T, ulong p)
   if (DEBUGLEVEL) timer_printf(&ti,"Barrett");
   if (!ispcyc)
   {
-    Xm = FpXQ_powers(monomial(gen_1,n,get_FpX_var(T2)),p-1,T2,q);
+    Xm = FpXQ_powers(pol_xn(n,get_FpX_var(T2)),p-1,T2,q);
     if (DEBUGLEVEL) timer_printf(&ti,"Xm");
   } else
     Xm = cgetg(1,t_VEC);
@@ -1023,6 +1092,155 @@ Flxq_ellcard_Harley(GEN a4, GEN a6, GEN T, ulong p)
   return gerepileupto(av, subii(addis(powuu(p,n),1),t));
 }
 
+static void
+liftcurve(GEN J, GEN T, GEN q, ulong p, long N, GEN *A4, GEN *A6)
+{
+  pari_sp av = avma;
+  GEN r = ZpXQ_inv(Z_ZX_sub(utoi(1728),J),T,utoi(p),N);
+  GEN g = FpXQ_mul(J,r,T,q);
+  *A4 = FpX_mulu(g,3,q);
+  *A6 = FpX_mulu(g,2,q);
+  gerepileall(av,2,A4,A6);
+}
+
+static GEN
+getc5(GEN H, GEN A40, GEN A60, GEN A41, GEN A61, GEN T, GEN q, ulong p, long N)
+{
+  long d = lg(H)-1;
+  GEN s1 = gel(H,d-1), s2 = gel(H,d-2), s3 = d<5 ? pol_0(varn(T)): gel(H,d-3);
+  GEN s12 = FpXQ_sqr(s1,T,q);
+  GEN h2 = ZX_sub(ZX_shifti(s2,1),s12); /*2*s2-s1^2*/
+  GEN h3 = ZX_sub(FpXQ_mul(ZX_add(h2,s2),s1,T,q),ZX_mulu(s3,3));
+                                        /*3*s2*s1-s1^3-3s3*/
+  GEN alpha= ZX_sub(ZX_mulu(h2,30), ZX_mulu(A40,5*p-6)); /* 30*h2+A40*(6-5*p)*/
+  GEN beta = ZX_sub(ZX_sub(ZX_mulu(FpXQ_mul(A40,s1,T,q),42),ZX_mulu(A60,14*p-15)),
+                    ZX_mulu(h3,70)); /* 42*A40*s1-A60*(14*p-15)-70*h3 */
+  GEN u2 = FpXQ_mul(FpXQ_mul(A41,beta,T,q),
+                    ZpXQ_inv(FpXQ_mul(A61,alpha,T,q),T,utoi(p),N),T,q);
+  return u2;
+}
+
+static GEN
+ZpXQX_liftrootmod_vald(GEN f, GEN H, long v, GEN T, GEN p, long e)
+{
+  pari_sp av = avma, av2, lim;
+  GEN pv = p, q, qv, W, df, Tq, fr, dfr;
+  ulong mask;
+  pari_timer ti;
+  if (e <= v+1) return H;
+  df = RgX_deriv(f);
+  if (v) { pv = powiu(p,v); qv = mulii(pv,p); df = ZXX_Z_divexact(df, pv); }
+  else qv = p;
+  mask = quadratic_prec_mask(e-v);
+  Tq = FpXT_red(T, qv); dfr = FpXQX_red(df, Tq, p);
+  if (DEBUGLEVEL) timer_start(&ti);
+  W = FpXQXQ_inv(FpXQX_rem(dfr, H, Tq, p), H, Tq, p); /* 1/f'(a) mod (T,p) */
+  if (DEBUGLEVEL) timer_printf(&ti,"FpXQXQ_inv");
+  q = p;
+  av2 = avma; lim = stack_lim(av2, 2);
+  for (;;)
+  {
+    GEN u, fa, qv, q2v, Tq2, fadH;
+    GEN H2 = H, q2 = q;
+    q = sqri(q);
+    if (mask & 1) q = diviiexact(q,p);
+    mask >>= 1;
+    if (v) { qv = mulii(q, pv); q2v = mulii(q2, pv); }
+    else { qv = q; q2v = q2; }
+    Tq2 = FpXT_red(T, q2v); Tq = FpXT_red(T, qv);
+    fr = FpXQX_red(f, Tq, qv);
+    fa = FpXQX_rem(fr, H, Tq, qv);
+    fa = ZXX_Z_divexact(fa, q2v);
+    fadH = FpXQXQ_mul(RgX_deriv(H),fa,H,Tq2,q2);
+    H = FpXX_add(H, gmul(FpXQXQ_mul(W, fadH, H, Tq2, q2v), q2), qv);
+    if (mask == 1) return gerepileupto(av, H);
+    dfr = FpXQX_rem(FpXQX_red(df, Tq, q),H,Tq,q);
+    u = ZXX_Z_divexact(ZXX_Z_add_shallow(FpXQXQ_mul(W,dfr,H,Tq,q),gen_m1),q2);
+    W = gsub(W,gmul(FpXQXQ_mul(u,W,H2,Tq2,q2),q2));
+    if (low_stack(lim, stack_lim(av2,2)))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"ZpXQX_liftroot, e = %ld", e);
+      gerepileall(av2, 3, &H, &W, &q);
+    }
+  }
+}
+
+static GEN
+get_H1(GEN A41, GEN A61, GEN T2, ulong p)
+{
+  GEN q = utoi(p), T = FpXT_red(T2,q);
+  GEN pol = FpXQ_elldivpol(FpX_red(A41,q),FpX_red(A61,q),p,T,q);
+  return FpXQX_normalize(RgX_deflate(pol,p),T,q);
+}
+
+static GEN
+Flxq_ellcard_Harley(GEN a4, GEN a6, GEN T, ulong p)
+{
+  pari_sp av = avma, av2;
+  pari_timer ti;
+  long n = get_Flx_degree(T), N = (n+5)/2;
+  GEN q = powuu(p, N);
+  GEN T2, j, t;
+  GEN J1,A40,A41,A60,A61, sqx,Xm;
+  GEN pol, h1, H;
+  GEN c2, tc2, c2p, Nc2, Np;
+  long ispcyc = zx_is_pcyc(get_Flx_mod(T));
+  timer_start(&ti);
+  if (!ispcyc)
+  {
+    T2 = Flx_canonlift(get_Flx_mod(T),N,p);
+    if (DEBUGLEVEL) timer_printf(&ti,"Teich");
+  } else
+    T2 = Flx_to_ZX(get_Flx_mod(T));
+  T2 = FpX_get_red(T2, q); T = ZXT_to_FlxT(T2, p);
+  av2 = avma;
+  if (DEBUGLEVEL) timer_printf(&ti,"Barrett");
+  if (!ispcyc)
+  {
+    Xm = FpXQ_powers(pol_xn(n,get_FpX_var(T2)),p-1,T2,q);
+    if (DEBUGLEVEL) timer_printf(&ti,"Xm");
+  } else
+    Xm = cgetg(1,t_VEC);
+  if (DEBUGLEVEL) timer_printf(&ti,"Xm");
+  j = Flxq_ellj(a4,a6,T,p);
+  sqx = Flxq_powers(Flxq_lroot(polx_Flx(T[1]), T, p), p-1, T, p);
+  J1 = lift_isogeny(polmodular_ZM(p, 0), Flx_to_ZX(j), N, Xm, T2,sqx,T,p);
+  if (DEBUGLEVEL) timer_printf(&ti,"Lift isogeny");
+  liftcurve(J1,T2,q,p,N,&A41,&A61);
+  A40 = ZpXQ_frob(A41, Xm, T2, q, p);
+  A60 = ZpXQ_frob(A61, Xm, T2, q, p);
+  if (DEBUGLEVEL) timer_printf(&ti,"liftcurve");
+  pol = FpXQ_elldivpol(A40,A60,p,T2,q);
+  if (DEBUGLEVEL) timer_printf(&ti,"p-division");
+  h1 = get_H1(A41,A61,T2,p);
+  H = ZpXQX_liftrootmod_vald(pol,h1,1,T2,utoi(p),N);
+  q = diviuexact(q,p); N--;
+  if (DEBUGLEVEL) timer_printf(&ti,"kernel");
+  c2 = getc5(H,A40,A60,A41,A61,T2,q,p,N);
+  if (DEBUGLEVEL) timer_printf(&ti,"c^2");
+  if (!ispcyc)
+  {
+    c2p = Flx_to_ZX(Flxq_inv(ZX_to_Flx(c2,p),T,p));
+    tc2 = Teichmuller_lift(c2p,Xm, T2,sqx,T,p,N);
+    if (DEBUGLEVEL) timer_printf(&ti,"teichmuller");
+    c2 = FpX_rem(FpX_mul(tc2,c2,q),T2,q);
+  }
+  c2 = gerepileupto(av2, c2);
+  q = powuu(p, N);
+  Nc2 = (ispcyc? ZpXQ_sqrtnorm_pcyc: ZpXQ_sqrtnorm)(c2, T2, q, utoi(p), N);
+  if (DEBUGLEVEL) timer_printf(&ti,"Norm");
+  Np = get_norm(a4,a6,T,p,N);
+  if (ispcyc)
+  {
+    GEN Ncpi = utoi(Fl_inv(umodiu(Nc2,p), p));
+    GEN tNc2 = Zp_sqrtnlift(gen_1, subss(p,1), Ncpi, utoi(p), N);
+    if (DEBUGLEVEL) timer_printf(&ti,"Teichmuller/Fp");
+    Nc2 = Fp_mul(Nc2,tNc2,q);
+  }
+  t = Fp_center(Fp_mul(Nc2,Np,q),q,shifti(q,-1));
+  return gerepileupto(av, subii(addis(powuu(p,n),1),t));
+}
+
 /***************************************************************************/
 /*                                                                         */
 /*                          Shanks Mestre                                  */
@@ -1039,14 +1257,14 @@ closest_lift(GEN a, GEN b, GEN h)
 static GEN
 FlxqE_find_order(GEN f, GEN h, GEN bound, GEN B, GEN a4, GEN T, ulong p)
 {
-  pari_sp av = avma, av1, lim;
+  pari_sp av = avma, av1;
   pari_timer Ti;
   long s = itos( gceil(gsqrt(gdiv(bound,B),DEFAULTPREC)) ) >> 1;
   GEN tx, ti;
   GEN fh = FlxqE_mul(f, h, a4, T, p);
   GEN F, P = fh, fg;
   long i;
-  if (DEBUGLEVEL) timer_start(&Ti);
+  if (DEBUGLEVEL >= 6) timer_start(&Ti);
   if (ell_is_inf(fh)) return h;
   F = FlxqE_mul(f, B, a4, T, p);
   if (s < 3)
@@ -1062,26 +1280,26 @@ FlxqE_find_order(GEN f, GEN h, GEN bound, GEN B, GEN a4, GEN T, ulong p)
   }
   tx = cgetg(s+1,t_VECSMALL);
   /* Baby Step/Giant Step */
-  av1 = avma; lim = stack_lim(av1,3);
+  av1 = avma;
   for (i=1; i<=s; i++)
   { /* baby steps */
     tx[i] = hash_GEN(gel(P, 1));
     P = FlxqE_add(P, F, a4, T, p); /* h.f + i.F */
     if (ell_is_inf(P)) return gerepileupto(av, addii(h, mului(i,B)));
-    if (low_stack(lim, stack_lim(av1,3)))
+    if (gc_needed(av1,3))
     {
-      if(DEBUGMEM>1) pari_warn(warnmem,"[ellap3] baby steps, i=%ld",i);
+      if(DEBUGMEM>1) pari_warn(warnmem,"[Flxq_ellcard] baby steps, i=%ld",i);
       P = gerepileupto(av1,P);
     }
   }
-  if (DEBUGLEVEL) timer_printf(&Ti, "[ellap3] baby steps, s = %ld",s);
+  if (DEBUGLEVEL >= 6) timer_printf(&Ti, "[Flxq_ellcard] baby steps, s = %ld",s);
   /* giant steps: fg = s.F */
   fg = gerepileupto(av1, FlxqE_sub(P, fh, a4, T, p));
   if (ell_is_inf(fg)) return gerepileupto(av,mului(s,B));
   ti = vecsmall_indexsort(tx); /* = permutation sorting tx */
   tx = perm_mul(tx,ti);
-  if (DEBUGLEVEL) timer_printf(&Ti, "[ellap3] sorting");
-  av1 = avma; lim = stack_lim(av1,3);
+  if (DEBUGLEVEL >= 6) timer_printf(&Ti, "[Flxq_ellcard] sorting");
+  av1 = avma;
   for (P=fg, i=1; ; i++)
   {
     long k = hash_GEN(gel(P,1));
@@ -1093,7 +1311,8 @@ FlxqE_find_order(GEN f, GEN h, GEN bound, GEN B, GEN a4, GEN T, ulong p)
       {
         long j = ti[r]-1;
         GEN Q = FlxqE_add(FlxqE_mul(F, stoi(j), a4, T, p), fh, a4, T, p);
-        if (DEBUGLEVEL) timer_printf(&Ti, "[ellap3] giant steps, i = %ld",i);
+        if (DEBUGLEVEL >= 6)
+          timer_printf(&Ti, "[Flxq_ellcard] giant steps, i = %ld",i);
         if (Flx_equal(gel(P,1), gel(Q,1)))
         {
           if (Flx_equal(gel(P,2), gel(Q,2))) i = -i;
@@ -1102,9 +1321,9 @@ FlxqE_find_order(GEN f, GEN h, GEN bound, GEN B, GEN a4, GEN T, ulong p)
       }
     }
     P = FlxqE_add(P,fg,a4,T,p);
-    if (low_stack(lim, stack_lim(av1,3)))
+    if (gc_needed(av1,3))
     {
-      if(DEBUGMEM>1) pari_warn(warnmem,"[ellap3] giants steps, i=%ld",i);
+      if(DEBUGMEM>1) pari_warn(warnmem,"[Flxq_ellcard] giants steps, i=%ld",i);
       P = gerepileupto(av1,P);
     }
   }
@@ -1115,7 +1334,7 @@ Flx_next(GEN t, ulong p)
 {
   long i;
   for(i=2;;i++)
-    if ((ulong)t[i]==p-1)
+    if (uel(t,i)==p-1)
       t[i]=0;
     else
     {
@@ -1172,14 +1391,38 @@ Flxq_ellcard_naive(GEN a4, GEN a6, GEN T, ulong p)
   return a;
 }
 
+/* assume T irreducible mod p, m = (q-1)/(p-1) */
+static int
+Flxq_kronecker(GEN x, GEN m, GEN T, ulong p)
+{
+  pari_sp av;
+  ulong z;
+  if (lgpol(x) == 0) return 0;
+  av = avma;
+  z = Flxq_pow(x, m, T, p)[2];
+  avma = av; return krouu(z, p);
+}
+
+/* Find x such that kronecker(u = x^3+a4x+a6, p) is KRO.
+ * Return point [x*u,u^2] on E (KRO=1) / E^twist (KRO=-1) */
+static GEN
+Flxq_ellpoint(long KRO, GEN a4, GEN a6, GEN m, long n, long vn, GEN T, ulong p)
+{
+  for(;;)
+  {
+    GEN x = random_Flx(n,vn,p);
+    GEN u = Flx_add(a6, Flxq_mul(Flx_add(a4, Flxq_sqr(x,T,p), p), x, T,p), p);
+    if (Flxq_kronecker(u, m,T,p) == KRO)
+      return mkvec2(Flxq_mul(u,x, T,p), Flxq_sqr(u, T,p));
+  }
+}
+
 static GEN
 Flxq_ellcard_Shanks(GEN a4, GEN a6, GEN q, GEN T, ulong p)
 {
   pari_sp av = avma;
-  long vn = get_Flx_var(T);
-  long i, twistp, twistpold=0;
-  GEN h,f, ta4, u, x, A, B;
-  long n = get_Flx_degree(T);
+  long vn = get_Flx_var(T), n = get_Flx_degree(T), KRO = -1;
+  GEN h,f, ta4, A, B, m;
   GEN q1p = addsi(1, q), q2p = shifti(q1p, 1);
   GEN bound = addis(sqrti(gmul2n(q,4)), 1); /* ceil( 4sqrt(q) ) */
   /* once #E(Flxq) is know mod B >= bound, it is completely determined */
@@ -1190,42 +1433,30 @@ Flxq_ellcard_Shanks(GEN a4, GEN a6, GEN q, GEN T, ulong p)
   case 1:  A = gen_0; B = gen_2; break;
   default: A = gen_1; B = gen_2; break; /* 0 */
   }
-  h = closest_lift(A, B, q1p);
+  m = diviuexact(subiu(powuu(p,n), 1), p-1);
   for(;;)
   {
-    do
-    { /* look for points alternatively on E and its quadratic twist E' */
-      x = random_Flx(n,vn,p);
-      u = Flx_add(a6, Flxq_mul(Flx_add(a4, Flxq_sqr(x, T, p), p), x , T, p), p);
-      twistp = Flxq_issquare(u, T, p);
-    } while (twistp == twistpold);
-    twistpold = twistp;
+    h = closest_lift(A, B, q1p);
     /* [ux, u^2] is on E_u: y^2 = x^3 + c4 u^2 x + c6 u^3
-     * E_u isomorphic to E (resp. E') iff twistp = 1 (resp. -1)
+     * E_u isomorphic to E (resp. E') iff KRO = 1 (resp. -1)
      * #E(F_p) = p+1 - a_p, #E'(F_p) = p+1 + a_p
      *
      * #E_u(Flxq) = A (mod B),  h is close to #E_u(Flxq) */
+    KRO = -KRO;
+    f = Flxq_ellpoint(KRO, a4,a6, m,n,vn, T,p);
 
-    f = mkvec2(Flxq_mul(u, x, T, p), Flxq_sqr(u, T, p));
     ta4 = Flxq_mul(a4, gel(f,2), T, p); /* a4 for E_u */
     h = FlxqE_find_order(f, h, bound, B, ta4,T,p);
     h = FlxqE_order(f, h, ta4, T, p);
     /* h | #E_u(Flxq) = A (mod B) */
-    if (B == gen_1)
-      B = h;
-    else
-      A = Z_chinese_all(A, gen_0, B, h, &B);
-
-    i = (cmpii(B, bound) < 0);
-    /* If we are not done, update A mod B for the _next_ curve, isomorphic to
+    A = Z_chinese_all(A, gen_0, B, h, &B);
+    if (cmpii(B, bound) >= 0) break;
+    /* not done, update A mod B for the _next_ curve, isomorphic to
      * the quadratic twist of this one */
-    if (i) A = remii(subii(q2p,A), B); /* #E(Fq)+#E'(Fq) = 2q+2 */
-
-    /* h = A mod B, closest lift to p+1 */
-    h = closest_lift(A, B, q1p);
-    if (!i) break;
+    A = remii(subii(q2p,A), B); /* #E(Fq)+#E'(Fq) = 2q+2 */
   }
-  return gerepileuptoint(av, twistp? h: subii(shifti(q1p,1),h));
+  h = closest_lift(A, B, q1p);
+  return gerepileuptoint(av, KRO == 1? h: subii(q2p,h));
 }
 
 static GEN
@@ -1249,40 +1480,59 @@ F3xq_ellcard(GEN a2, GEN a6, GEN T)
       if (umodiu(t, 3)!=1) t = negi(t);
       return Flx_equal1(a2) || Flxq_issquare(a2,T,3) ? subii(q1,t): addii(q1,t);
     }
-    else return Flxq_ellcard_Harley(mkvec(a2), a6, T, 3);
+    else return Flxq_ellcard_Kohel(mkvec(a2), a6, T, 3);
   }
 }
 
 static GEN
-Flxq_ellcard_Satoh(GEN a4, GEN a6, GEN T, ulong p)
+Flxq_ellcard_Satoh(GEN a4, GEN a6, GEN j, GEN T, ulong p)
 {
   long n = get_Flx_degree(T);
   if (n <= 2)
     return utoi(Flxq_ellcard_naive(a4, a6, T, p));
   else
   {
-    GEN q1 = addis(powuu(p, get_Flx_degree(T)), 1), t;
-    GEN j = Flxq_ellj(a4,a6,T,p);
-    if (Flx_equal1(Flxq_powu(j, p*p-1, T, p)))
-    {
-      GEN sk, sA4, u;
-      GEN P = Flxq_minpoly(j, T, p);
-      long dP = degpol(P); /* dP <= 2 */
-      GEN q = powuu(p,dP);
-      GEN k = mkvecsmall3(T[1], 1728%p, p-1);
-      GEN kj = Flx_shift(k, 1), k2j = Flxq_mul(kj, k, P, p);
-      GEN A4 = Flx_triple(kj,p), A6 = Flx_double(k2j,p);
-      GEN tP = addis(q, 1 - Flxq_ellcard_naive(A4, A6, P, p));
-      t = elltrace_extension(tP, n/dP, q);
-      sk = Flx_Fl_add(Flx_neg(j,p),1728%p,p);
-      sA4 = Flx_triple(Flxq_mul(sk,j,T,p),p);
-      u = Flxq_div(a4,sA4, T, p);
-      return Flxq_is2npower(u, 2, T, p) ? subii(q1,t): addii(q1,t);
+    GEN jp = Flxq_powu(j, p, T, p);
+    GEN s = Flx_add(j, jp, p);
+    if (degpol(s) <= 0)
+    { /* it is assumed j not in F_p */
+      GEN m = Flxq_mul(j, jp, T, p);
+      if (degpol(m) <= 0)
+      {
+        GEN q = sqru(p);
+        GEN q1 = addis(powuu(p, get_Flx_degree(T)), 1);
+        GEN sk = Flx_Fl_add(Flx_neg(j, p), 1728%p, p);
+        GEN sA4 = Flx_triple(Flxq_mul(sk, j, T, p), p);
+        GEN u = Flxq_div(a4, sA4, T, p);
+        ulong ns = lgpol(s) ? Fl_neg(s[2], p): 0UL;
+        GEN P = mkvecsmall4(T[1], m[2], ns, 1L);
+        GEN A4, A6, t, tP;
+        Flxq_ellj_to_a4a6(polx_Flx(T[1]), P, p, &A4, &A6);
+        tP = addis(q, 1 - Flxq_ellcard_naive(A4, A6, P, p));
+        t = elltrace_extension(tP, n>>1, q);
+        return Flxq_is2npower(u, 2, T, p) ? subii(q1,t): addii(q1,t);
+      }
     }
+    if (p<=7 || p==13 ) return Flxq_ellcard_Kohel(a4, a6, T, p);
     else return Flxq_ellcard_Harley(a4, a6, T, p);
   }
 }
 
+static GEN
+Flxq_ellcard_Kedlaya(GEN a4, GEN a6, GEN T, ulong p)
+{
+  pari_sp av = avma;
+  GEN H = mkpoln(4, gen_1, gen_0, Flx_to_ZX(a4), Flx_to_ZX(a6));
+  GEN Tp = Flx_to_ZX(get_Flx_mod(T));
+  long n = degpol(Tp), e = ((p < 16 ? n+1: n)>>1)+1;
+  GEN M = ZlXQX_hyperellpadicfrobenius(H, Tp, p, e);
+  GEN N = ZpXQM_prodFrobenius(M, Tp, utoi(p), e);
+  GEN q = powuu(p, e);
+  GEN tp = Fq_add(gcoeff(N,1,1), gcoeff(N,2,2), Tp, q);
+  GEN t = Fp_center(typ(tp)==t_INT ? tp: leading_coeff(tp), q, shifti(q,-1));
+  return gerepileupto(av, subii(addis(powuu(p, n), 1), t));
+}
+
 GEN
 Flxq_ellj(GEN a4, GEN a6, GEN T, ulong p)
 {
@@ -1305,6 +1555,24 @@ Flxq_ellj(GEN a4, GEN a6, GEN T, ulong p)
   }
 }
 
+void
+Flxq_ellj_to_a4a6(GEN j, GEN T, ulong p, GEN *pt_a4, GEN *pt_a6)
+{
+  ulong zagier = 1728 % p;
+  if (lgpol(j)==0)
+    { *pt_a4 = pol0_Flx(T[1]); *pt_a6 =pol1_Flx(T[1]); }
+  else if (lgpol(j)==1 && uel(j,2) == zagier)
+    { *pt_a4 = pol1_Flx(T[1]); *pt_a6 =pol0_Flx(T[1]); }
+  else
+  {
+    GEN k = Flx_Fl_add(Flx_neg(j, p), zagier, p);
+    GEN kj = Flxq_mul(k, j, T, p);
+    GEN k2j = Flxq_mul(kj, k, T, p);
+    *pt_a4 = Flx_triple(kj, p);
+    *pt_a6 = Flx_double(k2j, p);
+  }
+}
+
 static GEN
 F3xq_ellcardj(GEN a4, GEN a6, GEN T, GEN q, long n)
 {
@@ -1415,16 +1683,17 @@ Flxq_ellcard(GEN a4, GEN a6, GEN T, ulong p)
     r = Fp_ffellcard(utoi(Flx_eval(a4,0,p)),utoi(Flx_eval(a6,0,p)),q,n,utoi(p));
   else if (degpol(J=Flxq_ellj(a4,a6,T,p))<=0)
     r = Flxq_ellcardj(a4,a6,lgpol(J)?J[2]:0,T,q,p,n);
-  else if (p <= 7 || p==13)
-    r = Flxq_ellcard_Satoh(a4,a6,T,p);
+  else if (p <= 7)
+    r = Flxq_ellcard_Satoh(a4, a6, J, T, p);
   else if (cmpis(q,100)<0)
     r = utoi(Flxq_ellcard_naive(a4, a6, T, p));
+  else if (p == 13 || (7*p <= (ulong)10*n && (BITS_IN_LONG==64 || p <= 103)))
+    r = Flxq_ellcard_Satoh(a4, a6, J, T, p);
+  else if (p <= (ulong)2*n)
+    r = Flxq_ellcard_Kedlaya(a4, a6, T, p);
   else if (expi(q)<=62)
     r = Flxq_ellcard_Shanks(a4, a6, q, T, p);
   else
-  {
     r = Fq_ellcard_SEA(Flx_to_ZX(a4),Flx_to_ZX(a6),q,Flx_to_ZX(T),utoi(p),0);
-    if (!r) pari_err_PACKAGE("seadata");
-  }
   return gerepileuptoint(av, r);
 }
