@@ -47,6 +47,19 @@ FpC_center(GEN z, GEN p, GEN pov2)
   return x;
 }
 
+/* assume 0 <= u < p and ps2 = p>>1 */
+INLINE void
+Fp_center_inplace(GEN u, GEN p, GEN ps2)
+{ if (abscmpii(u,ps2) > 0) subiiz(u,p,u); }
+
+void
+FpC_center_inplace(GEN z, GEN p, GEN ps2)
+{
+  long i,l = lg(z);
+  for (i=1; i<l; i++)
+    Fp_center_inplace(gel(z,i), p, ps2);
+}
+
 GEN
 Flv_center(GEN z, ulong p, ulong ps2)
 {
@@ -74,6 +87,12 @@ FpM_center(GEN z, GEN p, GEN pov2)
   return x;
 }
 
+void
+FpM_center_inplace(GEN z, GEN p, GEN pov2)
+{
+  long i, l = lg(z);
+  for (i=1; i<l; i++) FpC_center_inplace(gel(z,i), p, pov2);
+}
 GEN
 Flm_center(GEN z, ulong p, ulong ps2)
 {
@@ -104,6 +123,16 @@ FpV_add(GEN x, GEN y, GEN p)
   for (i = 1; i < lx; i++) gel(z, i) = Fp_add(gel(x, i), gel(y, i), p);
   return z;
 }
+GEN
+FpM_add(GEN x, GEN y, GEN p)
+{
+  long lx = lg(x), j;
+  GEN z;
+  if (lx == 1) return cgetg(1, t_MAT);
+  z = cgetg(lx, t_MAT);
+  for (j = 1; j < lx; j++) gel(z,j) = FpC_add(gel(x,j), gel(y,j), p);
+  return z;
+}
 
 GEN
 Flv_add(GEN x, GEN y, ulong p)
@@ -156,6 +185,15 @@ FpV_sub(GEN x, GEN y, GEN p)
   return z;
 }
 
+GEN
+FpM_sub(GEN x, GEN y, GEN p)
+{
+  long i, l = lg(x);
+  GEN z = cgetg(l, t_MAT);
+  for (i = 1; i < l; i++) gel(z, i) = FpC_sub(gel(x, i), gel(y, i), p);
+  return z;
+}
+
 GEN
 Flv_sub(GEN x, GEN y, ulong p)
 {
@@ -190,6 +228,24 @@ Flm_Fl_add(GEN x, ulong y, ulong p)
   return z;
 }
 
+GEN
+Flm_add(GEN x, GEN y, ulong p)
+{
+  long i, l = lg(x);
+  GEN z = cgetg(l,t_MAT);
+  for (i = 1; i < l; i++) gel(z,i) = Flv_add(gel(x,i),gel(y,i),p);
+  return z;
+}
+
+GEN
+Flm_sub(GEN x, GEN y, ulong p)
+{
+  long i, l = lg(x);
+  GEN z = cgetg(l, t_MAT);
+  for (i = 1; i < l; i++) gel(z, i) = Flv_sub(gel(x, i), gel(y, i), p);
+  return z;
+}
+
 /********************************************************************/
 /**                                                                **/
 /**                           MULTIPLICATION                       **/
@@ -204,7 +260,7 @@ FpC_Fp_mul(GEN x, GEN y, GEN p)
   return z;
 }
 GEN
-Flc_Fl_mul(GEN x, ulong y, ulong p)
+Flv_Fl_mul(GEN x, ulong y, ulong p)
 {
   long i, l = lg(x);
   GEN z = cgetg(l, t_VECSMALL);
@@ -212,26 +268,40 @@ Flc_Fl_mul(GEN x, ulong y, ulong p)
   return z;
 }
 GEN
-Flc_Fl_div(GEN x, ulong y, ulong p)
+Flv_Fl_div(GEN x, ulong y, ulong p)
 {
-  return Flc_Fl_mul(x, Fl_inv(y, p), p);
+  return Flv_Fl_mul(x, Fl_inv(y, p), p);
 }
 void
-Flc_Fl_div_inplace(GEN x, ulong y, ulong p)
+Flv_Fl_div_inplace(GEN x, ulong y, ulong p)
 {
-  Flc_Fl_mul_inplace(x, Fl_inv(y, p), p);
+  Flv_Fl_mul_inplace(x, Fl_inv(y, p), p);
+}
+GEN
+FpM_Fp_mul(GEN X, GEN c, GEN p) {
+  long i, j, h, l = lg(X);
+  GEN A = cgetg(l, t_MAT);
+  if (l == 1) return A;
+  h = lgcols(X);
+  for (j=1; j<l; j++)
+  {
+    GEN a = cgetg(h, t_COL), x = gel(X, j);
+    for (i = 1; i < h; i++) gel(a,i) = Fp_mul(gel(x,i), c, p);
+    gel(A,j) = a;
+  }
+  return A;
 }
 
 /* x *= y */
 void
-Flc_Fl_mul_part_inplace(GEN x, ulong y, ulong p, long l)
+Flv_Fl_mul_part_inplace(GEN x, ulong y, ulong p, long l)
 {
   long i;
   for (i=1;i<=l;i++) x[i] = Fl_mul(x[i], y, p);
 }
 void
-Flc_Fl_mul_inplace(GEN x, ulong y, ulong p)
-{ Flc_Fl_mul_part_inplace(x, y, p, lg(x)-1); }
+Flv_Fl_mul_inplace(GEN x, ulong y, ulong p)
+{ Flv_Fl_mul_part_inplace(x, y, p, lg(x)-1); }
 
 /* set y *= x */
 void
@@ -263,15 +333,31 @@ Flm_Fl_mul(GEN y, ulong x, ulong p)
     }
   return z;
 }
+
+GEN
+Flv_neg(GEN v, ulong p)
+{
+  long i, m = lg(v);
+  GEN c = cgetg(m, t_VECSMALL);
+  for(i=1; i<m; i++) uel(c,i) = Fl_neg(uel(v,i), p);
+  return c;
+}
+
+void
+Flv_neg_inplace(GEN v, ulong p)
+{
+  long i;
+  for (i = 1; i < lg(v); ++i)
+    v[i] = Fl_neg(v[i], p);
+}
+
 GEN
 Flm_neg(GEN y, ulong p)
 {
-  long i, j, m = lgcols(y), l = lg(y);
+  long j, l = lg(y);
   GEN z = cgetg(l, t_MAT);
-  for(j=1; j<l; j++) {
-    GEN c = cgetg(m, t_VECSMALL); gel(z,j) = c;
-    for(i=1; i<m; i++) c[i] = Fl_neg(ucoeff(y,i,j), p);
-  }
+  for(j=1; j<l; j++)
+    gel(z,j) = Flv_neg(gel(y,j), p);
   return z;
 }
 
@@ -331,22 +417,29 @@ zm_mul(GEN x, GEN y)
 static ulong
 Flmrow_Flc_mul_SMALL(GEN x, GEN y, ulong p, long lx, long i)
 {
-  ulong c = ucoeff(x,i,1) * y[1];
+  ulong c = ucoeff(x,i,1) * uel(y,1);
   long k;
   for (k = 2; k < lx; k++) {
-    c += ucoeff(x,i,k) * y[k];
+    c += ucoeff(x,i,k) * uel(y,k);
     if (c & HIGHBIT) c %= p;
   }
   return c % p;
 }
+
 static ulong
-Flmrow_Flc_mul(GEN x, GEN y, ulong p, long lx, long i)
+Flmrow_Flc_mul_i(GEN x, GEN y, ulong p, ulong pi, long lx, long i)
 {
-  ulong c = Fl_mul(ucoeff(x,i,1), y[1], p);
-  long k;
-  for (k = 2; k < lx; k++)
-    c = Fl_add(c, Fl_mul(ucoeff(x,i,k), y[k], p), p);
-  return c;
+  ulong l0, l1, v1, h0, h1;
+  long k = 1;
+  LOCAL_OVERFLOW;
+  LOCAL_HIREMAINDER;
+  l1 = mulll(ucoeff(x,i,k), uel(y,k)); h1 = hiremainder; v1 = 0;
+  while (++k < lx) {
+    l0 = mulll(ucoeff(x,i,k), uel(y,k)); h0 = hiremainder;
+    l1 = addll(l0, l1); h1 = addllx(h0, h1); v1 += overflow;
+  }
+  if (v1 == 0) return remll_pre(h1, l1, p, pi);
+  else return remlll_pre(v1, h1, l1, p, pi);
 }
 
 static GEN
@@ -378,22 +471,35 @@ FpM_FpC_mul_i(GEN x, GEN y, long lx, long l, GEN p)
   }
   return z;
 }
+
+static void
+__Flm_Flc_mul_i_SMALL(GEN z, GEN x, GEN y, long lx, long l, ulong p)
+{
+  long i;
+  for (i = 1; i < l; i++) z[i] = Flmrow_Flc_mul_SMALL(x, y, p, lx, i);
+}
 static GEN
 Flm_Flc_mul_i_SMALL(GEN x, GEN y, long lx, long l, ulong p)
 {
   GEN z = cgetg(l,t_VECSMALL);
-  long i;
-  for (i = 1; i < l; i++) z[i] = Flmrow_Flc_mul_SMALL(x, y, p, lx, i);
+  __Flm_Flc_mul_i_SMALL(z, x, y, lx, l, p);
   return z;
 }
+
+static void
+__Flm_Flc_mul_i(GEN z, GEN x, GEN y, long lx, long l, ulong p, ulong pi)
+{
+  long i;
+  for (i = 1; i < l; i++) z[i] = Flmrow_Flc_mul_i(x, y, p, pi, lx, i);
+}
 static GEN
-Flm_Flc_mul_i(GEN x, GEN y, long lx, long l, ulong p)
+Flm_Flc_mul_i(GEN x, GEN y, long lx, long l, ulong p, ulong pi)
 {
   GEN z = cgetg(l,t_VECSMALL);
-  long i;
-  for (i=1; i<l; i++) z[i] = Flmrow_Flc_mul(x, y, p, lx, i);
+  __Flm_Flc_mul_i(z, x, y, lx, l, p, pi);
   return z;
 }
+
 INLINE GEN
 F2m_F2c_mul_i(GEN x, GEN y, long lx, long l)
 {
@@ -420,7 +526,7 @@ FpM_mul(GEN x, GEN y, GEN p)
   if (lgefint(p) == 3)
   {
     pari_sp av = avma;
-    ulong pp = (ulong)p[2];
+    ulong pp = uel(p,2);
     if (pp == 2)
     {
       x = ZM_to_F2m(x);
@@ -456,8 +562,9 @@ Flm_mul(GEN x, GEN y, ulong p)
     for (j=1; j<ly; j++)
       gel(z,j) = Flm_Flc_mul_i_SMALL(x, gel(y,j), lx, l, p);
   } else {
+    ulong pi = get_Fl_red(p);
     for (j=1; j<ly; j++)
-      gel(z,j) = Flm_Flc_mul_i(x, gel(y,j), lx, l, p);
+      gel(z,j) = Flm_Flc_mul_i(x, gel(y,j), lx, l, p, pi);
   }
   return z;
 }
@@ -518,7 +625,7 @@ FpM_powu(GEN x, ulong n, GEN p)
   if (lgefint(p) == 3)
   {
     pari_sp av = avma;
-    ulong pp = (ulong)p[2];
+    ulong pp = uel(p,2);
     GEN z;
     if (pp == 2)
       z = F2m_to_ZM(F2m_powu(ZM_to_F2m(x),n));
@@ -568,20 +675,71 @@ FpV_dotsquare(GEN x, GEN p)
   for (i=2; i<lx; i++) c = addii(c, sqri(gel(x,i)));
   return gerepileuptoint(av, modii(c,p));
 }
+
+INLINE ulong
+Flv_dotproductspec_SMALL(GEN x, GEN y, ulong p, long lx)
+{
+  ulong c = uel(x,0) * uel(y,0);
+  long k;
+  for (k = 1; k < lx; k++) {
+    c += uel(x,k) * uel(y,k);
+    if (c & HIGHBIT) c %= p;
+  }
+  return c % p;
+}
+
+INLINE ulong
+Flv_dotproductspec_i(GEN x, GEN y, ulong p, ulong pi, long lx)
+{
+  ulong l0, l1, v1, h0, h1;
+  long i = 0;
+  LOCAL_OVERFLOW;
+  LOCAL_HIREMAINDER;
+  l1 = mulll(uel(x,i), uel(y,i)); h1 = hiremainder; v1 = 0;
+  while (++i < lx) {
+    l0 = mulll(uel(x,i), uel(y,i)); h0 = hiremainder;
+    l1 = addll(l0, l1); h1 = addllx(h0, h1); v1 += overflow;
+  }
+  if (v1 == 0) return remll_pre(h1, l1, p, pi);
+  else return remlll_pre(v1, h1, l1, p, pi);
+}
+
 ulong
 Flv_dotproduct(GEN x, GEN y, ulong p)
 {
-  long i, lx = lg(x);
-  ulong c;
+  long lx = lg(x)-1;
   if (lx == 1) return 0;
-  c = Fl_mul(x[1], y[1], p);
-  for (i=2; i<lx; i++) c = Fl_add(c, Fl_mul(x[i], y[i], p), p);
-  return c;
+  if (SMALL_ULONG(p))
+    return Flv_dotproductspec_SMALL(x+1, y+1, p, lx);
+  else
+    return Flv_dotproductspec_i(x+1, y+1, p, get_Fl_red(p), lx);
 }
+
 ulong
-F2v_dotproduct(GEN x0, GEN y0)
+Flv_dotproduct_pre(GEN x, GEN y, ulong p, ulong pi)
+{
+  long lx = lg(x)-1;
+  if (lx == 1) return 0;
+  if (SMALL_ULONG(p))
+    return Flv_dotproductspec_SMALL(x+1, y+1, p, lx);
+  else
+    return Flv_dotproductspec_i(x+1, y+1, p, pi, lx);
+}
+
+ulong
+Flx_dotproduct(GEN x, GEN y, ulong p)
+{
+  long lx = minss(lgpol(x), lgpol(y));
+  if (lx == 0) return 0;
+  if (SMALL_ULONG(p))
+    return Flv_dotproductspec_SMALL(x+2, y+2, p, lx);
+  else
+    return Flv_dotproductspec_i(x+2, y+2, p, get_Fl_red(p), lx);
+}
+
+ulong
+F2v_dotproduct(GEN x, GEN y)
 {
-  GEN x = x0, y = y0;
   long i, lx = lg(x);
   ulong c;
   if (lx == 2) return 0;
@@ -615,8 +773,39 @@ Flm_Flc_mul(GEN x, GEN y, ulong p)
   else if (SMALL_ULONG(p))
     return Flm_Flc_mul_i_SMALL(x, y, lx, l, p);
   else
-    return Flm_Flc_mul_i(x, y, lx, l, p);
+    return Flm_Flc_mul_i(x, y, lx, l, p, get_Fl_red(p));
 }
+
+GEN
+Flm_Flc_mul_pre(GEN x, GEN y, ulong p, ulong pi)
+{
+  long l, lx = lg(x);
+  GEN z;
+  if (lx==1) return cgetg(1,t_VECSMALL);
+  l = lgcols(x);
+  z = cgetg(l, t_VECSMALL);
+  if (SMALL_ULONG(p))
+    __Flm_Flc_mul_i_SMALL(z, x, y, lx, l, p);
+  else
+    __Flm_Flc_mul_i(z, x, y, lx, l, p, pi);
+  return z;
+}
+
+GEN
+Flm_Flc_mul_pre_Flx(GEN x, GEN y, ulong p, ulong pi, long sv)
+{
+  long l, lx = lg(x);
+  GEN z;
+  if (lx==1) return pol0_Flx(sv);
+  l = lgcols(x);
+  z = cgetg(l + 1, t_VECSMALL); z[1] = sv;
+  if (SMALL_ULONG(p))
+    __Flm_Flc_mul_i_SMALL(z + 1, x, y, lx, l, p);
+  else
+    __Flm_Flc_mul_i(z + 1, x, y, lx, l, p, pi);
+  return Flx_normalize(z, l + 1);
+}
+
 GEN
 F2m_F2c_mul(GEN x, GEN y)
 {
@@ -673,7 +862,7 @@ Flm_transpose(GEN x)
   GEN y;
   if (lx == 1) return cgetg(1,t_MAT);
   dx = lgcols(x); y = cgetg(dx,t_MAT);
-  for (i=1; i<dx; i++) gel(y,i) = row_Flm(x,i);
+  for (i=1; i<dx; i++) gel(y,i) = Flm_row(x,i);
   return y;
 }
 
@@ -765,6 +954,14 @@ ZM_to_Flm(GEN x, ulong p)
   for (j=1; j<n; j++) gel(y,j) = ZV_to_Flv(gel(x,j), p);
   return y;
 }
+GEN
+ZMV_to_FlmV(GEN z, ulong m)
+{
+  long i, l = lg(z);
+  GEN x = cgetg(l,t_VEC);
+  for (i=1; i<l; i++) gel(x,i) = ZM_to_Flm(gel(z,i), m);
+  return x;
+}
 
 /*                          TO INTMOD                        */
 static GEN
@@ -933,7 +1130,7 @@ gen_FpM_Wiedemann(void *E, GEN (*f)(void*, GEN), GEN B, GEN p)
   if (ZV_equal0(B)) return zerocol(n);
   while (++col <= n)
   {
-    pari_sp btop = avma, av, lim;
+    pari_sp btop = avma, av;
     long i, lQ;
     GEN V, Q, M, W = B;
     GEN b = cgetg(m+2, t_POL);
@@ -941,12 +1138,12 @@ gen_FpM_Wiedemann(void *E, GEN (*f)(void*, GEN), GEN B, GEN p)
     gel(b, 2) = gel(W, col);
     for (i = 3; i<m+2; i++)
       gel(b, i) = cgeti(lgefint(p));
-    av = avma; lim = stack_lim(av,1);
+    av = avma;
     for (i = 3; i<m+2; i++)
     {
       W = f(E, W);
       affii(gel(W, col),gel(b, i));
-      if (low_stack(lim, stack_lim(av,2)))
+      if (gc_needed(av,1))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"Wiedemann: first loop, %ld",i);
         W = gerepileupto(av, W);
@@ -954,17 +1151,17 @@ gen_FpM_Wiedemann(void *E, GEN (*f)(void*, GEN), GEN B, GEN p)
     }
     b = FpX_renormalize(b, m+2);
     if (lgpol(b)==0) {avma = btop; continue; }
-    M = FpX_halfgcd(b, monomial(gen_1, m, 0), p);
+    M = FpX_halfgcd(b, pol_xn(m, 0), p);
     Q = FpX_neg(FpX_normalize(gcoeff(M, 2, 1),p),p);
     W = B; lQ =lg(Q);
     if (DEBUGLEVEL) err_printf("Wiedemann: deg. minpoly: %ld\n",lQ-3);
     V = FpC_Fp_mul(W, gel(Q, lQ-2), p);
-    av = avma; lim = stack_lim(av,1);
+    av = avma;
     for (i = lQ-3; i > 1; i--)
     {
       W = f(E, W);
       V = ZC_lincomb(gen_1, gel(Q,i), V, W);
-      if (low_stack(lim, stack_lim(av,2)))
+      if (gc_needed(av,1))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"Wiedemann: second loop, %ld",i);
         gerepileall(av, 2, &V, &W);
@@ -1059,7 +1256,7 @@ FpV_FpMs_mul(GEN B, GEN M, GEN p) { return FpV_red(ZV_zMs_mul(B, M), p); }
 GEN
 ZlM_gauss(GEN a, GEN b, ulong p, long e, GEN C)
 {
-  pari_sp av = avma, lim = stack_lim(av, 2), av2;
+  pari_sp av = avma, av2;
   GEN xi, xb, pi = gen_1, P;
   long i;
   if (!C) {
@@ -1074,7 +1271,7 @@ ZlM_gauss(GEN a, GEN b, ulong p, long e, GEN C)
   {
     pi = muliu(pi, p); /* = p^(i-1) */
     b = ZM_Z_divexact(ZM_sub(b, ZM_nm_mul(a, xi)), P);
-    if (low_stack(lim, stack_lim(av,2)))
+    if (gc_needed(av,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"ZlM_gauss. i=%ld",i);
       gerepileall(av2,3, &pi,&b,&xb);
@@ -1109,7 +1306,7 @@ GEN
 gen_ZpM_Dixon(void *E, GEN (*f)(void*, GEN), GEN B, GEN p, long e)
 {
   struct wrapper_modp_s W;
-  pari_sp av = avma, lim = stack_lim(av, 2);
+  pari_sp av = avma;
   GEN xb, xi, pi = gen_1;
   long i;
   W.E = E;
@@ -1122,7 +1319,7 @@ gen_ZpM_Dixon(void *E, GEN (*f)(void*, GEN), GEN B, GEN p, long e)
   {
     pi = mulii(pi, p); /* = p^(i-1) */
     B = ZC_Z_divexact(ZC_sub(B, f(E, xi)), p);
-    if (low_stack(lim, stack_lim(av,2)))
+    if (gc_needed(av,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"gen_ZpM_Dixon. i=%ld",i);
       gerepileall(av,3, &pi,&B,&xb);
@@ -1138,7 +1335,7 @@ gen_ZpM_Dixon(void *E, GEN (*f)(void*, GEN), GEN B, GEN p, long e)
 static GEN
 vecprow(GEN A, GEN prow)
 {
-  return mkvec2(vecpermute(prow,gel(A,1)), gel(A,2));
+  return mkvec2(vecsmallpermute(prow,gel(A,1)), gel(A,2));
 }
 
 /* Solve the equation MX = A. Return either a solution as a t_COL,
@@ -1245,20 +1442,19 @@ FpMs_structelim_back(GEN M, GEN V, GEN prow, GEN p)
 
 /* Return a linear form Y such that YM is essentially 0 */
 GEN
-FpMs_leftkernel_elt(GEN M, long nbrow, GEN p)
+FpMs_leftkernel_elt_col(GEN M, long nbcol, long nbrow, GEN p)
 {
   pari_sp av = avma, av2;
   GEN pcol, prow;
-  long nbi=lg(M)-1;
   long i, n;
   GEN Mp, B, MB, R, Rp;
   pari_timer ti;
   struct wrapper_modp_s W;
   if (DEBUGLEVEL) timer_start(&ti);
-  RgMs_structelim(M, nbrow, cgetg(1,t_VECSMALL), &pcol, &prow);
+  RgMs_structelim_col(M, nbcol, nbrow, cgetg(1,t_VECSMALL), &pcol, &prow);
   if (!pcol) { avma = av; return NULL; }
   if (DEBUGLEVEL)
-    timer_printf(&ti,"structured elimination (%ld -> %ld)",nbi,lg(pcol)-1);
+    timer_printf(&ti,"structured elimination (%ld -> %ld)",nbcol,lg(pcol)-1);
   n = lg(pcol)-1;
   Mp = cgetg(n+1, t_MAT);
   for(i=1; i<=n; i++)
@@ -1297,3 +1493,9 @@ FpMs_leftkernel_elt(GEN M, long nbrow, GEN p)
     return gerepilecopy(av, R);
   }
 }
+
+GEN
+FpMs_leftkernel_elt(GEN M, long nbrow, GEN p)
+{
+  return FpMs_leftkernel_elt_col(M, lg(M)-1, nbrow, p);
+}
