@@ -52,7 +52,7 @@ get_FpX_degree(GEN T) { return typ(T)==t_VEC? degpol(gel(T,2)): degpol(T); }
 static ulong
 to_Flx(GEN *P, GEN *Q, GEN p)
 {
-  ulong pp = (ulong)p[2];
+  ulong pp = uel(p,2);
   *P = ZX_to_Flx(*P, pp);
   *Q = ZX_to_Flx(*Q, pp); return pp;
 }
@@ -60,7 +60,7 @@ to_Flx(GEN *P, GEN *Q, GEN p)
 static ulong
 to_Flxq(GEN *P, GEN *T, GEN p)
 {
-  ulong pp = (ulong)p[2];
+  ulong pp = uel(p,2);
   if (P) *P = ZX_to_Flx(*P, pp);
   *T = ZXT_to_FlxT(*T, pp); return pp;
 }
@@ -111,7 +111,7 @@ FpXT_red(GEN z, GEN p)
 GEN
 FpX_normalize(GEN z, GEN p)
 {
-  GEN p1 = leading_term(z);
+  GEN p1 = leading_coeff(z);
   if (lg(z) == 2 || equali1(p1)) return z;
   return FpX_Fp_mul_to_monic(z, Fp_inv(p1,p), p);
 }
@@ -290,6 +290,7 @@ FpX_mulu(GEN y, ulong x,GEN p)
 {
   GEN z;
   long i, l;
+  x = umodui(x, p);
   if (!x) return zeropol(varn(y));
   z = cgetg_copy(y, &l); z[1] = y[1];
   for(i=2; i<l; i++) gel(z,i) = Fp_mulu(gel(y,i), x, p);
@@ -324,10 +325,46 @@ FpX_Fp_mul_to_monic(GEN y,GEN x,GEN p)
   gel(z,l-1) = gen_1; return z;
 }
 
+struct _FpXQ {
+  GEN T, p, aut;
+};
+
+static GEN
+_FpX_sqr(void *data, GEN x)
+{
+  struct _FpXQ *D = (struct _FpXQ*)data;
+  return FpX_sqr(x, D->p);
+}
+static GEN
+_FpX_mul(void *data, GEN x, GEN y)
+{
+  struct _FpXQ *D = (struct _FpXQ*)data;
+  return FpX_mul(x,y, D->p);
+}
+
+GEN
+FpX_powu(GEN x, ulong n, GEN p)
+{
+  struct _FpXQ D;
+  if (n==0) return pol_1(varn(x));
+  D.p = p;
+  return gen_powu(x, n, (void *)&D, _FpX_sqr, _FpX_mul);
+}
+
+GEN
+FpX_halve(GEN y, GEN p)
+{
+  GEN z;
+  long i, l;
+  z = cgetg_copy(y, &l); z[1] = y[1];
+  for(i=2; i<l; i++) gel(z,i) = Fp_halve(gel(y,i), p);
+  return z;
+}
+
 static GEN
 FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
 {
-  long vx, dx, dy, dz, i, j, sx, lr;
+  long vx, dx, dy, dy1, dz, i, j, sx, lr;
   pari_sp av0, av;
   GEN z,p1,rem,lead;
 
@@ -346,7 +383,7 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
     }
     return pol_0(vx);
   }
-  lead = leading_term(y);
+  lead = leading_coeff(y);
   if (!dy) /* y is constant */
   {
     if (pr && pr != ONLY_DIVIDES)
@@ -377,13 +414,14 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
   avma = av0;
   z=cgetg(dz+3,t_POL); z[1] = x[1];
   x += 2; y += 2; z += 2;
+  for (dy1=dy-1; dy1>=0 && !signe(gel(y, dy1)); dy1--);
 
   p1 = gel(x,dx); av = avma;
   gel(z,dz) = lead? gerepileuptoint(av, Fp_mul(p1,lead, p)): icopy(p1);
   for (i=dx-1; i>=dy; i--)
   {
     av=avma; p1=gel(x,i);
-    for (j=i-dy+1; j<=i && j<=dz; j++)
+    for (j=i-dy1; j<=i && j<=dz; j++)
       p1 = subii(p1, mulii(gel(z,j),gel(y,i-j)));
     if (lead) p1 = mulii(p1,lead);
     gel(z,i-dy) = gerepileuptoint(av,modii(p1, p));
@@ -394,7 +432,7 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
   for (sx=0; ; i--)
   {
     p1 = gel(x,i);
-    for (j=0; j<=i && j<=dz; j++)
+    for (j=maxss(0,i-dy1); j<=i && j<=dz; j++)
       p1 = subii(p1, mulii(gel(z,j),gel(y,i-j)));
     p1 = modii(p1,p); if (signe(p1)) { sx = 1; break; }
     if (!i) break;
@@ -414,7 +452,7 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
   for (i--; i>=0; i--)
   {
     av=avma; p1 = gel(x,i);
-    for (j=0; j<=i && j<=dz; j++)
+    for (j=maxss(0,i-dy1); j<=i && j<=dz; j++)
       p1 = subii(p1, mulii(gel(z,j),gel(y,i-j)));
     gel(rem,i) = gerepileuptoint(av, modii(p1,p));
   }
@@ -438,6 +476,43 @@ FpX_div_by_X_x(GEN a, GEN x, GEN p, GEN *r)
   return z;
 }
 
+static GEN
+_FpX_divrem(void * E, GEN x, GEN y, GEN *r)
+{
+  struct _FpXQ *D = (struct _FpXQ*) E;
+  return FpX_divrem(x, y, D->p, r);
+}
+static GEN
+_FpX_add(void * E, GEN x, GEN y) {
+  struct _FpXQ *D = (struct _FpXQ*) E;
+  return FpX_add(x, y, D->p);
+}
+
+static struct bb_ring FpX_ring = { _FpX_add,_FpX_mul,_FpX_sqr };
+
+GEN
+FpX_digits(GEN x, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  struct _FpXQ D;
+  long d = degpol(T), n = (lgpol(x)+d-1)/d;
+  GEN z;
+  D.p = p;
+  z = gen_digits(x,T,n,(void *)&D, &FpX_ring, _FpX_divrem);
+  return gerepileupto(av, z);
+}
+
+GEN
+FpX_fromdigits(GEN x, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  struct _FpXQ D;
+  GEN z;
+  D.p = p;
+  z = gen_fromdigits(x,T,(void *)&D, &FpX_ring);
+  return gerepileupto(av, z);
+}
+
 long
 FpX_valrem(GEN x, GEN t, GEN p, GEN *py)
 {
@@ -458,7 +533,7 @@ FpX_valrem(GEN x, GEN t, GEN p, GEN *py)
 static GEN
 FpX_halfgcd_basecase(GEN a, GEN b, GEN p)
 {
-  pari_sp av=avma, lim = stack_lim(av,2);
+  pari_sp av=avma;
   GEN u,u1,v,v1;
   long vx = varn(a);
   long n = lgpol(a)>>1;
@@ -470,7 +545,7 @@ FpX_halfgcd_basecase(GEN a, GEN b, GEN p)
     a = b; b = r; swap(u,u1); swap(v,v1);
     u1 = FpX_sub(u1, FpX_mul(u, q, p), p);
     v1 = FpX_sub(v1, FpX_mul(v, q ,p), p);
-    if (low_stack(lim,stack_lim(av,2)))
+    if (gc_needed(av,2))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"FpX_halfgcd (d = %ld)",degpol(b));
       gerepileall(av,6, &a,&b,&u1,&v1,&u,&v);
@@ -493,14 +568,22 @@ FpXM_FpX_mul2(GEN M, GEN x, GEN y, GEN p)
   return res;
 }
 
-/*TODO: implement Strassen 7 multiplications formula (p is large) */
 static GEN
-FpXM_mul2(GEN M, GEN N, GEN p)
+FpXM_mul2(GEN A, GEN B, GEN p)
 {
-  GEN res = cgetg(3, t_MAT);
-  gel(res, 1) = FpXM_FpX_mul2(M,gcoeff(N,1,1),gcoeff(N,2,1),p);
-  gel(res, 2) = FpXM_FpX_mul2(M,gcoeff(N,1,2),gcoeff(N,2,2),p);
-  return res;
+  GEN A11=gcoeff(A,1,1),A12=gcoeff(A,1,2), B11=gcoeff(B,1,1),B12=gcoeff(B,1,2);
+  GEN A21=gcoeff(A,2,1),A22=gcoeff(A,2,2), B21=gcoeff(B,2,1),B22=gcoeff(B,2,2);
+  GEN M1 = FpX_mul(FpX_add(A11,A22, p), FpX_add(B11,B22, p), p);
+  GEN M2 = FpX_mul(FpX_add(A21,A22, p), B11, p);
+  GEN M3 = FpX_mul(A11, FpX_sub(B12,B22, p), p);
+  GEN M4 = FpX_mul(A22, FpX_sub(B21,B11, p), p);
+  GEN M5 = FpX_mul(FpX_add(A11,A12, p), B22, p);
+  GEN M6 = FpX_mul(FpX_sub(A21,A11, p), FpX_add(B11,B12, p), p);
+  GEN M7 = FpX_mul(FpX_sub(A12,A22, p), FpX_add(B21,B22, p), p);
+  GEN T1 = FpX_add(M1,M4, p), T2 = FpX_sub(M7,M5, p);
+  GEN T3 = FpX_sub(M1,M2, p), T4 = FpX_add(M3,M6, p);
+  retmkmat2(mkcol2(FpX_add(T1,T2, p), FpX_add(M2,M4, p)),
+            mkcol2(FpX_add(M3,M5, p), FpX_add(T3,T4, p)));
 }
 
 /* Return [0,1;1,-q]*M */
@@ -580,11 +663,11 @@ FpX_halfgcd(GEN x, GEN y, GEN p)
 static GEN
 FpX_gcd_basecase(GEN a, GEN b, GEN p)
 {
-  pari_sp av = avma, av0=avma, lim = stack_lim(av0,2);
+  pari_sp av = avma, av0=avma;
   while (signe(b))
   {
     GEN c;
-    if (low_stack(lim,stack_lim(av0,2)))
+    if (gc_needed(av0,2))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"FpX_gcd (d = %ld)",degpol(b));
       gerepileall(av0,2, &a,&b);
@@ -637,7 +720,7 @@ FpX_gcd_check(GEN x, GEN y, GEN p)
   b = FpX_red(y, p);
   while (signe(b))
   {
-    GEN lead = leading_term(b);
+    GEN lead = leading_coeff(b);
     GEN g = gcdii(lead,p);
     if (!equali1(g)) return gerepileuptoint(av,g);
     c = FpX_rem(a,b,p); a=b; b=c;
@@ -648,7 +731,7 @@ FpX_gcd_check(GEN x, GEN y, GEN p)
 static GEN
 FpX_extgcd_basecase(GEN a, GEN b, GEN p, GEN *ptu, GEN *ptv)
 {
-  pari_sp av=avma, lim = stack_lim(av,2);
+  pari_sp av=avma;
   GEN u,v,d,d1,v1;
   long vx = varn(a);
   d = a; d1 = b;
@@ -659,7 +742,7 @@ FpX_extgcd_basecase(GEN a, GEN b, GEN p, GEN *ptu, GEN *ptv)
     v = FpX_sub(v,FpX_mul(q,v1,p),p);
     u=v; v=v1; v1=u;
     u=r; d=d1; d1=u;
-    if (low_stack(lim,stack_lim(av,2)))
+    if (gc_needed(av,2))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"FpX_extgcd (d = %ld)",degpol(d));
       gerepileall(av,5, &d,&d1,&u,&v,&v1);
@@ -759,6 +842,18 @@ random_FpX(long d1, long v, GEN p)
   return FpX_renormalize(y,d);
 }
 
+GEN
+FpX_dotproduct(GEN x, GEN y, GEN p)
+{
+  long i, l = minss(lg(x), lg(y));
+  pari_sp av;
+  GEN c;
+  if (l == 2) return gen_0;
+  av = avma; c = mulii(gel(x,2),gel(y,2));
+  for (i=3; i<l; i++) c = addii(c, mulii(gel(x,i),gel(y,i)));
+  return gerepileuptoint(av, modii(c,p));
+}
+
 /* Evaluation in Fp
  * x a ZX and y an Fp, return x(y) mod p
  *
@@ -770,8 +865,8 @@ FpX_eval(GEN x,GEN y,GEN p)
   pari_sp av;
   GEN p1,r,res;
   long j, i=lg(x)-1;
-  if (i<=2)
-    return (i==2)? modii(gel(x,2),p): gen_0;
+  if (i<=2 || !signe(y))
+    return (i==1)? gen_0: modii(gel(x,2),p);
   res=cgeti(lgefint(p));
   av=avma; p1=gel(x,i);
   /* specific attention to sparse polynomials (see poleval)*/
@@ -818,7 +913,7 @@ GEN
 FpX_resultant(GEN a, GEN b, GEN p)
 {
   long da,db,dc;
-  pari_sp av, lim;
+  pari_sp av;
   GEN c,lb, res = gen_1;
 
   if (!signe(a) || !signe(b)) return gen_0;
@@ -830,7 +925,7 @@ FpX_resultant(GEN a, GEN b, GEN p)
     if (both_odd(da,db)) res = subii(p, res);
   }
   if (!da) return gen_1; /* = res * a[2] ^ db, since 0 <= db <= da = 0 */
-  av = avma; lim = stack_lim(av,2);
+  av = avma;
   while (db)
   {
     lb = gel(b,db+2);
@@ -840,7 +935,7 @@ FpX_resultant(GEN a, GEN b, GEN p)
 
     if (both_odd(da,db)) res = subii(p, res);
     if (!equali1(lb)) res = Fp_mul(res, Fp_powu(lb, da - dc, p), p);
-    if (low_stack(lim,stack_lim(av,2)))
+    if (gc_needed(av,2))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"FpX_resultant (da = %ld)",da);
       gerepileall(av,3, &a,&b,&res);
@@ -852,22 +947,28 @@ FpX_resultant(GEN a, GEN b, GEN p)
   return gerepileuptoint(av, res);
 }
 
+/* disc P = (-1)^(n(n-1)/2) lc(P)^(n - deg P' - 2) Res(P,P'), n = deg P */
 GEN
-FpX_disc(GEN x, GEN p)
+FpX_disc(GEN P, GEN p)
 {
   pari_sp av = avma;
-  GEN L, D = FpX_resultant(x, FpX_deriv(x,p), p);
+  GEN L, dP = FpX_deriv(P,p), D = FpX_resultant(P, dP, p);
+  long dd;
   if (!D || !signe(D)) return gen_0;
-  L = leading_term(x); if (!equali1(L)) D = Fp_div(D,L,p);
-  if (degpol(x) & 2) D = Fp_neg(D,p);
+  dd = degpol(P) - 2 - degpol(dP); /* >= -1; > -1 iff p | deg(P) */
+  L = leading_coeff(P);
+  if (dd && !equali1(L))
+    D = (dd == -1)? Fp_div(D,L,p): Fp_mul(D, Fp_powu(L, dd, p), p);
+  if (degpol(P) & 2) D = Fp_neg(D ,p);
   return gerepileuptoint(av, D);
 }
 
-static GEN _FpX_mul(void *p,GEN a,GEN b){return FpX_mul(a,b,(GEN)p);}
 GEN
 FpXV_prod(GEN V, GEN p)
 {
-  return divide_conquer_assoc(V, (void *)p, &_FpX_mul);
+  struct _FpXQ D;
+  D.p = p;
+  return gen_product(V, (void *)&D, &_FpX_mul);
 }
 
 GEN
@@ -1181,6 +1282,172 @@ FpX_rem(GEN x, GEN T, GEN p)
   }
 }
 
+static GEN
+deg2pol_shallow(GEN x2, GEN x1, GEN x0, long v)
+{
+  GEN x = cgetg(5,t_POL);
+  x[1] = evalsigne(1) | evalvarn(v);
+  gel(x,2) = x0;
+  gel(x,3) = x1;
+  gel(x,4) = x2;
+  return normalizepol_lg(x,5);
+}
+
+static GEN
+FpV_producttree(GEN xa, GEN s, GEN p, long vs)
+{
+  long n = lg(xa)-1;
+  long m = n==1 ? 1: expu(n-1)+1;
+  long i, j, k, ls = lg(s);
+  GEN T = cgetg(m+1, t_VEC);
+  GEN t = cgetg(ls, t_VEC);
+  for (j=1, k=1; j<ls; k+=s[j++])
+    gel(t, j) = s[j] == 1 ?
+             deg1pol(gen_1, Fp_neg(gel(xa,k), p), vs):
+             deg2pol_shallow(gen_1,
+               Fp_neg(Fp_add(gel(xa,k), gel(xa,k+1), p), p),
+               Fp_mul(gel(xa,k), gel(xa,k+1), p), vs);
+  gel(T,1) = t;
+  for (i=2; i<=m; i++)
+  {
+    GEN u = gel(T, i-1);
+    long n = lg(u)-1;
+    GEN t = cgetg(((n+1)>>1)+1, t_VEC);
+    for (j=1, k=1; k<n; j++, k+=2)
+      gel(t, j) = FpX_mul(gel(u, k), gel(u, k+1), p);
+    gel(T, i) = t;
+  }
+  return T;
+}
+
+static GEN
+FpX_FpV_multieval_tree(GEN P, GEN xa, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  long i,j,k;
+  long m = lg(T)-1;
+  GEN t;
+  GEN Tp = cgetg(m+1, t_VEC);
+  gel(Tp, m) = mkvec(P);
+  for (i=m-1; i>=1; i--)
+  {
+    GEN u = gel(T, i);
+    GEN v = gel(Tp, i+1);
+    long n = lg(u)-1;
+    t = cgetg(n+1, t_VEC);
+    for (j=1, k=1; k<n; j++, k+=2)
+    {
+      gel(t, k)   = FpX_rem(gel(v, j), gel(u, k), p);
+      gel(t, k+1) = FpX_rem(gel(v, j), gel(u, k+1), p);
+    }
+    gel(Tp, i) = t;
+  }
+  {
+    GEN R = cgetg(lg(xa), t_VEC);
+    GEN u = gel(T, i+1);
+    GEN v = gel(Tp, i+1);
+    long n = lg(u)-1;
+    for (j=1, k=1; j<=n; j++)
+    {
+      long c, d = degpol(gel(u,j));
+      for (c=1; c<=d; c++, k++)
+        gel(R,k) = FpX_eval(gel(v, j), gel(xa,k), p);
+    }
+    return gerepileupto(av, R);
+  }
+}
+
+static GEN
+FpVV_polint_tree(GEN T, GEN R, GEN s, GEN xa, GEN ya, GEN p, long vs)
+{
+  pari_sp av = avma;
+  long m = lg(T)-1;
+  long i, j, k, ls = lg(s);
+  GEN Tp = cgetg(m+1, t_VEC);
+  GEN t = cgetg(ls, t_VEC);
+  for (j=1, k=1; j<ls; k+=s[j++])
+    if (s[j]==2)
+    {
+      GEN a = Fp_mul(gel(ya,k), gel(R,k), p);
+      GEN b = Fp_mul(gel(ya,k+1), gel(R,k+1), p);
+      gel(t, j) = deg1pol(Fp_add(a, b, p),
+              Fp_neg(Fp_add(Fp_mul(gel(xa,k), b, p ),
+              Fp_mul(gel(xa,k+1), a, p), p), p), vs);
+    }
+    else
+      gel(t, j) = scalarpol(Fp_mul(gel(ya,k), gel(R,k), p), vs);
+  gel(Tp, 1) = t;
+  for (i=2; i<=m; i++)
+  {
+    GEN u = gel(T, i-1);
+    GEN t = cgetg(lg(gel(T,i)), t_VEC);
+    GEN v = gel(Tp, i-1);
+    long n = lg(v)-1;
+    for (j=1, k=1; k<n; j++, k+=2)
+      gel(t, j) = FpX_add(ZX_mul(gel(u, k), gel(v, k+1)),
+                          ZX_mul(gel(u, k+1), gel(v, k)), p);
+    gel(Tp, i) = t;
+  }
+  return gerepilecopy(av, gmael(Tp,m,1));
+}
+
+GEN
+FpX_FpV_multieval(GEN P, GEN xa, GEN p)
+{
+  pari_sp av = avma;
+  GEN s = producttree_scheme(lg(xa)-1);
+  GEN T = FpV_producttree(xa, s, p, varn(P));
+  return gerepileupto(av, FpX_FpV_multieval_tree(P, xa, T, p));
+}
+
+GEN
+FpV_polint(GEN xa, GEN ya, GEN p, long vs)
+{
+  pari_sp av = avma;
+  GEN s = producttree_scheme(lg(xa)-1);
+  GEN T = FpV_producttree(xa, s, p, vs);
+  long m = lg(T)-1;
+  GEN P = FpX_deriv(gmael(T, m, 1), p);
+  GEN R = FpV_inv(FpX_FpV_multieval_tree(P, xa, T, p), p);
+  return gerepileupto(av, FpVV_polint_tree(T, R, s, xa, ya, p, vs));
+}
+
+GEN
+FpV_FpM_polint(GEN xa, GEN ya, GEN p, long vs)
+{
+  pari_sp av = avma;
+  GEN s = producttree_scheme(lg(xa)-1);
+  GEN T = FpV_producttree(xa, s, p, vs);
+  long i, m = lg(T)-1, l = lg(ya)-1;
+  GEN P = FpX_deriv(gmael(T, m, 1), p);
+  GEN R = FpV_inv(FpX_FpV_multieval_tree(P, xa, T, p), p);
+  GEN M = cgetg(l+1, t_VEC);
+  for (i=1; i<=l; i++)
+    gel(M,i) = FpVV_polint_tree(T, R, s, xa, gel(ya,i), p, vs);
+  return gerepileupto(av, M);
+}
+
+GEN
+FpV_invVandermonde(GEN L, GEN den, GEN p)
+{
+  pari_sp av = avma;
+  long i, n = lg(L);
+  GEN M, R;
+  GEN s = producttree_scheme(n-1);
+  GEN tree = FpV_producttree(L, s, p, 0);
+  long m = lg(tree)-1;
+  GEN T = gmael(tree, m, 1);
+  R = FpV_inv(FpX_FpV_multieval_tree(FpX_deriv(T, p), L, tree, p), p);
+  if (den) R = FpC_Fp_mul(R, den, p);
+  M = cgetg(n, t_MAT);
+  for (i = 1; i < n; i++)
+  {
+    GEN P = FpX_Fp_mul(FpX_div_by_X_x(T, gel(L,i), p, NULL), gel(R,i), p);
+    gel(M,i) = RgX_to_RgC(P, n-1);
+  }
+  return gerepilecopy(av, M);
+}
+
 /***********************************************************************/
 /**                                                                   **/
 /**                              FpXQ                                 **/
@@ -1238,10 +1505,6 @@ FpXQ_div(GEN x,GEN y,GEN T,GEN p)
   return gerepileupto(av, FpXQ_mul(x,FpXQ_inv(y,T,p),T,p));
 }
 
-struct _FpXQ {
-  GEN T, p, mg, aut;
-};
-
 static GEN
 _FpXQ_add(void *data, GEN x, GEN y)
 {
@@ -1249,6 +1512,12 @@ _FpXQ_add(void *data, GEN x, GEN y)
   return ZX_add(x, y);
 }
 static GEN
+_FpXQ_sub(void *data, GEN x, GEN y)
+{
+  (void) data;
+  return ZX_sub(x, y);
+}
+static GEN
 _FpXQ_cmul(void *data, GEN P, long a, GEN x)
 {
   (void) data;
@@ -1285,7 +1554,31 @@ _FpXQ_red(void *data, GEN x)
   return FpX_red(x,D->p);
 }
 
-static struct bb_algebra FpXQ_algebra = { _FpXQ_red,_FpXQ_add,_FpXQ_mul,_FpXQ_sqr,_FpXQ_one,_FpXQ_zero };
+static struct bb_algebra FpXQ_algebra = { _FpXQ_red, _FpXQ_add, _FpXQ_sub,
+       _FpXQ_mul, _FpXQ_sqr, _FpXQ_one, _FpXQ_zero };
+
+const struct bb_algebra *
+get_FpXQ_algebra(void **E, GEN T, GEN p)
+{
+  GEN z = new_chunk(sizeof(struct _FpXQ));
+  struct _FpXQ *e = (struct _FpXQ *) z;
+  e->T = FpX_get_red(T, p);
+  e->p  = p; *E = (void*)e;
+  return &FpXQ_algebra;
+}
+
+static struct bb_algebra FpX_algebra = { _FpXQ_red, _FpXQ_add, _FpXQ_sub,
+       _FpX_mul, _FpX_sqr, _FpXQ_one, _FpXQ_zero };
+
+const struct bb_algebra *
+get_FpX_algebra(void **E, GEN p, long v)
+{
+  GEN z = new_chunk(sizeof(struct _FpXQ));
+  struct _FpXQ *e = (struct _FpXQ *) z;
+  e->T = pol_x(v);
+  e->p  = p; *E = (void*)e;
+  return &FpX_algebra;
+}
 
 /* x,pol in Z[X], p in Z, n in Z, compute lift(x^n mod (p, pol)) */
 GEN
@@ -1343,7 +1636,7 @@ FpXQ_powers(GEN x, long l, GEN T, GEN p)
     GEN z = FlxV_to_ZXV(Flxq_powers(x, l, T, pp));
     return gerepileupto(av, z);
   }
-  use_sqr = (degpol(x)<<1)>=get_FpX_degree(T);
+  use_sqr = 2*degpol(x)>=get_FpX_degree(T);
   D.T = FpX_get_red(T,p); D.p = p;
   return gen_powers(x, l, use_sqr, (void*)&D, &_FpXQ_sqr, &_FpXQ_mul,&_FpXQ_one);
 }
@@ -1354,6 +1647,19 @@ FpXQ_matrix_pow(GEN y, long n, long m, GEN P, GEN l)
   return RgXV_to_RgM(FpXQ_powers(y,m-1,P,l),n);
 }
 
+GEN
+FpX_Frobenius(GEN T, GEN p)
+{
+  return FpXQ_pow(pol_x(get_FpX_var(T)), p, T, p);
+}
+
+GEN
+FpX_matFrobenius(GEN T, GEN p)
+{
+  long n = get_FpX_degree(T);
+  return FpXQ_matrix_pow(FpX_Frobenius(T, p), n, n, T, p);
+}
+
 GEN
 FpX_FpXQV_eval(GEN Q, GEN x, GEN T, GEN p)
 {
@@ -1374,11 +1680,31 @@ FpX_FpXQ_eval(GEN Q, GEN x, GEN T, GEN p)
     GEN z = Flx_Flxq_eval(ZX_to_Flx(Q, pp), x, T, pp);
     return Flx_to_ZX_inplace(gerepileuptoleaf(av, z));
   }
-  use_sqr = (degpol(x)<<1) >= get_FpX_degree(T);
+  use_sqr = 2*degpol(x) >= get_FpX_degree(T);
   D.T = FpX_get_red(T,p); D.p = p;
   return gen_bkeval(Q,degpol(Q),x,use_sqr,(void*)&D,&FpXQ_algebra,_FpXQ_cmul);
 }
 
+GEN
+FpXC_FpXQV_eval(GEN P, GEN x, GEN T, GEN p)
+{
+  long i, l = lg(P);
+  GEN res = cgetg(l, t_COL);
+  for (i=1; i<l; i++)
+    gel(res,i) = FpX_FpXQV_eval(gel(P,i), x, T, p);
+  return res;
+}
+
+GEN
+FpXM_FpXQV_eval(GEN Q, GEN x, GEN T, GEN p)
+{
+  long i, l = lg(Q);
+  GEN y = cgetg(l, t_MAT);
+  for (i=1; i<l; i++)
+    gel(y,i) = FpXC_FpXQV_eval(gel(Q,i), x, T, p);
+  return y;
+}
+
 GEN
 FpXQ_autpowers(GEN aut, long f, GEN T, GEN p)
 {
@@ -1421,17 +1747,45 @@ FpXQ_autpow(GEN x, ulong n, GEN T, GEN p)
   return gen_powu(x,n,(void*)&D,FpXQ_autpow_sqr,FpXQ_autpow_mul);
 }
 
+static GEN
+FpXQ_auttrace_mul(void *E, GEN x, GEN y)
+{
+  struct _FpXQ *D = (struct _FpXQ*)E;
+  GEN T = D->T, p = D->p;
+  GEN phi1 = gel(x,1), a1 = gel(x,2);
+  GEN phi2 = gel(y,1), a2 = gel(y,2);
+  ulong d = brent_kung_optpow(maxss(degpol(phi2),degpol(a2)),2,1);
+  GEN V1 = FpXQ_powers(phi1, d, T, p);
+  GEN phi3 = FpX_FpXQV_eval(phi2, V1, T, p);
+  GEN aphi = FpX_FpXQV_eval(a2, V1, T, p);
+  GEN a3 = FpX_add(a1, aphi, p);
+  return mkvec2(phi3, a3);
+}
+
+static GEN
+FpXQ_auttrace_sqr(void *E, GEN x)
+{ return FpXQ_auttrace_mul(E, x, x); }
+
+GEN
+FpXQ_auttrace(GEN x, ulong n, GEN T, GEN p)
+{
+  struct _FpXQ D;
+  D.T = FpX_get_red(T, p); D.p = p;
+  return gen_powu(x,n,(void*)&D,FpXQ_auttrace_sqr,FpXQ_auttrace_mul);
+}
+
 static GEN
 FpXQ_autsum_mul(void *E, GEN x, GEN y)
 {
   struct _FpXQ *D = (struct _FpXQ*)E;
+  GEN T = D->T, p = D->p;
   GEN phi1 = gel(x,1), a1 = gel(x,2);
   GEN phi2 = gel(y,1), a2 = gel(y,2);
-  ulong d = brent_kung_optpow(maxss(degpol(phi1),degpol(a1)),2,1);
-  GEN V2 = FpXQ_powers(phi2,d,D->T,D->p);
-  GEN phi3 = FpX_FpXQV_eval(phi1,V2,D->T,D->p);
-  GEN aphi = FpX_FpXQV_eval(a1,V2,D->T,D->p);
-  GEN a3 = FpXQ_mul(aphi,a2,D->T,D->p);
+  ulong d = brent_kung_optpow(maxss(degpol(phi2),degpol(a2)),2,1);
+  GEN V1 = FpXQ_powers(phi1, d, T, p);
+  GEN phi3 = FpX_FpXQV_eval(phi2, V1, T, p);
+  GEN aphi = FpX_FpXQV_eval(a2, V1, T, p);
+  GEN a3 = FpXQ_mul(a1, aphi, T, p);
   return mkvec2(phi3, a3);
 }
 static GEN
@@ -1446,6 +1800,33 @@ FpXQ_autsum(GEN x, ulong n, GEN T, GEN p)
   return gen_powu(x,n,(void*)&D,FpXQ_autsum_sqr,FpXQ_autsum_mul);
 }
 
+static GEN
+FpXQM_autsum_mul(void *E, GEN x, GEN y)
+{
+  struct _FpXQ *D = (struct _FpXQ*)E;
+  GEN T = D->T, p = D->p;
+  GEN phi1 = gel(x,1), a1 = gel(x,2);
+  GEN phi2 = gel(y,1), a2 = gel(y,2);
+  long g = lg(a2)-1, dT = get_FpX_degree(T);
+  ulong d = brent_kung_optpow(dT-1, g*g+1, 1);
+  GEN V1 = FpXQ_powers(phi1, d, T, p);
+  GEN phi3 = FpX_FpXQV_eval(phi2, V1, T, p);
+  GEN aphi = FpXM_FpXQV_eval(a2, V1, T, p);
+  GEN a3 = FqM_mul(a1, aphi, T, p);
+  return mkvec2(phi3, a3);
+}
+static GEN
+FpXQM_autsum_sqr(void *E, GEN x)
+{ return FpXQM_autsum_mul(E, x, x); }
+
+GEN
+FpXQM_autsum(GEN x, ulong n, GEN T, GEN p)
+{
+  struct _FpXQ D;
+  D.T = FpX_get_red(T, p); D.p = p;
+  return gen_powu(x, n, (void*)&D, FpXQM_autsum_sqr, FpXQM_autsum_mul);
+}
+
 static long
 bounded_order(GEN p, GEN b, long k)
 {
@@ -1484,7 +1865,7 @@ FpXQ_pow_Frobenius(GEN x, GEN n, GEN aut, GEN T, GEN p)
   q = powiu(p, d);
   if (dvdii(q, n))
   {
-    long vn = logint(an,p,NULL)-1;
+    long vn = logint(an,p);
     GEN autvn = vn==1 ? aut: FpXQ_autpow(aut,vn,T,p);
     z = FpX_FpXQ_eval(x,autvn,T,p);
   } else
@@ -1516,16 +1897,16 @@ FpXQ_issquare(GEN x, GEN T, GEN p)
 {
   pari_sp av;
   long res;
-  if (lg(x) == 2 || equalui(2, p)) return 1;
+  if (lg(x) == 2 || absequalui(2, p)) return 1;
   if (lg(x) == 3) return Fq_issquare(gel(x,2), T, p);
   /* Ng = g^((q-1)/(p-1)) */
-  av = avma; res = kronecker(FpX_resultant(T,x,p), p) == 1;
+  av = avma; res = kronecker(FpXQ_norm(x,T,p), p) == 1;
   avma = av; return res;
 }
 int
 Fp_issquare(GEN x, GEN p)
 {
-  if (equalui(2, p)) return 1;
+  if (absequalui(2, p)) return 1;
   return kronecker(x, p) == 1;
 }
 /* assume T irreducible mod p */
@@ -1536,27 +1917,19 @@ Fq_issquare(GEN x, GEN T, GEN p)
   return (T && ! odd(get_FpX_degree(T))) || Fp_issquare(x, p);
 }
 
-static GEN
-famat_Z_gcd(GEN M, GEN n)
+long
+Fq_ispower(GEN x, GEN K, GEN T, GEN p)
 {
-  pari_sp av=avma;
-  long i, j, l=lgcols(M);
-  GEN F=cgetg(3,t_MAT);
-  gel(F,1)=cgetg(l,t_COL);
-  gel(F,2)=cgetg(l,t_COL);
-  for (i=1, j=1; i<l; i++)
-  {
-    GEN p = gcoeff(M,i,1);
-    GEN e = gminsg(Z_pval(n,p),gcoeff(M,i,2));
-    if (signe(e))
-    {
-      gcoeff(F,j,1)=p;
-      gcoeff(F,j,2)=e;
-      j++;
-    }
-  }
-  setlg(gel(F,1),j); setlg(gel(F,2),j);
-  return gerepilecopy(av,F);
+  pari_sp av = avma;
+  long d;
+  GEN Q;
+  if (!T) return Fp_ispower(x,K,p);
+  d = get_FpX_degree(T);
+  if (!umodui(d, K)) return 1;
+  Q = subiu(powiu(p,d), 1);
+  Q = diviiexact(Q, gcdii(Q, K));
+  d = gequal1(Fq_pow(x, Q, T,p));
+  avma = av; return d;
 }
 
 /* discrete log in FpXQ for a in Fp^*, g in FpXQ^* of order ord */
@@ -1570,7 +1943,7 @@ Fp_FpXQ_log(GEN a, GEN g, GEN o, GEN T, GEN p)
   /* p > 2 */
 
   ordp = subis(p, 1); /* even */
-  ord  = dlog_get_ord(o);
+  ord  = get_arith_Z(o);
   if (!ord) ord = T? subis(powiu(p, get_FpX_degree(T)), 1): ordp;
   if (equalii(a, ordp)) /* -1 */
     return gerepileuptoint(av, shifti(ord,-1));
@@ -1584,7 +1957,7 @@ Fp_FpXQ_log(GEN a, GEN g, GEN o, GEN T, GEN p)
       q = diviiexact(ord,ordp);
       g = FpXQ_pow(g,q,T,p);
     }
-    g = constant_term(g);
+    g = constant_coeff(g);
   }
   n_q = Fp_log(a,g,op,p);
   if (lg(n_q)==1) return gerepileuptoleaf(av, n_q);
@@ -1618,15 +1991,16 @@ _FpXQ_easylog(void *E, GEN a, GEN g, GEN ord)
 {
   struct _FpXQ *s=(struct _FpXQ*) E;
   if (degpol(a)) return NULL;
-  return Fp_FpXQ_log(constant_term(a),g,ord,s->T,s->p);
+  return Fp_FpXQ_log(constant_coeff(a),g,ord,s->T,s->p);
 }
 
 static const struct bb_group FpXQ_star={_FpXQ_mul,_FpXQ_pow,_FpXQ_rand,hash_GEN,ZX_equal,ZX_equal1,_FpXQ_easylog};
 
-const struct bb_group *get_FpXQ_star(void **E, GEN T, GEN p)
+const struct bb_group *
+get_FpXQ_star(void **E, GEN T, GEN p)
 {
   struct _FpXQ *e = (struct _FpXQ *) stack_malloc(sizeof(struct _FpXQ));
-  e->T = T; e->p  = p; e->aut =  FpXQ_pow(pol_x(get_FpX_var(T)), p, T, p);
+  e->T = T; e->p  = p; e->aut =  FpX_Frobenius(T, p);
   *E = (void*)e; return &FpXQ_star;
 }
 
@@ -1651,22 +2025,47 @@ FpXQ_order(GEN a, GEN ord, GEN T, GEN p)
 GEN
 FpXQ_log(GEN a, GEN g, GEN ord, GEN T, GEN p)
 {
+  pari_sp av=avma;
   if (lgefint(p)==3)
   {
-    pari_sp av=avma;
-    ulong pp = to_Flxq(&a, &T, p);
-    GEN z = Flxq_log(a, ZX_to_Flx(g, pp), ord, T, pp);
-    return gerepileuptoint(av,z);
+    if (uel(p,2) == 2)
+    {
+      GEN z = F2xq_log(ZX_to_F2x(a), ZX_to_F2x(g), ord,
+                                     ZX_to_F2x(get_FpX_mod(T)));
+      return gerepileuptoleaf(av, z);
+    }
+    else
+    {
+      ulong pp = to_Flxq(&a, &T, p);
+      GEN z = Flxq_log(a, ZX_to_Flx(g, pp), ord, T, pp);
+      return gerepileuptoleaf(av, z);
+    }
   }
   else
   {
     void *E;
     const struct bb_group *S = get_FpXQ_star(&E,T,p);
     GEN z = gen_PH_log(a,g,ord,E,S);
-    return z? z: cgetg(1,t_VEC);
+    return gerepileuptoleaf(av, z);
   }
 }
 
+GEN
+Fq_log(GEN a, GEN g, GEN ord, GEN T, GEN p)
+{
+  if (!T) return Fp_log(a,g,ord,p);
+  if (typ(g) == t_INT)
+  {
+    if (typ(a) == t_POL)
+    {
+      if (degpol(a)) return cgetg(1,t_VEC);
+      a = gel(a,2);
+    }
+    return Fp_log(a,g,ord,p);
+  }
+  return typ(a) == t_INT? Fp_FpXQ_log(a,g,ord,T,p): FpXQ_log(a,g,ord,T,p);
+}
+
 GEN
 FpXQ_sqrtn(GEN a, GEN n, GEN T, GEN p, GEN *zeta)
 {
@@ -1681,12 +2080,22 @@ FpXQ_sqrtn(GEN a, GEN n, GEN T, GEN p, GEN *zeta)
   }
   if (lgefint(p)==3)
   {
-    ulong pp = to_Flxq(&a, &T, p);
-    z = Flxq_sqrtn(a, n, T, pp, zeta);
-    if (!z) return NULL;
-    if (!zeta) return Flx_to_ZX_inplace(gerepileuptoleaf(av, z));
-    z = Flx_to_ZX(z);
-    *zeta=Flx_to_ZX(*zeta);
+    if (uel(p,2) == 2)
+    {
+      z = F2xq_sqrtn(ZX_to_F2x(a), n, ZX_to_F2x(get_Flx_mod(T)), zeta);
+      if (!z) return NULL;
+      z = F2x_to_ZX(z);
+      if (!zeta) return gerepileuptoleaf(av, z);
+      *zeta=F2x_to_ZX(*zeta);
+    } else
+    {
+      ulong pp = to_Flxq(&a, &T, p);
+      z = Flxq_sqrtn(a, n, T, pp, zeta);
+      if (!z) return NULL;
+      if (!zeta) return Flx_to_ZX_inplace(gerepileuptoleaf(av, z));
+      z = Flx_to_ZX(z);
+      *zeta=Flx_to_ZX(*zeta);
+    }
   }
   else
   {
@@ -1713,7 +2122,7 @@ FpXQ_norm(GEN x, GEN TB, GEN p)
   pari_sp av = avma;
   GEN T = get_FpX_mod(TB);
   GEN y = FpX_resultant(T, x, p);
-  GEN L = leading_term(T);
+  GEN L = leading_coeff(T);
   if (gequal1(L) || signe(x)==0) return y;
   return gerepileupto(av, Fp_div(y, Fp_pows(L, degpol(x), p), p));
 }
@@ -1734,30 +2143,113 @@ GEN
 FpXQ_charpoly(GEN x, GEN T, GEN p)
 {
   pari_sp ltop=avma;
-  long v;
+  long vT, v = fetch_var();
   GEN R;
   T = leafcopy(get_FpX_mod(T));
-  v = varn(T); setvarn(T, MAXVARN);
-  x = leafcopy(x); setvarn(x, MAXVARN);
-  R = FpX_FpXY_resultant(T, deg1pol_shallow(gen_1,FpX_neg(x,p),v),p);
-  return gerepileupto(ltop,R);
+  vT = varn(T); setvarn(T, v);
+  x = leafcopy(x); setvarn(x, v);
+  R = FpX_FpXY_resultant(T, deg1pol_shallow(gen_1,FpX_neg(x,p),vT),p);
+  (void)delete_var(); return gerepileupto(ltop,R);
+}
+
+/* Computing minimal polynomial :                         */
+/* cf Shoup 'Efficient Computation of Minimal Polynomials */
+/*          in Algebraic Extensions of Finite Fields'     */
+
+static GEN
+FpXn_mul(GEN a, GEN b, long n, GEN p)
+{
+  return FpX_red(RgXn_red_shallow(ZX_mul(a, b), n), p);
+}
+
+/* Let v a linear form, return the linear form z->v(tau*z)
+   that is, v*(M_tau) */
+
+static GEN
+FpXQ_transmul_init(GEN tau, GEN T, GEN p)
+{
+  GEN bht;
+  GEN h, Tp = get_FpX_red(T, &h);
+  long n = degpol(Tp), vT = varn(Tp);
+  GEN ft = FpX_recipspec(Tp+2, n+1, n+1);
+  GEN bt = FpX_recipspec(tau+2, lgpol(tau), n);
+  setvarn(ft, vT); setvarn(bt, vT);
+  if (h)
+    bht = FpXn_mul(bt, h, n-1, p);
+  else
+  {
+    GEN bh = FpX_div(RgX_shift_shallow(tau, n-1), T, p);
+    bht = FpX_recipspec(bh+2, lgpol(bh), n-1);
+    setvarn(bht, vT);
+  }
+  return mkvec3(bt, bht, ft);
+}
+
+static GEN
+FpXQ_transmul(GEN tau, GEN a, long n, GEN p)
+{
+  pari_sp ltop = avma;
+  GEN t1, t2, t3, vec;
+  GEN bt = gel(tau, 1), bht = gel(tau, 2), ft = gel(tau, 3);
+  if (signe(a)==0) return pol_0(varn(a));
+  t2 = RgX_shift_shallow(FpX_mul(bt, a, p),1-n);
+  if (signe(bht)==0) return gerepilecopy(ltop, t2);
+  t1 = RgX_shift_shallow(FpX_mul(ft, a, p),-n);
+  t3 = FpXn_mul(t1, bht, n-1, p);
+  vec = FpX_sub(t2, RgX_shift_shallow(t3, 1), p);
+  return gerepileupto(ltop, vec);
 }
 
 GEN
 FpXQ_minpoly(GEN x, GEN T, GEN p)
 {
-  pari_sp ltop=avma;
-  GEN G,R=FpXQ_charpoly(x, T, p);
-  GEN dR=FpX_deriv(R,p);
-  while (signe(dR)==0)
+  pari_sp ltop = avma;
+  long vT, n;
+  GEN v_x, g, tau;
+  if (lgefint(p)==3)
   {
-    R  = RgX_deflate(R,itos(p));
-    dR = FpX_deriv(R,p);
+    ulong pp = to_Flxq(&x, &T, p);
+    GEN g = Flxq_minpoly(x, T, pp);
+    return gerepileupto(ltop, Flx_to_ZX(g));
   }
-  G=FpX_gcd(R,dR,p);
-  G=FpX_normalize(G,p);
-  G=FpX_div(R,G,p);
-  return gerepileupto(ltop,G);
+  vT = get_FpX_var(T);
+  n = get_FpX_degree(T);
+  g = pol_1(vT);
+  tau = pol_1(vT);
+  T = FpX_get_red(T, p);
+  x = FpXQ_red(x, T, p);
+  v_x = FpXQ_powers(x, usqrt(2*n), T, p);
+  while(signe(tau) != 0)
+  {
+    long i, j, m, k1;
+    GEN M, v, tr;
+    GEN g_prime, c;
+    if (degpol(g) == n) { tau = pol_1(vT); g = pol_1(vT); }
+    v = random_FpX(n, vT, p);
+    tr = FpXQ_transmul_init(tau, T, p);
+    v = FpXQ_transmul(tr, v, n, p);
+    m = 2*(n-degpol(g));
+    k1 = usqrt(m);
+    tr = FpXQ_transmul_init(gel(v_x,k1+1), T, p);
+    c = cgetg(m+2,t_POL);
+    c[1] = evalsigne(1)|evalvarn(vT);
+    for (i=0; i<m; i+=k1)
+    {
+      long mj = minss(m-i, k1);
+      for (j=0; j<mj; j++)
+        gel(c,m+1-(i+j)) = FpX_dotproduct(v, gel(v_x,j+1), p);
+      v = FpXQ_transmul(tr, v, n, p);
+    }
+    c = FpX_renormalize(c, m+2);
+    /* now c contains <v,x^i> , i = 0..m-1  */
+    M = FpX_halfgcd(pol_xn(m, vT), c, p);
+    g_prime = gmael(M, 2, 2);
+    if (degpol(g_prime) < 1) continue;
+    g = FpX_mul(g, g_prime, p);
+    tau = FpXQ_mul(tau, FpX_FpXQV_eval(g_prime, v_x, T, p), T, p);
+  }
+  g = FpX_normalize(g,p);
+  return gerepilecopy(ltop,g);
 }
 
 GEN
@@ -1766,9 +2258,9 @@ FpXQ_conjvec(GEN x, GEN T, GEN p)
   pari_sp av=avma;
   long i;
   long n = get_FpX_degree(T), v = varn(x);
-  GEN M = FpXQ_matrix_pow(FpXQ_pow(pol_x(v),p,T,p),n,n,T,p);
+  GEN M = FpX_matFrobenius(T, p);
   GEN z = cgetg(n+1,t_COL);
-  gel(z,1) = RgX_to_RgV(x,n);
+  gel(z,1) = RgX_to_RgC(x,n);
   for (i=2; i<=n; i++) gel(z,i) = FpM_FpC_mul(M,gel(z,i-1),p);
   gel(z,1) = x;
   for (i=2; i<=n; i++) gel(z,i) = RgV_to_RgX(gel(z,i),v);
@@ -1785,7 +2277,7 @@ gener_FpXQ_i(GEN T, GEN p, GEN p_1, GEN Lp, GEN Lq)
 {
   pari_sp av;
   long vT = varn(T), f = degpol(T), l = lg(Lq);
-  GEN F = FpXQ_pow(pol_x(vT), p, T, p); /* Frobenius */
+  GEN F = FpX_Frobenius(T, p);
   int p_is_2 = is_pm1(p_1);
   for (av = avma;; avma = av)
   {
@@ -1837,7 +2329,6 @@ gener_FpXQ(GEN T, GEN p, GEN *po)
     g = gener_Flxq(T, pp, po);
     if (!po) return Flx_to_ZX_inplace(gerepileuptoleaf(av, g));
     g = Flx_to_ZX(g);
-    gel(*po,2) = Flx_to_ZX(gel(*po,2));
     gerepileall(av, 2, &g, po);
     return g;
   }
@@ -1898,7 +2389,7 @@ gener_FpXQ_local(GEN T, GEN p, GEN L)
   for (i=1; i < l; i++)
   {
     GEN a, b, ell = gel(L,i);
-    if (equaliu(ell,2)) continue;
+    if (absequaliu(ell,2)) continue;
     a = dvmdii(Q, ell, &b);
     if (b == gen_0)
       gel(Lp,ip++) = a;
