@@ -43,30 +43,38 @@ RgV_is_ZMV(GEN V)
 /**                                                                **/
 /********************************************************************/
 /*           GENERIC  MULTIPLICATION involving zc/zm                */
+
+/* x[i,] * y */
+static GEN
+RgMrow_zc_mul_i(GEN x, GEN y, long c, long i)
+{
+  pari_sp av = avma;
+  GEN s = NULL;
+  long j;
+  for (j=1; j<c; j++)
+  {
+    long t = y[j];
+    if (!t) continue;
+    if (!s) { s = gmulgs(gcoeff(x,i,j),t); continue; }
+    switch(t)
+    {
+      case  1: s = gadd(s, gcoeff(x,i,j)); break;
+      case -1: s = gsub(s, gcoeff(x,i,j)); break;
+      default: s = gadd(s, gmulgs(gcoeff(x,i,j), t)); break;
+    }
+  }
+  if (!s) { avma = av; return gen_0; }
+  return gerepileupto(av, s);
+}
+GEN
+RgMrow_zc_mul(GEN x, GEN y, long i) { return RgMrow_zc_mul_i(x,y,lg(y),i); }
 /* x non-empty t_MAT, y a compatible zc (dimension > 0). */
 static GEN
 RgM_zc_mul_i(GEN x, GEN y, long c, long l)
 {
-  long i, j;
-  pari_sp av;
-  GEN z = cgetg(l,t_COL), s;
-
-  for (i=1; i<l; i++)
-  {
-    av = avma; s = gmulgs(gcoeff(x,i,1),y[1]);
-    for (j=2; j<c; j++)
-    {
-      long t = y[j];
-      switch(t)
-      {
-        case  0: break;
-        case  1: s = gadd(s, gcoeff(x,i,j)); break;
-        case -1: s = gsub(s, gcoeff(x,i,j)); break;
-        default: s = gadd(s, gmulgs(gcoeff(x,i,j), t)); break;
-      }
-    }
-    gel(z,i) = gerepileupto(av,s);
-  }
+  GEN z = cgetg(l,t_COL);
+  long i;
+  for (i = 1; i < l; i++) gel(z,i) = RgMrow_zc_mul_i(x,y,c,i);
   return z;
 }
 GEN
@@ -109,14 +117,14 @@ GEN
 RgV_dotsquare(GEN x)
 {
   long i, lx = lg(x);
-  pari_sp av = avma, lim = stack_lim(av,3);
+  pari_sp av = avma;
   GEN z;
   if (lx == 1) return gen_0;
   z = gsqr(gel(x,1));
   for (i=2; i<lx; i++)
   {
     z = gadd(z, gsqr(gel(x,i)));
-    if (low_stack(lim,stack_lim(av,3)))
+    if (gc_needed(av,3))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"RgV_dotsquare, i = %ld",i);
       z = gerepileupto(av, z);
@@ -129,7 +137,7 @@ RgV_dotsquare(GEN x)
 static GEN
 RgV_dotproduct_i(GEN x, GEN y, long lx)
 {
-  pari_sp av = avma, lim = stack_lim(av,3);
+  pari_sp av = avma;
   long i;
   GEN z;
   if (lx == 1) return gen_0;
@@ -137,7 +145,7 @@ RgV_dotproduct_i(GEN x, GEN y, long lx)
   for (i=2; i<lx; i++)
   {
     z = gadd(z, gmul(gel(x,i), gel(y,i)));
-    if (low_stack(lim,stack_lim(av,3)))
+    if (gc_needed(av,3))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"RgV_dotproduct, i = %ld",i);
       z = gerepileupto(av, z);
@@ -181,6 +189,35 @@ RgV_sumpart2(GEN v, long m, long n)
   p = gel(v,m); for (i=m+1; i<=n; i++) p = gadd(p, gel(v,i));
   return p;
 }
+GEN
+RgM_sumcol(GEN A)
+{
+  long i,j,m,l = lg(A);
+  GEN v;
+
+  if (l == 1) return cgetg(1,t_MAT);
+  if (l == 2) return gcopy(gel(A,1));
+  m = lgcols(A);
+  v = cgetg(m, t_COL);
+  for (i = 1; i < m; i++)
+  {
+    pari_sp av = avma;
+    GEN s = gcoeff(A,i,1);
+    for (j = 2; j < l; j++) s = gadd(s, gcoeff(A,i,j));
+    gel(v, i) = gerepileupto(av, s);
+  }
+  return v;
+}
+
+static GEN
+_gmul(void *data, GEN x, GEN y)
+{ (void)data; return gmul(x,y); }
+
+GEN
+RgV_prod(GEN x)
+{
+  return gen_product(x, NULL, _gmul);
+}
 
 /*                    ADDITION SCALAR + MATRIX                     */
 /* x square matrix, y scalar; create the square matrix x + y*Id */
@@ -269,6 +306,36 @@ RgC_Rg_add(GEN x, GEN y)
   for (k = 2; k < lx; k++) gel(z,k) = gcopy(gel(x,k));
   return z;
 }
+GEN
+RgC_Rg_sub(GEN x, GEN y)
+{
+  long k, lx = lg(x);
+  GEN z = cgetg(lx, t_COL);
+  if (lx == 1)
+  {
+    if (isintzero(y)) return z;
+    pari_err_TYPE2("-",x,y);
+  }
+  gel(z,1) = gsub(gel(x,1), y);
+  for (k = 2; k < lx; k++) gel(z,k) = gcopy(gel(x,k));
+  return z;
+}
+/* a - x */
+GEN
+Rg_RgC_sub(GEN a, GEN x)
+{
+  long k, lx = lg(x);
+  GEN z = cgetg(lx,t_COL);
+  if (lx == 1)
+  {
+    if (isintzero(a)) return z;
+    pari_err_TYPE2("-",a,x);
+  }
+  gel(z,1) = gsub(a, gel(x,1));
+  for (k = 2; k < lx; k++) gel(z,k) = gneg(gel(x,k));
+  return z;
+}
+
 
 static GEN
 RgC_add_i(GEN x, GEN y, long lx)
@@ -495,6 +562,8 @@ RgM_mul(GEN x, GEN y)
   lx = lg(x);
   if (lx != lgcols(y)) pari_err_OP("operation 'RgM_mul'", x,y);
   if (lx == 1) return zeromat(0,ly-1);
+  if (RgM_is_ZM(x) && RgM_is_ZM(y))
+    return ZM_mul(x, y);
   if (is_modular_mul(x,y,&z)) return gerepileupto(av, z);
   if (RgM_is_FFM(x, &ffx) && RgM_is_FFM(y, &ffy)) {
     if (!FF_samefield(ffx, ffy))
@@ -594,10 +663,12 @@ RgM_sqr(GEN x)
 {
   pari_sp av = avma;
   long j, lx = lg(x);
-  GEN z;
+  GEN z, ffx = NULL;
   if (lx == 1) return cgetg(1, t_MAT);
   if (lx != lgcols(x)) pari_err_OP("operation 'RgM_mul'", x,x);
+  if (RgM_is_ZM(x))         return ZM_sqr(x);
   if (is_modular_sqr(x,&z)) return gerepileupto(av, z);
+  if (RgM_is_FFM(x, &ffx))  return FFM_mul(x, x, ffx);
   z = cgetg(lx, t_MAT);
   for (j=1; j<lx; j++) gel(z,j) = RgM_RgC_mul_i(x, gel(x,j), lx, lx);
   return z;
@@ -606,6 +677,9 @@ RgM_sqr(GEN x)
 static GEN
 _RgM_add(void *E, GEN x, GEN y) { (void)E; return RgM_add(x, y); }
 
+static GEN
+_RgM_sub(void *E, GEN x, GEN y) { (void)E; return RgM_sub(x, y); }
+
 static GEN
 _RgM_cmul(void *E, GEN P, long a, GEN x) { (void)E; return RgM_Rg_mul(x,gel(P,a+2)); }
 
@@ -624,7 +698,8 @@ _RgM_zero(void *E) { long *n = (long*) E; return zeromat(*n,*n); }
 static GEN
 _RgM_red(void *E, GEN x) { (void)E; return x; }
 
-static struct bb_algebra RgM_algebra = { _RgM_red,_RgM_add,_RgM_mul,_RgM_sqr,_RgM_one,_RgM_zero };
+static struct bb_algebra RgM_algebra = { _RgM_red, _RgM_add, _RgM_sub,
+       _RgM_mul, _RgM_sqr, _RgM_one, _RgM_zero };
 
 /* generates the list of powers of x of degree 0,1,2,...,l*/
 GEN
@@ -799,6 +874,20 @@ RgM_isidentity(GEN x)
   return 1;
 }
 
+long
+RgC_is_ei(GEN x)
+{
+  long i, j = 0, l = lg(x);
+  for (i = 1; i < l; i++)
+  {
+    GEN c = gel(x,i);
+    if (gequal0(c)) continue;
+    if (!gequal1(c) || j) return 0;
+    j = i;
+  }
+  return j;
+}
+
 int
 RgM_isdiagonal(GEN x)
 {
@@ -845,3 +934,15 @@ RgM_det_triangular(GEN mat)
   return av==avma? gcopy(s): gerepileupto(av,s);
 }
 
+GEN
+RgV_kill0(GEN v)
+{
+  long i, l;
+  GEN w = cgetg_copy(v, &l);
+  for (i = 1; i < l; i++)
+  {
+    GEN a = gel(v,i);
+    gel(w,i) = gequal0(a) ? NULL: a;
+  }
+  return w;
+}
