@@ -25,6 +25,73 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /*                                                                 */
 /*******************************************************************/
 
+static GEN
+Flm_charpoly_i(GEN x, ulong p)
+{
+  long lx = lg(x), r, i;
+  GEN H, y = cgetg(lx+1, t_VEC);
+  gel(y,1) = pol1_Flx(0); H = Flm_hess(x, p);
+  for (r = 1; r < lx; r++)
+  {
+    pari_sp av2 = avma;
+    ulong a = 1;
+    GEN z, b = zero_Flx(0);
+    for (i = r-1; i; i--)
+    {
+      a = Fl_mul(a, ucoeff(H,i+1,i), p);
+      if (!a) break;
+      b = Flx_add(b, Flx_Fl_mul(gel(y,i), Fl_mul(a,ucoeff(H,i,r),p), p), p);
+    }
+    z = Flx_sub(Flx_shift(gel(y,r), 1),
+                Flx_Fl_mul(gel(y,r), ucoeff(H,r,r), p), p);
+    /* (X - H[r,r])y[r] - b */
+    gel(y,r+1) = gerepileuptoleaf(av2, Flx_sub(z, b, p));
+  }
+  return gel(y,lx);
+}
+
+GEN
+Flm_charpoly(GEN x, ulong p)
+{
+  pari_sp av = avma;
+  return gerepileuptoleaf(av, Flm_charpoly_i(x,p));
+}
+
+GEN
+FpM_charpoly(GEN x, GEN p)
+{
+  pari_sp av = avma;
+  long lx, r, i;
+  GEN y, H;
+
+  if (lgefint(p) == 3)
+  {
+    ulong pp = p[2];
+    y = Flx_to_ZX(Flm_charpoly_i(ZM_to_Flm(x,pp), pp));
+    return gerepileupto(av, y);
+  }
+  lx = lg(x); y = cgetg(lx+1, t_VEC);
+  gel(y,1) = pol_1(0); H = FpM_hess(x, p);
+  for (r = 1; r < lx; r++)
+  {
+    pari_sp av2 = avma;
+    GEN z, a = gen_1, b = pol_0(0);
+    for (i = r-1; i; i--)
+    {
+      a = Fp_mul(a, gcoeff(H,i+1,i), p);
+      if (!signe(a)) break;
+      b = ZX_add(b, ZX_Z_mul(gel(y,i), Fp_mul(a,gcoeff(H,i,r),p)));
+    }
+    b = FpX_red(b, p);
+    z = FpX_sub(RgX_shift_shallow(gel(y,r), 1),
+                FpX_Fp_mul(gel(y,r), gcoeff(H,r,r), p), p);
+    z = FpX_sub(z,b,p);
+    if (r+1 == lx) { gel(y,lx) = z; break; }
+    gel(y,r+1) = gerepileupto(av2, z); /* (X - H[r,r])y[r] - b */
+  }
+  return gerepileupto(av, gel(y,lx));
+}
+
 GEN
 charpoly0(GEN x, long v, long flag)
 {
@@ -76,8 +143,13 @@ easychar(GEN x, long v)
     }
 
     case t_POLMOD:
-      return RgXQ_charpoly(gel(x,2), gel(x,1), v);
-
+    {
+      GEN A = gel(x,2), T = gel(x,1);
+      if (typ(A)==t_POL && RgX_is_QX(A) && RgX_is_ZX(T))
+        return QXQ_charpoly(A, T, v);
+      else
+        return RgXQ_charpoly(A, T, v);
+    }
     case t_MAT:
       lx=lg(x);
       if (lx==1) return pol_1(v);
@@ -95,7 +167,7 @@ RgM_Fp_charpoly(GEN x, GEN p, long v)
   if (lgefint(p) == 3)
   {
     ulong pp = itou(p);
-    T = Flm_charpoly(RgM_to_Flm(x, pp), pp);
+    T = Flm_charpoly_i(RgM_to_Flm(x, pp), pp);
     T = Flx_to_ZX(T);
   }
   else
@@ -278,19 +350,21 @@ GEN
 adj(GEN x)
 {
   GEN y;
-  (void)caradj(x, MAXVARN, &y); return y;
+  (void)caradj(x, fetch_var(), &y);
+  (void)delete_var(); return y;
 }
 
 GEN
 adjsafe(GEN x)
 {
-  const long v = MAXVARN;
+  const long v = fetch_var();
   pari_sp av = avma;
-  GEN C;
+  GEN C, A;
   if (typ(x) != t_MAT) pari_err_TYPE("matadjoint",x);
   if (lg(x) < 3) return gcopy(x);
   C = charpoly(x,v);
-  return gerepileupto(av, RgM_adj_from_char(x, v, C));
+  A = RgM_adj_from_char(x, v, C);
+  (void)delete_var(); return gerepileupto(av, A);
 }
 
 GEN
@@ -304,6 +378,303 @@ matadjoint0(GEN x, long flag)
   pari_err_FLAG("matadjoint"); return NULL; /* not reached */
 }
 
+/*******************************************************************/
+/*                                                                 */
+/*                       Frobenius form                            */
+/*                                                                 */
+/*******************************************************************/
+
+/* The following section implement a mix of Ozello and Storjohann algorithms
+
+P. Ozello, doctoral thesis (in French):
+Calcul exact des formes de Jordan et de Frobenius d'une matrice, Chapitre 2
+http://tel.archives-ouvertes.fr/tel-00323705
+
+A. Storjohann,  Diss. ETH No. 13922
+Algorithms for Matrix Canonical Forms, Chapter 9
+https://cs.uwaterloo.ca/~astorjoh/diss2up.pdf
+
+We use Storjohann Lemma 9.14 (step1, step2, step3) Ozello theorem 4,
+and Storjohann Lemma 9.18
+*/
+
+/* Elementary transforms */
+
+/* M <- U^(-1) M U, U = E_{i,j}(k) => U^(-1) = E{i,j}(-k)
+ * P = U * P */
+static void
+transL(GEN M, GEN P, GEN k, long i, long j)
+{
+  long l, n = lg(M)-1;
+  for(l=1; l<=n; l++) /* M[,j]-=k*M[,i] */
+    gcoeff(M,l,j) = gsub(gcoeff(M,l,j), gmul(gcoeff(M,l,i), k));
+  for(l=1; l<=n; l++) /* M[i,]+=k*M[j,] */
+    gcoeff(M,i,l) = gadd(gcoeff(M,i,l), gmul(gcoeff(M,j,l), k));
+  if (P)
+    for(l=1; l<=n; l++)
+      gcoeff(P,i,l) = gadd(gcoeff(P,i,l), gmul(gcoeff(P,j,l), k));
+}
+
+/* j = a or b */
+static void
+transD(GEN M, GEN P, long a, long b, long j)
+{
+  long l, n;
+  GEN k = gcoeff(M,a,b), ki;
+
+  if (gequal1(k)) return;
+  ki = ginv(k); n = lg(M)-1;
+  for(l=1; l<=n; l++)
+    if (l!=j)
+    {
+      gcoeff(M,l,j) = gmul(gcoeff(M,l,j), k);
+      gcoeff(M,j,l) = (j==a && l==b)? gen_1: gmul(gcoeff(M,j,l), ki);
+    }
+  if (P)
+    for(l=1; l<=n; l++)
+      gcoeff(P,j,l) = gmul(gcoeff(P,j,l), ki);
+}
+
+static void
+transS(GEN M, GEN P, long i, long j)
+{
+  long l, n = lg(M)-1;
+  swap(gel(M,i), gel(M,j));
+  for (l=1; l<=n; l++)
+    swap(gcoeff(M,i,l), gcoeff(M,j,l));
+  if (P)
+    for (l=1; l<=n; l++)
+      swap(gcoeff(P,i,l), gcoeff(P,j,l));
+}
+
+/* Convert companion matrix to polynomial*/
+static GEN
+minpoly_polslice(GEN M, long i, long j, long v)
+{
+  long k, d = j+1-i;
+  GEN P = cgetg(d+3,t_POL);
+  P[1] = evalsigne(1)|evalvarn(v);
+  for (k=0; k<d; k++)
+    gel(P,k+2) = gneg(gcoeff(M,i+k, j));
+  gel(P,d+2) = gen_1;
+  return P;
+}
+
+static GEN
+minpoly_listpolslice(GEN M, GEN V, long v)
+{
+  long i, n = lg(M)-1, nb = lg(V)-1;
+  GEN W = cgetg(nb+1, t_VEC);
+  for (i=1; i<=nb; i++)
+    gel(W,i) = minpoly_polslice(M, V[i], i < nb? V[i+1]-1: n, v);
+  return W;
+}
+
+static int
+minpoly_dvdslice(GEN M, long i, long j, long k)
+{
+  pari_sp av = avma;
+  long r = signe(RgX_rem(minpoly_polslice(M, i, j-1, 0),
+                        minpoly_polslice(M, j, k, 0)));
+  avma = av; return r==0;
+}
+
+static void
+RgM_replace(GEN M, GEN M2)
+{
+  long n = lg(M)-1, m = nbrows(M), i, j;
+  for(i=1; i<=n; i++)
+    for(j=1; j<=m; j++)
+      gcoeff(M, i, j) = gcoeff(M2, i, j);
+}
+
+static void
+gerepilemat2_inplace(pari_sp av, GEN M, GEN P)
+{
+  GEN M2 = M, P2 = P;
+  gerepileall(av, P ? 2: 1, &M2, &P2);
+  RgM_replace(M, M2);
+  if (P) RgM_replace(P, P2);
+}
+
+/* Lemma 9.14 */
+static long
+weakfrobenius_step1(GEN M, GEN P, long j0)
+{
+  pari_sp av = avma;
+  long n = lg(M)-1, k, j;
+  for (j = j0; j < n; ++j)
+  {
+    if (gequal0(gcoeff(M, j+1, j)))
+    {
+      for (k = j+2; k <= n; ++k)
+        if (!gequal0(gcoeff(M,k,j))) break;
+      if (k > n) return j;
+      transS(M, P, k, j+1);
+    }
+    transD(M, P, j+1, j, j+1);
+    /* Now M[j+1,j] = 1 */
+    for (k = 1; k <= n; ++k)
+      if (k != j+1 && !gequal0(gcoeff(M,k,j))) /* zero M[k,j] */
+      {
+        transL(M, P, gneg(gcoeff(M,k,j)), k, j+1);
+        gcoeff(M,k,j) = gen_0; /* avoid approximate 0 */
+      }
+    if (gc_needed(av,1))
+    {
+      if (DEBUGMEM > 1)
+        pari_warn(warnmem,"RgM_minpoly stage 1: j0=%ld, j=%ld", j0, j);
+      gerepilemat2_inplace(av, M, P);
+    }
+  }
+  return n;
+}
+
+static void
+weakfrobenius_step2(GEN M, GEN P, long j)
+{
+  pari_sp av = avma;
+  long i, k, n = lg(M)-1;
+  for(i=j; i>=2; i--)
+  {
+    for(k=j+1; k<=n; k++)
+      if (!gequal0(gcoeff(M,i,k)))
+        transL(M, P, gcoeff(M,i,k), i-1, k);
+    if (gc_needed(av,1))
+    {
+      if (DEBUGMEM > 1)
+        pari_warn(warnmem,"RgM_minpoly stage 2: j=%ld, i=%ld", j, i);
+      gerepilemat2_inplace(av, M, P);
+    }
+  }
+}
+
+static long
+weakfrobenius_step3(GEN M, GEN P, long j0, long j)
+{
+  long i, k, n = lg(M)-1;
+  if (j == n) return 0;
+  if (gequal0(gcoeff(M, j0, j+1)))
+  {
+    for (k=j+2; k<=n; k++)
+      if (!gequal0(gcoeff(M, j0, k))) break;
+    if (k > n) return 0;
+    transS(M, P, k, j+1);
+  }
+  transD(M, P, j0, j+1, j+1);
+  for (i=j+2; i<=n; i++)
+    if (!gequal0(gcoeff(M, j0, i)))
+      transL(M, P, gcoeff(M, j0, i),j+1, i);
+  return 1;
+}
+
+/* flag: 0 -> full Frobenius from , 1 -> weak Frobenius form */
+static GEN
+RgM_Frobenius(GEN M, long flag, GEN *pt_P, GEN *pt_v)
+{
+  pari_sp av = avma, av2, ltop;
+  long n = lg(M)-1, eps, j0 = 1, nb = 0;
+  GEN v, P;
+  v = cgetg(n+1, t_VECSMALL);
+  ltop = avma;
+  P = pt_P ? matid(n): NULL;
+  M = RgM_shallowcopy(M);
+  av2 = avma;
+  while (j0 <= n)
+  {
+    long j = weakfrobenius_step1(M, P, j0);
+    weakfrobenius_step2(M, P, j);
+    eps = weakfrobenius_step3(M, P, j0, j);
+    if (eps == 0)
+    {
+      v[++nb] = j0;
+      if (flag == 0 && nb > 1 && !minpoly_dvdslice(M, v[nb-1], j0, j))
+      {
+        j = j0; j0 = v[nb-1]; nb -= 2;
+        transL(M, P, gen_1, j, j0); /*lemma 9.18*/
+      } else
+        j0 = j+1;
+    }
+    else
+      transS(M, P, j0, j+1); /*theorem 4*/
+    if (gc_needed(av,1))
+    {
+      if (DEBUGMEM > 1)
+        pari_warn(warnmem,"weakfrobenius j0=%ld",j0);
+      gerepilemat2_inplace(av2, M, P);
+    }
+  }
+  fixlg(v, nb+1);
+  if (pt_v) *pt_v = v;
+  gerepileall(pt_v ? ltop: av, P? 2: 1, &M, &P);
+  if (pt_P) *pt_P = P;
+  return M;
+}
+
+static GEN
+RgM_minpoly(GEN M, long v)
+{
+  pari_sp av = avma;
+  GEN V, W;
+  M = RgM_Frobenius(M, 1, NULL, &V);
+  W = minpoly_listpolslice(M, V, v);
+  if (varncmp(v,gvar2(W)) >= 0)
+    pari_err_PRIORITY("matfrobenius", M, "<=", v);
+  return gerepileupto(av, RgX_normalize(glcm0(W, NULL)));
+}
+
+GEN
+Frobeniusform(GEN V, long n)
+{
+  long i, j, k;
+  GEN M = zeromatcopy(n,n);
+  for (k=1,i=1;i<lg(V);i++,k++)
+  {
+    GEN  P = gel(V,i);
+    long d = degpol(P);
+    if (k+d-1 > n) pari_err_PREC("matfrobenius");
+    for (j=0; j<d-1; j++, k++) gcoeff(M,k+1,k) = gen_1;
+    for (j=0; j<d; j++) gcoeff(M,k-j,k) = gneg(gel(P, 1+d-j));
+  }
+  return M;
+}
+
+GEN
+matfrobenius(GEN M, long flag, long v)
+{
+  long n;
+  if (typ(M)!=t_MAT) pari_err_TYPE("matfrobenius",M);
+  if (v < 0) v = 0;
+  n = lg(M)-1;
+  if (n && lgcols(M)!=n+1) pari_err_DIM("matfrobenius");
+  if (flag > 2) pari_err_FLAG("matfrobenius");
+  switch (flag)
+  {
+  case 0:
+    return RgM_Frobenius(M, 0, NULL, NULL);
+  case 1:
+    {
+      pari_sp av = avma;
+      GEN V, W, F;
+      F = RgM_Frobenius(M, 0, NULL, &V);
+      W = minpoly_listpolslice(F, V, v);
+      if (varncmp(v, gvar2(W)) >= 0)
+        pari_err_PRIORITY("matfrobenius", M, "<=", v);
+      return gerepileupto(av, W);
+    }
+  case 2:
+    {
+      GEN P, F, R = cgetg(3, t_VEC);
+      F = RgM_Frobenius(M, 0, &P, NULL);
+      gel(R,1) = F; gel(R,2) = P;
+      return R;
+    }
+  default:
+    pari_err_FLAG("matfrobenius");
+  }
+  return NULL; /*NOT REACHED*/
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                       MINIMAL POLYNOMIAL                        */
@@ -322,7 +693,7 @@ easymin(GEN x, long v)
   dR=RgX_deriv(R);
   if (!lgpol(dR)) {avma=ltop; return NULL;}
   G=RgX_gcd(R,dR);
-  G=RgX_Rg_div(G,leading_term(G));
+  G=RgX_Rg_div(G,leading_coeff(G));
   return gerepileupto(ltop, RgX_div(R,G));
 }
 
@@ -338,7 +709,7 @@ minpoly(GEN x, long v)
       setvarn(p1,v); return gerepileupto(ltop,p1);
   }
 
-  P=easymin(x,v);
+  P = easymin(x,v);
   if (P) return P;
   if (typ(x)==t_POLMOD)
   {
@@ -348,7 +719,7 @@ minpoly(GEN x, long v)
   }
   if (typ(x)!=t_MAT) pari_err_TYPE("minpoly",x);
   if (lg(x) == 1) return pol_1(v);
-  return gerepilecopy(ltop,gel(matfrobenius(x,1,v),1));
+  return RgM_minpoly(x,v);
 }
 
 /*******************************************************************/
@@ -359,14 +730,14 @@ minpoly(GEN x, long v)
 GEN
 hess(GEN x)
 {
-  pari_sp av = avma, lim;
+  pari_sp av = avma;
   long lx = lg(x), m, i, j;
 
   if (typ(x) != t_MAT) pari_err_TYPE("hess",x);
   if (lx == 1) return cgetg(1,t_MAT);
   if (lgcols(x) != lx) pari_err_DIM("hess");
 
-  x = RgM_shallowcopy(x); lim = stack_lim(av,2);
+  x = RgM_shallowcopy(x);
   for (m=2; m<lx-1; m++)
   {
     GEN t = NULL;
@@ -385,7 +756,7 @@ hess(GEN x)
         gcoeff(x,i,j) = gsub(gcoeff(x,i,j), gmul(c,gcoeff(x,m,j)));
       for (j=1; j<lx; j++)
         gcoeff(x,j,m) = gadd(gcoeff(x,j,m), gmul(c,gcoeff(x,j,i)));
-      if (low_stack(lim, stack_lim(av,2)))
+      if (gc_needed(av,2))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"hess, m = %ld", m);
         gerepileall(av,2, &x, &t);
@@ -428,7 +799,7 @@ Flm_hess(GEN x, ulong p)
 GEN
 FpM_hess(GEN x, GEN p)
 {
-  pari_sp av = avma, lim;
+  pari_sp av = avma;
   long lx = lg(x), m, i, j;
   if (lx == 1) return cgetg(1,t_MAT);
   if (lgcols(x) != lx) pari_err_DIM("hess");
@@ -438,7 +809,7 @@ FpM_hess(GEN x, GEN p)
     x = Flm_hess(ZM_to_Flm(x, pp), pp);
     return gerepileupto(av, Flm_to_ZM(x));
   }
-  x = RgM_shallowcopy(x); lim = stack_lim(av,2);
+  x = RgM_shallowcopy(x);
   for (m=2; m<lx-1; m++)
   {
     GEN t = NULL;
@@ -457,7 +828,7 @@ FpM_hess(GEN x, GEN p)
         gcoeff(x,i,j) = Fp_sub(gcoeff(x,i,j), Fp_mul(c,gcoeff(x,m,j),p), p);
       for (j=1; j<lx; j++)
         gcoeff(x,j,m) = Fp_add(gcoeff(x,j,m), Fp_mul(c,gcoeff(x,j,i),p), p);
-      if (low_stack(lim, stack_lim(av,2)))
+      if (gc_needed(av,2))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"hess, m = %ld", m);
         gerepileall(av,2, &x, &t);
@@ -494,125 +865,104 @@ carhess(GEN x, long v)
   return fix_pol(av, gel(y,lx));
 }
 
-GEN
-FpM_charpoly(GEN x, GEN p)
-{
-  pari_sp av = avma;
-  long lx, r, i;
-  GEN y, H;
-
-  if (lgefint(p) == 3)
-  {
-    ulong pp = p[2];
-    y = Flx_to_ZX(Flm_charpoly(ZM_to_Flm(x,pp), pp));
-    return gerepileupto(av, y);
-  }
-  lx = lg(x); y = cgetg(lx+1, t_VEC);
-  gel(y,1) = pol_1(0); H = FpM_hess(x, p);
-  for (r = 1; r < lx; r++)
-  {
-    pari_sp av2 = avma;
-    GEN z, a = gen_1, b = pol_0(0);
-    for (i = r-1; i; i--)
-    {
-      a = Fp_mul(a, gcoeff(H,i+1,i), p);
-      if (!signe(a)) break;
-      b = ZX_add(b, ZX_Z_mul(gel(y,i), Fp_mul(a,gcoeff(H,i,r),p)));
-    }
-    b = FpX_red(b, p);
-    z = FpX_sub(RgX_shift_shallow(gel(y,r), 1),
-                FpX_Fp_mul(gel(y,r), gcoeff(H,r,r), p), p);
-    z = FpX_sub(z,b,p);
-    if (r+1 == lx) { gel(y,lx) = z; break; }
-    gel(y,r+1) = gerepileupto(av2, z); /* (X - H[r,r])y[r] - b */
-  }
-  return gerepileupto(av, gel(y,lx));
-}
-GEN
-Flm_charpoly(GEN x, long p)
-{
-  pari_sp av;
-  long lx, r, i;
-  GEN y, H;
-
-  lx = lg(x); av = avma; y = cgetg(lx+1, t_VEC);
-  gel(y,1) = pol1_Flx(0); H = Flm_hess(x, p);
-  for (r = 1; r < lx; r++)
-  {
-    pari_sp av2 = avma;
-    ulong a = 1;
-    GEN z, b = zero_Flx(0);
-    for (i = r-1; i; i--)
-    {
-      a = Fl_mul(a, ucoeff(H,i+1,i), p);
-      if (!a) break;
-      b = Flx_add(b, Flx_Fl_mul(gel(y,i), Fl_mul(a,ucoeff(H,i,r),p), p), p);
-    }
-    z = Flx_sub(Flx_shift(gel(y,r), 1),
-                Flx_Fl_mul(gel(y,r), ucoeff(H,r,r), p), p);
-    /* (X - H[r,r])y[r] - b */
-    gel(y,r+1) = gerepileuptoleaf(av2, Flx_sub(z, b, p));
-  }
-  return gerepileuptoleaf(av, gel(y,lx));
-}
-
-/* s = max_k binomial(n,k) (kB^2)^(k/2),  B = |M|oo. Return ceil(log2(s)) */
+/* Bound for sup norm of charpoly(M/dM), M integral: let B = |M|oo / |dM|,
+ *   s = max_k binomial(n,k) (kB^2)^(k/2),
+ * return ceil(log2(s)) */
 static double
-charpoly_bound(GEN M)
+charpoly_bound(GEN M, GEN dM)
 {
   pari_sp av = avma;
-  GEN s = real_0(LOWDEFAULTPREC), bin, B2 = itor(sqri(ZM_supnorm(M)), LOWDEFAULTPREC);
+  GEN B = itor(ZM_supnorm(M), LOWDEFAULTPREC);
+  GEN s = real_0(LOWDEFAULTPREC), bin, B2;
   long n = lg(M)-1, k;
   double d;
   bin = gen_1;
+  if (dM) B = divri(B, dM);
+  B2 = sqrr(B);
   for (k = n; k >= (n+1)>>1; k--)
   {
     GEN t = mulri(powruhalf(mulur(k, B2), k), bin);
-    if (absr_cmp(t, s) > 0) s = t;
+    if (abscmprr(t, s) > 0) s = t;
     bin = diviuexact(muliu(bin, k), n-k+1);
   }
   d = dbllog2(s); avma = av; return ceil(d);
 }
 
-GEN
-ZM_charpoly(GEN M)
+/* Return char_{M/d}(X) = d^(-n) char_M(dX) modulo p. Assume dp = d mod p. */
+static GEN
+QM_charpoly_Flx(GEN M, ulong dp, ulong p)
 {
-  pari_timer T;
   pari_sp av = avma;
-  long l = lg(M), n = l-1, bit;
-  GEN q = NULL, H = NULL, Hp;
+  GEN H = Flm_charpoly_i(ZM_to_Flm(M,p), p);
+  if (dp) H = Flx_rescale(H, Fl_inv(dp,p), p);
+  return gerepileuptoleaf(av, H);
+}
+
+static int
+ZX_CRT(GEN *H, GEN Hp, GEN *q, ulong p, long bit)
+{
+  if (!*H)
+  {
+    *H = ZX_init_CRT(Hp, p, 0);
+    if (DEBUGLEVEL>5)
+      err_printf("charpoly mod %lu, bound = 2^%ld\n", p, expu(p));
+    if (expu(p) > bit) return 1;
+    *q = utoipos(p);
+  }
+  else
+  {
+    int stable = ZX_incremental_CRT(H, Hp, q,p);
+    if (DEBUGLEVEL>5)
+      err_printf("charpoly mod %lu (stable=%ld), bound = 2^%ld\n",
+                 p, stable, expi(*q));
+    if (stable && expi(*q) > bit) return 1;
+  }
+  return 0;
+}
+
+/* Assume M a square ZM, dM integer. Return charpoly(M / dM) in Z[X] */
+static GEN
+QM_charpoly_ZX_i(GEN M, GEN dM, long bit)
+{
+  long n = lg(M)-1;
+  GEN q = NULL, H = NULL;
   forprime_t S;
   ulong p;
   if (!n) return pol_1(0);
 
-  bit = (long)charpoly_bound(M) + 1;
-  if (DEBUGLEVEL>5) {
-    err_printf("ZM_charpoly: bit-bound 2^%ld\n", bit);
-    timer_start(&T);
-  }
-  init_modular(&S);
+  if (bit < 0) bit = (long)charpoly_bound(M, dM) + 1;
+  if (DEBUGLEVEL>5) err_printf("ZM_charpoly: bit-bound 2^%ld\n", bit);
+  init_modular_big(&S);
   while ((p = u_forprime_next(&S)))
   {
-    Hp = Flm_charpoly(ZM_to_Flm(M, p), p);
-    if (!H)
-    {
-      H = ZX_init_CRT(Hp, p, 0);
-      if (DEBUGLEVEL>5)
-        timer_printf(&T, "charpoly mod %lu, bound = 2^%ld", p, expu(p));
-      if (expu(p) > bit) break;
-      q = utoipos(p);
-    }
-    else
-    {
-      int stable = ZX_incremental_CRT(&H, Hp, &q,p);
-      if (DEBUGLEVEL>5)
-        timer_printf(&T, "charpoly mod %lu (stable=%ld), bound = 2^%ld",
-                     p, stable, expi(q));
-      if (stable && expi(q) > bit) break;
-    }
+    ulong dMp = 0;
+    GEN Hp;
+    if (dM && !(dMp = umodiu(dM, p))) continue;
+    Hp = QM_charpoly_Flx(M, dMp, p);
+    if (ZX_CRT(&H, Hp, &q,p, bit)) break;
   }
-  if (!p) pari_err_OVERFLOW("ZM_charpoly [ran out of primes]");
-  return gerepilecopy(av, H);
+  if (!p) pari_err_OVERFLOW("charpoly [ran out of primes]");
+  return H;
+}
+GEN
+QM_charpoly_ZX_bound(GEN M, long bit)
+{
+  pari_sp av = avma;
+  GEN dM; M = Q_remove_denom(M, &dM);
+  return gerepilecopy(av, QM_charpoly_ZX_i(M, dM, bit));
+}
+GEN
+QM_charpoly_ZX(GEN M)
+{
+  pari_sp av = avma;
+  GEN dM; M = Q_remove_denom(M, &dM);
+  return gerepilecopy(av, QM_charpoly_ZX_i(M, dM, -1));
+}
+GEN
+ZM_charpoly(GEN M)
+{
+  pari_sp av = avma;
+  return gerepilecopy(av, QM_charpoly_ZX_i(M, NULL, -1));
 }
 
 /*******************************************************************/
@@ -625,9 +975,9 @@ carberkowitz(GEN x, long v)
 {
   long lx, i, j, k, r;
   GEN V, S, C, Q;
-  pari_sp av0, av, lim;
+  pari_sp av0, av;
   if ((V = easychar(x,v))) return V;
-  lx = lg(x); av0 = avma; lim = stack_lim(av0,1);
+  lx = lg(x); av0 = avma;
   V = cgetg(lx+1, t_VEC);
   S = cgetg(lx+1, t_VEC);
   C = cgetg(lx+1, t_VEC);
@@ -660,7 +1010,7 @@ carberkowitz(GEN x, long v)
     av2 = avma; t = gmul(gcoeff(x,r,1), gel(S,1));
     for (j = 2; j < r; j++) t = gadd(t, gmul(gcoeff(x,r,j), gel(S,j)));
     gel(C,r+1) = gerepileupto(av2, t);
-    if (low_stack(lim, stack_lim(av0,1)))
+    if (gc_needed(av0,1))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"carberkowitz");
       gerepileall(av, 2, &C, &V);
@@ -735,7 +1085,7 @@ cxquadnorm(GEN q, long prec)
 static GEN
 gnorml2_i(GEN x, long prec)
 {
-  pari_sp av, lim;
+  pari_sp av;
   long i, lx;
   GEN s;
 
@@ -757,12 +1107,12 @@ gnorml2_i(GEN x, long prec)
       return NULL; /* not reached */
   }
   if (lx == 1) return gen_0;
-  av = avma; lim = stack_lim(av,1);
+  av = avma;
   s = gnorml2(gel(x,1));
   for (i=2; i<lx; i++)
   {
     s = gadd(s, gnorml2(gel(x,i)));
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,1))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"gnorml2");
       s = gerepileupto(av, s);
@@ -777,13 +1127,13 @@ static GEN pnormlp(GEN,GEN,long);
 static GEN
 pnormlpvec(long i0, GEN x, GEN p, long prec)
 {
-  pari_sp av = avma, lim = stack_lim(av,1);
+  pari_sp av = avma;
   long i, lx = lg(x);
   GEN s = gen_0;
   for (i=i0; i<lx; i++)
   {
     s = gadd(s, pnormlp(gel(x,i),p,prec));
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,1))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"gnormlp, i = %ld", i);
       s = gerepileupto(av, s);
@@ -811,7 +1161,8 @@ GEN
 gnormlp(GEN x, GEN p, long prec)
 {
   pari_sp av = avma;
-  if (!p) return gsupnorm(x, prec);
+  if (!p || (typ(p) == t_INFINITY && inf_get_sign(p) > 0))
+    return gsupnorm(x, prec);
   if (gsigne(p) <= 0) pari_err_DOMAIN("normlp", "p", "<=", gen_0, p);
   if (is_scalar_t(typ(x))) return gabs(x, prec);
   if (typ(p) == t_INT)
@@ -1001,6 +1352,7 @@ gtrace(GEN x)
       return normalizepol_lg(y, lx);
 
     case t_SER:
+      if (ser_isexactzero(x)) return gcopy(x);
       y = cgetg_copy(x, &lx); y[1] = x[1];
       for (i=2; i<lx; i++) gel(y,i) = gtrace(gel(x,i));
       return normalize(y);
@@ -1042,7 +1394,7 @@ gtrace(GEN x)
 GEN
 qfgaussred_positive(GEN a)
 {
-  pari_sp av = avma, lim=stack_lim(av,1);
+  pari_sp av = avma;
   GEN b;
   long i,j,k, n = lg(a);
 
@@ -1071,7 +1423,7 @@ qfgaussred_positive(GEN a)
       for (j=i; j<n; j++)
         gcoeff(b,i,j) = gsub(gcoeff(b,i,j), gmul(c,gcoeff(b,k,j)));
     }
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,1))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"qfgaussred_positive");
       b=gerepilecopy(av,b);
@@ -1093,7 +1445,7 @@ suitable(GEN x, long k, GEN *pp, long *pi)
     case t_FRAC: return 1;
     case t_REAL: {
       GEN p = *pp;
-      if (signe(x) && (!p || absr_cmp(p, x) < 0)) { *pp = x; *pi = k; }
+      if (signe(x) && (!p || abscmprr(p, x) < 0)) { *pp = x; *pi = k; }
       return 0;
     }
     default: return !gequal0(x);
@@ -1107,7 +1459,7 @@ static GEN
 gaussred(GEN a, long signature)
 {
   GEN r, ak, al;
-  pari_sp av, av1, lim;
+  pari_sp av, av1;
   long n = lg(a), i, j, k, l, sp, sn, t;
 
   if (typ(a) != t_MAT) pari_err_TYPE("gaussred",a);
@@ -1117,7 +1469,7 @@ gaussred(GEN a, long signature)
 
   av = avma;
   r = const_vecsmall(n, 1);
-  av1= avma; lim = stack_lim(av1,1);
+  av1= avma;
   a = RgM_shallowcopy(a);
   t = n; sp = sn = 0;
   while (t)
@@ -1145,6 +1497,11 @@ gaussred(GEN a, long signature)
           gcoeff(a,i,j) = gsub(gcoeff(a,i,j), gmul(c,gcoeff(a,k,j)));
       }
       gcoeff(a,k,k) = p;
+      if (gc_needed(av1,1))
+      {
+        if(DEBUGMEM>1) pari_warn(warnmem,"gaussred (t = %ld)", t);
+        a = gerepilecopy(av1, a);
+      }
     }
     else
     { /* all remaining diagonal coeffs are currently 0 */
@@ -1186,7 +1543,7 @@ gaussred(GEN a, long signature)
         gcoeff(a,l,k) = gen_m1;
         gcoeff(a,k,k) = gmul2n(p,-1);
         gcoeff(a,l,l) = gneg(gcoeff(a,k,k));
-        if (low_stack(lim, stack_lim(av1,1)))
+        if (gc_needed(av1,1))
         {
           if(DEBUGMEM>1) pari_warn(warnmem,"gaussred");
           a = gerepilecopy(av1, a);
@@ -1269,7 +1626,7 @@ jacobi(GEN a, long prec)
   {
     pari_sp av2 = avma;
     GEN x, y, t, c, s, u;
-    /* compute associated rotation in the plane formed by basis vectors number
+    /* compute attached rotation in the plane formed by basis vectors number
      * p and q */
     x = subrr(gel(L,q),gel(L,p));
     if (signe(x))
@@ -1333,15 +1690,15 @@ matrixqz0(GEN x,GEN p)
   if (!p) return QM_minors_coprime(x,NULL);
   if (typ(p) != t_INT) pari_err_TYPE("QM_minors_coprime",p);
   if (signe(p)>=0) return QM_minors_coprime(x,p);
-  if (equaliu(p,1)) return QM_ImZ_hnf(x); /* p = -1 */
-  if (equaliu(p,2)) return QM_ImQ_hnf(x); /* p = -2 */
+  if (absequaliu(p,1)) return QM_ImZ_hnf(x); /* p = -1 */
+  if (absequaliu(p,2)) return QM_ImQ_hnf(x); /* p = -2 */
   pari_err_FLAG("QM_minors_coprime"); return NULL; /* not reached */
 }
 
 GEN
 QM_minors_coprime(GEN x, GEN D)
 {
-  pari_sp av = avma, av1, lim;
+  pari_sp av = avma, av1;
   long i, j, m, n, lP;
   GEN P, y;
 
@@ -1369,7 +1726,7 @@ QM_minors_coprime(GEN x, GEN D)
     if (is_pm1(D)) { avma = av2; return ZM_copy(x); }
   }
   P = gel(Z_factor(D), 1); lP = lg(P);
-  av1 = avma; lim = stack_lim(av1,1);
+  av1 = avma;
   for (i=1; i < lP; i++)
   {
     GEN p = gel(P,i), pov2 = shifti(p, -1);
@@ -1379,14 +1736,14 @@ QM_minors_coprime(GEN x, GEN D)
       long lM = lg(M);
       if (lM==1) break;
 
-      M = FpM_center(M, p, pov2);
+      FpM_center_inplace(M, p, pov2);
       N = ZM_Z_divexact(ZM_mul(x,M), p);
       for (j=1; j<lM; j++)
       {
         long k = n; while (!signe(gcoeff(M,k,j))) k--;
         gel(x,k) = gel(N,j);
       }
-      if (low_stack(lim, stack_lim(av1,1)))
+      if (gc_needed(av1,1))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"QM_minors_coprime, p = %Ps", p);
         x = gerepilecopy(av1, x); pov2 = shifti(p, -1);
@@ -1461,9 +1818,8 @@ QC_elem(GEN aj, GEN ak, GEN A, long j, long k)
 static GEN
 QM_imZ_hnf_aux(GEN A)
 {
-  pari_sp av = avma, lim = stack_lim(av,1);
+  pari_sp av = avma;
   long i,j,k,n,m;
-  GEN a;
 
   n = lg(A);
   if (n == 1) return cgetg(1,t_MAT);
@@ -1476,6 +1832,7 @@ QM_imZ_hnf_aux(GEN A)
   m = lgcols(A);
   for (i=1; i<m; i++)
   {
+    GEN b;
     for (j = k = 1; j<n; j++)
     {
       GEN a = gcoeff(A,i,j);
@@ -1485,13 +1842,13 @@ QM_imZ_hnf_aux(GEN A)
       /* zero a = Aij  using  b = Aik */
       QC_elem(a, gcoeff(A,i,k), A, j,k);
     }
-    a = gcoeff(A,i,k);
-    if (!gequal0(a))
+    b = gcoeff(A,i,k);
+    if (!gequal0(b))
     {
-      a = Q_denom(a);
-      if (!is_pm1(a)) gel(A,k) = RgC_Rg_mul(gel(A,k), a);
+      b = Q_denom(b);
+      if (!is_pm1(b)) gel(A,k) = RgC_Rg_mul(gel(A,k), b);
     }
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"QM_imZ_hnf_aux");
       A = gerepilecopy(av,A);
@@ -1510,14 +1867,14 @@ QM_ImZ_hnf(GEN x)
 GEN
 QM_ImQ_hnf(GEN x)
 {
-  pari_sp av = avma, av1, lim;
+  pari_sp av = avma, av1;
   long j,j1,k,m,n;
   GEN c;
 
   n = lg(x); if (n==1) return gcopy(x);
   m = lgcols(x); x = RgM_shallowcopy(x);
   c = zero_zv(n-1);
-  av1 = avma; lim = stack_lim(av1,1);
+  av1 = avma;
   for (k=1; k<m; k++)
   {
     j=1; while (j<n && (c[j] || gequal0(gcoeff(x,k,j)))) j++;
@@ -1530,7 +1887,7 @@ QM_ImQ_hnf(GEN x)
         GEN t = gcoeff(x,k,j1);
         if (!gequal0(t)) gel(x,j1) = RgC_sub(gel(x,j1), RgC_Rg_mul(gel(x,j),t));
       }
-    if (low_stack(lim, stack_lim(av1,1)))
+    if (gc_needed(av1,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"QM_ImQ_hnf");
       x = gerepilecopy(av1,x);
