@@ -29,7 +29,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 GEN
 vecsum(GEN v)
 {
-  pari_sp av = avma, lim;
+  pari_sp av = avma;
   long i, l;
   GEN p;
   if (!is_vec_t(typ(v)))
@@ -38,11 +38,10 @@ vecsum(GEN v)
   if (l == 1) return gen_0;
   p = gel(v,1);
   if (l == 2) return gcopy(p);
-  lim = stack_lim(av, 2);
   for (i=2; i<l; i++)
   {
     p = gadd(p, gel(v,i));
-    if (low_stack(lim, stack_lim(av, 2)))
+    if (gc_needed(av, 2))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"sum");
       p = gerepileupto(av, p);
@@ -351,21 +350,21 @@ static long
 vecslice_parse_arg(long lA, long *y1, long *y2, long *skip)
 {
   *skip=0;
-  if (!*y1)
+  if (*y1==LONG_MAX)
   {
-    if (*y2)
+    if (*y2!=LONG_MAX)
     {
       if (*y2<0) *y2 += lA;
-      if (*y2<=0 || *y2>=lA)
+      if (*y2<0 || *y2==LONG_MAX || *y2>=lA)
         pari_err_DIM("_[..]");
       *skip=*y2;
     }
     *y1 = 1; *y2 = lA-1;
   }
-  else if (!*y2) *y2 = *y1;
-  if (*y1<0) *y1 += lA;
+  else if (*y2==LONG_MAX) *y2 = *y1;
+  if (*y1<=0) *y1 += lA;
   if (*y2<0) *y2 += lA;
-  if (*y1<=0 || *y1>*y2 || *y2>=lA) pari_err_DIM("_[..]");
+  if (*y1<=0 || *y1>*y2+1 || *y2>=lA) pari_err_DIM("_[..]");
   return *y2 - *y1 + 2 - !!*skip;
 }
 
@@ -442,7 +441,8 @@ matslice0(GEN A, long x1, long x2, long y1, long y2)
 {
   GEN B;
   long i, lB, lA = lg(A), t, skip, rskip, rlB;
-  long is_col = y1 && !y2, is_row = x1 && !x2;
+  long is_col = y1!=LONG_MAX && y2==LONG_MAX;
+  long is_row = x1!=LONG_MAX && x2==LONG_MAX;
   GEN (*slice)(GEN A, long t, long lB, long y1, long skip);
   if (typ(A)!=t_MAT) pari_err_TYPE("_[_.._,_.._]",A);
   lB = vecslice_parse_arg(lA, &y1, &y2, &skip);
@@ -549,14 +549,14 @@ genselect(void *E, long (*f)(void* E, GEN x), GEN A)
   {
     case t_LIST:
       z = list_data(A);
-      if (!z) y = listcreate();
+      if (!z) y = mklist();
       else
       {
         GEN B;
         y = cgetg(3, t_LIST);
         v = genindexselect(E, f, z);
         B = extract_copy(z, v);
-        list_nmax(y) = lg(B)-1;
+        y[1] = lg(B)-1;
         list_data(y) = B;
       }
       break;
@@ -571,10 +571,17 @@ genselect(void *E, long (*f)(void* E, GEN x), GEN A)
   clone_unlock(A); return y;
 }
 
+static void
+check_callgen1(GEN f, const char *s)
+{
+  if (typ(f) != t_CLOSURE || closure_is_variadic(f)  || closure_arity(f) < 1)
+    pari_err_TYPE(s, f);
+}
+
 GEN
 select0(GEN f, GEN x, long flag)
 {
-  if (typ(f) != t_CLOSURE || closure_arity(f) < 1) pari_err_TYPE("select", f);
+  check_callgen1(f, "select");
   switch(flag)
   {
     case 0: return genselect((void *) f, gp_callbool, x);
@@ -591,7 +598,7 @@ parselect(GEN C, GEN D, long flag)
   long lv, l = lg(D), i, pending = 0, workid;
   GEN V, worker, done;
   struct pari_mt pt;
-  if (typ(C) != t_CLOSURE || closure_arity(C) < 1) pari_err_TYPE("parapply",C);
+  check_callgen1(C, "parselect");
   if (!is_vec_t(typ(D))) pari_err_TYPE("parapply",D);
   V = cgetg(l, t_VECSMALL); av = avma;
   worker = strtoclosure("_parapply_worker", 1, C);
@@ -636,6 +643,19 @@ vecapply1(void *E, GEN (*f)(void* E, GEN x), GEN x)
   for (i=1; i<lx; i++) gel(y,i) = f(E, gel(x,i));
   return y;
 }
+static GEN
+mapapply1(void *E, GEN (*f)(void* E, GEN x), GEN x)
+{
+  long i, lx;
+  GEN y = cgetg_copy(x, &lx);
+  for (i=1; i<lx; i++)
+  {
+    GEN xi = gel(x, i);
+    gel(y,i) = mkvec2(mkvec2(gcopy(gmael(xi, 1, 1)), f(E,gmael(xi, 1, 2))),
+                             gcopy(gel(xi, 2)));
+  }
+  return y;
+}
 /* as genapply, but treat A [ t_VEC,t_COL, or t_MAT] as a t_VEC */
 GEN
 vecapply(void *E, GEN (*f)(void* E, GEN x), GEN x)
@@ -653,16 +673,29 @@ genapply(void *E, GEN (*f)(void* E, GEN x), GEN x)
   clone_lock(x);
   switch(tx) {
     case t_POL: y = normalizepol(vecapply2(E,f,x)); break;
-    case t_SER: y = normalize(vecapply2(E,f,x)); break;
+    case t_SER:
+      y = ser_isexactzero(x)? gcopy(x): normalize(vecapply2(E,f,x));
+      break;
     case t_LIST:
-      z = list_data(x);
-      if (!z)
-        y = listcreate();
-      else
       {
-        y = cgetg(3, t_LIST);
-        list_nmax(y) = lg(z)-1;
-        list_data(y) = vecapply1(E,f,z);
+        long t = list_typ(x);
+        z = list_data(x);
+        if (!z)
+          y = mklist_typ(t);
+        else
+        {
+          y = cgetg(3, t_LIST);
+          y[1] = evaltyp(t)|evallg(lg(z)-1);
+          switch(t)
+          {
+          case t_LIST_RAW:
+            list_data(y) = vecapply1(E,f,z);
+            break;
+          case t_LIST_MAP:
+            list_data(y) = mapapply1(E,f,z);
+            break;
+          }
+        }
       }
       break;
     case t_MAT:
@@ -680,7 +713,7 @@ genapply(void *E, GEN (*f)(void* E, GEN x), GEN x)
 GEN
 apply0(GEN f, GEN x)
 {
-  if (typ(f) != t_CLOSURE || closure_arity(f) < 1) pari_err_TYPE("apply",f);
+  check_callgen1(f, "apply");
   return genapply((void *) f, gp_call, x);
 }
 
@@ -718,7 +751,7 @@ parapply(GEN C, GEN D)
   long l = lg(D), i, pending = 0, workid;
   GEN V, worker, done;
   struct pari_mt pt;
-  if (typ(C) != t_CLOSURE || closure_arity(C) < 1) pari_err_TYPE("parapply",C);
+  check_callgen1(C, "parapply");
   if (!is_vec_t(typ(D))) pari_err_TYPE("parapply",D);
   worker = strtoclosure("_parapply_worker", 1, C);
   V = cgetg(l, typ(D));
@@ -733,6 +766,34 @@ parapply(GEN C, GEN D)
   return gerepilecopy(av, V);
 }
 
+GEN
+genfold(void *E, GEN (*f)(void* E, GEN x, GEN y), GEN x)
+{
+  pari_sp av = avma;
+  GEN z;
+  long i, l = lg(x);
+  if (!is_vec_t(typ(x))|| l==1  ) pari_err_TYPE("fold",x);
+  clone_lock(x);
+  z = gel(x,1);
+  for (i=2; i<l; i++)
+  {
+    z = f(E,z,gel(x,i));
+    if (gc_needed(av, 2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"fold");
+      z = gerepilecopy(av, z);
+    }
+  }
+  clone_unlock(x);
+  return gerepilecopy(av, z);
+}
+
+GEN
+fold0(GEN f, GEN x)
+{
+  if (typ(f) != t_CLOSURE || closure_arity(f) < 2) pari_err_TYPE("apply",f);
+  return genfold((void *) f, gp_call2, x);
+}
 /*******************************************************************/
 /*                                                                 */
 /*                     SCALAR-MATRIX OPERATIONS                    */
@@ -748,6 +809,8 @@ gtomat(GEN x)
   switch(typ(x))
   {
     case t_LIST:
+      if (list_typ(x)==t_LIST_MAP)
+        return maptomat(x);
       x = list_data(x);
       if (!x) return cgetg(1, t_MAT);
       /* fall through */
@@ -772,7 +835,7 @@ gtomat(GEN x)
       lx = lg(x);
       if (lx == 1) return cgetg(1, t_MAT);
       if (typ(gel(x,1)) == t_VEC) {
-        long j, h = lgcols(x);
+        long j, h = lg(gel(x,1));
         for (i=2; i<lx; i++) {
           if (typ(gel(x,i)) != t_VEC || lg(gel(x,i)) != h) break;
         }
