@@ -221,11 +221,10 @@ R_from_QR(GEN x, long prec)
 /**             QR Factorization via Gram-Schmidt                  **/
 /********************************************************************/
 
-#if 0
-/* return Gram-Schmidt orthogonal basis (f) associated to (e), B is the
+/* return Gram-Schmidt orthogonal basis (f) attached to (e), B is the
  * vector of the (f_i . f_i)*/
 GEN
-gram_schmidt(GEN e, GEN *ptB)
+RgM_gram_schmidt(GEN e, GEN *ptB)
 {
   long i,j,lx = lg(e);
   GEN f = RgM_shallowcopy(e), B, iB;
@@ -250,7 +249,27 @@ gram_schmidt(GEN e, GEN *ptB)
   }
   *ptB = B; return f;
 }
-#endif
+
+/* Assume B an LLL-reduced basis, t a vector. Apply Babai's nearest plane
+ * algorithm to (B,t) */
+GEN
+RgM_Babai(GEN B, GEN t)
+{
+  GEN C, N, G = RgM_gram_schmidt(B, &N), b = t;
+  long j, n = lg(B)-1;
+
+  C = cgetg(n+1,t_COL);
+  for (j = n; j > 0; j--)
+  {
+    GEN c = gdiv( RgV_dotproduct(b, gel(G,j)), gel(N,j) );
+    long e;
+    c = grndtoi(c,&e);
+    if (e >= 0) return NULL;
+    if (signe(c)) b = RgC_sub(b, RgC_Rg_mul(gel(G,j), c));
+    gel(C,j) = c;
+  }
+  return C;
+}
 
 /********************************************************************/
 /**                                                                **/
@@ -383,7 +402,7 @@ lllintpartialall(GEN m, long flag)
    /* For each pair of column vectors v and w in mid * tm2,
     * try to replace (v, w) by (v, v - q*w) for some q.
     * We compute all inner products and check them repeatedly. */
-    const pari_sp av3 = avma, lim = stack_lim(av3,2);
+    const pari_sp av3 = avma;
     long i, j, npass = 0, e = LONG_MAX;
     GEN dot = cgetg(ncol+1, t_MAT); /* scalar products */
 
@@ -421,7 +440,7 @@ lllintpartialall(GEN m, long flag)
                                     mulii(q, gcoeff(dot,k2,k1)));
           for (d = 1; d <= ncol; d++) gcoeff(dot,k1,d) = gcoeff(dot,d,k1);
         } /* for ijdif */
-        if (low_stack(lim, stack_lim(av3,2)))
+        if (gc_needed(av3,2))
         {
           if(DEBUGMEM>1) pari_warn(warnmem,"lllintpartialall");
           gerepileall(av3, 2, &dot,&tm2);
@@ -453,13 +472,6 @@ lllintpartialall(GEN m, long flag)
         swap(gcoeff(dot,i,i), gcoeff(dot,s,s));
       }
     }
-    i = 1;
-    while (i <= ncol && !signe(gcoeff(dot,i,i))) i++;
-    if (i > 1)
-    {
-      tm2 += (i - 1);
-      tm2[0] = evaltyp(t_MAT)|evallg(ncol - i);
-    }
   } /* local block */
   return gerepileupto(av, ZM_mul(tm1? tm1: mid, tm2));
 }
@@ -495,7 +507,7 @@ static void
 choose_params(GEN P, GEN N, GEN X, GEN B, long *pdelta, long *pt)
 {
   long d = degpol(P);
-  GEN P0 = leading_term(P);
+  GEN P0 = leading_coeff(P);
   double logN = gtodouble(glog(N, DEFAULTPREC));
   double tau, beta, rho;
   long delta, t;
@@ -565,7 +577,7 @@ zncoppersmith(GEN P0, GEN N, GEN X, GEN B)
   if (d < 0) pari_err_ROOTS0("zncoppersmith");
   if (B && typ(B) != t_INT) B = gceil(B);
 
-  if (cmpiu(X, X_SMALL) <= 0)
+  if (abscmpiu(X, X_SMALL) <= 0)
     return gerepileupto(av, do_exhaustive(P0, N, itos(X), B));
 
   if (B && equalii(B,N)) B = NULL;
@@ -801,7 +813,9 @@ Xadic_lindep(GEN x)
   if (lx == 1) return cgetg(1,t_COL);
   vx = gvar(x);
   v = gvaluation(x, pol_x(vx));
-  if (v) x = gmul(x, monomial(gen_1, -v, vx)); else x = shallowcopy(x);
+  if (!v)         x = shallowcopy(x);
+  else if (v > 0) x = gdiv(x, pol_xn(v, vx));
+  else            x = gmul(x, pol_xn(-v, vx));
   /* all t_SER have valuation >= 0 */
   for (i=1; i<lx; i++)
   {
@@ -820,6 +834,22 @@ Xadic_lindep(GEN x)
   m = RgXV_to_RgM(x, prec);
   return gerepileupto(av, deplin(m));
 }
+static GEN
+vec_lindep(GEN x)
+{
+  pari_sp av = avma;
+  long i, l = lg(x); /* > 1 */
+  long t = typ(gel(x,1)), h = lg(gel(x,1));
+  GEN m = cgetg(l, t_MAT);
+  for (i = 1; i < l; i++)
+  {
+    GEN y = gel(x,i);
+    if (lg(y) != h || typ(y) != t) pari_err_TYPE("lindep",x);
+    if (t != t_COL) y = shallowtrans(y); /* Sigh */
+    gel(m,i) = y;
+  }
+  return gerepileupto(av, deplin(m));
+}
 
 GEN
 lindep(GEN x) { return lindep2(x, 0); }
@@ -837,6 +867,8 @@ lindep0(GEN x,long bit)
       case t_POL:
       case t_RFRAC:
       case t_SER: return Xadic_lindep(x);
+      case t_VEC:
+      case t_COL: return vec_lindep(x);
     }
   return lindep2(x, bit);
 }
@@ -867,7 +899,7 @@ algdep0(GEN x, long n, long bit)
     y = lindep2(y, bit);
   if (lg(y) == 1) pari_err(e_DOMAIN,"algdep", "degree(x)",">", stoi(n), x);
   y = RgV_to_RgX(y, 0);
-  if (signe(leading_term(y)) > 0) return gerepilecopy(av, y);
+  if (signe(leading_coeff(y)) > 0) return gerepilecopy(av, y);
   return gerepileupto(av, ZX_neg(y));
 }
 
@@ -897,8 +929,7 @@ seralgdep(GEN s, long p, long r)
   for (i = 2; i <= p; i++) gel(S,i) = gmul(gel(S,i-1), s);
   v = cgetg(r*p+1, t_VEC); /* v[r*n+m+1] = s^n * y^m */
   /* n = 0 */
-  for (m = 0; m < r; m++)
-    gel(v, m + 1) = monomial(gen_1, m, vy);
+  for (m = 0; m < r; m++) gel(v, m+1) = pol_xn(m, vy);
   for(n=1; n < p; n++)
     for (m = 0; m < r; m++)
     {
@@ -938,38 +969,18 @@ minim_alloc(long n, double ***q, GEN *x, double **y,  double **z, double **v)
   for (i=1; i<n; i++) (*q)[i] = dalloc(s);
 }
 
-/* If V depends linearly from the columns of the matrix, return 0.
- * Otherwise, update INVP and L and return 1. No GC. */
-static int
-addcolumntomatrix(GEN V, GEN invp, GEN L)
+static GEN
+ZC_canon(GEN V)
 {
-  GEN a = RgM_zc_mul(invp,V);
-  long i,j,k, n = lg(invp);
+  long l = lg(V), j;
+  for (j = 1; j < l  &&  signe(gel(V,j)) == 0; ++j);
+  return (j < l  &&  signe(gel(V,j)) < 0)? ZC_neg(V): V;
+}
 
-  if (DEBUGLEVEL>6)
-  {
-    err_printf("adding vector = %Ps\n",V);
-    err_printf("vector in new basis = %Ps\n",a);
-    err_printf("list = %Ps\n",L);
-    err_printf("base change matrix =\n%Ps\n", invp);
-  }
-  k = 1; while (k<n && (L[k] || gequal0(gel(a,k)))) k++;
-  if (k == n) return 0;
-  L[k] = 1;
-  for (i=k+1; i<n; i++) gel(a,i) = gdiv(gneg_i(gel(a,i)),gel(a,k));
-  for (j=1; j<=k; j++)
-  {
-    GEN c = gel(invp,j), ck = gel(c,k);
-    if (gequal0(ck)) continue;
-    gel(c,k) = gdiv(ck, gel(a,k));
-    if (j==k)
-      for (i=k+1; i<n; i++)
-        gel(c,i) = gmul(gel(a,i), ck);
-    else
-      for (i=k+1; i<n; i++)
-        gel(c,i) = gadd(gel(c,i), gmul(gel(a,i), ck));
-  }
-  return 1;
+static GEN
+ZM_zc_mul_canon(GEN u, GEN x)
+{
+  return ZC_canon(ZM_zc_mul(u,x));
 }
 
 struct qfvec
@@ -984,17 +995,24 @@ err_minim(GEN a)
                   strtoGENstr("positive definite"),a);
 }
 
+static GEN
+minim_lll(GEN a, GEN *u)
+{
+  *u = lllgramint(a);
+  if (lg(*u) != lg(a)) err_minim(a);
+  return qf_apply_ZM(a,*u);
+}
+
 static void
 forqfvec_init_dolll(struct qfvec *qv, GEN a, long dolll)
 {
   GEN r, u;
-  if (typ(a) != t_MAT || !RgM_is_ZM(a)) pari_err_TYPE("qfminim",a);
-  if (dolll)
+  if (!dolll) u = NULL;
+  else
   {
-    u = lllgramint(a);
-    if (lg(u) != lg(a)) err_minim(a);
-    a = qf_apply_ZM(a,u);
-  } else u = NULL;
+    if (typ(a) != t_MAT || !RgM_is_ZM(a)) pari_err_TYPE("qfminim",a);
+    a = minim_lll(a, &u);
+  }
   qv->a = RgM_gtofp(a, DEFAULTPREC);
   r = qfgaussred_positive(qv->a);
   if (!r)
@@ -1009,12 +1027,10 @@ forqfvec_init_dolll(struct qfvec *qv, GEN a, long dolll)
 
 static void
 forqfvec_init(struct qfvec *qv, GEN a)
-{
-  forqfvec_init_dolll(qv, a, 1);
-}
+{ forqfvec_init_dolll(qv, a, 1); }
 
 static void
-forqfvec(void *E, long (*fun)(void *, GEN, GEN, double), struct qfvec *qv, GEN BORNE)
+forqfvec_i(void *E, long (*fun)(void *, GEN, GEN, double), struct qfvec *qv, GEN BORNE)
 {
   GEN x, a = qv->a, r = qv->r, u = qv->u;
   long n = lg(a), i, j, k;
@@ -1077,12 +1093,22 @@ forqfvec(void *E, long (*fun)(void *, GEN, GEN, double), struct qfvec *qv, GEN B
   }
 }
 
+void
+forqfvec(void *E, long (*fun)(void *, GEN, GEN, double), GEN a, GEN BORNE)
+{
+  pari_sp av = avma;
+  struct qfvec qv;
+  forqfvec_init(&qv, a);
+  forqfvec_i(E, fun, &qv, BORNE);
+  avma = av;
+}
+
 static long
 _gp_forqf(void *E, GEN u, GEN x, double p/*unused*/)
 {
   pari_sp av = avma;
   (void)p;
-  set_lex(-1, ZM_zc_mul(u, x));
+  set_lex(-1, ZM_zc_mul_canon(u, x));
   closure_evalvoid((GEN)E);
   avma = av;
   return loop_break();
@@ -1095,30 +1121,32 @@ forqfvec0(GEN a, GEN BORNE, GEN code)
   struct qfvec qv;
   forqfvec_init(&qv, a);
   push_lex(gen_0, code);
-  forqfvec((void*) code, &_gp_forqf, &qv, BORNE);
+  forqfvec_i((void*) code, &_gp_forqf, &qv, BORNE);
   pop_lex(1);
   avma = av;
 }
 
+enum { min_ALL = 0, min_FIRST, min_VECSMALL, min_VECSMALL2 };
+
 /* Minimal vectors for the integral definite quadratic form: a.
  * Result u:
  *   u[1]= Number of vectors of square norm <= BORNE
  *   u[2]= maximum norm found
- *   u[3]= list of vectors found (at most STOCKMAX)
+ *   u[3]= list of vectors found (at most STOCKMAX, unless NULL)
  *
- *  If BORNE = gen_0: Minimal non-zero vectors.
+ *  If BORNE = NULL: Minimal non-zero vectors.
  *  flag = min_ALL,   as above
  *  flag = min_FIRST, exits when first suitable vector is found.
- *  flag = min_PERF,  only compute rank of the family of v.v~ (v min.)
- *  flag = min_VECSMALL, return a t_VECSMALL of (half) the number of vectors for each norm
- *  flag = min_VECSMALL2, same but count only vectors with even norm, and shift the answer
- */
+ *  flag = min_VECSMALL, return a t_VECSMALL of (half) the number of vectors
+ *  for each norm
+ *  flag = min_VECSMALL2, same but count only vectors with even norm, and shift
+ *  the answer */
 static GEN
 minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
 {
-  GEN x, u, r, L, gnorme, invp, V;
+  GEN x, u, r, L, gnorme;
   long n = lg(a), i, j, k, s, maxrank, sBORNE;
-  pari_sp av = avma, av1, lim;
+  pari_sp av = avma, av1;
   double p,maxnorm,BOUND,*v,*y,*z,**q;
   const double eps = 1e-10;
   int stockall = 0;
@@ -1147,7 +1175,6 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
       pari_err_TYPE("minim0 [negative number of vectors]",STOCKMAX);
   }
 
-  L = V = invp = NULL; /* gcc -Wall */
   switch(flag)
   {
     case min_VECSMALL:
@@ -1158,31 +1185,32 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
       if (n == 1) return L;
       break;
     case min_FIRST:
-      if (n == 1) return cgetg(1,t_VEC);
-      if (!sBORNE && BORNE) return cgetg(1, t_VEC);
+      if (n == 1 || (!sBORNE && BORNE)) return cgetg(1,t_VEC);
+      L = NULL; /* gcc -Wall */
       break;
-    case min_PERF:
-      if (n == 1) return gen_0;
-      break;
-    default:
+    case min_ALL:
       if (n == 1 || (!sBORNE && BORNE))
         retmkvec3(gen_0, gen_0, cgetg(1, t_MAT));
+      L = new_chunk(1+maxrank);
       break;
+    default:
+      return NULL;
   }
   minim_alloc(n, &q, &x, &y, &z, &v);
-  av1 = avma;
 
   forqfvec_init_dolll(&qv, a, dolll);
+  av1 = avma;
   r = qv.r;
   u = qv.u;
   n--;
   for (j=1; j<=n; j++)
   {
     v[j] = rtodbl(gcoeff(r,j,j));
-    for (i=1; i<j; i++) q[i][j] = rtodbl(gcoeff(r,i,j));
+    for (i=1; i<j; i++) q[i][j] = rtodbl(gcoeff(r,i,j)); /* |.| <= 1/2 */
   }
 
-  if (!sBORNE)
+  if (sBORNE) maxnorm = 0.;
+  else
   {
     GEN B = gcoeff(a,1,1);
     long t = 1;
@@ -1193,26 +1221,13 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
     }
     if (flag == min_FIRST) return gerepilecopy(av, mkvec2(B, gel(u,t)));
     maxnorm = -1.; /* don't update maxnorm */
+    if (is_bigint(B)) return NULL;
     sBORNE = itos(B);
   }
-  else
-    maxnorm = 0.;
   BOUND = sBORNE * (1 + eps);
-  if ((long)BOUND != sBORNE) pari_err_PREC( "qfminim");
+  if ((long)BOUND != sBORNE) return NULL;
 
-  switch(flag)
-  {
-    case min_ALL:
-      L = new_chunk(1+maxrank);
-      break;
-    case min_PERF:
-      avma = av1;
-      maxrank = (n*(n+1)) >> 1;
-      L = zero_zv(maxrank);
-      V = cgetg(1+maxrank, t_VECSMALL);
-  }
-
-  s = 0; av1 = avma; lim = stack_lim(av1,1);
+  s = 0;
   k = n; y[n] = z[n] = 0;
   x[n] = (long)sqrt(BOUND/v[n]);
   for(;;x[1]--)
@@ -1261,7 +1276,8 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
     switch(flag)
     {
       case min_FIRST:
-        return gerepilecopy(av, mkvec2(roundr(dbltor(p)), ZM_zc_mul(u,x)));
+        if (dolll) x = ZM_zc_mul_canon(u,x);
+        return gerepilecopy(av, mkvec2(roundr(dbltor(p)), x));
 
       case min_ALL:
         if (s > maxrank && stockall) /* overflow */
@@ -1275,45 +1291,13 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
         break;
 
       case min_VECSMALL:
-        {
-          ulong norm = (ulong)(p + 0.5);
-          L[norm]++;
-        }
+        { ulong norm = (ulong)(p + 0.5); L[norm]++; }
         break;
 
       case min_VECSMALL2:
-        {
-          ulong norm = (ulong)(p + 0.5);
-          if ((norm&1) == 0) L[norm>>1]++;
-        }
+        { ulong norm = (ulong)(p + 0.5); if (!odd(norm)) L[norm>>1]++; }
         break;
 
-      case min_PERF:
-      {
-        pari_sp av2;
-        long I;
-
-        if (s == 1) {
-          invp = matid(maxrank);
-          for (i = 1; i <= maxrank; i++) L[i] = 0;
-        }
-        /* must go till the end in case we find a smallest vector last */
-        if (s > maxrank) { s = maxrank; continue; }
-        av2 = avma;
-        for (i = I = 1; i<=n; i++)
-          for (j=i; j<=n; j++,I++) V[I] = x[i]*x[j];
-        if (! addcolumntomatrix(V,invp,L))
-        {
-          if (DEBUGLEVEL>1) { err_printf("."); err_flush(); }
-          s--; avma=av2; continue;
-        }
-        if (DEBUGLEVEL>1) { err_printf("*"); err_flush(); }
-        if (low_stack(lim, stack_lim(av1,1)))
-        {
-          if(DEBUGMEM>1) pari_warn(warnmem,"minim0, rank>=%ld",s);
-          invp = gerepilecopy(av1, invp);
-        }
-      }
     }
   }
   switch(flag)
@@ -1323,23 +1307,21 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
     case min_VECSMALL:
     case min_VECSMALL2:
       avma = (pari_sp)L; return L;
-    case min_PERF:
-      if (DEBUGLEVEL>1) { err_printf("\n"); err_flush(); }
-      avma = av; return stoi(s);
   }
   r = (maxnorm >= 0) ? roundr(dbltor(maxnorm)): stoi(sBORNE);
   k = minss(s,maxrank);
   L[0] = evaltyp(t_MAT) | evallg(k + 1);
   if (dolll)
-    for (j=1; j<=k; j++)
-      gel(L,j) = ZM_zc_mul(u, gel(L,j));
+    for (j=1; j<=k; j++) gel(L,j) = ZM_zc_mul_canon(u, gel(L,j));
   return gerepilecopy(av, mkvec3(stoi(s<<1), r, L));
 }
 
 static GEN
 minim0(GEN a, GEN BORNE, GEN STOCKMAX, long flag)
 {
-  return minim0_dolll(a, BORNE, STOCKMAX, flag, 1);
+  GEN v = minim0_dolll(a, BORNE, STOCKMAX, flag, 1);
+  if (!v) pari_err_PREC("qfminim");
+  return v;
 }
 
 GEN
@@ -1372,26 +1354,102 @@ qfminim0(GEN a, GEN borne, GEN stockmax, long flag, long prec)
 
 GEN
 minim(GEN a, GEN borne, GEN stockmax)
-{
-  return minim0(a,borne,stockmax,min_ALL);
-}
+{ return minim0(a,borne,stockmax,min_ALL); }
 
 GEN
 minim_raw(GEN a, GEN BORNE, GEN STOCKMAX)
-{
-  return minim0_dolll(a, BORNE, STOCKMAX, min_ALL, 0);
-}
+{ return minim0_dolll(a, BORNE, STOCKMAX, min_ALL, 0); }
 
 GEN
 minim2(GEN a, GEN borne, GEN stockmax)
+{ return minim0(a,borne,stockmax,min_FIRST); }
+
+/* If V depends linearly from the columns of the matrix, return 0.
+ * Otherwise, update INVP and L and return 1. No GC. */
+static int
+addcolumntomatrix(GEN V, GEN invp, GEN L)
 {
-  return minim0(a,borne,stockmax,min_FIRST);
+  long i,j,k, n = lg(invp);
+  GEN a = cgetg(n, t_COL), ak = NULL, mak;
+
+  for (k = 1; k < n; k++)
+    if (!L[k])
+    {
+      ak = RgMrow_zc_mul(invp, V, k);
+      if (!gequal0(ak)) break;
+    }
+  if (k == n) return 0;
+  L[k] = 1;
+  mak = gneg_i(ak);
+  for (i=k+1; i<n; i++)
+    gel(a,i) = gdiv(RgMrow_zc_mul(invp, V, i), mak);
+  for (j=1; j<=k; j++)
+  {
+    GEN c = gel(invp,j), ck = gel(c,k);
+    if (gequal0(ck)) continue;
+    gel(c,k) = gdiv(ck, ak);
+    if (j==k)
+      for (i=k+1; i<n; i++)
+        gel(c,i) = gmul(gel(a,i), ck);
+    else
+      for (i=k+1; i<n; i++)
+        gel(c,i) = gadd(gel(c,i), gmul(gel(a,i), ck));
+  }
+  return 1;
 }
 
 GEN
 perf(GEN a)
 {
-  return minim0(a,NULL,NULL,min_PERF);
+  pari_sp av = avma;
+  GEN u, L;
+  long r, s, k, l, n = lg(a)-1;
+
+  if (!n) return gen_0;
+  if (typ(a) != t_MAT || !RgM_is_ZM(a)) pari_err_TYPE("qfperfection",a);
+  a = minim_lll(a, &u);
+  L = minim_raw(a,NULL,NULL);
+  r = (n*(n+1)) >> 1;
+  if (L)
+  {
+    GEN D, V, invp;
+    L = gel(L, 3); l = lg(L);
+    if (l == 2) { avma = av; return gen_1; }
+
+    D = zero_zv(r);
+    V = cgetg(r+1, t_VECSMALL);
+    invp = matid(r);
+    s = 0;
+    for (k = 1; k < l; k++)
+    {
+      pari_sp av2 = avma;
+      GEN x = gel(L,k);
+      long i, j, I;
+      for (i = I = 1; i<=n; i++)
+        for (j=i; j<=n; j++,I++) V[I] = x[i]*x[j];
+      if (!addcolumntomatrix(V,invp,D)) avma = av2;
+      else if (++s == r) break;
+    }
+  }
+  else
+  {
+    GEN M;
+    L = fincke_pohst(a,NULL,-1, DEFAULTPREC, NULL);
+    if (!L) pari_err_PREC("qfminim");
+    L = gel(L, 3); l = lg(L);
+    if (l == 2) { avma = av; return gen_1; }
+    M = cgetg(l, t_MAT);
+    for (k = 1; k < l; k++)
+    {
+      GEN x = gel(L,k), c = cgetg(r+1, t_COL);
+      long i, I, j;
+      for (i = I = 1; i<=n; i++)
+        for (j=i; j<=n; j++,I++) gel(c,I) = mulii(gel(x,i), gel(x,j));
+      gel(M,k) = c;
+    }
+    s = ZM_rank(M);
+  }
+ avma = av; return utoipos(s);
 }
 
 static GEN
@@ -1510,7 +1568,7 @@ static GEN
 smallvectors(GEN q, GEN BORNE, long maxnum, FP_chk_fun *CHECK)
 {
   long N = lg(q), n = N-1, i, j, k, s, stockmax, checkcnt = 1;
-  pari_sp av, av1, lim;
+  pari_sp av, av1;
   GEN inc, S, x, y, z, v, p1, alpha, norms;
   GEN norme1, normax1, borne1, borne2;
   GEN (*check)(void *,GEN) = CHECK? CHECK->f: NULL;
@@ -1524,7 +1582,7 @@ smallvectors(GEN q, GEN BORNE, long maxnum, FP_chk_fun *CHECK)
   v = cgetg(N,t_VEC);
   inc = const_vecsmall(n, 1);
 
-  av = avma; lim = stack_lim(av,2);
+  av = avma;
   stockmax = stockall? 2000: maxnum;
   norms = cgetg(check?(stockmax+1): 1,t_VEC); /* unused if (!check) */
   S = cgetg(stockmax+1,t_VEC);
@@ -1600,7 +1658,7 @@ smallvectors(GEN q, GEN BORNE, long maxnum, FP_chk_fun *CHECK)
         if (++k > n) goto END;
       }
 
-      if (low_stack(lim, stack_lim(av,2)))
+      if (gc_needed(av,2))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"smallvectors");
         if (stockmax) S = clonefill(S, s, stockmax);
