@@ -2056,7 +2056,7 @@ get_NR1D(long Nf, long clhray, long degk, long nz, GEN fadkabs, GEN idealrel)
 {
   long n, R1;
   GEN dlk;
-  if (nz < 0) mkvec3(gen_0,gen_0,gen_0); /*EMPTY*/
+  if (nz < 0) return mkvec3(gen_0,gen_0,gen_0); /*EMPTY*/
   n  = clhray * degk;
   R1 = clhray * nz;
   dlk = factordivexact(factorpow(Z_factor(utoipos(Nf)),clhray), idealrel);
