@@ -30,6 +30,13 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 else\
   gerepilecoeffssp((pari_sp)z, tetpil, z+1, 2); }
 
+static void
+warn_coercion(GEN x, GEN y, GEN z)
+{
+  if (DEBUGLEVEL)
+   pari_warn(warner,"coercing quotient rings; moduli %Ps and %Ps -> %Ps",x,y,z);
+}
+
 static long
 kro_quad(GEN x, GEN y)
 {
@@ -199,7 +206,7 @@ static GEN
 div_intmod_same(GEN z, GEN X, GEN x, GEN y)
 {
   if (lgefint(X) == 3) {
-    ulong m = (ulong)X[2], u = Fl_div(itou(x), itou(y), m);
+    ulong m = uel(X,2), u = Fl_div(itou(x), itou(y), m);
     avma = (pari_sp)z; gel(z,2) = utoi(u);
   }
   else
@@ -221,7 +228,7 @@ rfrac_denom_mul_scal(GEN d, GEN y)
   GEN D = RgX_Rg_mul(d, y);
   if (lg(D) != lg(d))
   { /* try to generate a meaningful diagnostic */
-    D = gdiv(leading_term(d), y); /* should fail */
+    D = gdiv(leading_coeff(d), y); /* should fail */
     pari_err_INV("gred_rfrac", y); /* better than nothing */
   }
   return D;
@@ -243,6 +250,7 @@ gred_rfrac_simple(GEN n, GEN d)
   }
 
   cd = content(d);
+  while (typ(n) == t_POL && !degpol(n)) n = gel(n,2);
   cn = (typ(n) == t_POL && varn(n) == varn(d))? content(n): n;
   if (!gequal1(cd)) {
     d = RgX_Rg_div(d,cd);
@@ -316,7 +324,7 @@ fix_rfrac(GEN x, long d)
 
 /* assume d != 0 */
 static GEN
-gred_rfrac2_i(GEN n, GEN d)
+gred_rfrac2(GEN n, GEN d)
 {
   GEN y, z;
   long v, vd, vn;
@@ -330,7 +338,7 @@ gred_rfrac2_i(GEN n, GEN d)
   {
     if (varncmp(vd, gvar(n)) >= 0) return gdiv(n,d);
     if (varncmp(vd, gvar2(n)) < 0) return gred_rfrac_simple(n,d);
-    pari_err_BUG("gred_rfrac2_i [incompatible variables]");
+    pari_err_BUG("gred_rfrac2 [incompatible variables]");
   }
   vn = varn(n);
   if (varncmp(vd, vn) < 0) return gred_rfrac_simple(n,d);
@@ -355,41 +363,44 @@ gred_rfrac2_i(GEN n, GEN d)
   return fix_rfrac(gred_rfrac_simple(n,d), v);
 }
 
+/* x,y t_INT, return x/y in reduced form */
 GEN
-gred_rfrac2(GEN x1, GEN x2)
+Qdivii(GEN x, GEN y)
 {
   pari_sp av = avma;
-  return gerepileupto(av, gred_rfrac2_i(x1, x2));
-}
+  GEN r, q;
 
-/* x1,x2 t_INT, return x1/x2 in reduced form */
-GEN
-gred_frac2(GEN x1, GEN x2)
-{
-  GEN r, y = dvmdii(x1,x2,&r);
-  pari_sp av;
-
-  if (r == gen_0) return y; /* gen_0 intended */
-  av = avma; r = gcdii(x2,r);
+  if (is_pm1(y)) return (signe(y) < 0)? negi(x): icopy(x);
+  if (is_pm1(x)) {
+    long s = signe(y);
+    if (!s) pari_err_INV("gdiv",y);
+    if (signe(x) < 0) s = -s;
+    q = cgetg(3, t_FRAC);
+    gel(q,1) = s<0? gen_m1: gen_1;
+    gel(q,2) = absi(y); return q;
+  }
+  q = dvmdii(x,y,&r);
+  if (r == gen_0) return q; /* gen_0 intended */
+  r = gcdii(y,r);
   if (lgefint(r) == 3)
   {
-    ulong rr = r[2];
+    ulong t = r[2];
     avma = av;
-    if (rr == 1) y = mkfraccopy(x1, x2);
+    if (t == 1) q = mkfraccopy(x,y);
     else
     {
-      y = cgetg(3,t_FRAC);
-      gel(y,1) = diviuexact(x1, rr);
-      gel(y,2) = diviuexact(x2, rr);
+      q = cgetg(3,t_FRAC);
+      gel(q,1) = diviuexact(x,t);
+      gel(q,2) = diviuexact(y,t);
     }
   }
   else
-  { /* rare: r left on stack for efficiency */
-    y = cgetg(3,t_FRAC);
-    gel(y,1) = diviiexact(x1,r);
-    gel(y,2) = diviiexact(x2,r);
+  { /* rare: r and q left on stack for efficiency */
+    q = cgetg(3,t_FRAC);
+    gel(q,1) = diviiexact(x,r);
+    gel(q,2) = diviiexact(y,r);
   }
-  normalize_frac(y); return y;
+  normalize_frac(q); return q;
 }
 
 /*******************************************************************/
@@ -544,7 +555,7 @@ conjvec(GEN x,long prec)
           pari_err_TYPE("conjvec [not a rational t_POL]",x);
         retconst_col(lx-3, gcopy(x));
       }
-      if (!RgX_is_QX(x)) pari_err_TYPE("conjvec", x);
+      RgX_check_QX(x,"conjvec");
       av = avma;
       if (varn(x) != varn(T)) pari_err_VAR("conjvec",x,T);
       r = cleanroots(T,prec);
@@ -611,19 +622,19 @@ addsub_pp(GEN x, GEN y, GEN (*op)(GEN,GEN))
   gel(z,3) = icopy(mod);
   gel(z,4) = icopy(u); return z;
 }
-
+/* Rg_to_Fp(t_FRAC) without GC */
+static GEN
+Q_to_Fp(GEN x, GEN p)
+{ return mulii(gel(x,1), Fp_inv(gel(x,2),p)); }
 /* return x + y, where y t_PADIC and x is a non-zero t_INT or t_FRAC */
 static GEN
 addQp(GEN x, GEN y)
 {
   pari_sp av = avma;
-  long tx,d,r,e, vy = valp(y), py = precp(y);
-  GEN z,q,p1,p2,mod,u, p = gel(y,2);
+  long d, r, e, vy = valp(y), py = precp(y);
+  GEN z, q, mod, u, p = gel(y,2);
 
-  tx = typ(x);
-  e = (tx == t_INT)? Z_pvalrem(x,p,&p1)
-                   : Z_pvalrem(gel(x,1),p,&p1) -
-                     Z_pvalrem(gel(x,2),p,&p2);
+  e = Q_pvalrem(x, p, &x);
   d = vy - e; r = d + py;
   if (r <= 0) { avma = av; return gcopy(y); }
   mod = gel(y,3);
@@ -634,23 +645,21 @@ addQp(GEN x, GEN y)
   {
     q = powiu(p,d);
     mod = mulii(mod, q);
-    u   = mulii(u, q);
-    if (tx != t_INT && !is_pm1(p2)) p1 = mulii(p1, Fp_inv(p2,mod));
-    u = addii(u, p1);
+    if (typ(x) != t_INT) x = Q_to_Fp(x, mod);
+    u = addii(x,  mulii(u, q));
   }
   else if (d < 0)
   {
     q = powiu(p,-d);
-    if (tx != t_INT && !is_pm1(p2)) p1 = mulii(p1, Fp_inv(p2,mod));
-    p1 = mulii(p1, q);
-    u = addii(u, p1);
+    if (typ(x) != t_INT) x = Q_to_Fp(x, mod);
+    u = addii(u, mulii(x, q));
     r = py; e = vy;
   }
   else
   {
     long c;
-    if (tx != t_INT && !is_pm1(p2)) p1 = mulii(p1, Fp_inv(p2,mod));
-    u = addii(u, p1);
+    if (typ(x) != t_INT) x = Q_to_Fp(x, mod);
+    u = addii(u, x);
     if (!signe(u) || (c = Z_pvalrem(u,p,&u)) >= r)
     {
       avma = av; return zeropadic(p,e+r);
@@ -662,8 +671,8 @@ addQp(GEN x, GEN y)
       e += c;
     }
   }
-  u = modii(u, mod);
-  avma = av; z = cgetg(5,t_PADIC);
+  u = modii(u, mod); avma = av;
+  z = cgetg(5,t_PADIC);
   z[1] = evalprecp(r) | evalvalp(e);
   gel(z,2) = icopy(p);
   gel(z,3) = icopy(mod);
@@ -682,6 +691,7 @@ addsub_polmod(GEN X, GEN Y, GEN x, GEN y, GEN(*op)(GEN,GEN))
   if (vx==vy) {
     pari_sp av;
     gel(z,1) = RgX_gcd(X,Y); av = avma;
+    warn_coercion(X,Y,gel(z,1));
     gel(z,2) = gerepileupto(av, gmod(op(x, y), gel(z,1))); return z;
   }
   if (varncmp(vx, vy) < 0)
@@ -699,23 +709,12 @@ addsub_polmod_scal(GEN Y, GEN y, GEN x, GEN(*op)(GEN,GEN))
   gel(z,2) = op(y, x); return z;
 }
 
-/* check y[a..b-1] and set signe to 1 if one coeff is non-0, 0 otherwise
- * For t_POL and t_SER */
-static GEN
-NORMALIZE_i(GEN y, long a, long b)
-{
-  long i;
-  for (i = a; i < b; i++)
-    if (!gequal0(gel(y,i))) { setsigne(y, 1); return y; }
-  setsigne(y, 0); return y;
-}
 /* typ(y) == t_SER, x "scalar" [e.g object in lower variable] */
 static GEN
 add_ser_scal(GEN y, GEN x)
 {
-  long i, j, l, ly, vy;
-  pari_sp av;
-  GEN z, t;
+  long i, l, ly, vy;
+  GEN z;
 
   if (isrationalzero(x)) return gcopy(y);
   ly = lg(y);
@@ -728,47 +727,54 @@ add_ser_scal(GEN y, GEN x)
     for (i = 2; i <= 1-l; i++) gel(z,i) = gcopy(gel(y,i));
     gel(z,i) = gadd(x,gel(y,i)); i++;
     for (     ; i < ly; i++)   gel(z,i) = gcopy(gel(y,i));
-    return NORMALIZE_i(z, 2, ly);
+    return z;
   }
   vy = varn(y);
   if (l > 0)
   {
-    ly += l; y -= l; z = cgetg(ly,t_SER);
-    z[1] = evalsigne(1) | _evalvalp(0) | evalvarn(vy);
-    gel(z,2) = gcopy(x);
+    if (ser_isexactzero(y))
+      return scalarser(ly == 2? x: gadd(x,gel(y,2)), vy, l);
+    y -= l; ly += l;
+    z = cgetg(ly,t_SER);
+    x = gcopy(x);
     for (i=3; i<=l+1; i++) gel(z,i) = gen_0;
-    for (   ; i < ly; i++) gel(z,i) = gcopy(gel(y,i));
-    if (gequal0(x)) return normalize(z);
-    return z;
   }
-  /* l = 0, ly >= 3, !isrationalzero(x) */
-  av = avma; z = cgetg(ly,t_SER);
-  x = gadd(x, gel(y,2));
-  if (!isrationalzero(x))
+  else
+  { /* l = 0, ly >= 3. Also OK if ser_isexactzero(y) */
+    z = cgetg(ly,t_SER);
+    x = gadd(x, gel(y,2));
+    i = 3;
+  }
+  for (; i<ly; i++) gel(z,i) = gcopy(gel(y,i));
+  gel(z,2) = x;
+  z[1] = evalsigne(1) | _evalvalp(0) | evalvarn(vy);
+  return gequal0(x)? normalize(z): z;
+}
+static long
+_serprec(GEN x) { return ser_isexactzero(x)? 2: lg(x); }
+/* x,y t_SER in the same variable: x+y */
+static GEN
+ser_add(GEN x, GEN y)
+{
+  long i, lx,ly, n = valp(y) - valp(x);
+  GEN z;
+  if (n < 0) { n = -n; swap(x,y); }
+  /* valp(x) <= valp(y) */
+  lx = _serprec(x);
+  if (lx == 2) /* don't lose type information */
+    return scalarser(gadd(RgX_get_0(x), RgX_get_0(y)), varn(x), valp(x));
+  ly = _serprec(y) + n; if (lx < ly) ly = lx;
+  if (n)
   {
-    z[1] = evalsigne(1) | _evalvalp(0) | evalvarn(vy);
-    gel(z,2) = x;
-    for (i=3; i<ly; i++) gel(z,i) = gcopy(gel(y,i));
-    if (gequal0(x)) return normalize(z);
-    return z;
+    if (n+2 > lx) return gcopy(x);
+    z = cgetg(ly,t_SER);
+    for (i=2; i<=n+1; i++) gel(z,i) = gcopy(gel(x,i));
+    for (   ; i < ly; i++) gel(z,i) = gadd(gel(x,i),gel(y,i-n));
+  } else {
+    z = cgetg(ly,t_SER);
+    for (i=2; i < ly; i++) gel(z,i) = gadd(gel(x,i),gel(y,i));
   }
-  avma = av; /* first coeff is rational 0 */
-  i = 3;
-  while (i<ly && isrationalzero(gel(y,i))) i++;
-  if (i == ly) return zeroser(vy, ly-2);
-  t = gel(y,i);
-  while (i<ly && isexactzero(gel(y,i))) i++;
-  if (i == ly)
-  {
-    z = cgetg(3, t_SER);
-    z[1] = evalsigne(0) | _evalvalp(i-2) | evalvarn(vy);
-    gel(z,2) = gcopy(t); return gerepileupto(av, z);
-  }
-
-  i -= 2; ly -= i; y += i;
-  z = cgetg(ly,t_SER); z[1] = evalvalp(i) | evalvarn(vy);
-  for (j = 2; j < ly; j++) gel(z,j) = gcopy(gel(y,j));
-  return NORMALIZE_i(z, 2, ly);
+  z[1] = x[1]; return normalize(z);
 }
 /* typ(y) == RFRAC, x polynomial in same variable or "scalar" */
 static GEN
@@ -944,7 +950,7 @@ add_rfrac(GEN x, GEN y)
 GEN
 gadd(GEN x, GEN y)
 {
-  long tx = typ(x), ty = typ(y), vx, vy, lx, ly, i, l;
+  long tx = typ(x), ty = typ(y), vx, vy, lx, i, l;
   pari_sp av, tetpil;
   GEN z, p1;
 
@@ -957,6 +963,7 @@ gadd(GEN x, GEN y)
       if (X==Y || equalii(X,Y))
         return add_intmod_same(z, X, gel(x,2), gel(y,2));
       gel(z,1) = gcdii(X,Y);
+      warn_coercion(X,Y,gel(z,1));
       av = avma; p1 = addii(gel(x,2),gel(y,2));
       gel(z,2) = gerepileuptoint(av, remii(p1, gel(z,1))); return z;
     }
@@ -998,22 +1005,7 @@ gadd(GEN x, GEN y)
         if (varncmp(vx, vy) < 0) return add_ser_scal(x, y);
         else                     return add_ser_scal(y, x);
       }
-      l = valp(y) - valp(x);
-      if (l < 0) { l = -l; swap(x,y); }
-      /* valp(x) <= valp(y) */
-      lx = lg(x);
-      ly = lg(y) + l; if (lx < ly) ly = lx;
-      if (l)
-      {
-        if (l > lx-2) return gcopy(x);
-        z = cgetg(ly,t_SER);
-        for (i=2; i<=l+1; i++) gel(z,i) = gcopy(gel(x,i));
-        for (   ; i < ly; i++) gel(z,i) = gadd(gel(x,i),gel(y,i-l));
-      } else {
-        z = cgetg(ly,t_SER);
-        for (i=2; i < ly; i++) gel(z,i) = gadd(gel(x,i),gel(y,i));
-      }
-      z[1] = x[1]; return normalize(z);
+      return ser_add(x, y);
     case t_RFRAC:
       vx = varn(gel(x,2));
       vy = varn(gel(y,2));
@@ -1169,7 +1161,7 @@ gadd(GEN x, GEN y)
           i = lg(y) + valp(y) - RgX_val(x);
           if (i < 3) return gcopy(y);
 
-          p1 = RgX_to_ser(x,i); y = gadd(p1,y);
+          p1 = RgX_to_ser(x,i); y = ser_add(p1,y);
           settyp(p1, t_VECSMALL); /* p1 left on stack */
           return y;
 
@@ -1276,6 +1268,7 @@ gsub(GEN x, GEN y)
       if (X==Y || equalii(X,Y))
         return sub_intmod_same(z, X, gel(x,2), gel(y,2));
       gel(z,1) = gcdii(X,Y);
+      warn_coercion(X,Y,gel(z,1));
       av = avma; p1 = subii(gel(x,2),gel(y,2));
       gel(z,2) = gerepileuptoint(av, modii(p1, gel(z,1))); return z;
     }
@@ -1341,7 +1334,12 @@ static GEN
 mul_ser_scal(GEN y, GEN x) {
   long ly, i;
   GEN z;
-  if (isrationalzero(x)) return zeropol(varn(y));
+  if (isexactzero(x)) return gmul(RgX_get_0(y), x);
+  if (ser_isexactzero(y))
+  {
+    if (lg(y) == 2) return gcopy(y);
+    return scalarser(gmul(x,gel(y,2)), varn(y), valp(y));
+  }
   z = cgetg_copy(y, &ly); z[1] = y[1];
   for (i = 2; i < ly; i++) gel(z,i) = gmul(x,gel(y,i));
   return normalize(z);
@@ -1366,7 +1364,7 @@ mul_rfrac_scal(GEN n, GEN d, GEN x)
       d = gmul(d, gmodulo(gen_1, gel(x,1)));
       return gerepileupto(av, gdiv(n,d));
   }
-  z = gred_rfrac2_i(x, d);
+  z = gred_rfrac2(x, d);
   n = simplify_shallow(n);
   if (typ(z) == t_RFRAC)
   {
@@ -1413,8 +1411,8 @@ mul_rfrac(GEN x1, GEN x2, GEN y1, GEN y2)
   GEN z, X, Y;
   pari_sp av = avma;
 
-  X = gred_rfrac2_i(x1, y2);
-  Y = gred_rfrac2_i(y1, x2);
+  X = gred_rfrac2(x1, y2);
+  Y = gred_rfrac2(y1, x2);
   if (typ(X) == t_RFRAC)
   {
     if (typ(Y) == t_RFRAC) {
@@ -1437,7 +1435,7 @@ static GEN
 div_rfrac_pol(GEN x1, GEN x2, GEN y2)
 {
   pari_sp av = avma;
-  GEN X = gred_rfrac2_i(x1, y2);
+  GEN X = gred_rfrac2(x1, y2);
   if (typ(X) == t_RFRAC && varn(gel(X,2)) == varn(x2))
   {
     x2 = RgX_mul(gel(X,2), x2);
@@ -1577,6 +1575,7 @@ mul_polmod(GEN X, GEN Y, GEN x, GEN y)
   if (vx==vy) {
     pari_sp av;
     gel(z,1) = RgX_gcd(X,Y); av = avma;
+    warn_coercion(X,Y,gel(z,1));
     gel(z,2) = gerepileupto(av, gmod(gmul(x, y), gel(z,1)));
     return z;
   }
@@ -1847,7 +1846,9 @@ gmul(GEN x, GEN y)
       z = cgetg(3,t_INTMOD);
       if (X==Y || equalii(X,Y))
         return mul_intmod_same(z, X, gel(x,2), gel(y,2));
-      gel(z,1) = gcdii(X,Y); av = avma; p1 = mulii(gel(x,2),gel(y,2));
+      gel(z,1) = gcdii(X,Y);
+      warn_coercion(X,Y,gel(z,1));
+      av = avma; p1 = mulii(gel(x,2),gel(y,2));
       gel(z,2) = gerepileuptoint(av, remii(p1, gel(z,1))); return z;
     }
     case t_FRAC:
@@ -1908,7 +1909,7 @@ gmul(GEN x, GEN y)
         }
       }
       else
-        y = RgX_mullow(x, y, lx-2);
+        y = RgXn_mul(x, y, lx-2);
       z = fill_ser(z, y);
       return gerepilecopy((pari_sp)(z + lx), z);
     }
@@ -2134,7 +2135,7 @@ ff_poltype(GEN *x, GEN *p, GEN *pol)
 {
   GEN Q, P = *x;
   if (!signe(P)) return 0;
-  if (!RgX_is_FpXQX(P,pol,p) || !*p || !*pol)
+  if (!RgX_is_FpXQX(P,pol,p) || !*p || !*pol || typ(*pol)!=t_POL)
     return 0;
   Q = RgX_to_FpX(*pol, *p);
   P = RgX_to_FpXQX(P, Q, *p);
@@ -2143,6 +2144,7 @@ ff_poltype(GEN *x, GEN *p, GEN *pol)
   return 1;
 }
 
+/* return a non-normalized result */
 GEN
 sqr_ser_part(GEN x, long l1, long l2)
 {
@@ -2208,9 +2210,9 @@ gsqr(GEN x)
 
     case t_PADIC:
       z = cgetg(5,t_PADIC);
-      i = (equaliu(gel(x,2), 2) && signe(gel(x,4)))? 1: 0;
+      i = (absequaliu(gel(x,2), 2) && signe(gel(x,4)))? 1: 0;
       if (i && precp(x) == 1) i = 2; /* (1 + O(2))^2 = 1 + O(2^3) */
-      z[1] = evalprecp(precp(x)+i) | evalvalp(valp(x) << 1);
+      z[1] = evalprecp(precp(x)+i) | evalvalp(valp(x)*2);
       gel(z,2) = icopy(gel(x,2));
       gel(z,3) = shifti(gel(x,3), i); av = avma;
       gel(z,4) = gerepileuptoint(av, remii(sqri(gel(x,4)), gel(z,3)));
@@ -2261,6 +2263,11 @@ gsqr(GEN x)
 
     case t_SER:
       lx = lg(x);
+      if (ser_isexactzero(x)) {
+        GEN z = gcopy(x);
+        setvalp(z, 2*valp(x));
+        return z;
+      }
       if (lx < 40)
         return normalize( sqr_ser_part(x, 0, lx-3) );
       else
@@ -2279,7 +2286,7 @@ gsqr(GEN x)
           }
         }
         else
-          x = RgX_sqrlow(x, lx-2);
+          x = RgXn_sqr(x, lx-2);
         z = fill_ser(z, x);
         return gerepilecopy(av, z);
       }
@@ -2336,7 +2343,13 @@ div_rfrac(GEN x, GEN y)
 static GEN
 div_ser_scal(GEN x, GEN y) {
   long i, lx;
-  GEN z = cgetg_copy(x, &lx); z[1] = x[1];
+  GEN z;
+  if (ser_isexactzero(x))
+  {
+    if (lg(x) == 2) return gcopy(x);
+    return scalarser(gdiv(gel(x,2), y), varn(x), valp(x));
+  }
+  z = cgetg_copy(x, &lx); z[1] = x[1];
   for (i=2; i<lx; i++) gel(z,i) = gdiv(gel(x,i),y);
   return normalize(z);
 }
@@ -2404,7 +2417,11 @@ div_ser(GEN x, GEN y, long vx)
   GEN y_lead, p1, p2, z;
 
   if (!signe(y)) pari_err_INV("div_ser", y);
-  if (lx == 2) return zeroser(vx, l);
+  if (ser_isexactzero(x))
+  {
+    if (lx == 2) return zeroser(vx, l);
+    return scalarser(gmul(gel(x,2),RgX_get_0(y)), varn(x), l);
+  }
   y_lead = gel(y,2);
   if (gequal0(y_lead)) /* normalize denominator if leading term is 0 */
   {
@@ -2510,24 +2527,16 @@ gdiv(GEN x, GEN y)
   if (tx == ty) switch(tx)
   {
     case t_INT:
-      if (is_pm1(y)) return (signe(y) < 0)? negi(x): icopy(x);
-      if (is_pm1(x)) {
-        long s = signe(y);
-        if (!s) pari_err_INV("gdiv",y);
-        if (signe(x) < 0) s = -s;
-        z = cgetg(3, t_FRAC);
-        gel(z,1) = s<0? gen_m1: gen_1;
-        gel(z,2) = absi(y); return z;
-      }
-      return gred_frac2(x,y);
+      return Qdivii(x,y);
 
     case t_REAL: return divrr(x,y);
     case t_INTMOD: { GEN X = gel(x,1), Y = gel(y,1);
       z = cgetg(3,t_INTMOD);
       if (X==Y || equalii(X,Y))
         return div_intmod_same(z, X, gel(x,2), gel(y,2));
-      gel(z,1) = gcdii(X,Y); av = avma;
-      p1 = mulii(gel(x,2), Fp_inv(gel(y,2), gel(z,1)));
+      gel(z,1) = gcdii(X,Y);
+      warn_coercion(X,Y,gel(z,1));
+      av = avma; p1 = mulii(gel(x,2), Fp_inv(gel(y,2), gel(z,1)));
       gel(z,2) = gerepileuptoint(av, remii(p1, gel(z,1))); return z;
     }
     case t_FRAC: {
@@ -2588,7 +2597,8 @@ gdiv(GEN x, GEN y)
       }
       if (!signe(y)) pari_err_INV("gdiv",y);
       if (lg(y) == 3) return RgX_Rg_div(x,gel(y,2));
-      return gred_rfrac2(x,y);
+      av = avma;
+      return gerepileupto(av, gred_rfrac2(x,y));
 
     case t_SER:
       vx = varn(x);
@@ -2671,7 +2681,11 @@ gdiv(GEN x, GEN y)
         return gerepile(av, tetpil, gdiv(p2,p1));
     }
   }
-  if (gequal0(y) && ty != t_MAT) pari_err_INV("gdiv",y);
+  if (gequal0(y))
+  {
+    if (is_matvec_t(tx) && lg(x) == 1) return gcopy(x);
+    if (ty != t_MAT) pari_err_INV("gdiv",y);
+  }
 
   if (is_const_t(tx) && is_const_t(ty)) switch(tx)
   {
@@ -2983,7 +2997,8 @@ gmulsg(long s, GEN y)
       return normalizepol_lg(z, ly);
 
     case t_SER:
-      if (!s) return zeropol(varn(y));
+      if (ser_isexactzero(y)) return gcopy(y);
+      if (!s) return RgX_get_0(y);
       z = cgetg_copy(y, &ly); z[1]=y[1];
       for (i=2; i<ly; i++) gel(z,i) = gmulsg(s,gel(y,i));
       return normalize(z);
@@ -3012,12 +3027,16 @@ gmulsg(long s, GEN y)
 GEN
 gdivgs(GEN x, long s)
 {
-  long lx, i;
+  long tx = typ(x), lx, i;
   pari_sp av;
   GEN z, y, p1;
 
-  if (!s) pari_err_INV("gdivgs",gen_0);
-  switch(typ(x))
+  if (!s)
+  {
+    if (is_matvec_t(tx) && lg(x) == 1) return gcopy(x);
+    pari_err_INV("gdivgs",gen_0);
+  }
+  switch(tx)
   {
     case t_INT:
       av = avma; z = divis_rem(x,s,&i);
@@ -3179,6 +3198,7 @@ gmul2n(GEN x, long n)
       for (i=2; i<lx; i++) gel(z,i) = gmul2n(gel(x,i),n);
       return normalizepol_lg(z, lx); /* needed if char = 2 */
     case t_SER:
+      if (ser_isexactzero(x)) return gcopy(x);
       z = cgetg_copy(x, &lx); z[1] = x[1];
       for (i=2; i<lx; i++) gel(z,i) = gmul2n(gel(x,i),n);
       return normalize(z); /* needed if char = 2 */
@@ -3202,50 +3222,6 @@ gmul2n(GEN x, long n)
 /*                              INVERSE                            */
 /*                                                                 */
 /*******************************************************************/
-GEN
-inv_ser(GEN b)
-{
-  pari_sp av = avma, av2, lim;
-  long j, lold, l = lg(b), e = valp(b), v = varn(b);
-  GEN y, x = cgetg(l, t_SER), a = leafcopy(b);
-  ulong mask = quadratic_prec_mask(l - 2);
-
-  if (!signe(b)) pari_err_INV("inv_ser",b);
-
-  for (j = 3; j < l; j++) gel(x,j) = gen_0;
-  gel(x,2) = ginv(gel(b,2));
-  a[1] = x[1] = _evalvalp(0) | evalvarn(v) | evalsigne(1);
-  av2 = avma; lim = stack_lim(av2, 2);
-  lold = 1;
-  while (mask > 1)
-  {
-    long lnew = lold << 1;
-    GEN z;
-
-    if (mask & 1) lnew--;
-    mask >>= 1;
-    setlg(a, lnew + 2);
-    setlg(x, lnew + 2);
-    /* TODO: gmul(a,x) should be a half product (the higher half is known) */
-    z = gmul(a,x); /* = 1 + O(t^lold) */
-    y = cgetg(lnew-lold + 2, t_SER);
-    y[1] = _evalvalp(lold) | evalvarn(v) | evalsigne(1);
-    for (j = 2; j < 2+lnew-lold; j++) gel(y,j) = gel(z,j+lold);
-    /* y = a*x - 1; */
-    y = gsub(x, gmul(x, y));
-    for (j = lold+2; j < lnew+2; j++) x[j] = y[j];
-    if (low_stack(lim, stack_lim(av2,2)))
-    {
-      if(DEBUGMEM>1) pari_warn(warnmem,"inv_ser");
-      y = gerepilecopy(av2, x);
-      for (j = 2; j < lnew+2; j++) x[j] = y[j];
-    }
-    lold = lnew;
-  }
-  x[1] = evalvalp(valp(x)-e) | evalvarn(v) | evalsigne(1);
-  return gerepilecopy(av, x);
-}
-
 static GEN
 inv_polmod(GEN T, GEN x)
 {
@@ -3275,14 +3251,20 @@ inv_polmod(GEN T, GEN x)
         a = FpX_to_mod(a, p);
         a = gerepileupto(av, a);
       }
+      else if (RgX_is_ZX(T) && RgX_is_QX(x))
+      {
+        avma = av;
+        a = QXQ_inv(x, T);
+      }
       else {
         avma = av;
-        a = RgXQ_inv(x, gel(z,1));
+        a = RgXQ_inv(x, T);
       }
     }
   }
   gel(z,2) = a; return z;
 }
+
 GEN
 ginv(GEN x)
 {
@@ -3373,15 +3355,9 @@ ginv(GEN x)
         togglesign(gel(y,2));
       return y;
     case t_MAT:
-    {
-      GEN ff = NULL;
-      if (RgM_is_FFM(x,&ff))
-        y = FFM_inv(x, ff);
-      else
-        y = RgM_inv(x);
+      y = RgM_inv(x);
       if (!y) pari_err_INV("ginv",x);
       return y;
-    }
     case t_VECSMALL:
     {
       long i, lx = lg(x)-1;
