@@ -146,14 +146,14 @@ FpX_ratlift(GEN P, GEN mod, GEN amax, GEN bmax, GEN denom)
  * rarely succeed, and even if it does the polynomial we get has sensible
  * coefficients, so the full division will not be too costly.
  *
- * If not NULL, den must a a multiple of the denominator of the gcd,
+ * If not NULL, den must be a multiple of the denominator of the gcd,
  * for example the discriminant of T.
  *
  * NOTE: if T is not irreducible, nfgcd may loop forever, esp. if gcd | T */
 GEN
 nfgcd_all(GEN P, GEN Q, GEN T, GEN den, GEN *Pnew)
 {
-  pari_sp btop, st_lim, ltop = avma;
+  pari_sp btop, ltop = avma;
   GEN lP, lQ, M, dsol, R, bo, sol, mod = NULL;
   long vP = varn(P), vT = varn(T), dT = degpol(T), dM = 0, dR;
   forprime_t S;
@@ -162,20 +162,20 @@ nfgcd_all(GEN P, GEN Q, GEN T, GEN den, GEN *Pnew)
   if (!signe(Q)) { if (Pnew) *Pnew = pol_1(vT);   return gcopy(P); }
   /*Compute denominators*/
   if (!den) den = ZX_disc(T);
-  lP = leading_term(P);
-  lQ = leading_term(Q);
+  lP = leading_coeff(P);
+  lQ = leading_coeff(Q);
   if ( !((typ(lP)==t_INT && is_pm1(lP)) || (typ(lQ)==t_INT && is_pm1(lQ))) )
     den = mulii(den, gcdii(ZX_resultant(lP, T), ZX_resultant(lQ, T)));
 
-  init_modular(&S);
-  btop = avma; st_lim = stack_lim(btop, 1);
+  init_modular_small(&S);
+  btop = avma;
   for(;;)
   {
     ulong p = u_forprime_next(&S);
     if (!p) pari_err_OVERFLOW("nfgcd [ran out of primes]");
     /*Discard primes dividing disc(T) or lc(PQ) */
     if (!umodiu(den, p)) continue;
-    if (DEBUGLEVEL>5) err_printf("nfgcd: p=%d\n",p);
+    if (DEBUGLEVEL>5) err_printf("nfgcd: p=%lu\n",p);
     /*Discard primes when modular gcd does not exist*/
     if ((R = FlxqX_safegcd(ZXX_to_FlxX(P,p,vT),
                            ZXX_to_FlxX(Q,p,vT),
@@ -187,7 +187,7 @@ nfgcd_all(GEN P, GEN Q, GEN T, GEN den, GEN *Pnew)
     R = FlxX_to_Flm(R, dT);
     /* previous primes divided Res(P/gcd, Q/gcd)? Discard them. */
     if (!mod || dR < dM) { M = ZM_init_CRT(R, p); mod = utoipos(p); dM = dR; continue; }
-    if (low_stack(st_lim, stack_lim(btop, 1)))
+    if (gc_needed(btop, 1))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"nfgcd");
       gerepileall(btop, 2, &M, &mod);
@@ -218,6 +218,23 @@ GEN
 nfgcd(GEN P, GEN Q, GEN T, GEN den)
 { return nfgcd_all(P,Q,T,den,NULL); }
 
+int
+nfissquarefree(GEN nf, GEN x)
+{
+  pari_sp av = avma;
+  GEN g, y = RgX_deriv(x);
+  if (RgX_is_rational(x))
+    g = QX_gcd(x, y);
+  else
+  {
+    GEN T = get_nfpol(nf,&nf);
+    x = Q_primpart( liftpol_shallow(x) );
+    y = Q_primpart( liftpol_shallow(y) );
+    g = nfgcd(x, y, T, nf? nf_get_index(nf): NULL);
+  }
+  avma = av; return (degpol(g) == 0);
+}
+
 /*******************************************************************/
 /*             FACTOR OVER (Z_K/pr)[X] --> FqX_factor              */
 /*******************************************************************/
@@ -252,7 +269,7 @@ nffactormod(GEN nf, GEN x, GEN pr)
 static GEN
 QXQX_normalize(GEN P, GEN T)
 {
-  GEN P0 = leading_term(P);
+  GEN P0 = leading_coeff(P);
   long t = typ(P0);
   if (t == t_POL)
   {
@@ -267,7 +284,7 @@ QXQX_normalize(GEN P, GEN T)
 static GEN
 RgX_int_normalize(GEN P)
 {
-  GEN P0 = leading_term(P);
+  GEN P0 = leading_coeff(P);
   /* cater for t_POL */
   if (typ(P0) == t_POL)
   {
@@ -286,64 +303,51 @@ static GEN
 proper_nf(GEN nf)
 { return (lg(nf) == 3)? gel(nf,1): nf; }
 
+/* if *pnf = NULL replace if by a "quick" K = nfinit(T), ensuring maximality
+ * by small primes only. Return a multiplicative bound for the denominator of
+ * algebraic integers in Z_K in terms of K.zk */
 static GEN
 fix_nf(GEN *pnf, GEN *pT, GEN *pA)
 {
-  GEN den = gen_1;
-  if (!*pnf)
-  {
-    GEN fa, P, q, D, T = *pT;
-    GEN nf, NF = nfinitall(T, nf_PARTIALFACT, DEFAULTPREC);
-    *pnf = nf = proper_nf(NF);
-    if (nf != NF) { /* t_POL defining base field changed (not monic) */
-      long i, l;
-      GEN A = *pA, a = cgetg_copy(A, &l);
-      GEN rev = gel(NF,2), pow, dpow;
+  GEN nf, NF, fa, P, Q, q, D, T = *pT;
+  nfmaxord_t S;
+  long i, l;
 
-      *pT = T = nf_get_pol(nf); /* need to update T */
-      pow = QXQ_powers(lift_intern(rev), degpol(T)-1, T);
-      pow = Q_remove_denom(pow, &dpow);
-      a[1] = A[1];
-      for (i=2; i<l; i++) {
-        GEN c = gel(A,i);
-        if (typ(c) == t_POL) c = QX_ZXQV_eval(c, pow, dpow);
-        gel(a,i) = c;
-      }
-      *pA = Q_primpart(a); /* need to update A */
+  if (*pnf) return gen_1;
+  nfmaxord(&S, T, nf_PARTIALFACT);
+  NF = nfinit_complete(&S, 0, DEFAULTPREC);
+  *pnf = nf = proper_nf(NF);
+  if (nf != NF) { /* t_POL defining base field changed (not monic) */
+    GEN A = *pA, a = cgetg_copy(A, &l);
+    GEN rev = gel(NF,2), pow, dpow;
+
+    *pT = T = nf_get_pol(nf); /* need to update T */
+    pow = QXQ_powers(lift_shallow(rev), degpol(T)-1, T);
+    pow = Q_remove_denom(pow, &dpow);
+    a[1] = A[1];
+    for (i=2; i<l; i++) {
+      GEN c = gel(A,i);
+      if (typ(c) == t_POL) c = QX_ZXQV_eval(c, pow, dpow);
+      gel(a,i) = c;
     }
+    *pA = Q_primpart(a); /* need to update A */
+  }
 
-    D = nf_get_disc(nf);
-    if (is_pm1(D)) return gen_1;
-    fa = absi_factor_limit(D, 0);
-    P = gel(fa,1); q = gel(P, lg(P)-1);
-    if (!BPSW_psp(q)) den = q; /* nf_get_disc(nf) may be incorrect */
-  }
-  return den;
-}
-
-/* set B = A/gcd(A,A'), squarefree */
-static GEN
-get_nfsqff_data(GEN *pnf, GEN *pT, GEN *pA, GEN *pB, GEN *ptbad)
-{
-  GEN den, bad, A = *pA, T = *pT;
-  long n = degpol(T);
-  if (nfsqff_use_Trager(n, degpol(A)))
+  D = nf_get_disc(nf);
+  if (is_pm1(D)) return gen_1;
+  fa = absZ_factor_limit(D, 0);
+  P = gel(fa,1); q = gel(P, lg(P)-1);
+  if (BPSW_psp(q)) return gen_1;
+  /* nf_get_disc(nf) may be incorrect */
+  P = nf_get_ramified_primes(nf);
+  l = lg(P);
+  Q = q; q = gen_1;
+  for (i = 1; i < l; i++)
   {
-    *pnf = T;
-    bad = den = ZX_disc(T);
-    if (is_pm1(leading_term(T))) den = indexpartial(T, den);
+    GEN p = gel(P,i);
+    if (Z_pvalrem(Q, p, &Q) && !BPSW_psp(p)) q = mulii(q, p);
   }
-  else
-  {
-    den = fix_nf(pnf, pT, pA);
-    bad = nf_get_index(*pnf);
-    if (den != gen_1) bad = mulii(bad, den);
-    A = *pA;
-    T = *pT;
-  }
-  (void)nfgcd_all(A, RgX_deriv(A), T, bad, pB);
-  if( ptbad) *ptbad = bad;
-  return den;
+  return q;
 }
 
 /* lt(A) is an integer; ensure it is not a constant t_POL. In place */
@@ -355,13 +359,41 @@ ensure_lt_INT(GEN A)
   while (typ(lt) != t_INT) gel(A,n) = lt = gel(lt,2);
 }
 
+/* set B = A/gcd(A,A'), squarefree */
+static GEN
+get_nfsqff_data(GEN *pnf, GEN *pT, GEN *pA, GEN *pB, GEN *ptbad)
+{
+  GEN den, bad, D, B, A = *pA, T = *pT;
+  long n = degpol(T);
+
+  A = Q_primpart( QXQX_normalize(A, T) );
+  if (nfsqff_use_Trager(n, degpol(A)))
+  {
+    *pnf = T;
+    bad = den = ZX_disc(T);
+    if (is_pm1(leading_coeff(T))) den = indexpartial(T, den);
+  }
+  else
+  {
+    den = fix_nf(pnf, &T, &A);
+    bad = nf_get_index(*pnf);
+    if (den != gen_1) bad = mulii(bad, den);
+  }
+  D = nfgcd_all(A, RgX_deriv(A), T, bad, &B);
+  if (degpol(D)) B = Q_primpart( QXQX_normalize(B, T) );
+  if (ptbad) *ptbad = bad;
+  *pA = A;
+  *pB = B; ensure_lt_INT(B);
+  *pT = T; return den;
+}
+
 /* return the roots of pol in nf */
 GEN
 nfroots(GEN nf,GEN pol)
 {
   pari_sp av = avma;
   GEN z, A, B, T, den;
-  long d;
+  long d, dT;
 
   if (!nf) return nfrootsQ(pol);
   T = get_nfpol(nf, &nf);
@@ -376,30 +408,32 @@ nfroots(GEN nf,GEN pol)
     A = mkpolmod(gneg_i(gel(A,2)), T);
     return gerepilecopy(av, mkvec(A));
   }
-  if (degpol(T) == 1) return gerepileupto(av, nfrootsQ(simplify_shallow(A)));
+  dT = degpol(T);
+  if (dT == 1) return gerepileupto(av, nfrootsQ(simplify_shallow(A)));
 
-  A = Q_primpart(A);
   den = get_nfsqff_data(&nf, &T, &A, &B, NULL);
-  if (degpol(B) != d) B = Q_primpart( QXQX_normalize(B, T) );
-  ensure_lt_INT(B);
-  z = nfsqff(nf,B, ROOTS, den);
+  if (RgX_is_ZX(B))
+  {
+    GEN v = gel(ZX_factor(B), 1);
+    long i, l = lg(v), p = mael(factoru(dT),1,1); /* smallest prime divisor */
+    z = cgetg(1, t_VEC);
+    for (i = 1; i < l; i++)
+    {
+      GEN b = gel(v,i); /* irreducible / Q */
+      long db = degpol(b);
+      if (db != 1 && degpol(b) < p) continue;
+      z = shallowconcat(z, nfsqff(nf, b, ROOTS, den));
+    }
+  }
+  else
+    z = nfsqff(nf,B, ROOTS, den);
   z = gerepileupto(av, QXQV_to_mod(z, T));
   gen_sort_inplace(z, (void*)&cmp_RgX, &cmp_nodata, NULL);
   return z;
 }
 
-/* assume x is squarefree monic in nf.zk[X] */
-int
-nfissplit(GEN nf, GEN x)
-{
-  pari_sp av = avma;
-  long l;
-  nf = checknf(nf);
-  if (typ(x) != t_POL) pari_err_TYPE("nfissplit",x);
-  x = RgX_nffix("nfissplit", nf_get_pol(nf), x, 1);
-  l = lg(nfsqff(nf, x, ROOTS_SPLIT, gen_1));
-  avma = av; return l != 1;
-}
+static GEN
+_norml2(GEN x) { return RgC_fpnorml2(x, DEFAULTPREC); }
 
 /* return a minimal lift of elt modulo id, as a ZC */
 static GEN
@@ -420,7 +454,7 @@ nf_bestlift(GEN elt, GEN bound, nflift_t *L)
     elt = scalarcol(elt, l-1);
   }
   u = ZC_sub(elt, ZM_ZC_mul(L->prk, u));
-  if (bound && gcmp(RgC_fpnorml2(u,DEFAULTPREC), bound) > 0) u = NULL;
+  if (bound && gcmp(_norml2(u), bound) > 0) u = NULL;
   return u;
 }
 
@@ -451,16 +485,16 @@ nf_bestlift_to_pol(GEN elt, GEN bound, nflift_t *L)
 /* return the T->powden * (lift of pol with coefficients of T2-norm <= C)
  * if it exists. */
 static GEN
-nf_pol_lift(GEN pol, GEN bound, nfcmbf_t *T)
+nf_pol_lift(GEN pol, GEN bound, nflift_t *L)
 {
   long i, l = lg(pol);
-  GEN t, x = cgetg(l,t_POL);
+  GEN x = cgetg(l,t_POL);
 
   x[1] = pol[1];
-  gel(x,l-1) = mul_content(gel(pol,l-1), T->L->topowden);
+  gel(x,l-1) = mul_content(gel(pol,l-1), L->topowden);
   for (i=l-2; i>1; i--)
   {
-    t = nf_bestlift_to_pol(gel(pol,i), bound, T->L);
+    GEN t = nf_bestlift_to_pol(gel(pol,i), bound, L);
     if (!t) return NULL;
     gel(x,i) = t;
   }
@@ -496,7 +530,7 @@ fact_from_sqff(GEN rep, GEN A, GEN B, GEN y, GEN T, GEN bad)
     }
     else
     { /* compute valuations mod a prime of degree 1 (avoid coeff explosion) */
-      GEN quo, p, r, Bp, lb = leading_term(B), E = cgetalloc(t_VECSMALL,n+1);
+      GEN quo, p, r, Bp, lb = leading_coeff(B), E = cgetalloc(t_VECSMALL,n+1);
       pari_sp av1 = avma;
       ulong pp;
       long j;
@@ -559,9 +593,9 @@ nffactor(GEN nf,GEN pol)
     avma = (pari_sp)(rep + 3);
     return (dA == 0)? trivial_fact(): zerofact(varn(pol));
   }
-  A = Q_primpart( QXQX_normalize(A, T) );
   if (dA == 1) {
     GEN c;
+    A = Q_primpart( QXQX_normalize(A, T) );
     A = gerepilecopy(av, A); c = gel(A,2);
     if (typ(c) == t_POL && degpol(c) > 0) gel(A,2) = mkpolmod(c, ZX_copy(T));
     gel(rep,1) = mkcol(A);
@@ -571,9 +605,19 @@ nffactor(GEN nf,GEN pol)
 
   den = get_nfsqff_data(&nf, &T, &A, &B, &bad);
   if (DEBUGLEVEL>2) timer_printf(&ti, "squarefree test");
-  if (degpol(B) != dA) B = Q_primpart( QXQX_normalize(B, T) );
-  ensure_lt_INT(B);
-  y = nfsqff(nf,B, 0, den);
+  if (RgX_is_ZX(B))
+  {
+    GEN v = gel(ZX_factor(B), 1);
+    long i, l = lg(v);
+    y = cgetg(1, t_VEC);
+    for (i = 1; i < l; i++)
+    {
+      GEN b = gel(v,i); /* irreducible / Q */
+      y = shallowconcat(y, nfsqff(nf, b, 0, den));
+    }
+  }
+  else
+    y = nfsqff(nf,B, 0, den);
   if (DEBUGLEVEL>3) err_printf("number of factor(s) found: %ld\n", lg(y)-1);
 
   fact_from_sqff(rep, A, B, y, T, bad);
@@ -587,55 +631,46 @@ arch_for_T2(GEN G, GEN x)
   return (typ(x) == t_COL)? RgM_RgC_mul(G,x)
                           : RgC_Rg_mul(gel(G,1),x);
 }
-static GEN
-arch_for_T2_prec(GEN G, GEN x, long prec)
-{
-  return (typ(x) == t_COL)? RgM_RgC_mul(G, RgC_gtofp(x,prec))
-                          : RgC_Rg_mul(gel(G,1), gtofp(x, prec));
-}
 
-/* return a bound for T_2(P), P | polbase in C[X]
- * NB: Mignotte bound: A | S ==>
+/* polbase a zkX with t_INT leading coeff; return a bound for T_2(P),
+ * P | polbase in C[X]. NB: Mignotte bound: A | S ==>
  *  |a_i| <= binom(d-1, i-1) || S ||_2 + binom(d-1, i) lc(S)
  *
  * Apply to sigma(S) for all embeddings sigma, then take the L_2 norm over
- * sigma, then take the sup over i.
- **/
+ * sigma, then take the sup over i */
 static GEN
 nf_Mignotte_bound(GEN nf, GEN polbase)
-{
-  GEN G = nf_get_G(nf), lS = leading_term(polbase); /* t_INT */
-  GEN p1, C, N2, matGS, binlS, bin;
-  long prec, i, j, d = degpol(polbase), n = nf_get_degree(nf), r1 = nf_get_r1(nf);
+{ GEN lS = leading_coeff(polbase); /* t_INT */
+  GEN p1, C, N2, binlS, bin;
+  long prec = nf_get_prec(nf), n = nf_get_degree(nf), r1 = nf_get_r1(nf);
+  long i, j, d = degpol(polbase);
 
   binlS = bin = vecbinome(d-1);
   if (!isint1(lS)) binlS = ZC_Z_mul(bin,lS);
 
   N2 = cgetg(n+1, t_VEC);
-  prec = gprecision(G);
   for (;;)
   {
-    nffp_t F;
+    GEN G = nf_get_G(nf), matGS = cgetg(d+2, t_MAT);
 
-    matGS = cgetg(d+2, t_MAT);
     for (j=0; j<=d; j++) gel(matGS,j+1) = arch_for_T2(G, gel(polbase,j+2));
     matGS = shallowtrans(matGS);
     for (j=1; j <= r1; j++) /* N2[j] = || sigma_j(S) ||_2 */
     {
-      GEN c = sqrtr( RgC_fpnorml2(gel(matGS,j), DEFAULTPREC) );
+      GEN c = sqrtr( _norml2(gel(matGS,j)) );
       gel(N2,j) = c; if (!signe(c)) goto PRECPB;
     }
     for (   ; j <= n; j+=2)
     {
-      GEN q1 = RgC_fpnorml2(gel(matGS,j  ), DEFAULTPREC);
-      GEN q2 = RgC_fpnorml2(gel(matGS,j+1), DEFAULTPREC);
+      GEN q1 = _norml2(gel(matGS, j));
+      GEN q2 = _norml2(gel(matGS, j+1));
       GEN c = sqrtr( gmul2n(addrr(q1, q2), -1) );
       gel(N2,j) = gel(N2,j+1) = c; if (!signe(c)) goto PRECPB;
     }
-    if (j > n) break; /* done */
+    break; /* done */
 PRECPB:
     prec = precdbl(prec);
-    remake_GM(nf, &F, prec); G = F.G;
+    nf = nfnewprec_shallow(nf, prec);
     if (DEBUGLEVEL>1) pari_warn(warnprec, "nf_factor_bound", prec);
   }
 
@@ -663,14 +698,14 @@ PRECPB:
 static GEN
 nf_Beauzamy_bound(GEN nf, GEN polbase)
 {
-  GEN lt, C, s, G = nf_get_G(nf), POL, bin;
-  long d = degpol(polbase), n = nf_get_degree(nf), prec   = MEDDEFAULTPREC;
+  GEN lt, C, s, POL, bin;
+  long d = degpol(polbase), n = nf_get_degree(nf), prec = nf_get_prec(nf);
   bin = vecbinome(d);
   POL = polbase + 2;
   /* compute [POL]_2 */
   for (;;)
   {
-    nffp_t F;
+    GEN G = nf_get_G(nf);
     long i;
 
     s = real_0(prec);
@@ -678,19 +713,18 @@ nf_Beauzamy_bound(GEN nf, GEN polbase)
     {
       GEN c = gel(POL,i);
       if (gequal0(c)) continue;
-      c = gnorml2(arch_for_T2_prec(G, c, prec));
+      c = _norml2(arch_for_T2(G,c));
       if (!signe(c)) goto PRECPB;
       /* s += T2(POL[i]) / binomial(d,i) */
       s = addrr(s, divri(c, gel(bin,i+1)));
     }
     break;
-
 PRECPB:
     prec = precdbl(prec);
-    remake_GM(nf, &F, prec); G = F.G;
+    nf = nfnewprec_shallow(nf, prec);
     if (DEBUGLEVEL>1) pari_warn(warnprec, "nf_factor_bound", prec);
   }
-  lt = leading_term(polbase);
+  lt = leading_coeff(polbase);
   s = mulri(s, muliu(sqri(lt), n));
   C = powruhalf(stor(3,DEFAULTPREC), 3 + 2*d); /* 3^{3/2 + d} */
   return divrr(mulrr(C, s), mulur(d, mppi(DEFAULTPREC)));
@@ -748,9 +782,8 @@ static GEN
 L2_bound(GEN nf, GEN den)
 {
   GEN M, L, prep, T = nf_get_pol(nf), tozk = nf_get_invzk(nf);
-  long prec;
-
-  prec = nbits2prec(bit_accuracy(ZX_max_lg(T)) + bit_accuracy(ZM_max_lg(tozk)));
+  long bit = bit_accuracy(ZX_max_lg(T)) + bit_accuracy(ZM_max_lg(tozk));
+  long prec = nbits2prec(bit + degpol(T));
   (void)initgaloisborne(nf, den, prec, &L, &prep, NULL);
   M = vandermondeinverse(L, RgX_gtofp(T,prec), den, prep);
   return RgM_fpnorml2(RgM_mul(tozk,M), DEFAULTPREC);
@@ -901,7 +934,7 @@ static void
 init_div_data(div_data *D, GEN pol, nflift_t *L)
 {
   GEN C = mul_content(L->topowden, L->dn);
-  GEN C2lt, Clt, lc = leading_term(pol), lt = is_pm1(lc)? NULL: absi(lc);
+  GEN C2lt, Clt, lc = leading_coeff(pol), lt = is_pm1(lc)? NULL: absi(lc);
   if (C)
   {
     GEN C2 = sqri(C);
@@ -1020,7 +1053,7 @@ nextK:
       if (T1)
       { /* d-1 test */
         t = get_trace(ind, T1);
-        if (rtodbl(RgC_fpnorml2(t,DEFAULTPREC)) > Bhigh)
+        if (rtodbl(_norml2(t)) > Bhigh)
         {
           if (DEBUGLEVEL>6) err_printf(".");
           avma = av; goto NEXT;
@@ -1029,7 +1062,7 @@ nextK:
       if (T2)
       { /* d-2 test */
         t = get_trace(ind, T2);
-        if (rtodbl(RgC_fpnorml2(t,DEFAULTPREC)) > Bhigh)
+        if (rtodbl(_norml2(t)) > Bhigh)
         {
           if (DEBUGLEVEL>3) err_printf("|");
           avma = av; goto NEXT;
@@ -1043,7 +1076,7 @@ nextK:
         if (y) q = gmul(y, q);
         y = FqX_centermod(q, Tpk, pk, pks2);
       }
-      y = nf_pol_lift(y, bound, T);
+      y = nf_pol_lift(y, bound, T->L);
       if (!y)
       {
         if (DEBUGLEVEL>3) err_printf("@");
@@ -1135,9 +1168,9 @@ nf_chk_factors(nfcmbf_t *T, GEN P, GEN M_L, GEN famod, GEN pk)
   GEN Tpk = T->L->Tpk;
   div_data D;
 
-  piv = special_pivot(M_L);
+  piv = ZM_hnf_knapsack(M_L);
   if (!piv) return NULL;
-  if (DEBUGLEVEL>3) err_printf("special_pivot output:\n%Ps\n",piv);
+  if (DEBUGLEVEL>3) err_printf("ZM_hnf_knapsack output:\n%Ps\n",piv);
 
   r  = lg(piv)-1;
   list = cgetg(r+1, t_VEC);
@@ -1148,7 +1181,7 @@ nf_chk_factors(nfcmbf_t *T, GEN P, GEN M_L, GEN famod, GEN pk)
     if (DEBUGLEVEL) err_printf("nf_LLL_cmbf: checking factor %ld\n", i);
     y = chk_factors_get(D.lt, famod, gel(piv,i), Tpk, pk);
 
-    if (! (y = nf_pol_lift(y, bound, T)) ) return NULL;
+    if (! (y = nf_pol_lift(y, bound, T->L)) ) return NULL;
     y = gerepilecopy(av, y);
     /* y is the candidate factor */
     pol = RgXQX_divrem(D.C2ltpol, y, nfT, ONLY_DIVIDES);
@@ -1235,7 +1268,7 @@ init_proj(nflift_t *L, GEN nfT)
   {
     GEN coTp = FpX_div(FpX_red(nfT, L->p), L->Tp,  L->p); /* Tp's cofactor */
     GEN z, proj;
-    z = ZpX_liftfact(nfT, mkvec2(L->Tp, coTp), NULL,  L->p, L->k, L->pk);
+    z = ZpX_liftfact(nfT, mkvec2(L->Tp, coTp), L->pk, L->p, L->k);
     L->Tpk = gel(z,1);
     proj = get_proj_modT(L->topow, L->Tpk, L->pk);
     if (L->topowden)
@@ -1338,11 +1371,11 @@ nf_LLL_cmbf(nfcmbf_t *T, long rec)
   long i, C, tmax, n0;
   GEN lP, Bnorm, Tra, T2, TT, CM_L, m, list, ZERO, Btra;
   double Bhigh;
-  pari_sp av, av2, lim;
+  pari_sp av, av2;
   long ti_LLL = 0, ti_CF = 0;
   pari_timer ti2, TI;
 
-  lP = absi(leading_term(P));
+  lP = absi(leading_coeff(P));
   if (is_pm1(lP)) lP = NULL;
 
   n0 = lg(famod) - 1;
@@ -1355,7 +1388,7 @@ nf_LLL_cmbf(nfcmbf_t *T, long rec)
   Bnorm = dbltor( n0 * C * C + Bhigh );
   ZERO = zeromat(n0, dnf);
 
-  av = avma; lim = stack_lim(av, 1);
+  av = avma;
   TT = cgetg(n0+1, t_VEC);
   Tra  = cgetg(n0+1, t_MAT);
   for (i=1; i<=n0; i++) TT[i] = 0;
@@ -1369,7 +1402,7 @@ nf_LLL_cmbf(nfcmbf_t *T, long rec)
 
     /* bound for f . S_k(genuine factor) = ZC * bound for T_2(S_tnew) */
     Btra = mulrr(ZC, mulur(dP*dP, normlp(Br, 2*tnew, dnf)));
-    bmin = logint(ceil_safe(sqrtr(Btra)), gen_2, NULL);
+    bmin = logint(ceil_safe(sqrtr(Btra)), gen_2) + 1;
     if (DEBUGLEVEL>2)
       err_printf("\nLLL_cmbf: %ld potential factors (tmax = %ld, bmin = %ld)\n",
                  r, tmax, bmin);
@@ -1381,7 +1414,7 @@ nf_LLL_cmbf(nfcmbf_t *T, long rec)
 
       bestlift_init((L->k)<<1, T->nf, T->pr, Btra, L);
       polred = ZqX_normalize(T->polbase, lP, L);
-      famod = ZpX_liftfact(polred, famod, L->Tpk, L->p, L->k, L->pk);
+      famod = ZqX_liftfact(polred, famod, L->Tpk, L->pk, L->p, L->k);
       for (i=1; i<=n0; i++) TT[i] = 0;
     }
     for (i=1; i<=n0; i++)
@@ -1468,7 +1501,7 @@ AGAIN:
       if (list) break;
     }
     CM_L = gerepilecopy(av2, CM_L);
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"nf_LLL_cmbf");
       gerepileall(av, L->Tpk? 9: 8,
@@ -1490,7 +1523,7 @@ nf_combine_factors(nfcmbf_t *T, GEN polred, long klim)
   pari_timer ti;
 
   if (DEBUGLEVEL>2) timer_start(&ti);
-  T->fact = ZpX_liftfact(polred, T->fact, L->Tpk, L->p, L->k, L->pk);
+  T->fact = ZqX_liftfact(polred, T->fact, L->Tpk, L->pk, L->p, L->k);
   if (DEBUGLEVEL>2) timer_printf(&ti, "Hensel lift");
   res = nfcmbf(T, klim, &maxK, &done);
   if (DEBUGLEVEL>2) timer_printf(&ti, "Naive recombination");
@@ -1511,8 +1544,7 @@ nf_combine_factors(nfcmbf_t *T, GEN polred, long klim)
 }
 
 static GEN
-nf_DDF_roots(GEN pol, GEN polred, GEN nfpol, GEN init_fa, long nbf,
-              long fl, nflift_t *L)
+nf_DDF_roots(GEN pol, GEN polred, GEN nfpol, long fl, nflift_t *L)
 {
   GEN z, Cltx_r, ltdn;
   long i, m;
@@ -1520,16 +1552,7 @@ nf_DDF_roots(GEN pol, GEN polred, GEN nfpol, GEN init_fa, long nbf,
 
   init_div_data(&D, pol, L);
   ltdn = mul_content(D.lt, L->dn);
-  if (L->Tpk)
-  {
-    int cof = (degpol(pol) > nbf); /* non trivial cofactor ? */
-    z = FqX_split_roots(init_fa, L->Tp, L->p, cof? polred: NULL);
-    z = ZpX_liftfact(polred, z, L->Tpk, L->p, L->k, L->pk);
-    if (cof) setlg(z, lg(z)-1); /* remove cofactor */
-    z = roots_from_deg1(z);
-  }
-  else
-    z = rootpadicfast(polred, L->p, L->k);
+  z = ZqX_roots(polred, L->Tpk, L->p, L->k);
   Cltx_r = deg1pol_shallow(D.Clt? D.Clt: gen_1, NULL, varn(pol));
   for (m=1,i=1; i<lg(z); i++)
   {
@@ -1582,9 +1605,9 @@ static long
 get_maxf(long nfdeg)
 {
   long maxf = 1;
-  if      (nfdeg >= 45) maxf =16;
-  else if (nfdeg >= 30) maxf = 8;
-  else if (nfdeg >= 15) maxf = 4;
+  if      (nfdeg >= 45) maxf =32;
+  else if (nfdeg >= 30) maxf =16;
+  else if (nfdeg >= 15) maxf = 8;
   return maxf;
 }
 
@@ -1599,7 +1622,7 @@ get_maxf(long nfdeg)
  *   Tp: polynomial defining Fq/Fp */
 static long
 nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
-              GEN *lt, GEN *Fa, GEN *pr, GEN *Tp)
+              GEN *lt, GEN *pr, GEN *Tp)
 {
   GEN nfpol = nf_get_pol(nf), bad = mulii(nf_get_disc(nf), nf_get_index(nf));
   long maxf, nfdeg = degpol(nfpol), dpol = degpol(polbase), nbf = 0;
@@ -1608,10 +1631,9 @@ nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
   pari_timer ti_pr;
 
   if (DEBUGLEVEL>3) timer_start(&ti_pr);
-  *lt  = leading_term(polbase); /* t_INT */
+  *lt  = leading_coeff(polbase); /* t_INT */
   if (gequal1(*lt)) *lt = NULL;
   *pr = NULL;
-  *Fa = NULL;
   *Tp = NULL;
 
   maxf = get_maxf(nfdeg);
@@ -1619,7 +1641,7 @@ nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
   /* select pr such that pol has the smallest number of factors, ct attempts */
   while ((pp = u_forprime_next(&S)))
   {
-    GEN aT, apr, ap, amodpr, red, r, fa = NULL;
+    GEN aT, apr, ap, amodpr, red, r;
     long anbf;
     ulong ltp = 0;
     pari_sp av2 = avma;
@@ -1631,7 +1653,7 @@ nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
     if (!r) continue;
 
     ap = utoipos(pp);
-    apr = primedec_apply_kummer(nf, Flx_to_ZX(r), 1, ap);
+    apr = idealprimedec_kummer(nf, Flx_to_ZX(r), 1, ap);
     amodpr = zk_to_Fq_init(nf,&apr,&aT,&ap);
 
     /* second step : evaluate factorisation mod apr */
@@ -1647,8 +1669,8 @@ nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
     {
       if (ltp) red = FqX_normalize(red, aT,ap);
       if (!FqX_is_squarefree(red,aT,ap)) { avma = av2; continue; }
-      anbf = fl == FACTORS? FqX_split_by_degree(&fa, red, aT, ap)
-                          : FqX_split_deg1(&fa, red, aT, ap);
+      anbf = fl == FACTORS? FqX_nbfact(red, aT, ap)
+                          : FqX_nbroots(red, aT, ap);
     }
     if (fl == ROOTS_SPLIT && anbf < dpol) return anbf;
     if (anbf <= 1)
@@ -1656,6 +1678,9 @@ nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
       if (fl == FACTORS) return anbf; /* irreducible */
       if (!anbf) return 0; /* no root */
     }
+    if (DEBUGLEVEL>3)
+      err_printf("%3ld %s at prime\n  %Ps\nTime: %ld\n",
+                 anbf, fl == FACTORS?"factors": "roots", apr, timer_delay(&ti_pr));
 
     if (!nbf || anbf < nbf
              || (anbf == nbf && pr_get_f(apr) > pr_get_f(*pr)))
@@ -1663,12 +1688,8 @@ nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
       nbf = anbf;
       *pr = apr;
       *Tp = aT;
-      *Fa = fa;
     }
     else avma = av2;
-    if (DEBUGLEVEL>3)
-      err_printf("%3ld %s at prime\n  %Ps\nTime: %ld\n",
-                 anbf, fl == FACTORS?"factors": "roots", apr, timer_delay(&ti_pr));
     if (--ct <= 0) break;
   }
   if (!nbf) pari_err_OVERFLOW("nf_pick_prime [ran out of primes]");
@@ -1688,7 +1709,7 @@ nfsqff_trager(GEN u, GEN T, GEN dent)
   fa = ZX_DDF(Q_primpart(n)); lx = lg(fa);
   if (lx == 2) return mkcol(u);
 
-  tmonic = is_pm1(leading_term(T));
+  tmonic = is_pm1(leading_coeff(T));
   P = cgetg(lx,t_COL);
   x0 = deg1pol_shallow(stoi(-k), gen_0, varn(T));
   mx0 = deg1pol_shallow(stoi(k), gen_0, varn(T));
@@ -1718,7 +1739,7 @@ polfnf(GEN a, GEN T)
 
   if (typ(a)!=t_POL) pari_err_TYPE("polfnf",a);
   if (typ(T)!=t_POL) pari_err_TYPE("polfnf",T);
-  T = Q_primpart(T); tmonic = is_pm1(leading_term(T));
+  T = Q_primpart(T); tmonic = is_pm1(leading_coeff(T));
   RgX_check_ZX(T,"polfnf");
   A = Q_primpart( QXQX_normalize(RgX_nffix("polfnf",T,a,1), T) );
   dA = degpol(A);
@@ -1755,7 +1776,7 @@ static GEN
 nfsqff(GEN nf, GEN pol, long fl, GEN den)
 {
   long n, nbf, dpol = degpol(pol);
-  GEN pr, C0, polbase, init_fa = NULL;
+  GEN pr, C0, polbase;
   GEN N2, res, polred, lt, nfpol = typ(nf)==t_POL?nf:nf_get_pol(nf);
   nfcmbf_t T;
   nflift_t L;
@@ -1790,7 +1811,7 @@ nfsqff(GEN nf, GEN pol, long fl, GEN den)
   }
 
   polbase = RgX_to_nfX(nf, pol);
-  nbf = nf_pick_prime(5, nf, polbase, fl, &lt, &init_fa, &pr, &L.Tp);
+  nbf = nf_pick_prime(5, nf, polbase, fl, &lt, &pr, &L.Tp);
   if (fl == ROOTS_SPLIT && nbf < dpol) return cgetg(1,t_VEC);
   if (nbf <= 1)
   {
@@ -1832,25 +1853,12 @@ nfsqff(GEN nf, GEN pol, long fl, GEN den)
   polred = ZqX_normalize(polbase, lt, &L); /* monic */
 
   if (fl != FACTORS) {
-    GEN z = nf_DDF_roots(pol, polred, nfpol, init_fa, nbf, fl, &L);
+    GEN z = nf_DDF_roots(pol, polred, nfpol, fl, &L);
     if (lg(z) == 1) return cgetg(1, t_VEC);
     return z;
   }
 
-  {
-    pari_sp av = avma;
-    if (L.Tp)
-      res = FqX_split_all(init_fa, L.Tp, L.p);
-    else
-    {
-      long d;
-      res = cgetg(dpol + 1, t_VEC); gel(res,1) = FpX_red(polred,L.p);
-      d = FpX_split_Berlekamp((GEN*)(res + 1), L.p);
-      setlg(res, d + 1);
-    }
-    gen_sort_inplace(res, (void*)&cmp_RgX, &gen_cmp_RgX, NULL);
-    T.fact  = gerepilecopy(av, res);
-  }
+  T.fact = gel(FqX_factor(polred, L.Tp, L.p), 1);
   if (DEBUGLEVEL>2) timer_printf(&ti, "splitting mod %Ps", pr);
   T.pr = pr;
   T.L  = &L;
@@ -1865,12 +1873,13 @@ nfsqff(GEN nf, GEN pol, long fl, GEN den)
 
 /* assume pol monic in nf.zk[X] */
 GEN
-nfroots_split(GEN nf, GEN pol)
+nfroots_if_split(GEN *pnf, GEN pol)
 {
-  GEN T = get_nfpol(nf,&nf), den = fix_nf(&nf, &T, &pol);
+  GEN T = get_nfpol(*pnf,pnf), den = fix_nf(pnf, &T, &pol);
   pari_sp av = avma;
-  GEN z = gerepilecopy(av, nfsqff(nf, pol, ROOTS_SPLIT, den));
-  return (lg(z) == 1)? NULL: mkvec2(z, nf);
+  GEN z = nfsqff(*pnf, pol, ROOTS_SPLIT, den);
+  if (lg(z) == 1) { avma = av; return NULL; }
+  return gerepilecopy(av, z);
 }
 
 /*******************************************************************/
@@ -1901,8 +1910,6 @@ typedef struct {
   nflift_t *L;
 } prklift_t;
 
-/* FIXME: check that all primes dividing n are ramified ! */
-
 /* Choose prime ideal unramified with "large" inertia degree */
 static void
 nf_pick_prime_for_units(GEN nf, prklift_t *P)
@@ -1922,7 +1929,7 @@ nf_pick_prime_for_units(GEN nf, prklift_t *P)
   }
   if (!r) pari_err_OVERFLOW("nf_pick_prime [ran out of primes]");
   ap = utoipos(pp);
-  apr = primedec_apply_kummer(nf, Flx_to_ZX(r), 1, ap);
+  apr = idealprimedec_kummer(nf, Flx_to_ZX(r), 1, ap);
   amodpr = zk_to_Fq_init(nf,&apr,&aT,&ap);
   P->pr = apr;
   P->q = pr_norm(apr);
@@ -1949,26 +1956,27 @@ mybestlift_bound(GEN C)
   return ceil(log(gtodouble(C)) / 0.2) + 3;
 }
 
-/* Returns the roots of the n_cyclo-th cyclotomic polynomial
- * if it splits, NULL otherwise */
+/* simplified nf_DDF_roots: polcyclo(n) monic in ZX either splits or has no
+ * root in nf.
+ * Return a root or NULL (no root) */
 static GEN
-nfcyclo_root(GEN nf, long n_cyclo, prklift_t *P)
+nfcyclo_root(long n, GEN nfpol, nflift_t *L)
 {
-  pari_sp av = avma;
-  GEN init_fa = NULL; /* factors mod pr */
-  GEN z, nfpol = nf_get_pol(nf), pol = polcyclo(n_cyclo, MAXVARN);
-  long nbf, deg = degpol(pol); /* = eulerphi(n_cyclo) */
-  if (P->L->Tp)
-    nbf = FqX_split_deg1(&init_fa, pol, P->L->Tp, P->L->p);
-  else
-  {
-    ulong p = itou(P->L->p);
-    nbf = Flx_nbroots(ZX_to_Flx(pol,p), p);
-  }
-  if (nbf != deg) return NULL; /* no roots in residue field */
-  z = nf_DDF_roots(pol, pol, nfpol, init_fa, nbf, ROOTS_SPLIT, P->L);
-  if (lg(z) == 1) { avma = av; return NULL; } /* no roots */
-  return gel(z,1);
+  GEN q, r, Cltx_r, pol = polcyclo(n,0), gn = utoipos(n);
+  div_data D;
+
+  init_div_data(&D, pol, L);
+  (void)Fq_sqrtn(gen_1, gn, L->Tp, L->p, &r);
+  /* r primitive n-th root of 1 in Fq */
+  r = Zq_sqrtnlift(gen_1, gn, r, L->Tpk, L->p, L->k);
+  /* lt*dn*topowden * r = Clt * r */
+  r = nf_bestlift_to_pol(r, NULL, L);
+  Cltx_r = deg1pol_shallow(D.Clt? D.Clt: gen_1, gneg(r), varn(pol));
+  /* check P(r) == 0 */
+  q = RgXQX_divrem(D.C2ltpol, Cltx_r, nfpol, ONLY_DIVIDES); /* integral */
+  if (!q) return NULL;
+  if (D.Clt) r = gdiv(r, D.Clt);
+  return r;
 }
 
 /* Guesses the number of roots of unity in number field [nf].
@@ -1986,7 +1994,7 @@ guess_roots(GEN nf)
 
   (void)u_forprime_init(&S, 3, ULONG_MAX);
   av = avma;
-  /* result must be stationnary (counter c) for at least B loops */
+  /* result must be stationary (counter c) for at least B loops */
   for (l=1; (p = u_forprime_next(&S)); l++)
   {
     GEN old, F, pf_1, Tp;
@@ -2005,7 +2013,7 @@ guess_roots(GEN nf)
     old = nbroots;
     nbroots = nbroots? gcdii(pf_1, nbroots): pf_1;
     if (DEBUGLEVEL>5)
-      err_printf("p=%ld; gcf(f(P/p))=%ld; nbroots | %Ps",p, gcdf, nbroots);
+      err_printf("p=%lu; gcf(f(P/p))=%ld; nbroots | %Ps",p, gcdf, nbroots);
     /* if same result go on else reset the stop counter [c] */
     if (old && equalii(nbroots,old))
     { if (!is_bigint(nbroots) && ++c > B) break; }
@@ -2066,7 +2074,7 @@ rootsof1(GEN nf)
 {
   prklift_t P;
   nflift_t L;
-  GEN fa, LP, LE, C0, z, prim_root, disc;
+  GEN fa, LP, LE, C0, z, prim_root, disc, nfpol;
   pari_timer ti;
   long i, l, nbguessed, nbroots, nfdegree;
   pari_sp av;
@@ -2155,14 +2163,16 @@ rootsof1(GEN nf)
 
   /* Step 4 : actual computation of roots */
   nbroots = 2; prim_root = gen_m1;
+  nfpol = nf_get_pol(nf);
   for (i = 1; i < l; i++)
   { /* for all prime power factors of nbguessed, find a p^k-th root of unity */
     long k, p = LP[i];
-    for (k = LE[i]; k > 0; k--)
+    for (k = minss(LE[i], Z_lval(subiu(P.q,1UL),p)); k > 0; k--)
     { /* find p^k-th roots */
+      pari_sp av = avma;
       long pk = upowuu(p,k);
       if (pk==2) continue; /* no need to test second roots ! */
-      z = nfcyclo_root(nf,pk,&P);
+      z = nfcyclo_root(pk, nfpol, P.L);
       if (DEBUGLEVEL>2) timer_printf(&ti, "for factoring Phi_%ld^%ld", p,k);
       if (z) {
         if (DEBUGLEVEL>2) err_printf("  %ld-th root of unity found.\n", pk);
@@ -2170,6 +2180,7 @@ rootsof1(GEN nf)
         else     { nbroots *= pk; prim_root = nfmul(nf, prim_root,z); }
         break;
       }
+      avma = av;
       if (DEBUGLEVEL) pari_warn(warner,"rootsof1: wrong guess");
     }
   }
@@ -2177,23 +2188,25 @@ rootsof1(GEN nf)
 }
 
 static long
-nf_pm1(GEN y) {
-  GEN z = gel(y,1);
-  return (is_pm1(z) && ZV_isscalar(y))? signe(z): 0;
+zk_equal1(GEN y)
+{
+  if (typ(y) == t_INT) return equali1(y);
+  return equali1(gel(y,1)) && ZV_isscalar(y);
 }
+/* x^w = 1 */
 static GEN
 is_primitive_root(GEN nf, GEN fa, GEN x, long w)
 {
-  GEN y, exp = utoipos(2), pp = gel(fa,1);
-  long i,p, l = lg(pp);
+  GEN P = gel(fa,1);
+  long i, l = lg(P);
 
-  for (i=1; i<l; i++)
+  for (i = 1; i < l; i++)
   {
-    p = itos(gel(pp,i));
-    exp[2] = w / p; y = nfpow(nf,x,exp);
-    if (nf_pm1(y) > 0) /* y = 1 */
+    long p = itos(gel(P,i));
+    GEN y = nfpow_u(nf,x, w/p);
+    if (zk_equal1(y) > 0) /* y = 1 */
     {
-      if (p!=2 || !gequal1(gcoeff(fa,i,2))) return NULL;
+      if (p != 2 || !equali1(gcoeff(fa,i,2))) return NULL;
       x = gneg_i(x);
     }
   }
