@@ -360,7 +360,7 @@ polgraeffe(GEN p)
 /**                                                                **/
 /********************************************************************/
 
-/* Quick approximation to log2(|x); first define y s.t. |y-x| < 2^-32 then
+/* Quick approximation to log2(|x|); first define y s.t. |y-x| < 2^-32 then
  * return y rounded to 2 ulp. In particular, if result < 2^21, absolute error
  * is bounded by 2^-31. If result > 2^21, it is correct to 2 ulp */
 static double
@@ -407,6 +407,7 @@ dbllog2(GEN z)
   switch(typ(z))
   {
     case t_INT: return mydbllog2i(z);
+    case t_FRAC: return mydbllog2i(gel(z,1))-mydbllog2i(gel(z,2));
     case t_REAL: return mydbllog2r(z);
     default: /*t_COMPLEX*/
       x = dbllog2mp(gel(z,1));
@@ -499,13 +500,6 @@ myshiftic(GEN z, long e)
   return signe(z)? mpshift(z,e): gen_0;
 }
 
-/* as real_1 with precision in bits, not in words */
-static GEN
-myreal_1(long bit)
-{
-  if (bit < 0) bit = 0;
-  return real_1(nbits2prec(bit));
-}
 static GEN
 RgX_gtofp_bit(GEN q, long bit)
 {
@@ -667,7 +661,7 @@ logmax_modulus(GEN p, double tau)
 
   eps = - 1/log(1.5*tau2); /* > 0 */
   bit = (long) ((double) n*log2(1./tau2)+3*log2((double) n))+1;
-  gunr = myreal_1(bit+2*n);
+  gunr = real_1_bit(bit+2*n);
   aux = gdiv(gunr, gel(p,2+n));
   q = RgX_Rg_mul(p, aux); gel(q,2+n) = gunr;
   e = findpower(q);
@@ -894,55 +888,11 @@ fft(GEN Omega, GEN p, GEN f, long step, long l)
   for (i=0; i<l; i++) f[i] = ff[i+1];
 }
 
-/* e(1/N) */
-static GEN
-RUgen(long N, long bit)
-{
-  if (N == 2) return real_m1(nbits2prec(bit));
-  if (N == 4) return gen_I();
-  return expIr(divru(Pi2n(1, nbits2prec(bit)), N));
-}
-
-/* N=2^k. returns a vector RU which contains exp(2*i*k*Pi/N), k=0..N-1 */
-static GEN
-initRU(long N, long bit)
-{
-  GEN *RU, z = RUgen(N, bit);
-  long i, N2 = (N>>1), N4 = (N>>2), N8 = (N>>3);
-
-  RU = (GEN*)cgetg(N+1,t_VEC); RU++;
-
-  RU[0] = myreal_1(bit);
-  RU[1] = z;
-  for (i=1; i<N8; i++)
-  {
-    GEN t = RU[i];
-    RU[i+1] = gmul(z, t);
-    RU[N4-i] = mkcomplex(gel(t,2), gel(t,1));
-  }
-  for (i=0; i<N4; i++) RU[i+N4] = mulcxI(RU[i]);
-  for (i=0; i<N2; i++) RU[i+N2] = gneg(RU[i]);
-  return (GEN)RU;
-}
-
-/* as above, N arbitrary */
-static GEN
-initRUgen(long N, long bit)
-{
-  GEN *RU = (GEN*)cgetg(N+1,t_VEC), z = RUgen(N,bit);
-  long i, k = (N+3)>>1;
-  RU[0] = gen_1;
-  RU[1] = z;
-  for (i=2; i<k; i++) RU[i] = gmul(z, RU[i-1]);
-  for (   ; i<N; i++) RU[i] = gconj(RU[N-i]);
-  return (GEN)RU;
-}
-
 GEN
 FFTinit(long k, long prec)
 {
   if (k <= 0) pari_err_DOMAIN("FFTinit", "k", "<=", gen_0, stoi(k));
-  return initRU(1L << k, prec2nbits(prec)) - 1;
+  return grootsof1(1L << k, prec);
 }
 
 GEN
@@ -1000,13 +950,21 @@ abs_update(GEN x, double *mu) {
   return y;
 }
 
+static void
+initdft(GEN *Omega, GEN *prim, long N, long Lmax, long bit)
+{
+  long prec = nbits2prec(bit);
+  *Omega = grootsof1(Lmax, prec) + 1;
+  *prim = rootsof1u_cx(N, prec);
+}
+
 static void
 parameters(GEN p, long *LMAX, double *mu, double *gamma,
            int polreal, double param, double param2)
 {
-  GEN q, pc, Omega, A, RU, prim, g, ONE,TWO;
+  GEN q, pc, Omega, A, RU, prim, g, TWO;
   long n = degpol(p), bit, NN, K, i, j, Lmax;
-  pari_sp av2, av = avma, lim = stack_lim(av, 1);
+  pari_sp av2, av = avma;
 
   bit = gexpo(p) + (long)param2+8;
   Lmax = 4; while (Lmax <= n) Lmax <<= 1;
@@ -1015,9 +973,7 @@ parameters(GEN p, long *LMAX, double *mu, double *gamma,
   NN = Lmax*K;
   if (polreal) K = K/2+1;
 
-  Omega = initRU(Lmax,bit);
-  prim = RUgen(NN, bit);
-
+  initdft(&Omega, &prim, NN, Lmax, bit);
   q = mygprec(p,bit) + 2;
   A = cgetg(Lmax+1,t_VEC); A++;
   pc= cgetg(Lmax+1,t_VEC); pc++;
@@ -1026,10 +982,9 @@ parameters(GEN p, long *LMAX, double *mu, double *gamma,
 
   *mu = pariINFINITY;
   g = real_0_bit(-bit);
-  ONE = real_1(DEFAULTPREC);
   TWO = real2n(1, DEFAULTPREC);
   av2 = avma;
-  RU = myreal_1(bit);
+  RU = gen_1;
   for (i=0; i<K; i++)
   {
     if (i) {
@@ -1046,9 +1001,9 @@ parameters(GEN p, long *LMAX, double *mu, double *gamma,
     if (polreal && i>0 && i<K-1)
       for (j=0; j<Lmax; j++) g = addrr(g, divrr(TWO, abs_update(gel(A,j),mu)));
     else
-      for (j=0; j<Lmax; j++) g = addrr(g, divrr(ONE, abs_update(gel(A,j),mu)));
+      for (j=0; j<Lmax; j++) g = addrr(g, invr(abs_update(gel(A,j),mu)));
     RU = gmul(RU, prim);
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"parameters");
       gerepileall(av2,2, &g,&RU);
@@ -1066,9 +1021,8 @@ dft(GEN p, long k, long NN, long Lmax, long bit, GEN F, GEN H, long polreal)
   long n = degpol(p), i, j, K;
   pari_sp ltop;
 
-  Omega = initRU(Lmax,bit);
-  prim = RUgen(NN, bit);
-  RU = cgetg(n+2,t_VEC); RU++;
+  initdft(&Omega, &prim, NN, Lmax, bit);
+  RU = cgetg(n+2,t_VEC) + 1;
 
   K = NN/Lmax; if (polreal) K = K/2+1;
   q = mygprec(p,bit);
@@ -1088,7 +1042,7 @@ dft(GEN p, long k, long NN, long Lmax, long bit, GEN F, GEN H, long polreal)
   for (i=1; i<=k; i++) gel(W,i) = gel(U,i) = gen_0;
 
   gel(RU,0) = gen_1;
-  prim2 = myreal_1(bit);
+  prim2 = gen_1;
   for (i=0; i<K; i++)
   {
     gel(RU,1) = prim2;
@@ -1110,16 +1064,16 @@ dft(GEN p, long k, long NN, long Lmax, long bit, GEN F, GEN H, long polreal)
       {
         for (j=1; j<=k; j++)
         {
-          gel(W,j) = gadd(gel(W,j), gshift(real_i(gmul(gel(A,j+1),gel(RU,j+1))),1));
-          gel(U,j) = gadd(gel(U,j), gshift(real_i(gmul(gel(B,j),gel(RU,j))),1));
+          gel(W,j) = gadd(gel(W,j), gshift(mulreal(gel(A,j+1),gel(RU,j+1)),1));
+          gel(U,j) = gadd(gel(U,j), gshift(mulreal(gel(B,j),gel(RU,j)),1));
         }
       }
       else
       {
         for (j=1; j<=k; j++)
         {
-          gel(W,j) = gadd(gel(W,j), real_i(gmul(gel(A,j+1),gel(RU,j+1))));
-          gel(U,j) = gadd(gel(U,j), real_i(gmul(gel(B,j),gel(RU,j))));
+          gel(W,j) = gadd(gel(W,j), mulreal(gel(A,j+1),gel(RU,j+1)));
+          gel(U,j) = gadd(gel(U,j), mulreal(gel(B,j),gel(RU,j)));
         }
       }
     }
@@ -1154,14 +1108,14 @@ static GEN
 refine_H(GEN F, GEN G, GEN HH, long bit, long Sbit)
 {
   GEN H = HH, D, aux;
-  pari_sp ltop = avma, lim = stack_lim(ltop, 1);
+  pari_sp ltop = avma;
   long error, i, bit1, bit2;
 
   D = Rg_RgX_sub(gen_1, RgX_rem(RgX_mul(H,G),F)); error = gexpo(D);
   bit2 = bit + Sbit;
   for (i=0; error>-bit && i<NEWTON_MAX && error<=0; i++)
   {
-    if (low_stack(lim, stack_lim(ltop,1)))
+    if (gc_needed(ltop,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"refine_H");
       gerepileall(ltop,2, &D,&H);
@@ -1185,7 +1139,7 @@ refine_F(GEN p, GEN *F, GEN *G, GEN H, long bit, double gamma)
 {
   GEN f0, FF, GG, r, HH = H;
   long error, i, bit1 = 0, bit2, Sbit, Sbit2,  enh, normF, normG, n = degpol(p);
-  pari_sp av = avma, lim = stack_lim(av, 1);
+  pari_sp av = avma;
 
   FF = *F; GG = RgX_divrem(p, FF, &r);
   error = gexpo(r); if (error <= -bit) error = 1-bit;
@@ -1198,7 +1152,7 @@ refine_F(GEN p, GEN *F, GEN *G, GEN H, long bit, double gamma)
   for (i=0; error>-bit && i<NEWTON_MAX && error<=0; i++)
   {
     if (bit1 == bit2 && i >= 2) { Sbit += n; Sbit2 += n; bit2 += n; }
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"refine_F");
       gerepileall(av,4, &FF,&GG,&r,&HH);
@@ -1304,13 +1258,13 @@ scalepol(GEN p, GEN R, long bit)
 
 /* return (conj(a)X-1)^n * p[ (X-a) / (conj(a)X-1) ] */
 static GEN
-conformal_pol(GEN p, GEN a, long bit)
+conformal_pol(GEN p, GEN a)
 {
   GEN z, r, ma = gneg(a), ca = gconj(a);
   long n = degpol(p), i;
-  pari_sp av = avma, lim = stack_lim(av,2);
+  pari_sp av = avma;
 
-  z = mkpoln(2, ca, negr(myreal_1(bit)));
+  z = mkpoln(2, ca, gen_m1);
   r = scalarpol(gel(p,2+n), 0);
   for (i=n-1; ; i--)
   {
@@ -1318,7 +1272,7 @@ conformal_pol(GEN p, GEN a, long bit)
     r = gadd(r, gmul(z, gel(p,2+i)));
     if (i == 0) return gerepileupto(av, r);
     z = addmulXn(gmul(z,ca), gneg(z), 1); /* z *= conj(a)X - 1 */
-    if (low_stack(lim, stack_lim(av,2)))
+    if (gc_needed(av,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"conformal_pol");
       gerepileall(av,2, &r,&z);
@@ -1402,7 +1356,7 @@ conformal_mapping(double *radii, GEN ctr, GEN p, long k, long bit,
   a = gmul(mygprec(a,bit2), mygprec(ctr,bit2)); /* a = -ctr/2sqrt(3) */
 
   av = avma;
-  q = conformal_pol(mygprec(p,bit2), a, bit2);
+  q = conformal_pol(mygprec(p,bit2), a);
   for (i=1; i<=n; i++)
     if (radii[i] != UNDEF) /* update array radii */
     {
@@ -1427,8 +1381,8 @@ conformal_mapping(double *radii, GEN ctr, GEN p, long k, long bit,
   GG = scalepol(GG,R,bit2);
 
   a = mygprec(a,bit2);
-  FF = conformal_pol(FF,a,bit2);
-  GG = conformal_pol(GG,a,bit2);
+  FF = conformal_pol(FF,a);
+  GG = conformal_pol(GG,a);
 
   a = invr(subsr(1, gnorm(a)));
   FF = RgX_Rg_mul(FF, powru(a,k));
@@ -1519,7 +1473,7 @@ static void
 split_1(GEN p, long bit, GEN *F, GEN *G)
 {
   long i, imax, n = degpol(p), polreal = isreal(p), ep = gexpo(p), bit2 = bit+n;
-  GEN TWO, ctr, q, qq, FF, GG, v, gr, r, newq;
+  GEN ctr, q, qq, FF, GG, v, gr, r, newq;
   double lrmin, lrmax, lthick;
   const double LOG3 = 1.098613;
 
@@ -1528,10 +1482,9 @@ split_1(GEN p, long bit, GEN *F, GEN *G)
   q = scalepol(p,gr,bit2);
 
   bit2 = bit + gexpo(q) - ep + (long)((double)n*2.*log2(3.)+1);
-  TWO = myreal_1(bit2); setexpo(TWO,1);
   v = cgetg(5,t_VEC);
-  gel(v,1) = TWO;
-  gel(v,2) = negr(TWO);
+  gel(v,1) = gen_2;
+  gel(v,2) = gen_m2;
   gel(v,3) = mkcomplex(gen_0, gel(v,1));
   gel(v,4) = mkcomplex(gen_0, gel(v,2));
   q = mygprec(q,bit2); lthick = 0;
@@ -1585,7 +1538,7 @@ split_0_2(GEN p, long bit, GEN *F, GEN *G)
   {
     if (k > n/2) k = n/2;
     bit2 += k<<1;
-    FF = monomial(myreal_1(bit2), k, 0);
+    FF = pol_xn(k, 0);
     GG = RgX_shift_shallow(q, -k);
   }
   else
@@ -1631,7 +1584,7 @@ split_0(GEN p, long bit, GEN *F, GEN *G)
   if (k > 0)
   {
     if (k > n/2) k = n/2;
-    *F = monomial(myreal_1(bit), k, 0);
+    *F = pol_xn(k, 0);
     *G = RgX_shift_shallow(p, -k);
   }
   else
@@ -1788,15 +1741,14 @@ split_complete(GEN p, long bit, GEN roots_pol)
 }
 
 static GEN
-quickabs(GEN x)
+quicktofp(GEN x)
 {
   const long prec = DEFAULTPREC;
-  GEN y;
   switch(typ(x))
   {
-    case t_INT: y = itor(x, prec); setabssign(y); return y;
-    case t_REAL: y = rtor(x, prec); setabssign(y); return y;
-    case t_FRAC: y = fractor(x, prec); setabssign(y); return y;
+    case t_INT: return itor(x, prec);
+    case t_REAL: return rtor(x, prec);
+    case t_FRAC: return fractor(x, prec);
     case t_COMPLEX: {
       GEN a = gel(x,1), b = gel(x,2);
       /* avoid problem with 0, e.g. x = 0 + I*1e-100. We don't want |x| = 0. */
@@ -1805,32 +1757,75 @@ quickabs(GEN x)
       a = cxcompotor(a, prec);
       b = cxcompotor(b, prec); return sqrtr(addrr(sqrr(a), sqrr(b)));
     }
-    default: pari_err_TYPE("quickabs",x);
+    default: pari_err_TYPE("quicktofp",x);
       return NULL;/*not reached*/
   }
 
 }
 
-/* bound ln |largest root of p| */
+/* bound log_2 |largest root of p| (Fujiwara's bound) */
 double
-cauchy_bound(GEN p)
+fujiwara_bound(GEN p)
 {
   pari_sp av = avma;
   long i, n = degpol(p);
-  GEN invlc;
-  double Lmax = -pariINFINITY;
+  GEN cc;
+  double loglc, Lmax;
 
-  if (n <= 0) pari_err_CONSTPOL("cauchy_bound");
-  invlc = invr( quickabs(gel(p,n+2)) ); /* 1 / |lc(p)| */
-  for (i = 0; i < n; i++)
+  if (n <= 0) pari_err_CONSTPOL("fujiwara_bound");
+  loglc = mydbllog2r( quicktofp(gel(p,n+2)) ); /* log_2 |lc(p)| */
+  cc = gel(p, 2);
+  if (gequal0(cc))
+    Lmax = -pariINFINITY-1;
+  else
+    Lmax = (mydbllog2r(quicktofp(cc)) - loglc - 1) / n;
+  for (i = 1; i < n; i++)
   {
     GEN y = gel(p,i+2);
     double L;
     if (gequal0(y)) continue;
-    L = mydbllogr(mulrr(quickabs(y), invlc)) / (n-i);
+    L = (mydbllog2r(quicktofp(y)) - loglc) / (n-i);
     if (L > Lmax) Lmax = L;
   }
-  avma = av; return Lmax + LOG2;
+  avma = av; return Lmax + 1;
+}
+
+/* Fujiwara's bound, real roots. Based on the following remark: if
+ *   p = x^n + sum a_i x^i and q = x^n + sum min(a_i,0)x^i
+ * then for all x >= 0, p(x) >= q(x). Thus any bound for the (positive) roots
+ * of q is a bound for the positive roots of p. */
+double
+fujiwara_bound_real(GEN p, long sign)
+{
+  pari_sp av = avma;
+  GEN x;
+  long n = degpol(p), i, signodd, signeven;
+  double fb;
+  if (n <= 0) pari_err_CONSTPOL("fujiwara_bound");
+  x = shallowcopy(p);
+  if (gsigne(gel(x, n+2)) > 0)
+  {
+    signeven = 1;
+    signodd = sign;
+  }
+  else
+  {
+    signeven = -1;
+    signodd = -sign;
+  }
+  for (i = 0; i < n; i++)
+  {
+    if ((n - i) % 2)
+    {
+      if (gsigne(gel(x, i+2)) == signodd ) gel(x, i+2) = gen_0;
+    }
+    else
+    {
+      if (gsigne(gel(x, i+2)) == signeven) gel(x, i+2) = gen_0;
+    }
+  }
+  fb = fujiwara_bound(x);
+  avma = av; return fb;
 }
 
 static GEN
@@ -1891,8 +1886,8 @@ fix_roots(GEN r, GEN *m, long h, long bit)
   long i, j, k, l, prec;
   GEN allr, ro1;
   if (h == 1) return fix_roots1(r);
-  ro1 = initRUgen(h, bit);
   prec = nbits2prec(bit);
+  ro1 = grootsof1(h, prec) + 1;
   l = lg(r)-1;
   allr = cgetg(h*l+1, t_VEC);
   for (k=1,i=1; i<=l; i++)
@@ -1928,8 +1923,8 @@ all_roots(GEN p, long bit)
   long bit0,  bit2, i, e, h, n = degpol(p);
   pari_sp av;
 
-  pd = RgX_deflate_max(p, &h); lc = leading_term(pd);
-  e = (long)((2/LOG2) * cauchy_bound(pd)); if (e < 0) e = 0;
+  pd = RgX_deflate_max(p, &h); lc = leading_coeff(pd);
+  e = (long)(2 * fujiwara_bound(pd)); if (e < 0) e = 0;
   bit0 = bit + gexpo(pd) - gexpo(lc) + (long)log2(n/h)+1+e;
   bit2 = bit0; e = 0;
   for (av=avma,i=1;; i++,avma=av)
@@ -1940,7 +1935,7 @@ all_roots(GEN p, long bit)
     q[1] = evalsigne(1)|evalvarn(0);
     m = split_complete(q,bit2,roots_pol);
     roots_pol = fix_roots(roots_pol, &m, h, bit2);
-    q = mygprec_special(p,bit2); lc = leading_term(q);
+    q = mygprec_special(p,bit2); lc = leading_coeff(q);
     q[1] = evalsigne(1)|evalvarn(0);
     if (h > 1) m = gmul(m,lc);
 
@@ -2182,3 +2177,717 @@ QX_complex_roots(GEN p, long l)
   L = all_roots(Q_primpart(p), bit);
   return gerepileupto(av, clean_roots(L, l, bit, 1));
 }
+
+/********************************************************************/
+/**                                                                **/
+/**                REAL ROOTS OF INTEGER POLYNOMIAL                **/
+/**                                                                **/
+/********************************************************************/
+
+/* Count sign changes in the coefficients of (x+1)^deg(P)*P(1/(x+1))
+ * The inversion is implicit (we take coefficients backwards). Roots of P
+ * at 0 and 1 (mapped to oo and 0) are ignored here and must be dealt with
+ * by the caller */
+static long
+X2XP1(GEN P, long deg, int *root1, GEN *Premapped)
+{
+  const pari_sp av = avma;
+  GEN v = shallowcopy(P);
+  long i, j, vlim, nb, s;
+
+  for (i = 0, vlim = deg+2;;)
+  {
+    for (j = 2; j < vlim; j++) gel(v, j+1) = addii(gel(v, j), gel(v, j+1));
+    s = -signe(gel(v, vlim));
+    vlim--; i++; if (s) break;
+  }
+  if (vlim == deg+1) *root1 = 0;
+  else
+  {
+    *root1 = 1;
+    if (Premapped) setlg(v, vlim + 2);
+  }
+
+  nb = 0;
+  for (; i < deg; i++)
+  {
+    long s2 = -signe(gel(v, 2));
+    int flag = (s2 == s);
+    for (j = 2; j < vlim; j++)
+    {
+      gel(v, j+1) = addii(gel(v, j), gel(v, j+1));
+      if (flag) flag = (s2 != signe(gel(v, j+1)));
+    }
+    if (s == signe(gel(v, vlim)))
+    {
+      if (++nb >= 2) { avma = av; return 2; }
+      s = -s;
+    }
+    /* if flag is set there will be no further sign changes */
+    if (flag && (!Premapped || !nb)) goto END;
+    vlim--;
+    if (gc_needed(av, 3))
+    {
+      if (!Premapped) setlg(v, vlim + 2);
+      v = gerepileupto(av, v);
+      if (DEBUGMEM > 1) pari_warn(warnmem, "X2XP1");
+    }
+  }
+  if (s == signe(gel(v, vlim))) nb++;
+END:
+  if (Premapped && nb == 1) *Premapped = v; else avma = av;
+  return nb;
+}
+
+static long
+_intervalcmp(GEN x, GEN y)
+{
+  if (typ(x) == t_VEC) x = gel(x, 1);
+  if (typ(y) == t_VEC) y = gel(y, 1);
+  return gcmp(x, y);
+}
+
+static GEN
+_gen_nored(void *E, GEN x) { (void)E; return x; }
+static GEN
+_mp_add(void *E, GEN x, GEN y) { (void)E; return mpadd(x, y); }
+static GEN
+_mp_sub(void *E, GEN x, GEN y) { (void)E; return mpsub(x, y); }
+static GEN
+_mp_mul(void *E, GEN x, GEN y) { (void)E; return mpmul(x, y); }
+static GEN
+_mp_sqr(void *E, GEN x) { (void)E; return mpsqr(x); }
+static GEN
+_gen_one(void *E) { (void)E; return gen_1; }
+static GEN
+_gen_zero(void *E) { (void)E; return gen_0; }
+
+static struct bb_algebra mp_algebra = { _gen_nored, _mp_add, _mp_sub,
+                         _mp_mul, _mp_sqr, _gen_one, _gen_zero };
+
+static GEN
+_mp_cmul(void *E, GEN P, long a, GEN x) {(void)E; return mpmul(gel(P,a+2), x);}
+
+/* Split the polynom P in two parts, whose coeffs have constant sign:
+ * P(X) = X^D*Pp + Pm. Also compute the two parts of the derivative of P,
+ * Pprimem = Pm', Pprimep = X*Pp'+ D*Pp => P' = X^(D-1)*Pprimep + Pprimem;
+ * Pprimep[i] = (i+D) Pp[i]. Return D */
+static long
+split_pols(GEN P, GEN *pPp, GEN *pPm, GEN *pPprimep, GEN *pPprimem)
+{
+  long i, D, dP = degpol(P), s0 = signe(gel(P,2));
+  GEN Pp, Pm, Pprimep, Pprimem;
+  for(i=1; i <= dP; i++)
+    if (signe(gel(P, i+2)) == -s0) break;
+  D = i;
+  Pm = cgetg(D + 2, t_POL);
+  Pprimem = cgetg(D + 1, t_POL);
+  Pp = cgetg(dP-D + 3, t_POL);
+  Pprimep = cgetg(dP-D + 3, t_POL);
+  Pm[1] = Pp[1] = Pprimem[1] = Pprimep[1] = P[1];
+  for(i=0; i < D; i++)
+  {
+    GEN c = gel(P, i+2);
+    gel(Pm, i+2) = c;
+    if (i) gel(Pprimem, i+1) = mului(i, c);
+  }
+  for(; i <= dP; i++)
+  {
+    GEN c = gel(P, i+2);
+    gel(Pp, i+2-D) = c;
+    gel(Pprimep, i+2-D) = mului(i, c);
+  }
+  *pPm = normalizepol_lg(Pm, D+2);
+  *pPprimem = normalizepol_lg(Pprimem, D+1);
+  *pPp = normalizepol_lg(Pp, dP-D+3);
+  *pPprimep = normalizepol_lg(Pprimep, dP-D+3);
+  return dP - degpol(*pPp);
+}
+
+static GEN
+bkeval_single_power(long d, GEN V)
+{
+  long mp = lg(V) - 2;
+  if (d > mp) return gmul(gpowgs(gel(V, mp+1), d/mp), gel(V, (d%mp)+1));
+  return gel(V, d+1);
+}
+
+static GEN
+splitpoleval(GEN Pp, GEN Pm, GEN pows, long D, long bitprec)
+{
+  GEN vp = gen_bkeval_powers(Pp, degpol(Pp), pows, NULL, &mp_algebra, _mp_cmul);
+  GEN vm = gen_bkeval_powers(Pm, degpol(Pm), pows, NULL, &mp_algebra, _mp_cmul);
+  GEN xa = bkeval_single_power(D, pows);
+  GEN r;
+  if (!signe(vp)) return vm;
+  vp = gmul(vp, xa);
+  r = gadd(vp, vm);
+  if (gexpo(vp) - (signe(r)? gexpo(r): 0) > prec2nbits(realprec(vp)) - bitprec)
+    return NULL;
+  return r;
+}
+
+/* optimized Cauchy bound for P = X^D*Pp + Pm, D > deg(Pm) */
+static GEN
+splitcauchy(GEN Pp, GEN Pm, long prec)
+{
+  GEN S = gel(Pp,2), A = gel(Pm,2);
+  long i, lPm = lg(Pm), lPp = lg(Pp);
+  for (i=3; i < lPm; i++) { GEN c = gel(Pm,i); if (abscmpii(A, c) < 0) A = c; }
+  for (i=3; i < lPp; i++) S = addii(S, gel(Pp, i));
+  return subsr(1, rdivii(A, S, prec)); /* 1 + |Pm|_oo / |Pp|_1 */
+}
+
+/* Newton for polynom P, P(0)!=0, with unique sign change => one root in ]0,oo[
+ * P' has also at most one zero there */
+static GEN
+polsolve(GEN P, long bitprec)
+{
+  pari_sp av = avma;
+  GEN Pp, Pm, Pprimep, Pprimem, pows;
+  GEN Pprime, Pprime2;
+  GEN ra, rb, rc, rcold = NULL, Pc, Ppc;
+  long deg = degpol(P), D, rt;
+  long s0, cprec = DEFAULTPREC, prec = nbits2prec(bitprec);
+  long bitaddprec, addprec;
+  long expoold = LONG_MAX, iter;
+  if (deg == 1)
+    return gerepileuptoleaf(av, rdivii(negi(gel(P,2)), gel(P,3), prec));
+  Pprime = ZX_deriv(P);
+  Pprime2 = ZX_deriv(Pprime);
+  bitaddprec = 1 + 2*expu(deg); addprec = nbits2prec(bitaddprec);
+  D = split_pols(P, &Pp, &Pm, &Pprimep, &Pprimem); /* P = X^D*Pp + Pm */
+  s0 = signe(gel(P, 2));
+  rt = maxss(D, brent_kung_optpow(maxss(degpol(Pp), degpol(Pm)), 2, 1));
+  rb = splitcauchy(Pp, Pm, DEFAULTPREC);
+  for(;;)
+  {
+    pows = gen_powers(rb, rt, 1, NULL, _mp_sqr, _mp_mul, _gen_one);
+    Pc = splitpoleval(Pp, Pm, pows, D, bitaddprec);
+    if (!Pc) { cprec++; rb = rtor(rb, cprec); continue; }
+    if (signe(Pc) != s0) break;
+    shiftr_inplace(rb,1);
+  }
+  ra = NULL;
+  iter = 0;
+  for(;;)
+  {
+    GEN wdth;
+    iter++;
+    if (ra)
+      rc = shiftr(addrr(ra, rb), -1);
+    else
+      rc = shiftr(rb, -1);
+    do
+    {
+      pows = gen_powers(rc, rt, 1, NULL, _mp_sqr, _mp_mul, _gen_one);
+      Pc = splitpoleval(Pp, Pm, pows, D, bitaddprec+2);
+      if (Pc) break;
+      cprec++; rc = rtor(rc, cprec);
+    } while (1);
+    if (signe(Pc) == s0)
+      ra = rc;
+    else
+      rb = rc;
+    if (!ra) continue;
+    wdth = subrr(rb, ra);
+    if (!(iter % 8))
+    {
+      GEN m1 = poleval(Pprime, ra), M2;
+      if (signe(m1) == s0) continue;
+      M2 = poleval(Pprime2, rb);
+      if (abscmprr(gmul(M2, wdth), shiftr(m1, 1)) > 0) continue;
+      break;
+    }
+    else if (gexpo(wdth) <= -bitprec)
+      break;
+  }
+  rc = rb;
+  iter = 0;
+  for(;;)
+  {
+    long exponew;
+    GEN dist;
+    iter++;
+    rcold = rc;
+    pows = gen_powers(rc, rt, 1, NULL, _mp_sqr, _mp_mul, _gen_one);
+    Ppc = splitpoleval(Pprimep, Pprimem, pows, D-1, bitaddprec+4);
+    if (Ppc)
+      Pc = splitpoleval(Pp, Pm, pows, D, bitaddprec+4);
+    if (!Ppc || !Pc)
+    {
+      if (cprec >= prec+addprec)
+        cprec++;
+      else
+        cprec = minss(2*cprec, prec+addprec);
+      rc = rtor(rc, cprec); /* Backtrack one step */
+      continue;
+    }
+    dist = typ(Ppc) == t_REAL? divrr(Pc, Ppc): divri(Pc, Ppc);
+    rc = subrr(rc, dist);
+    if (cmprr(ra, rc) > 0 || cmprr(rb, rc) < 0)
+    {
+      if (cprec >= prec+addprec) break;
+      cprec = minss(2*cprec, prec+addprec);
+      rc = rtor(rcold, cprec); /* Backtrack one step */
+      continue;
+    }
+    if (expoold == LONG_MAX) { expoold = expo(dist); continue; }
+    exponew = expo(dist);
+    if (exponew < -bitprec - 1)
+    {
+      if (cprec >= prec+addprec) break;
+      cprec = minss(2*cprec, prec+addprec);
+      rc = rtor(rc, cprec); continue;
+    }
+    if (exponew > expoold - 2)
+    {
+      if (cprec >= prec+addprec) break;
+      cprec = minss(2*cprec, prec+addprec);
+      rc = rtor(rc, cprec);
+      expoold = LONG_MAX; continue;
+    }
+    expoold = exponew;
+  }
+  return gerepileuptoleaf(av, rtor(rc, prec));
+}
+
+static GEN
+usp(GEN Q0, long deg, long flag, long bitprec)
+{
+  const pari_sp av = avma;
+  GEN Q, sol, c, Lc, Lk;
+  long listsize = 64, nbr = 0, nb_todo, ind, deg0, indf, i, k, nb;
+
+
+  sol = const_col(deg, gen_0);
+  deg0 = deg;
+  Lc = const_vec(listsize, gen_0);
+  Lk = cgetg(listsize+1, t_VECSMALL);
+  c = gen_0;
+  k = Lk[1] = 0;
+  ind = 1; indf = 2;
+  Q = leafcopy(Q0);
+
+  nb_todo = 1;
+  while (nb_todo)
+  {
+    GEN nc = gel(Lc, ind), Qremapped;
+    pari_sp av2;
+    int root1;
+    if (Lk[ind] == k + 1)
+    {
+      deg0 = deg;
+      setlg(Q0, deg + 3);
+      Q0 = ZX_rescale2n(Q0, 1);
+      Q = Q_primpart(Q0);
+      c = gen_0;
+    }
+
+    if (!equalii(nc, c)) Q = ZX_translate(Q, subii(nc, c));
+
+    k = Lk[ind];
+    c = nc;
+    ind++;
+    nb_todo--;
+
+    if (equalii(gel(Q, 2), gen_0))
+    { /* Q(0) = 0 */
+      GEN s = gmul2n(c, -k);
+      long j;
+      for (j = 1; j <= nbr; j++)
+        if (gequal(gel(sol, j), s)) break;
+      if (j > nbr) gel(sol, ++nbr) = s;
+
+      deg0--;
+      for (j = 2; j <= deg0 + 2; j++) gel(Q, j) = gel(Q, j+1);
+      setlg(Q, j);
+    }
+
+    av2 = avma;
+    nb = X2XP1(Q, deg0, &root1, flag == 1 ? &Qremapped : NULL);
+
+    if      (nb == 0) /* no root in this open interval */;
+    else if (nb == 1) /* exactly one root */
+    {
+      GEN s = gen_0;
+      if (flag == 0)
+        s = mkvec2(gmul2n(c,-k), gmul2n(addiu(c,1),-k));
+      else if (flag == 1) /* Caveat: Qremapped is the reciprocal polynomial */
+      {
+        s = polsolve(Qremapped, bitprec+1);
+        s = divrr(s, addsr(1, s));
+        s = gmul2n(addir(c, s), -k);
+        s = rtor(s, nbits2prec(bitprec));
+      }
+      gel(sol, ++nbr) = gerepileupto(av2, s);
+    }
+    else
+    { /* unknown, add two nodes to refine */
+      if (indf + 2 > listsize)
+      {
+        if (ind>1)
+        {
+          for (i = ind; i < indf; i++)
+          {
+            gel(Lc, i-ind+1) = gel(Lc, i);
+            Lk[i-ind+1] = Lk[i];
+          }
+          indf -= ind-1;
+          ind = 1;
+        }
+        if (indf + 2 > listsize)
+        {
+          listsize *= 2;
+          Lc = vec_lengthen(Lc, listsize);
+          Lk = vecsmall_lengthen(Lk, listsize);
+        }
+        for (i = indf; i <= listsize; i++) gel(Lc, i) = gen_0;
+      }
+      nc = shifti(c, 1);
+      gel(Lc, indf) = nc;
+      gel(Lc, indf + 1) = addiu(nc, 1);
+      Lk[indf] = Lk[indf + 1] = k + 1;
+      indf += 2;
+      nb_todo += 2;
+    }
+    if (root1)
+    { /* Q(1) = 0 */
+      GEN s = gmul2n(addiu(c,1), -k);
+      long j;
+      for (j = 1; j <= nbr; j++)
+        if (gequal(gel(sol, j), s)) break;
+      if (j > nbr) gel(sol, ++nbr) = s;
+    }
+
+    if (gc_needed(av, 2))
+    {
+      gerepileall(av, 6, &Q0, &Q, &c, &Lc, &Lk, &sol);
+      if (DEBUGMEM > 1) pari_warn(warnmem, "ZX_Uspensky", avma);
+    }
+  }
+
+  setlg(sol, nbr+1);
+  return gerepilecopy(av, sol);
+}
+
+static GEN
+ZX_Uspensky_cst_pol(long nbz, long flag, long bitprec)
+{
+  switch(flag)
+  {
+    case 0:  return zerocol(nbz);
+    case 1:  retconst_col(nbz, real_0_bit(bitprec));
+    default: return utoi(nbz);
+  }
+}
+
+/* ZX_Uspensky(P, [a,a], flag) */
+static GEN
+ZX_Uspensky_equal(GEN P, GEN a, long flag)
+{
+  if (typ(a) != t_INFINITY && gequal0(poleval(P, a)))
+    return flag <= 1 ? mkcol(a): gen_1;
+  else
+    return flag <= 1 ? cgetg(1, t_COL) : gen_0;
+}
+
+GEN
+ZX_Uspensky(GEN P, GEN ab, long flag, long bitprec)
+{
+  pari_sp av = avma;
+  GEN a, b, sol = NULL, Pcur;
+  double fb;
+  long nbz, deg;
+
+  deg = degpol(P);
+  if (deg == 0) return flag <= 1 ? cgetg(1, t_COL) : gen_0;
+  if (ab)
+  {
+    if (typ(ab) == t_VEC)
+    {
+      if (lg(ab) != 3) pari_err_DIM("ZX_Uspensky");
+      a = gel(ab, 1);
+      b = gel(ab, 2);
+    }
+    else
+    {
+      a = ab;
+      b = mkoo();
+    }
+  }
+  else
+  {
+    a = mkmoo();
+    b = mkoo();
+  }
+  switch (gcmp(a, b))
+  {
+    case 1: avma = av; return flag <= 1 ? cgetg(1, t_COL) : gen_0;
+    case 0: return gerepilecopy(av, ZX_Uspensky_equal(P, a, flag));
+  }
+  nbz = ZX_valrem(P, &Pcur);
+  deg -= nbz;
+  if (!nbz) Pcur = P;
+  if (nbz && (gsigne(a) > 0 || gsigne(b) < 0)) nbz = 0;
+  if (deg == 0) { avma = av; return ZX_Uspensky_cst_pol(nbz, flag, bitprec); }
+  if (deg == 1)
+  {
+    sol = gdiv(gneg(gel(Pcur, 2)), pollead(Pcur, -1));
+    if (gcmp(a, sol) > 0 || gcmp(sol, b) > 0)
+    {
+      avma = av;
+      return ZX_Uspensky_cst_pol(nbz, flag, bitprec);
+    }
+    if (flag >= 2) { avma = av; return utoi(nbz+1); }
+    sol = gconcat(zerocol(nbz), mkcol(sol));
+    if (flag == 1) sol = RgC_gtofp(sol, nbits2prec(bitprec));
+    return gerepilecopy(av, sol);
+  }
+  switch(flag)
+  {
+    case 0:
+      sol = zerocol(nbz);
+      break;
+    case 1:
+      sol = const_col(nbz, real_0_bit(bitprec));
+      break;
+    /* case 2: nothing */
+  }
+
+  if (typ(a) == t_INFINITY && typ(b) != t_INFINITY && gsigne(b))
+  {
+    fb = fujiwara_bound_real(Pcur, -1);
+    if (fb > -pariINFINITY) a = negi(int2n((long)ceil(fb))); else a = gen_0;
+  }
+  if (typ(b) == t_INFINITY && typ(a) != t_INFINITY && gsigne(a))
+  {
+    fb = fujiwara_bound_real(Pcur, 1);
+    if (fb > -pariINFINITY) b = int2n((long)ceil(fb)); else b = gen_0;
+  }
+
+  if (typ(a) != t_INFINITY && typ(b) != t_INFINITY)
+  {
+    GEN den, diff, unscaledres, co, Pdiv, ascaled;
+    pari_sp av1;
+    long i;
+    if (gequal(a,b)) /* can occur if one of a,b was initially a t_INFINITY */
+      return gerepilecopy(av, ZX_Uspensky_equal(P, a, flag));
+    den = lcmii(Q_denom(a), Q_denom(b));
+    if (!is_pm1(den))
+    {
+      Pcur = ZX_rescale(Pcur, den);
+      ascaled = gmul(a, den);
+    }
+    else
+    {
+      den = NULL;
+      ascaled = a;
+    }
+    diff = subii(den ? gmul(b,den) : b, ascaled);
+    Pcur = ZX_unscale(ZX_translate(Pcur, ascaled), diff);
+    av1 = avma;
+    Pdiv = cgetg(deg+2, t_POL);
+    Pdiv[1] = Pcur[1];
+    co = gel(Pcur, deg+2);
+    for (i = deg; --i >= 0; )
+    {
+      gel(Pdiv, i+2) = co;
+      co = addii(co, gel(Pcur, i+2));
+    }
+    if (!signe(co))
+    {
+      Pcur = Pdiv;
+      deg--;
+      if (flag <= 1)
+        sol = gconcat(sol, b);
+      else
+        nbz++;
+    }
+    else
+      avma = av1;
+    unscaledres = usp(Pcur, deg, flag, bitprec);
+    if (flag <= 1)
+    {
+      for (i = 1; i < lg(unscaledres); i++)
+      {
+        GEN z = gmul(diff, gel(unscaledres, i));
+        if (typ(z) == t_VEC)
+        {
+          gel(z, 1) = gadd(ascaled, gel(z, 1));
+          gel(z, 2) = gadd(ascaled, gel(z, 2));
+        }
+        else
+          z = gadd(ascaled, z);
+        if (den) z = gdiv(z, den);
+        gel(unscaledres, i) = z;
+      }
+      sol = gconcat(sol, unscaledres);
+    }
+    else
+      nbz += lg(unscaledres) - 1;
+  }
+  if (typ(b) == t_INFINITY && (fb=fujiwara_bound_real(Pcur, 1)) > -pariINFINITY)
+  {
+    GEN Pcurp, unscaledres;
+    long bp = (long)ceil(fb);
+    if (bp < 0) bp = 0;
+    Pcurp = ZX_unscale2n(Pcur, bp);
+    unscaledres = usp(Pcurp, deg, flag, bitprec);
+    if (flag <= 1)
+      sol = shallowconcat(sol, gmul2n(unscaledres, bp));
+    else
+      nbz += lg(unscaledres)-1;
+  }
+  if (typ(a) == t_INFINITY && (fb=fujiwara_bound_real(Pcur,-1)) > -pariINFINITY)
+  {
+    GEN Pcurm, unscaledres;
+    long i, bm = (long)ceil(fb);
+    if (bm < 0) bm = 0;
+    Pcurm = ZX_unscale2n(ZX_unscale(Pcur, gen_m1), bm);
+    unscaledres = usp(Pcurm, deg, flag, bitprec);
+    if (flag <= 1)
+    {
+      for (i = 1; i < lg(unscaledres); i++)
+      {
+        GEN z = gneg(gmul2n(gel(unscaledres, i), bm));
+        if (typ(z) == t_VEC) swap(gel(z, 1), gel(z, 2));
+        gel(unscaledres, i) = z;
+      }
+      sol = shallowconcat(unscaledres, sol);
+    }
+    else
+      nbz += lg(unscaledres)-1;
+  }
+  if (flag >= 2) return utoi(nbz);
+  if (flag)
+    sol = sort(sol);
+  else
+    sol = gen_sort(sol, (void *)_intervalcmp, cmp_nodata);
+  return gerepileupto(av, sol);
+}
+
+/* x a scalar */
+static GEN
+rootsdeg0(GEN x)
+{
+  if (!is_rational_t(typ(x))) pari_err_TYPE("realroots",x);
+  if (gequal0(x)) pari_err_ROOTS0("realroots");
+  return cgetg(1,t_COL); /* constant polynomial */
+}
+static void
+checkbound(GEN a)
+{
+  switch(typ(a))
+  {
+    case t_INT: case t_FRAC: case t_INFINITY: break;
+    default: pari_err_TYPE("polrealroots", a);
+  }
+}
+static GEN
+check_ab(GEN ab)
+{
+  GEN a, b;
+  if (!ab) return NULL;
+  if (typ(ab) != t_VEC || lg(ab) != 3) pari_err_TYPE("polrootsreal",ab);
+  a = gel(ab,1); checkbound(a);
+  b = gel(ab,2); checkbound(b);
+  if (typ(a) == t_INFINITY && inf_get_sign(a) < 0 &&
+      typ(b) == t_INFINITY && inf_get_sign(b) > 0) ab = NULL;
+  return ab;
+}
+GEN
+realroots(GEN P, GEN ab, long prec)
+{
+  pari_sp av = avma;
+  long nrr = 0;
+  GEN sol = NULL, fa, ex;
+  long i, j, v;
+
+  ab = check_ab(ab);
+  if (typ(P) != t_POL) return rootsdeg0(P);
+  switch(degpol(P))
+  {
+    case -1: return rootsdeg0(gen_0);
+    case 0: return rootsdeg0(gel(P,2));
+  }
+  P = Q_primpart(P);
+  RgX_check_ZX(P,"realroots");
+  v = ZX_valrem(P,&P);
+  if (v && (!ab || (gsigne(gel(ab,1)) <= 0 && gsigne(gel(ab,2)) >= 0)))
+    sol = const_col(v, real_0(prec));
+  fa = ZX_squff(P, &ex);
+  for (i = 1; i < lg(fa); i++)
+  {
+    GEN Pi = gel(fa, i), soli, soli2 = NULL;
+    long n, nrri = 0, h;
+    if (ab)
+      h = 1;
+    else
+      Pi = ZX_deflate_max(Pi, &h);
+    soli = ZX_Uspensky(Pi, h%2 ? ab: gen_0, 1, prec2nbits(prec));
+    n = lg(soli);
+    if (!(h % 2)) soli2 = cgetg(n, t_COL);
+    for (j = 1; j < n; j++)
+    {
+      GEN elt = gel(soli, j); /* != 0 */
+      if (typ(elt) != t_REAL)
+      {
+        nrri++; if (h > 1 && !(h % 2)) nrri++;
+        elt = gtofp(elt, prec);
+        gel(soli, j) = elt;
+      }
+      if (h > 1)
+      {
+        GEN r;
+        if (h == 2)
+          r = sqrtr(elt);
+        else
+        {
+          if (signe(elt) < 0)
+            r = negr(sqrtnr(negr(elt), h));
+          else
+            r = sqrtnr(elt, h);
+        }
+        gel(soli, j) = r;
+        if (!(h % 2)) gel(soli2, j) = negr(r);
+      }
+    }
+    if (!(h % 2)) soli = shallowconcat(soli, soli2);
+    if (ex[i] > 1) soli = shallowconcat1( const_vec(ex[i], soli) );
+    sol = sol? shallowconcat(sol, soli): soli;
+    nrr += ex[i]*nrri;
+  }
+  if (!sol) { avma = av; return cgetg(1,t_COL); }
+
+  if (DEBUGLEVEL > 4)
+  {
+    err_printf("Number of real roots: %d\n", lg(sol)-1);
+    err_printf(" -- of which 2-integral: %ld\n", nrr);
+  }
+  return gerepileupto(av, sort(sol));
+}
+
+/* P non-constant, squarefree ZX */
+long
+ZX_sturm(GEN P)
+{
+  pari_sp av = avma;
+  long h, r;
+  P = ZX_deflate_max(P, &h);
+  if (odd(h))
+    r = itos(ZX_Uspensky(P, NULL, 2, 0));
+  else
+    r = 2*itos(ZX_Uspensky(P, gen_0, 2, 0));
+  avma = av; return r;
+}
+/* P non-constant, squarefree ZX */
+long
+ZX_sturmpart(GEN P, GEN ab)
+{
+  pari_sp av = avma;
+  long r;
+  if (!check_ab(ab)) return ZX_sturm(P);
+  r = itos(ZX_Uspensky(P, ab, 2, 0));
+  avma = av; return r;
+}
