@@ -158,9 +158,9 @@ pi_ramanujan(long prec)
 
   nmax = (long)(1 + prec2nbits(prec)/alpha2);
 #ifdef LONG_IS_64BIT
-  D = utoipos(10939058860032000); /* C^3/24 */
+  D = utoipos(10939058860032000UL); /* C^3/24 */
 #else
-  D = uutoi(2546948,495419392);
+  D = uutoi(2546948UL,495419392UL);
 #endif
   abpq_init(&S, nmax);
   S.a[0] = utoipos(A);
@@ -335,7 +335,7 @@ mpeuler(long prec) { return rtor(consteuler(prec), prec); }
 static GEN
 catalan(long prec)
 {
-  long i, nmax = bit_accuracy(prec) >> 1;
+  long i, nmax = prec2nbits(prec) >> 1;
   struct abpq_res R;
   struct abpq A;
   GEN u, v;
@@ -412,7 +412,7 @@ static GEN
 mpexp0(GEN x)
 {
   long e = expo(x);
-  return e >= 0? real_0_bit(e): real_1(nbits2prec(-e));
+  return e >= 0? real_0_bit(e): real_1_bit(-e);
 }
 static GEN
 powr0(GEN x)
@@ -483,6 +483,8 @@ _sqr(void *data /* ignored */, GEN x) { (void)data; return gsqr(x); }
 static GEN
 _mul(void *data /* ignored */, GEN x, GEN y) { (void)data; return gmul(x,y); }
 static GEN
+_one(void *x) { return gpowg0((GEN) x); }
+static GEN
 _sqri(void *data /* ignored */, GEN x) { (void)data; return sqri(x); }
 static GEN
 _muli(void *data /* ignored */, GEN x, GEN y) { (void)data; return mulii(x,y); }
@@ -490,6 +492,8 @@ static GEN
 _sqrr(void *data /* ignored */, GEN x) { (void)data; return sqrr(x); }
 static GEN
 _mulr(void *data /* ignored */, GEN x, GEN y) { (void)data; return mulrr(x,y); }
+static GEN
+_oner(void *data /* prec */) { return real_1( *(long*) data); }
 
 /* INTEGER POWERING (a^n for integer a != 0 and integer n > 0)
  *
@@ -670,8 +674,9 @@ upowuu(ulong p, ulong k)
     q *= q ;
     return odd(k)? q*p: q;
   }
-#endif
+#else
   return 0;
+#endif
 }
 
 typedef struct {
@@ -741,6 +746,13 @@ powru(GEN x, ulong n)
   return gerepileuptoleaf(av,y);
 }
 
+GEN
+powersr(GEN x, long n)
+{
+  long prec = realprec(x);
+  return gen_powers(x, n, 1, &prec, &_sqrr, &_mulr, &_oner);
+}
+
 /* x^(s/2), assume x t_REAL */
 GEN
 powrshalf(GEN x, long s)
@@ -818,7 +830,7 @@ powps(GEN x, long n)
   if (v == 0) mod = icopy(mod);
   else
   {
-    if (precp(x) == 1 && equaliu(p, 2)) v++;
+    if (precp(x) == 1 && absequaliu(p, 2)) v++;
     mod = mulii(mod, powiu(p,v));
     mod = gerepileuptoint((pari_sp)y, mod);
   }
@@ -1138,6 +1150,10 @@ gpow(GEN x, GEN n, long prec)
 
     case t_FFELT:
       return gerepileupto(av,FF_pow(FF_sqrtn(x,d,NULL),a));
+    default:
+      if (gequal(d, gen_2))
+        return gerepileupto(av, gmul(powgi(x, shifti(subis(a, 1), -1)),
+                                     gsqrt(x, prec)));
     }
   }
   i = (long) precision(n); if (i) prec=i;
@@ -1145,36 +1161,113 @@ gpow(GEN x, GEN n, long prec)
   return gerepileupto(av, gexp(y,prec));
 }
 
+GEN
+gpowers0(GEN x, long n, GEN x0)
+{
+  long i, l;
+  GEN V;
+  if (!x0) return gpowers(x,n);
+  if (n < 0) return cgetg(1,t_VEC);
+  l = n+2; V = cgetg(l, t_VEC); gel(V,1) = gcopy(x0);
+  for (i = 2; i < l; i++) gel(V,i) = gmul(gel(V,i-1),x);
+  return V;
+}
+
+GEN
+gpowers(GEN x, long n)
+{
+  if (n < 0) return cgetg(1,t_VEC);
+  return gen_powers(x, n, 1, (void*)x, &_sqr, &_mul, &_one);
+}
+
+/* return [q^1,q^4,...,q^{n^2}] */
+GEN
+gsqrpowers(GEN q, long n)
+{
+  pari_sp av = avma;
+  GEN L = gpowers0(gsqr(q), n, q); /* L[i] = q^(2i - 1), i <= n+1 */
+  GEN v = cgetg(n+1, t_VEC);
+  long i;
+  gel(v, 1) = gcopy(q);
+  for (i = 2; i <= n ; ++i) gel(v, i) = q = gmul(q, gel(L,i)); /* q^(i^2) */
+  return gerepileupto(av, v);
+}
+
+/* 4 | N. returns a vector RU which contains exp(2*i*k*Pi/N), k=0..N-1 */
+static GEN
+grootsof1_4(long N, long prec)
+{
+  GEN z, RU = cgetg(N+1,t_VEC), *v  = ((GEN*)RU) + 1;
+  long i, N2 = (N>>1), N4 = (N>>2), N8 = (N>>3);
+  /* z^N2 = -1, z^N4 = I; if z^k = a+I*b, then z^(N4-k) = I*conj(z) = b+a*I */
+
+  v[0] = gen_1; v[1] = z = rootsof1u_cx(N, prec);
+  if (odd(N4)) N8++;
+  for (i=1; i<N8; i++)
+  {
+    GEN t = v[i];
+    v[i+1] = gmul(z, t);
+    v[N4-i] = mkcomplex(gel(t,2), gel(t,1));
+  }
+  for (i=0; i<N4; i++) v[i+N4] = mulcxI(v[i]);
+  for (i=0; i<N2; i++) v[i+N2] = gneg(v[i]);
+  return RU;
+}
+
+/* as above, N arbitrary */
+GEN
+grootsof1(long N, long prec)
+{
+  GEN z, RU, *v;
+  long i, k;
+
+  if ((N & 3) == 0) return grootsof1_4(N, prec);
+  if (N == 1) return mkvec(gen_1);
+  k = (N+3)>>1;
+  RU = cgetg(N+1,t_VEC);
+  v  = ((GEN*)RU) + 1;
+  v[0] = gen_1; v[1] = z = rootsof1u_cx(N, prec);
+  for (i=2; i<k; i++) v[i] = gmul(z, v[i-1]);
+  for (   ; i<N; i++) v[i] = gconj(v[N-i]);
+  return RU;
+}
+
 /********************************************************************/
 /**                                                                **/
 /**                        RACINE CARREE                           **/
 /**                                                                **/
 /********************************************************************/
-/* assume x unit, precp(x) = pp > 3 */
-static GEN
-sqrt_2adic(GEN x, long pp)
+/* assume x unit, e = precp(x) */
+GEN
+Z2_sqrt(GEN x, long e)
 {
-  GEN z = mod16(x)==(signe(x)>=0?1:15)?gen_1: utoipos(3);
-  long zp;
-  pari_sp av, lim;
+  ulong r = signe(x)>=0?mod16(x):16-mod16(x);
+  GEN z;
+  long ez;
+  pari_sp av;
 
-  if (pp == 4) return z;
-  zp = 3; /* number of correct bits in z (compared to sqrt(x)) */
-
-  av = avma; lim = stack_lim(av,2);
+  switch(e)
+  {
+    case 1: return gen_1;
+    case 2: return (r & 3UL) == 1? gen_1: NULL;
+    case 3: return (r & 7UL) == 1? gen_1: NULL;
+    case 4: if (r == 1) return gen_1;
+            else return (r == 9)? utoipos(3): NULL;
+    default: if ((r&7UL) != 1) return NULL;
+  }
+  av = avma; z = (r==1)? gen_1: utoipos(3);
+  ez = 3; /* number of correct bits in z (compared to sqrt(x)) */
   for(;;)
   {
     GEN mod;
-    zp = (zp<<1) - 1;
-    if (zp > pp) zp = pp;
-    mod = int2n(zp);
-    z = addii(z, remi2n(mulii(x, Fp_inv(z,mod)), zp));
+    ez = (ez<<1) - 1;
+    if (ez > e) ez = e;
+    mod = int2n(ez);
+    z = addii(z, remi2n(mulii(x, Fp_inv(z,mod)), ez));
     z = shifti(z, -1); /* (z + x/z) / 2 */
-    if (pp == zp) return z;
-
-    if (zp < pp) zp--;
-
-    if (low_stack(lim,stack_lim(av,2)))
+    if (e == ez) return gerepileuptoint(av, z);
+    if (ez < e) ez--;
+    if (gc_needed(av,2))
     {
       if (DEBUGMEM > 1) pari_warn(warnmem,"Qp_sqrt");
       z = gerepileuptoint(av,z);
@@ -1183,39 +1276,6 @@ sqrt_2adic(GEN x, long pp)
 }
 
 /* x unit defined modulo p^e, e > 0 */
-static GEN
-Up_sqrt(GEN x, GEN p, long e)
-{
-  pari_sp av = avma;
-  if (equaliu(p,2))
-  {
-    long r = signe(x)>=0?mod8(x):8-mod8(x);
-    if (e <= 3)
-    {
-      switch(e) {
-      case 1: break;
-      case 2: if ((r&3) == 1) break;
-              return NULL;
-      case 3: if (r == 1) break;
-              return NULL;
-      }
-      return gen_1;
-    }
-    else
-    {
-      if (r != 1) return NULL;
-      return gerepileuptoint(av, sqrt_2adic(x, e));
-    }
-  }
-  else
-  {
-    GEN z = Fp_sqrt(x, p);
-    if (!z) return NULL;
-    if (e <= 1) return z;
-    return gerepileuptoint(av, Zp_sqrtlift(x, z, p, e));
-  }
-}
-
 GEN
 Qp_sqrt(GEN x)
 {
@@ -1230,9 +1290,9 @@ Qp_sqrt(GEN x)
   mod = gel(x,3);
   z   = gel(x,4); /* lift to t_INT */
   e >>= 1;
-  z = Up_sqrt(z, p, pp);
+  z = Zp_sqrt(z, p, pp);
   if (!z) return NULL;
-  if (equaliu(p,2))
+  if (absequaliu(p,2))
   {
     pp  = (pp <= 3) ? 1 : pp-1;
     mod = int2n(pp);
@@ -1247,16 +1307,16 @@ Qp_sqrt(GEN x)
 GEN
 Zn_sqrt(GEN d, GEN fn)
 {
-  pari_sp ltop = avma, btop, st_lim;
+  pari_sp ltop = avma, btop;
   GEN b = gen_0, m = gen_1;
   long j, np;
   if (typ(d) != t_INT) pari_err_TYPE("Zn_sqrt",d);
   if (typ(fn) == t_INT)
-    fn = absi_factor(fn);
+    fn = absZ_factor(fn);
   else if (!is_Z_factorpos(fn))
     pari_err_TYPE("Zn_sqrt",fn);
   np = nbrows(fn);
-  btop = avma; st_lim = stack_lim(btop, 1);
+  btop = avma;
   for (j = 1; j <= np; ++j)
   {
     GEN  bp, mp, pr, r;
@@ -1267,7 +1327,7 @@ Zn_sqrt(GEN d, GEN fn)
     else
     {
       if (odd(v)) return NULL;
-      bp = Up_sqrt(r, p, e-v);
+      bp = Zp_sqrt(r, p, e-v);
       if (!bp)    return NULL;
       if (v) bp = mulii(bp, powiu(p, v>>1L));
     }
@@ -1275,7 +1335,7 @@ Zn_sqrt(GEN d, GEN fn)
     pr = mulii(m, mp);
     b = Z_chinese_coprime(b, bp, m, mp, pr);
     m = pr;
-    if (low_stack(st_lim, stack_lim(btop, 1)))
+    if (gc_needed(btop, 1))
       gerepileall(btop, 2, &b, &m);
   }
   return gerepileupto(ltop, b);
@@ -1314,7 +1374,8 @@ sqrt_ser(GEN b, long prec)
     y = sqr_ser_part(x, lold, l-1) - lold;
     for (j = lold+2; j < l+2; j++) gel(y,j) = gsub(gel(y,j), gel(a,j));
     y += lold; setvalp(y, lold);
-    y = gsub(x, gdiv(y, x2)); /* = gmuloldn(gadd(x, gdiv(a,x)), -1); */
+    y = normalize(y);
+    y = gsub(x, gdiv(y, x2)); /* = gmul2n(gsub(x, gdiv(a,x)), -1); */
     for (j = lold+2; j < l+2; j++) gel(x,j) = gel(y,j);
     lold = l;
   }
@@ -1398,63 +1459,64 @@ gsqrt(GEN x, long prec)
 /**                          N-th ROOT                             **/
 /**                                                                **/
 /********************************************************************/
-/* exp(2Ipi/n), assume n positive t_INT */
-GEN
-rootsof1complex(GEN n, long prec)
+static void
+bug_logp(GEN p)
 {
-  pari_sp av = avma;
-  if (is_pm1(n)) return real_1(prec);
-  if (equaliu(n, 2)) return stor(-1, prec);
-  return gerepileupto(av, expIr( divri(Pi2n(1, prec), n) ));
+  if (!BPSW_psp(p)) pari_err_PRIME("p-adic log",p);
+  pari_err_BUG("log_p");
 }
-
-/*Only the O() of y is used*/
-GEN
-rootsof1padic(GEN n, GEN y)
-{
-  GEN z, r = cgetp(y), p = gel(y,2);
-  pari_sp av = avma;
-  z = rootsof1_Fp(n, p);
-  z = Zp_sqrtnlift(gen_1, n, z, p, precp(y));
-  affii(z, gel(r,4)); avma = av; return r;
-}
-
 /* Let x = 1 mod p and y := (x-1)/(x+1) = 0 (p). Then
  * log(x) = log(1+y) - log(1-y) = 2 \sum_{k odd} y^k / k.
  * palogaux(x) returns the last sum (not multiplied by 2) */
 static GEN
 palogaux(GEN x)
 {
-  long k,e,pp;
+  long i, k, e, pp, t;
   GEN y,s,y2, p = gel(x,2);
+  int is2 = absequaliu(p,2);
 
-  if (equalii(gen_1, gel(x,4)))
+  y = subiu(gel(x,4), 1);
+  if (!signe(y))
   {
     long v = valp(x)+precp(x);
-    if (equaliu(p,2)) v--;
+    if (is2) v--;
     return zeropadic(p, v);
   }
+  /* optimize t: log(x) = log(x^(p^t)) / p^t */
+  e = Z_pval(y, p); /* valp(y) = e >= 1; precp(y) = precp(x)-e */
+  if (!e) bug_logp(p);
+  if (is2)
+    t = sqrt( (double)(precp(x)-e) / e ); /* instead of (2*e) */
+  else
+    t = sqrt( (double)(precp(x)-e) / (e * (expi(p) + hammingweight(p))) );
+  for (i = 0; i < t; i++) x = gpow(x, p, 0);
+
   y = gdiv(gaddgs(x,-1), gaddgs(x,1));
   e = valp(y); /* > 0 */
-  if (e <= 0) {
-    if (!BPSW_psp(p)) pari_err_PRIME("p-adic log",p);
-    pari_err_BUG("log_p");
-  }
-  pp = e+precp(y);
-  if (equaliu(p,2)) pp--;
+  if (e <= 0) bug_logp(p);
+  pp = precp(y) + e;
+  if (is2) pp--;
   else
   {
     GEN p1;
-    for (p1=utoipos(e); cmpui(pp,p1) > 0; pp++) p1 = mulii(p1, p);
+    for (p1=utoipos(e); abscmpui(pp,p1) > 0; pp++) p1 = mulii(p1, p);
     pp -= 2;
   }
   k = pp/e; if (!odd(k)) k--;
-  y2 = gsqr(y); s = gdivgs(gen_1,k);
-  while (k > 2)
+  if (DEBUGLEVEL>5)
+    err_printf("logp: [pp,k,e,t] = [%ld,%ld,%ld,%ld]\n",pp,k,e,t);
+  if (k > 1)
   {
-    k -= 2; s = gadd(gmul(y2,s), gdivgs(gen_1,k));
+    y2 = gsqr(y); s = gdivgs(gen_1,k);
+    while (k > 2)
+    {
+      k -= 2;
+      s = gadd(gmul(y2,s), gdivgs(gen_1,k));
+    }
+    y = gmul(s,y);
   }
-  return gmul(s,y);
+  if (t) setvalp(y, valp(y) - t);
+  return y;
 }
 
 GEN
@@ -1465,7 +1527,7 @@ Qp_log(GEN x)
 
   if (!signe(a)) pari_err_DOMAIN("Qp_log", "argument", "=", gen_0, x);
   y = leafcopy(x); setvalp(y,0);
-  if (equaliu(p,2))
+  if (absequaliu(p,2))
     y = palogaux(gsqr(y));
   else if (gequal1(modii(a, p)))
     y = gmul2n(palogaux(y), 1);
@@ -1497,7 +1559,7 @@ Qp_sqrtn_ram(GEN x, long e)
     setvalp(x,0);
   }
   /*If p = 2, -1 is a root of 1 in U1: need extra check*/
-  if (equaliu(p, 2) && mod8(gel(x,4)) != 1) return NULL;
+  if (absequaliu(p, 2) && mod8(gel(x,4)) != 1) return NULL;
   a = Qp_log(x);
   va = valp(a) - e;
   if (va <= 0)
@@ -1552,7 +1614,7 @@ Qp_sqrtn(GEN x, GEN n, GEN *zetan)
   pari_sp av, tetpil;
   GEN q, p;
   long e;
-  if (equaliu(n, 2))
+  if (absequaliu(n, 2))
   {
     if (zetan) *zetan = gen_m1;
     if (signe(n) < 0) x = ginv(x);
@@ -1574,7 +1636,7 @@ Qp_sqrtn(GEN x, GEN n, GEN *zetan)
     if (signe(q) < 0) x = ginv(x);
     x = gerepileupto(av, x);
     if (zetan)
-      *zetan = (e && equaliu(p, 2))? gen_m1 /*-1 in Q_2*/
+      *zetan = (e && absequaliu(p, 2))? gen_m1 /*-1 in Q_2*/
                                    : gen_1;
     return x;
   }
@@ -1585,7 +1647,7 @@ Qp_sqrtn(GEN x, GEN n, GEN *zetan)
   if (zetan)
   {
     GEN *gptr[2];
-    if (e && equaliu(p, 2))/*-1 in Q_2*/
+    if (e && absequaliu(p, 2))/*-1 in Q_2*/
     {
       tetpil = avma; x = gcopy(x); *zetan = gneg(*zetan);
     }
@@ -1642,6 +1704,18 @@ sqrtnint(GEN a, long n)
   return gerepileuptoleaf(ltop, x);
 }
 
+GEN
+cbrtr_abs(GEN x)
+{
+  long prec = realprec(x), n = bit_accuracy(prec), e = expo(x), er = e / 3;
+  /* x 2^3(n-er) = b t_INT */
+  GEN b = mantissa2nr(x, (e-3*er+1) + n*2);
+  b = sqrtnint(b, 3);
+  b = itor(b, prec);
+  setexpo(b, expo(b)+er-n);
+  return b;
+}
+
 ulong
 usqrtn(ulong a, ulong n)
 {
@@ -1732,8 +1806,16 @@ gsqrtn(GEN x, GEN n, GEN *zetan, long prec)
         y = real_0_bit(b);
     }
     else
-      y = gerepileupto(av, gexp(gdiv(glog(x,prec), n), prec));
-    if (zetan) *zetan = rootsof1complex(n,prec);
+    {
+      long nn = itos_or_0(n);
+      if (tx == t_INT) { x = itor(x,prec); tx = t_REAL; }
+      if (nn > 0 && tx == t_REAL && signe(x) > 0)
+        y = sqrtnr(x, nn);
+      else
+        y = gexp(gdiv(glog(x,prec), n), prec);
+      y = gerepileupto(av, y);
+    }
+    if (zetan) *zetan = rootsof1_cx(n,prec);
     return y;
 
   case t_QUAD:
@@ -1835,6 +1917,7 @@ mpexpm1(GEN x)
   return gerepileupto(av, divrr(y, z));
 }
 
+static GEN serexp(GEN x, long prec);
 GEN
 gexpm1(GEN x, long prec)
 {
@@ -1842,6 +1925,26 @@ gexpm1(GEN x, long prec)
   {
     case t_REAL: return mpexpm1(x);
     case t_COMPLEX: return cxexpm1(x,prec);
+    case t_PADIC: return gsubgs(Qp_exp(x), 1);
+    default:
+    {
+      pari_sp av = avma;
+      long ey;
+      GEN y;
+      if (!(y = toser_i(x))) break;
+      ey = valp(y);
+      if (ey < 0) pari_err_DOMAIN("expm1","valuation", "<", gen_0, x);
+      if (gequal0(y)) return gcopy(y);
+      if (ey)
+        return gerepileupto(av, gsubgs(serexp(y,prec), 1));
+      else
+      {
+        GEN e1 = gexpm1(gel(y,2), prec), e = gaddgs(e1,1);
+        y = gmul(e, serexp(serchop0(y),prec));
+        gel(y,2) = e1;
+        return gerepilecopy(av, y);
+      }
+    }
   }
   return trans_eval("expm1",gexpm1,x,prec);
 }
@@ -1938,7 +2041,7 @@ Qp_exp_prec(GEN x)
 {
   long k, e = valp(x), n = e + precp(x);
   GEN p = gel(x,2);
-  int is2 = equaliu(p,2);
+  int is2 = absequaliu(p,2);
   if (e < 1 || (e == 1 && is2)) return -1;
   if (is2)
   {
@@ -2038,6 +2141,7 @@ serchop0(GEN s)
   long i, l = lg(s);
   GEN y;
   if (l == 2) return s;
+  if (l == 3 && isexactzero(gel(s,2))) return s;
   y = cgetg(l, t_SER); y[1] = s[1];
   gel(y,2) = gen_0; for (i=3; i <l; i++) gel(y,i) = gel(s,i);
   return normalize(y);
@@ -2214,12 +2318,55 @@ zellagmcx(GEN a0, GEN b0, GEN r, GEN t, long prec)
   return gerepileupto(av,gdiv(t,a1));
 }
 
+static long
+ser_cmp_expo(GEN A, GEN B)
+{
+  long e = -(long)HIGHEXPOBIT, d = valp(B) - valp(A);
+  long i, la = lg(A), v = varn(B);
+  for (i = 2; i < la; i++)
+  {
+    GEN a = gel(A,i), b;
+    long ei;
+    if (isexactzero(a)) continue;
+    b = polcoeff_i(B, i-2 + d, v);
+    ei = gexpo(a);
+    if (!isexactzero(b)) ei -= gexpo(b);
+    e = maxss(e, ei);
+  }
+  return e;
+}
+
+static GEN
+ser_agm1(GEN y, long prec)
+{
+  GEN a1 = y, b1 = gen_1;
+  long l = lg(y)-2, l2 = 6-prec2nbits(prec), eold = LONG_MAX;
+  for(;;)
+  {
+    GEN a = a1, p1;
+    a1 = gmul2n(gadd(a,b1),-1);
+    b1 = gsqrt(gmul(a,b1), prec);
+    p1 = gsub(b1,a1);
+    if (isinexactreal(p1))
+    {
+      long e = ser_cmp_expo(p1, b1);
+      if (e < l2 || e >= eold) break;
+      eold = e;
+    }
+    else
+    {
+      long ep = valp(p1)-valp(b1);
+      if (ep >= l && gequal0(p1)) break;
+    }
+  }
+  return a1;
+}
+
 /* agm(1,x) */
 static GEN
 agm1(GEN x, long prec)
 {
-  GEN p1, a1, b1, y;
-  long l, l2, ep;
+  GEN y;
   pari_sp av;
 
   if (gequal0(x)) return gcopy(x);
@@ -2236,34 +2383,26 @@ agm1(GEN x, long prec)
       return agm1cx(x, prec);
 
     case t_PADIC:
+    {
+      GEN a1 = x, b1 = gen_1;
+      long l = precp(x);
       av = avma;
-      a1 = x; b1 = gen_1; l = precp(x);
-      do
+      for(;;)
       {
-        GEN a = a1;
+        GEN a = a1, p1;
+        long ep;
         a1 = gmul2n(gadd(a,b1),-1);
         a = gmul(a,b1);
         b1 = Qp_sqrt(a); if (!b1) pari_err_SQRTN("Qp_sqrt",a);
         p1 = gsub(b1,a1); ep = valp(p1)-valp(b1);
         if (ep<=0) { b1 = gneg_i(b1); p1 = gsub(b1,a1); ep=valp(p1)-valp(b1); }
+        if (ep >= l || gequal0(p1)) return gerepilecopy(av,a1);
       }
-      while (ep<l && !gequal0(p1));
-      return gerepilecopy(av,a1);
+    }
 
     default:
       av = avma; if (!(y = toser_i(x))) break;
-      a1 = y; b1 = gen_1; l = lg(y)-2;
-      l2 = 5-prec2nbits(prec);
-      do
-      {
-        GEN a = a1;
-        a1 = gmul2n(gadd(a,b1),-1);
-        b1 = gsqrt(gmul(a,b1), prec);
-        p1 = gsub(b1,a1); ep = valp(p1)-valp(b1);
-      }
-      while (ep<l && !gequal0(p1)
-                  && (!isinexactreal(p1) || gexpo(p1) - gexpo(b1) >= l2));
-      return gerepilecopy(av,a1);
+      return gerepilecopy(av, ser_agm1(y, prec));
   }
   return trans_eval("agm",agm1,x,prec);
 }
@@ -2297,7 +2436,7 @@ atanhQ_split(ulong u, ulong v, long prec)
   struct abpq_res R;
   struct abpq A;
   /* satisfies (2n+1) (v/u)^2n > 2^bitprec */
-  nmax = bit_accuracy(prec) / (2*log2(d));
+  nmax = (long)(prec2nbits(prec) / (2*log2(d)));
   abpq_init(&A, nmax);
   A.a[0] = A.b[0] = gen_1;
   A.p[0] = utoipos(u);
@@ -2384,14 +2523,14 @@ logr_abs(GEN X)
   * 1-1/x and 1-x/2 ( crossover sqrt(2), worse ~ 0.29 ). To avoid an inverse,
   * we choose between x-1 and 1-x/2 ( crossover 4/3, worse ~ 0.33 ) */
   EX = expo(X);
-  u = (ulong)X[2];
+  u = uel(X,2);
   k = 2;
   if (u > (~0UL / 3) * 2) { /* choose 1-x/2 */
     EX++; u = ~u;
-    while (!u && ++k < l) { u = (ulong)X[k]; u = ~u; }
+    while (!u && ++k < l) { u = uel(X,k); u = ~u; }
   } else { /* choose x - 1 */
     u &= ~HIGHBIT; /* u - HIGHBIT, assuming HIGHBIT set */
-    while (!u && ++k < l) u = (ulong)X[k];
+    while (!u && ++k < l) u = uel(X,k);
   }
   if (k == l) return EX? mulsr(EX, mplog2(l)): real_0(l);
   z = cgetr(EX? l: l - (k-2)); ltop = avma;
@@ -2514,7 +2653,7 @@ Zp_teichmuller(GEN x, GEN p, long e, GEN pe)
   GEN q, z, p1;
   pari_sp av;
   ulong mask;
-  if (equaliu(p,2)) return (mod4(x) & 2)? subiu(pe,1): gen_1;
+  if (absequaliu(p,2)) return (mod4(x) & 2)? subiu(pe,1): gen_1;
   if (e == 1) return icopy(x);
   av = avma;
   p1 = subiu(p, 1);
@@ -2534,7 +2673,7 @@ Zp_teichmuller(GEN x, GEN p, long e, GEN pe)
     /* q <= qold^2 */
     if (lgefint(q) == 3)
     {
-      ulong Z = (ulong)z[2], Q = (ulong)q[2], P1 = (ulong)p1[2];
+      ulong Z = uel(z,2), Q = uel(q,2), P1 = uel(p1,2);
       ulong W = (Q-1) / P1; /* -1/(p-1) + O(qold) */
       ulong T = Fl_mul(W, Fl_powu(Z,P1,Q) - 1, Q);
       Z = Fl_mul(Z, 1 + T, Q);
@@ -2551,12 +2690,54 @@ Zp_teichmuller(GEN x, GEN p, long e, GEN pe)
 }
 
 GEN
-teich(GEN x)
+teichmullerinit(long p, long n)
+{
+  GEN t, pn, g, v;
+  ulong gp, tp;
+  long a, m;
+
+  if (p == 2) return mkvec(gen_1);
+  if (!uisprime(p)) pari_err_PRIME("teichmullerinit",utoipos(p));
+
+  m = p >> 1; /* (p-1)/2 */
+  tp= gp= pgener_Fl(p); /* order (p-1), gp^m = -1 */
+  pn = powuu(p, n);
+  v = cgetg(p, t_VEC);
+  t = g = Zp_teichmuller(utoipos(gp), utoipos(p), n, pn);
+  gel(v, 1) = gen_1;
+  gel(v, p-1) = subiu(pn,1);
+  for (a = 1; a < m; a++)
+  {
+    gel(v, tp) = t;
+    gel(v, p - tp) = Fp_neg(t, pn); /* g^(m+a) = -g^a */
+    if (a < m-1)
+    {
+      t = Fp_mul(t, g, pn); /* g^(a+1) */
+      tp = Fl_mul(tp, gp, p); /* t mod p  */
+    }
+  }
+  return v;
+}
+
+/* tab from teichmullerinit or NULL */
+GEN
+teichmuller(GEN x, GEN tab)
 {
   GEN p, q, y, z;
-  long n;
+  long n, tx = typ(x);
 
-  if (typ(x)!=t_PADIC) pari_err_TYPE("teichmuller",x);
+  if (!tab)
+  {
+    if (tx == t_VEC && lg(x) == 3)
+    {
+      p = gel(x,1);
+      q = gel(x,2);
+      if (typ(p) == t_INT && typ(q) == t_INT)
+        return teichmullerinit(itos(p), itos(q));
+    }
+  }
+  else if (typ(tab) != t_VEC) pari_err_TYPE("teichmuller",tab);
+  if (tx!=t_PADIC) pari_err_TYPE("teichmuller",x);
   z = gel(x,4);
   if (!signe(z)) return gcopy(x);
   p = gel(x,2);
@@ -2566,9 +2747,21 @@ teich(GEN x)
   y[1] = evalprecp(n) | _evalvalp(0);
   gel(y,2) = icopy(p);
   gel(y,3) = icopy(q);
-  gel(y,4) = Zp_teichmuller(z, p, n, q);
+  if (tab)
+  {
+    ulong pp = itou_or_0(p);
+    if (lg(tab) != (long)pp) pari_err_TYPE("teichmuller",tab);
+    z = gel(tab, umodiu(z, pp));
+    if (typ(z) != t_INT) pari_err_TYPE("teichmuller",tab);
+    z = remii(z, q);
+  }
+  else
+    z = Zp_teichmuller(z, p, n, q);
+  gel(y,4) = z;
   return y;
 }
+GEN
+teich(GEN x) { return teichmuller(x, NULL); }
 
 GEN
 glog(GEN x, long prec)
@@ -2601,6 +2794,14 @@ glog(GEN x, long prec)
     }
     case t_COMPLEX:
       if (ismpzero(gel(x,2))) return glog(gel(x,1), prec);
+      if (ismpzero(gel(x,1)))
+      {
+        GEN a = gel(x,2), b;
+        av = avma; b = Pi2n(-1,prec);
+        if (gsigne(a) < 0) { setsigne(b, -1); a = gabs(a,prec); }
+        a = isint1(a) ? gen_0: glog(a,prec);
+        return gerepilecopy(av, mkcomplex(a, b));
+      }
       l = precision(x); if (l > prec) prec = l;
       if (prec >= LOGAGMCX_LIMIT) return logagmcx(x, prec);
       y = cgetg(3,t_COMPLEX);
@@ -2647,7 +2848,7 @@ mpcosm1(GEN x, long *ptmod8)
       q = floorr( divrr(z,pitemp) ); /* round ( x / (Pi/2) ) */
       p = l+EXTRAPRECWORD; x = rtor(x,p);
     } else {
-      q = stoi((long)floor(rtodbl(x) / (PI/2) + 0.5));
+      q = stoi((long)floor(rtodbl(x) / (M_PI/2) + 0.5));
       p = l;
     }
     if (signe(q))
@@ -2660,10 +2861,10 @@ mpcosm1(GEN x, long *ptmod8)
   }
   /* a < 0 */
   b = signe(x); *ptmod8 = (b < 0)? 4 + n: n;
-  if (!b) return real_0_bit((expo(x)<<1) - 1);
+  if (!b) return real_0_bit(expo(x)*2 - 1);
 
   b = prec2nbits(l);
-  if (b + (a<<1) <= 0) {
+  if (b + 2*a <= 0) {
     y = sqrr(x); shiftr_inplace(y, -1); setsigne(y, -1);
     return y;
   }
@@ -2896,7 +3097,7 @@ mpsincos(GEN x, GEN *s, GEN *c)
   {
     long e = expo(x);
     *s = real_0_bit(e);
-    *c = e >= 0? real_0_bit(e): real_1(nbits2prec(-e));
+    *c = e >= 0? real_0_bit(e): real_1_bit(-e);
     return;
   }
 
@@ -3080,6 +3281,75 @@ gsincos(GEN x, GEN *s, GEN *c, long prec)
   pari_err_TYPE("gsincos",x);
 }
 
+/********************************************************************/
+/**                                                                **/
+/**                              SINC                              **/
+/**                                                                **/
+/********************************************************************/
+static GEN
+mpsinc(GEN x)
+{
+  pari_sp av = avma;
+  GEN s, c;
+
+  if (!signe(x)) {
+    long l = nbits2prec(-expo(x));
+    if (l < LOWDEFAULTPREC) l = LOWDEFAULTPREC;
+    return real_1(l);
+  }
+
+  mpsincos(x,&s,&c);
+  return gerepileuptoleaf(av, divrr(s,x));
+}
+
+GEN
+gsinc(GEN x, long prec)
+{
+  pari_sp av;
+  GEN r, u, v, y, u1, v1;
+  long i;
+
+  switch(typ(x))
+  {
+    case t_REAL: return mpsinc(x);
+    case t_COMPLEX:
+      if (isintzero(gel(x,1)))
+      {
+        av = avma;
+        return gerepileuptoleaf(av,gdiv(gsinh(gel(x,2),prec),gel(x,2)));
+      }
+      i = precision(x); if (!i) i = prec;
+      y = cgetc(i); av = avma;
+      r = gexp(gel(x,2),prec);
+      v1 = gmul2n(addrr(invr(r),r), -1); /* = cos(I*Im(x)) */
+      u1 = subrr(r, v1); /* = I*sin(I*Im(x)) */
+      gsincos(gel(x,1),&u,&v,prec);
+      affc_fixlg(gdiv(mkcomplex(gmul(v1,u), gmul(u1,v)), x), y);
+      avma = av; return y;
+
+    case t_INT:
+      if (!signe(x)) return real_1(prec); /*fall through*/
+    case t_FRAC:
+      y = cgetr(prec); av = avma;
+      affrr_fixlg(mpsinc(tofp_safe(x,prec)), y); avma = av; return y;
+
+    case t_PADIC:
+      if (gequal0(x)) return cvtop(gen_1, gel(x,2), valp(x));
+      av = avma; y = sin_p(x);
+      if (!y) pari_err_DOMAIN("gsinc(t_PADIC)","argument","",gen_0,x);
+      return gerepileuptoleaf(av,gdiv(y,x));
+
+    default:
+      av = avma; if (!(y = toser_i(x))) break;
+      if (gequal0(y)) return gerepileupto(av, gaddsg(1,y));
+      if (valp(y) < 0)
+        pari_err_DOMAIN("sinc","valuation", "<", gen_0, x);
+      gsincos(y,&u,&v,prec);
+      return gerepilecopy(av,gdiv(u,y));
+  }
+  return trans_eval("sinc",gsinc,x,prec);
+}
+
 /********************************************************************/
 /**                                                                **/
 /**                     TANGENT and COTANGENT                      **/
