@@ -82,7 +82,7 @@ mpatan(GEN x)
   if (e < -100)
     alpha = 1.65149612947 - e; /* log_2(Pi) - e */
   else
-    alpha = log2(PI / atan(rtodbl(p1)));
+    alpha = log2(M_PI / atan(rtodbl(p1)));
   beta = (double)(prec2nbits(l)>>1);
   delta = 1 + beta - alpha/2;
   if (delta <= 0) { n = 1; m = 0; }
@@ -370,7 +370,7 @@ mpcosh(GEN x)
 
   if (!signe(x)) { /* 1 + x */
     long e = expo(x);
-    return e >= 0? real_0_bit(e): real_1(nbits2prec(-e));
+    return e >= 0? real_0_bit(e): real_1_bit(-e);
   }
   av = avma;
   z = mpexp(x); z = addrr(z, invr(z)); shiftr_inplace(z, -1);
@@ -457,7 +457,7 @@ mptanh(GEN x)
 
   if (!s) return real_0_bit(expo(x));
   lx = realprec(x);
-  if (absr_cmp(x, stor(prec2nbits(lx), LOWDEFAULTPREC)) >= 0) {
+  if (abscmprr(x, stor(prec2nbits(lx), LOWDEFAULTPREC)) >= 0) {
     y = real_1(lx);
   } else {
     pari_sp av = avma;
@@ -497,6 +497,55 @@ gtanh(GEN x, long prec)
   }
   return trans_eval("tanh",gtanh,x,prec);
 }
+
+static GEN
+mpcotanh(GEN x)
+{
+  long lx, s = signe(x);
+  GEN y;
+
+  if (!s) pari_err_DOMAIN("cotan", "argument", "=", gen_0, x);
+
+  lx = realprec(x);
+  if (abscmprr(x, stor(prec2nbits(lx), LOWDEFAULTPREC)) >= 0) {
+    y = real_1(lx);
+  } else {
+    pari_sp av = avma;
+    long ex = expo(x);
+    GEN t;
+    if (ex < 1 - BITS_IN_LONG) x = rtor(x, lx + nbits2extraprec(-ex)-1);
+    t = exp1r_abs(gmul2n(x,1)); /* exp(|2x|) - 1 */
+    y = gerepileuptoleaf(av, divrr(addsr(2,t), t));
+  }
+  if (s < 0) togglesign(y); /* cotanh is odd */
+  return y;
+}
+
+GEN
+gcotanh(GEN x, long prec)
+{
+  pari_sp av;
+  GEN y, t;
+
+  switch(typ(x))
+  {
+    case t_REAL: return mpcotanh(x);
+    case t_COMPLEX:
+      if (isintzero(gel(x,1))) retmkcomplex(gen_0, gcotan(gel(x,2),prec));
+      /* fall through */
+    case t_PADIC:
+      av = avma;
+      t = gexpm1(gmul2n(x,1),prec);
+      return gerepileupto(av, gaddsg(1, gdivsg(2,t)));
+    default:
+      av = avma; if (!(y = toser_i(x))) break;
+      if (gequal0(y)) return gerepilecopy(av, y);
+      t = gexpm1(gmul2n(y,1),prec);
+      return gerepileupto(av, gaddsg(1, gdivsg(2,t)));
+  }
+  return trans_eval("cotanh",gcotanh,x,prec);
+}
+
 /********************************************************************/
 /**                                                                **/
 /**                     AREA HYPERBOLIC SINE                       **/
@@ -776,14 +825,14 @@ mpbern(long nb, long prec)
     /* Not cached, must compute */
     /* huge accuracy ? May as well compute exactly */
     if (n_is_small && (prec == LONG_MAX ||
-                       2*n * log(2*n) < prec2nbits_mul(prec, LOG2)))
+                       2*n * log((double)2*n) < prec2nbits_mul(prec, LOG2)))
       S = bernfrac_using_zeta(2*n);
     else
     {
 #ifdef LONG_IS_64BIT
-      const ulong mul_overflow = 3037000500;
+      const ulong mul_overflow = 3037000500UL;
 #else
-      const ulong mul_overflow = 46341;
+      const ulong mul_overflow = 46341UL;
 #endif
       ulong u = 8, v = 5, a = n-1, b = 2*n-3;
       n_is_small = 0;
@@ -910,7 +959,7 @@ faulhaber(long e, long v)
 GEN
 sumformal(GEN T, long v)
 {
-  pari_sp av = avma, av2, lim;
+  pari_sp av = avma, av2;
   long i, t, d;
   GEN R;
 
@@ -920,7 +969,7 @@ sumformal(GEN T, long v)
     return gerepileupto(av, monomialcopy(T, 1, v < 0? 0: v));
   if (t != t_POL) pari_err_TYPE("sumformal [not a t_POL]", T);
   if (v < 0) v = varn(T);
-  av2 = avma; lim = stack_lim(av2,3);
+  av2 = avma;
   R = gen_0;
   d = poldegree(T,v);
   for (i = d; i >= 0; i--)
@@ -928,7 +977,7 @@ sumformal(GEN T, long v)
     GEN c = polcoeff0(T, i, v);
     if (gequal0(c)) continue;
     R = gadd(R, gmul(c, faulhaber(i, v)));
-    if (low_stack(lim,stack_lim(av2,3)))
+    if (gc_needed(av2,3))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"sumformal, i = %ld/%ld", i,d);
       R = gerepileupto(av2, R);
@@ -971,11 +1020,11 @@ double
 darg(double s, double t)
 {
   double x;
-  if (!t) return (s>0)? 0.: PI;
-  if (!s) return (t>0)? PI/2: -PI/2;
+  if (!t) return (s>0)? 0.: M_PI;
+  if (!s) return (t>0)? M_PI/2: -M_PI/2;
   x = atan(t/s);
   return (s>0)? x
-              : ((t>0)? x+PI : x-PI);
+              : ((t>0)? x+M_PI : x-M_PI);
 }
 
 void
@@ -1003,12 +1052,32 @@ red_mod_2z(GEN x, GEN z)
 }
 #endif
 
+/* lngamma(1+z) = -Euler*z + sum_{i > 1} zeta(i)/i (-z)^i
+ * at relative precision prec, |z| < 1 is small */
+static GEN
+lngamma1(GEN z, long prec)
+{ /* sum_{i > l} |z|^(i-1) = |z|^l / (1-|z|) < 2^-B
+   * for l > (B+1) / |log2(|z|)| */
+  long i, l = ceil((bit_accuracy(prec) + 1) / - dbllog2(z));
+  GEN zet, me = mpeuler(prec), s = gen_0;
+  setsigne(me, -1); /* -Euler */
+  if (l <= 1) return gmul(me, z);
+  zet = veczeta(gen_1, gen_2, l-1, prec); /* z[i] = zeta(i+1) */
+  for (i = l; i > 1; i--)
+  {
+    GEN c = divru(gel(zet,i-1), i);
+    if (odd(i)) setsigne(c, -1);
+    s = gadd(gmul(s,z), c);
+  }
+  return gmul(z, gadd(gmul(s,z), me));
+}
+
 static GEN
 cxgamma(GEN s0, int dolog, long prec)
 {
   GEN s, u, a, y, res, tes, sig, tau, invn2, p1, nnx, pi, pi2, sqrtpi2;
   long i, lim, nn, esig, et;
-  pari_sp av, av2, avlim;
+  pari_sp av, av2;
   int funeq = 0;
   pari_timer T;
 
@@ -1037,7 +1106,7 @@ cxgamma(GEN s0, int dolog, long prec)
     iS = imag_i(S);
     if (et > 0 && l > 0)
     {
-      GEN t = gmul(iS, dbltor(PI / l)), logt = glog(t,LOWDEFAULTPREC);
+      GEN t = gmul(iS, dbltor(M_PI / l)), logt = glog(t,LOWDEFAULTPREC);
       la = gmul(t, logt);
       if      (gcmpgs(la, 3) < 0)   { logla = log(3.); la = stoi(3); }
       else if (gcmpgs(la, 150) > 0) { logla = rtodbl(logt); la = t; }
@@ -1050,7 +1119,7 @@ cxgamma(GEN s0, int dolog, long prec)
     lim = (long)ceil(l / (1.+ logla));
     if (lim == 0) lim = 1;
 
-    u = gmul(la, dbltor((lim-0.5)/PI));
+    u = gmul(la, dbltor((lim-0.5)/M_PI));
     l2 = gsub(gsqr(u), gsqr(iS));
     if (signe(l2) > 0)
     {
@@ -1066,15 +1135,22 @@ cxgamma(GEN s0, int dolog, long prec)
     double st = typ(s) == t_REAL? 0.0: rtodbl(imag_i(s));
     double la, l,l2,u,v, rlogs, ilogs;
 
-    if (dolog)
-    {
-      /* loggamma(1+u) ~ - Euler * u: cancellation if u is small */
-      if (fabs(ssig-1) + fabs(st) < 0.0001)
-      { /* s ~ 1, take care */
-        long e = gexpo(gsubgs(s,1));
-        prec += nbits2nlong(-e);
-        s = gprec_w(s, prec);
+    if (fabs(ssig-1) + fabs(st) < 0.0001)
+    { /* s ~ 1: loggamma(1+u) ~ - Euler * u, cancellation */
+      if (funeq) /* s0 ~ 0: use lngamma(s0)+log(s0) = lngamma(s0+1) */
+      {
+        if (dolog)
+          y = gsub(lngamma1(s0,prec), glog(s0,prec));
+        else
+          y = gdiv(gexp(lngamma1(s0,prec), prec), s0);
       }
+      else
+      {
+        if (isint1(s0)) { avma = av; return dolog? real_0(prec): real_1(prec); }
+        y = lngamma1(gsubgs(s0,1),prec);
+        if (!dolog) y = gexp(y,prec);
+      }
+      avma = av; return affc_fixlg(y, res);
     }
     dcxlog(ssig,st, &rlogs,&ilogs);
     /* Re (s - 1/2) log(s) */
@@ -1082,7 +1158,7 @@ cxgamma(GEN s0, int dolog, long prec)
     /* Im (s - 1/2) log(s) */
     v = (ssig - 0.5)*ilogs + st * rlogs;
     /* l2 = | (s - 1/2) log(s) - s + log(2Pi)/2 |^2 ~ |lngamma(s))|^2 */
-    u = u - ssig + log(2.*PI)/2;
+    u = u - ssig + log(2.*M_PI)/2;
     v = v - st;
     l2 = u*u + v*v;
     if (l2 < 0.000001) l2 = 0.000001;
@@ -1092,7 +1168,7 @@ cxgamma(GEN s0, int dolog, long prec)
     la = 3.; /* FIXME: heuristic... */
     if (st > 1 && l > 0)
     {
-      double t = st * PI / l;
+      double t = st * M_PI / l;
       la = t * log(t);
       if (la < 3) la = 3.;
       if (la > 150) la = t;
@@ -1100,7 +1176,7 @@ cxgamma(GEN s0, int dolog, long prec)
     lim = (long)ceil(l / (1.+ log(la)));
     if (lim == 0) lim = 1;
 
-    u = (lim-0.5) * la / PI;
+    u = (lim-0.5) * la / M_PI;
     l2 = u*u - st*st;
     if (l2 > 0)
     {
@@ -1113,7 +1189,7 @@ cxgamma(GEN s0, int dolog, long prec)
   }
   incrprec(prec);
 
-  av2 = avma; avlim = stack_lim(av2,3);
+  av2 = avma;
   y = s;
   if (typ(s0) == t_INT)
   {
@@ -1124,7 +1200,7 @@ cxgamma(GEN s0, int dolog, long prec)
       for (i=1; i < nn; i++)
       {
         y = mulri(y, addis(s0, i));
-        if (low_stack(avlim,stack_lim(av2,3)))
+        if (gc_needed(av2,3))
         {
           if(DEBUGMEM>1) pari_warn(warnmem,"gamma");
           y = gerepileuptoleaf(av2, y);
@@ -1135,7 +1211,7 @@ cxgamma(GEN s0, int dolog, long prec)
       for (i=1; i < nn; i++)
       {
         y = mulru(y, ss + i);
-        if (low_stack(avlim,stack_lim(av2,3)))
+        if (gc_needed(av2,3))
         {
           if(DEBUGMEM>1) pari_warn(warnmem,"gamma");
           y = gerepileuptoleaf(av2, y);
@@ -1149,7 +1225,7 @@ cxgamma(GEN s0, int dolog, long prec)
     for (i=1; i < nn; i++)
     {
       y = gmul(y, gaddgs(s,i));
-      if (low_stack(avlim,stack_lim(av2,3)))
+      if (gc_needed(av2,3))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"gamma");
         y = gerepileupto(av2, y);
@@ -1163,7 +1239,7 @@ cxgamma(GEN s0, int dolog, long prec)
     for (i=1; i < nn; i++)
     {
       y = gadd(y, glog(gaddgs(s,i), prec));
-      if (low_stack(avlim,stack_lim(av2,3)))
+      if (gc_needed(av2,3))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"gamma");
         y = gerepileupto(av2, y);
@@ -1174,7 +1250,7 @@ cxgamma(GEN s0, int dolog, long prec)
 
   nnx = gaddgs(s, nn);
   a = ginv(nnx); invn2 = gsqr(a);
-  av2 = avma; avlim = stack_lim(av2,3);
+  av2 = avma;
   mpbern(lim,prec);
   tes = divrunu(bernreal(2*lim,prec), 2*lim-1); /* B2l / (2l-1) 2l*/
   if (DEBUGLEVEL>5) timer_printf(&T,"Bernoullis");
@@ -1182,7 +1258,7 @@ cxgamma(GEN s0, int dolog, long prec)
   {
     u = divrunu(bernreal(i,prec), i-1); /* Bi / i(i-1) */
     tes = gadd(u, gmul(invn2,tes));
-    if (low_stack(avlim,stack_lim(av2,3)))
+    if (gc_needed(av2,3))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"gamma");
       tes = gerepileupto(av2, tes);
@@ -1402,7 +1478,7 @@ Qp_gamma(GEN x)
 {
   GEN n, m, N, p = gel(x,2);
   long s, e = precp(x);
-  if (equaliu(p, 2) && e == 2) e = 1;
+  if (absequaliu(p, 2) && e == 2) e = 1;
   if (valp(x) < 0) pari_err_DOMAIN("gamma","v_p(x)", "<", gen_0, x);
   n = gtrunc(x);
   m = gtrunc(gneg(x));
@@ -1413,11 +1489,26 @@ Qp_gamma(GEN x)
   return Qp_gamma_Dwork(x, itos(p));
 }
 
+/* gamma(1+x) - 1, |x| < 1 is "small" */
+GEN
+ggamma1m1(GEN x, long prec) { return gexpm1(lngamma1(x, prec), prec); }
+
+/* lngamma(y) with 0 constant term, using (lngamma y)' = y' psi(y) */
+static GEN
+serlngamma0(GEN y, long prec)
+{
+  GEN t;
+  if (valp(y)) pari_err_DOMAIN("lngamma","valuation", "!=", gen_0, y);
+  t = derivser(y);
+  /* don't compute psi if y'=0 */
+  if (signe(t)) t = gmul(t, gpsi(y,prec));
+  return integser(t);
+}
+
 GEN
 ggamma(GEN x, long prec)
 {
   pari_sp av;
-  long m;
   GEN y, z;
 
   switch(typ(x))
@@ -1426,40 +1517,64 @@ ggamma(GEN x, long prec)
       if (signe(x) <= 0)
         pari_err_DOMAIN("gamma","argument", "=",
                          strtoGENstr("non-positive integer"), x);
-      if (cmpiu(x,481177) > 0) pari_err_OVERFLOW("gamma");
+      if (abscmpiu(x,481177) > 0) pari_err_OVERFLOW("gamma");
       return mpfactr(itos(x) - 1, prec);
 
     case t_REAL: case t_COMPLEX:
       return cxgamma(x, 0, prec);
 
     case t_FRAC:
-      if (!equaliu(gel(x,2),2)) break;
-      z = gel(x,1); /* true argument is z/2 */
-      if (is_bigint(z) || labs(m = itos(z)) > 962354)
+    {
+      GEN a = gel(x,1), b = gel(x,2), c;
+      long m;
+      if (absequaliu(b,2))
       {
-        pari_err_OVERFLOW("gamma");
-        return NULL; /* not reached */
+        if (is_bigint(a) || labs(m = itos(a)) > 962354)
+        {
+          pari_err_OVERFLOW("gamma");
+          return NULL; /* not reached */
+        }
+        return gammahs(m-1, prec);
       }
-      return gammahs(m-1, prec);
+      av = avma; c = subii(a,b);
+      if (expi(c) - expi(b) < -10)
+      {
+        x = mkfrac(c,b);
+        if (lgefint(b) >= prec) x = fractor(x,prec);
+        y = mpexp(lngamma1(x, prec));
+      }
+      else
+      {
+        x = fractor(x, prec);
+        y = cxgamma(x, 0, prec);
+      }
+      return gerepileupto(av, y);
+    }
 
     case t_PADIC: return Qp_gamma(x);
     default:
       av = avma; if (!(y = toser_i(x))) break;
+      if (lg(y) == 2) pari_err_DOMAIN("gamma", "argument", "=", gen_0,y);
       /* exp(lngamma) */
-      if (valp(y)>0 || lg(y) == 2)
+      if (valp(y) > 0 || lg(y) == 2)
         z = gdiv(gexp(glngamma(gaddgs(y,1),prec),prec),y);
       else
-      { /* use fun eq. to avoid log singularity of lngamma at negative ints */
-        GEN Y = y, y0 = gel(y,2), t = ground(y0), pi = NULL;
-        if (gequal(y0, t) && typ(t) == t_INT && signe(t) < 0)
-        {
-          pi = mppi(prec);
-          Y = gsubsg(1, y);
+      {
+        GEN Y = y, y0 = simplify_shallow(gel(y,2));
+        z = NULL;
+        if (isint(y0, &y0))
+        { /* fun eq. avoids log singularity of lngamma at negative ints */
+          if (signe(y0) < 0) { Y = gsubsg(1, y); y0 = subsi(1, y0); }
+          if (abscmpiu(y0, 50) < 0) z = mpfact(itos(y0)-1); /* more precise */
         }
-        z = gexp(glngamma(Y,prec),prec);
-        if (pi)
-          z = gdiv(mpodd(t)? negr(pi): pi,
+        if (!z) z = ggamma(y0,prec);
+        z = gmul(z, gexp(serlngamma0(Y,prec),prec));
+        if (Y != y)
+        {
+          GEN pi = mppi(prec);
+          z = gdiv(mpodd(y0)? pi: negr(pi),
                    gmul(z, gsin(gmul(pi,serchop0(y)), prec)));
+        }
       }
       return gerepileupto(av, z);
   }
@@ -1482,8 +1597,8 @@ mpfactr(long n, long prec)
 GEN
 glngamma(GEN x, long prec)
 {
-  pari_sp av;
-  GEN y, p1;
+  pari_sp av = avma;
+  GEN y, y0, t;
 
   switch(typ(x))
   {
@@ -1491,35 +1606,41 @@ glngamma(GEN x, long prec)
       if (signe(x) <= 0)
         pari_err_DOMAIN("lngamma","argument", "=",
                          strtoGENstr("non-positive integer"), x);
-      if (cmpiu(x,200 + 50*(prec-2)) > 0) /* heuristic */
+      if (abscmpiu(x,200 + 50*(prec-2)) > 0) /* heuristic */
         return cxgamma(x, 1, prec);
-      av = avma;
       return gerepileuptoleaf(av, logr_abs( itor(mpfact(itos(x) - 1), prec) ));
     case t_FRAC:
     {
-      GEN a, b;
-      long e1, e2;
-      av = avma;
-      a = gel(x,1);
-      b = gel(x,2);
-      e1 = expi(subii(a,b)); e2 = expi(b);
-      if (e2 > e1) prec += nbits2nlong(e2 - e1);
-      x = fractor(x, prec);
-      return gerepileupto(av, cxgamma(x, 1, prec));
+      GEN a = gel(x,1), b = gel(x,2), c = subii(a,b);
+      long e = expi(b) - expi(c);
+      if (e > 10)
+      {
+        x = mkfrac(c,b);
+        if (lgefint(b) >= prec) x = fractor(x,prec + nbits2nlong(e));
+        y = lngamma1(x, prec);
+      }
+      else
+      {
+        x = fractor(x, e > 1? prec+EXTRAPRECWORD: prec);
+        y = cxgamma(x, 1, prec);
+      }
+      return gerepileupto(av, y);
     }
 
     case t_REAL: case t_COMPLEX:
       return cxgamma(x, 1, prec);
 
     default:
-      av = avma; if (!(y = toser_i(x))) break;
-      if (valp(y)) pari_err_DOMAIN("lngamma","valuation", "!=", gen_0, x);
-      /* (lngamma y)' = y' psi(y) */
-      p1 = integser(gmul(derivser(y), gpsi(y, prec)));
-      if (!gequal1(gel(y,2))) p1 = gadd(p1, glngamma(gel(y,2),prec));
-      return gerepileupto(av, p1);
+      if (!(y = toser_i(x))) break;
+      if (lg(y) == 2) pari_err_DOMAIN("lngamma", "argument", "=", gen_0,y);
+      t = serlngamma0(y,prec);
+      y0 = simplify_shallow(gel(y,2));
+      /* no constant term if y0 = 1 or 2 */
+      if (!isint(y0,&y0) || signe(y0) <= 0 || abscmpiu(y0,2) > 2)
+        t = gadd(t, glngamma(y0,prec));
+      return gerepileupto(av, t);
 
-    case t_PADIC: av = avma; return gerepileupto(av, Qp_log(Qp_gamma(x)));
+    case t_PADIC: return gerepileupto(av, Qp_log(Qp_gamma(x)));
   }
   return trans_eval("lngamma",glngamma,x,prec);
 }
@@ -1528,8 +1649,7 @@ glngamma(GEN x, long prec)
 /**                  PSI(x) = GAMMA'(x)/GAMMA(x)                   **/
 /**                                                                **/
 /********************************************************************/
-
-GEN
+static GEN
 cxpsi(GEN s0, long prec)
 {
   pari_sp av, av2;
@@ -1556,7 +1676,7 @@ cxpsi(GEN s0, long prec)
     lim = 2 + (long)ceil((prec2nbits_mul(prec, LOG2) - l) / (2*(1+log((double)la))));
     if (lim < 2) lim = 2;
 
-    l = (2*lim-1)*la / (2.*PI);
+    l = (2*lim-1)*la / (2.*M_PI);
     L = gsub(dbltor(l*l), gsqr(iS));
     if (signe(L) < 0) L = gen_0;
 
@@ -1579,7 +1699,7 @@ cxpsi(GEN s0, long prec)
     lim = 2 + (long)ceil((prec2nbits_mul(prec, LOG2) - l) / (2*(1+log((double)la))));
     if (lim < 2) lim = 2;
 
-    l = (2*lim-1)*la / (2.*PI);
+    l = (2*lim-1)*la / (2.*M_PI);
     l = l*l - st*st;
     if (l < 0.) l = 0.;
     nn = (long)ceil( sqrt(l) - ssig );
@@ -1616,18 +1736,19 @@ cxpsi(GEN s0, long prec)
   avma = av; return affc_fixlg(z, res);
 }
 
-/* psi(1+x) + O(x^n), x = pol_x(v) */
+/* n > 0; return psi(1+x) + O(x^n), x = pol_x(v) */
 static GEN
 serpsi1(long n, long v, long prec)
 {
   long i, l = n+3;
-  GEN g, s = cgetg(l, t_SER);
+  GEN z, g, s = cgetg(l, t_SER);
   s[1] = evalsigne(1)|evalvalp(0)|evalvarn(v);
   g = mpeuler(prec); setsigne(g, -1);
+  z = veczeta(gen_1, gen_2, n, prec); /* zeta(2..n) */
   gel(s,2) = g;
   for (i = 2; i < l-1; i++)
   {
-    GEN c = szeta(i, prec);
+    GEN c = gel(z,i-1); /* zeta(i) */
     if (odd(i)) setsigne(c, -1);
     gel(s,i+1) = c;
   }
@@ -1640,21 +1761,19 @@ tr(GEN T, GEN z0, long L)
   GEN s = RgX_to_ser(RgX_translate(T, z0), L+3);
   setvarn(s, 0); return s;
 }
-/* psi(z0+x) + O(x^n) */
+/* z0 a complex number with Re(z0) > 1/2; return psi(z0+x) + O(x^L)
+ * using Luke's rational approximation for psi(x) */
 static GEN
 serpsiz0(GEN z0, long L, long v, long prec)
 {
-  pari_sp av, lim;
+  pari_sp av;
   GEN A,A1,A2, B,B1,B2, Q;
   long n;
-
-  if (equali1(z0)) return serpsi1(L, v, prec);
-  /* otherwise use Luke's rational approximations for psi(x) */
   n = gprecision(z0); if (n) prec = n;
   z0 = gtofp(z0, prec + EXTRAPRECWORD);
   /* Start from n = 3; in Luke's notation, A2 := A_{n-2}, A1 := A_{n-1},
    * A := A_n. Same for B */
-  av = avma; lim = stack_lim(av,1);
+  av = avma;
   A2= gdivgs(mkpoln(2, gen_1, utoipos(6)), 2);
   B2 = scalarpol_shallow(utoipos(4), 0);
   A1= gdivgs(mkpoln(3, gen_1, utoipos(82), utoipos(96)), 6);
@@ -1689,10 +1808,10 @@ serpsiz0(GEN z0, long L, long v, long prec)
     a = gdiv(gadd(gadd(gmul(c1,A),gmul(c2,A1)),gmul(c3,A2)), c0);
     b = gdiv(gadd(gadd(gmul(c1,B),gmul(c2,B1)),gmul(c3,B2)), c0);
     Q = gdiv(a,b);
-    if (gexpo(gsub(Q,Q0)) < -bit_accuracy(prec)) break;
+    if (gexpo(gsub(Q,Q0)) < -prec2nbits(prec)) break;
     A2 = A1; A1 = A; A = a;
     B2 = B1; B1 = B; B = b;
-    if (low_stack(lim,stack_lim(av,1)))
+    if (gc_needed(av,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"serpsiz0, n = %ld", n);
       gerepileall(av, 7, &A,&A1,&A2, &B,&B1,&B2, &Q);
@@ -1702,12 +1821,42 @@ serpsiz0(GEN z0, long L, long v, long prec)
   setvarn(Q, v);
   return gadd(negr(mpeuler(prec)), Q);
 }
+/* sum (-1)^k*H(m,k)x^k + O(x^L); L > 0;
+ * H(m,k) = (-1)^{k * \delta_{m > 0}} sum_{1<=i<m} 1/i^(k+1) */
+static GEN
+Hseries(long m, long L, long v, long prec)
+{
+  long i, k, bit, l = L+3, M = m < 0? 1-m: m;
+  pari_sp av = avma;
+  GEN H = cgetg(l, t_SER);
+  H[1] = evalsigne(1)|evalvarn(v)|evalvalp(0);
+  prec++;
+  bit = -prec2nbits(prec);
+  for(k = 2; k < l; k++) gel(H,k) = gen_1; /* i=1 */
+  for (i = 2; i < M; i++)
+  {
+    GEN ik = invr(utor(i, prec));
+    for (k = 2; k < l; k++)
+    {
+      if (k > 2) { ik = divru(ik, i); if (expo(ik) < bit) break; }
+      gel(H,k) = gadd(gel(H,k), ik);
+    }
+    if (gc_needed(av,3))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"Hseries, i = %ld/%ld", i,M);
+      H = gerepilecopy(av, H);
+    }
+  }
+  if (m > 0)
+    for (k = 3; k < l; k+=2) togglesign_safe(&gel(H,k));
+  return H;
+}
+
 static GEN
 serpsi(GEN y, long prec)
 {
-  GEN Q, z0, Y;
+  GEN Q = NULL, z0, Y = y, Y2;
   long L = lg(y)-2, v  = varn(y), vy = valp(y);
-  int reflect;
 
   if (!L) pari_err_DOMAIN("psi", "argument", "=", gen_0,y);
   if (vy < 0) pari_err_DOMAIN("psi", "series valuation", "<", gen_0,y);
@@ -1715,22 +1864,42 @@ serpsi(GEN y, long prec)
     z0 = gen_0;
   else
   {
-    GEN t;
     z0 = simplify_shallow(gel(y,2));
-    t = ground(z0);
-    if (gequal(t,z0)) z0 = t;
+    (void)isint(z0, &z0);
   }
-  reflect = (gcmp(real_i(z0),ghalf) < 0); /* use reflection formula */
-  if (reflect) { z0 = gsubsg(1,z0); Y = gsubsg(1,y); } else Y = y;
-  Q = serpsiz0(z0, L, v, prec);
-  Q = gsubst(Q, v, serchop0(Y)); /* psi(Y) */
-  if (reflect)
-  { /* psi(y) = psi(Y) + Pi cotan(Pi Y) = psi(Y) - Pi cotan(Pi y) */
+  if (typ(z0) == t_INT && !is_bigint(z0))
+  {
+    long m = itos(z0);
+    if (abscmpiu(muluu(prec2nbits(prec),L), labs(m)) > 0)
+    { /* psi(m+x) = psi(1+x) + sum_{1 <= i < m} 1/(i+x) for m > 0
+                    psi(1+x) - sum_{0 <= i < -m} 1/(i+x) for m <= 0 */
+      GEN H = NULL;
+      if (m <= 0) L--; /* lose series accuracy due to 1/x term */
+      if (L)
+      {
+        Q = serpsi1(L, v, prec);
+        if (m && m != 1) { H = Hseries(m, L, v, prec); Q = gadd(Q, H); }
+        if (m <= 0) Q = gsub(Q, ginv(pol_x(v)));
+      }
+      else
+      {
+        Q = scalarser(gen_m1, v, 1);
+        setvalp(Q,-1);
+      }
+    }
+  }
+  if (!Q)
+  { /* use psi(1-y)=psi(y)+Pi*cotan(Pi*y) ? */
+    if (gcmp(real_i(z0),ghalf) < 0) { z0 = gsubsg(1,z0); Y = gsubsg(1,y); }
+    Q = serpsiz0(z0, L, v, prec);
+  }
+  Y2 = serchop0(Y); if (signe(Y2)) Q = gsubst(Q, v, Y2);
+  /* psi(z0 + Y2) = psi(Y) */
+  if (Y != y)
+  { /* psi(y) = psi(Y) + Pi cotan(Pi Y) */
     GEN pi = mppi(prec);
-    if (equali1(z0))
-      Q = gsub(Q, gmul(pi, gcotan(gmul(pi,y), prec)));
-    else
-      Q = gadd(Q, gmul(pi, gcotan(gmul(pi,Y), prec)));
+    if (typ(z0) == t_INT) Y = Y2; /* in this case cotan(Pi*Y2) = cotan(Pi*Y) */
+    Q = gadd(Q, gmul(pi, gcotan(gmul(pi,Y), prec)));
   }
   return Q;
 }
