@@ -42,157 +42,12 @@ const long DEFAULT_COLOR = 1, AXIS_COLOR = 2;
 INLINE long
 DTOL(double t) { return (long)(t + 0.5); }
 
-/********************************************************************/
-/**                                                                **/
-/**                         LOW-RES PLOT                           **/
-/**                                                                **/
-/********************************************************************/
-#define ISCR 64
-#define JSCR 22
-static char
-PICT(long j) {
-  switch(j%3) {
-    case 0:  return '_';
-    case 1:  return 'x';
-    default: return '"';
-  }
-}
-static char
-PICTZERO(long j) {
-  switch(j%3) {
-    case 0:  return ',';
-    case 1:  return '-';
-    default: return '`';
-  }
-}
-
-static char *
-dsprintf9(double d, char *buf)
-{
-  int i = 10;
-
-  while (--i >= 0) {
-    sprintf(buf, "%9.*g", i, d);
-    if (strlen(buf) <= 9) return buf;
-  }
-  return buf; /* Should not happen? */
-}
-
-typedef unsigned char screen[ISCR+1][JSCR+1];
-
-static void
-fill_gap(screen scr, long i, int jnew, int jpre)
-{
-  int mid, i_up, i_lo, up, lo;
-
-  if (jpre < jnew - 2) {
-    up = jnew - 1; i_up = i;
-    lo = jpre + 1; i_lo = i - 1;
-  } else if (jnew < jpre - 2) {
-    up = jpre - 1; i_up = i - 1;
-    lo = jnew + 1; i_lo = i;
-  } else return; /* if gap < 2, leave it as it is. */
-
-  mid = (jpre+jnew)/2;
-  if (mid>JSCR) mid=JSCR; else if (mid<0) mid=0;
-  if (lo<0) lo=0;
-  if (lo<=JSCR) while (lo <= mid) scr[i_lo][lo++] = ':';
-  if (up>JSCR) up=JSCR;
-  if (up>=0) while (up > mid) scr[i_up][up--] = ':';
-}
-
-static double
-todbl(GEN x) { return rtodbl(gtofp(x, LOWDEFAULTPREC)); }
-
-/* code is either a t_CLOSURE (from GP: ploth, etc.) or a t_POL/t_VEC of
- * two t_POLs from rectsplines */
 static GEN
 READ_EXPR(GEN code, GEN x) {
   if (typ(code)!=t_CLOSURE) return gsubst(code,0,x);
   set_lex(-1, x); return closure_evalgen(code);
 }
 
-void
-plot(GEN a, GEN b, GEN code, GEN ysmlu,GEN ybigu, long prec)
-{
-  const char BLANK = ' ', YY = '|', XX_UPPER = '\'', XX_LOWER = '.';
-  long jz, j, i, sig;
-  pari_sp av = avma, av2, limite;
-  int jnew, jpre = 0; /* for lint */
-  GEN x, dx;
-  double diff, dyj, ysml, ybig, y[ISCR+1];
-  screen scr;
-  char buf[80], z;
-
-  sig=gcmp(b,a); if (!sig) return;
-  if (sig<0) { x=a; a=b; b=x; }
-  x = gtofp(a, prec); push_lex(x, code);
-  dx = divru(gtofp(gsub(b,a),prec), ISCR-1);
-  for (j=1; j<=JSCR; j++) scr[1][j]=scr[ISCR][j]=YY;
-  for (i=2; i<ISCR; i++)
-  {
-    scr[i][1]   = XX_LOWER;
-    scr[i][JSCR]= XX_UPPER;
-    for (j=2; j<JSCR; j++) scr[i][j] = BLANK;
-  }
-  av2 = avma; limite=stack_lim(av2,1);
-  ysml = ybig = 0.; /* -Wall */
-  for (i=1; i<=ISCR; i++)
-  {
-    y[i] = gtodouble( READ_EXPR(code,x) );
-    if (i == 1)
-      ysml = ybig = y[1];
-    else
-    {
-      if (y[i] < ysml) ysml = y[i];
-      if (y[i] > ybig) ybig = y[i];
-    }
-    x = addrr(x,dx);
-    if (low_stack(limite, stack_lim(av2,1)))
-    {
-      if (DEBUGMEM>1) pari_warn(warnmem,"plot");
-      x = gerepileuptoleaf(av2, x);
-    }
-  }
-  avma = av;
-  if (ysmlu) ysml = gtodouble(ysmlu);
-  if (ybigu) ybig = gtodouble(ybigu);
-  diff = ybig - ysml;
-  if (!diff) { ybig += 1; diff= 1.; }
-  dyj = ((JSCR-1)*3+2) / diff;
-  /* work around bug in gcc-4.8 (32bit): plot(x=-5,5,sin(x)))) */
-  jz = 3 - (long)(ysml*dyj + 0.5); /* 3 - DTOL(ysml*dyj) */
-  z = PICTZERO(jz); jz /= 3;
-  for (i=1; i<=ISCR; i++, avma = av2)
-  {
-    if (0<=jz && jz<=JSCR) scr[i][jz]=z;
-    j = 3 + DTOL((y[i]-ysml)*dyj);
-    jnew = j/3;
-    if (i > 1) fill_gap(scr, i, jnew, jpre);
-    if (0<=jnew && jnew<=JSCR) scr[i][jnew] = PICT(j);
-    jpre = jnew;
-  }
-  pari_putc('\n');
-  pari_printf("%s ", dsprintf9(ybig, buf));
-  for (i=1; i<=ISCR; i++) pari_putc(scr[i][JSCR]);
-  pari_putc('\n');
-  for (j=(JSCR-1); j>=2; j--)
-  {
-    pari_puts("          ");
-    for (i=1; i<=ISCR; i++) pari_putc(scr[i][j]);
-    pari_putc('\n');
-  }
-  pari_printf("%s ", dsprintf9(ysml, buf));
-  for (i=1; i<=ISCR; i++)  pari_putc(scr[i][1]);
-  pari_putc('\n');
-  {
-    char line[10 + 32 + 32 + ISCR - 9];
-    sprintf(line, "%10s%-9.7g%*.7g\n"," ",todbl(a),ISCR-9,todbl(b));
-    pari_printf(line);
-  }
-  pop_lex(1);
-}
-
 /********************************************************************/
 /**                                                                **/
 /**                      RECTPLOT FUNCTIONS                        **/
@@ -227,8 +82,8 @@ free_graph(void)
     pari_free((void*)rgb_colors->table);
     pari_free((void*)rgb_colors);
   }
-  if (pari_colormap) pari_free(pari_colormap);
-  if (pari_graphcolors) pari_free(pari_graphcolors);
+  if (GP_DATA->colormap) pari_free(GP_DATA->colormap);
+  if (GP_DATA->graphcolors) pari_free(GP_DATA->graphcolors);
 }
 
 static PariRect *
@@ -388,7 +243,7 @@ rectrpoint(long ne, GEN x, GEN y)
 void
 rectcolor(long ne, long c)
 {
-  long n = lg(pari_colormap)-2;
+  long n = lg(GP_DATA->colormap)-2;
   check_rect(ne);
   if (c < 1) pari_err_DOMAIN("rectcolor", "color", "<", gen_1, stoi(c));
   if (c > n) pari_err_DOMAIN("rectcolor", "color", ">", stoi(n), stoi(c));
@@ -465,7 +320,7 @@ rectticks(PARI_plot *WW, long ne,
   if (dx > 1000 || dy > 1000)
     dxy = 1000; /* avoid overflow */
   else
-    dxy = (long)sqrt(dx*dx + dy*dy);
+    dxy = usqrt(dx*dx + dy*dy);
   nticks = (long) ((dxy + 2.5)/4);
   if (!nticks) return;
 
@@ -482,7 +337,7 @@ rectticks(PARI_plot *WW, long ne,
     l_min = l2, l_max = l1;
   minstep = (l_max - l_min)/(nticks + 1);
   maxstep = 2.5*(l_max - l_min);
-  step = exp(log(10) * floor(log10(minstep)));
+  step = exp(log(10.) * floor(log10(minstep)));
   if (!(flags & TICKS_ENDSTOO)) {
     double d = 2*(l_max - l_min)/dxy1;        /* Two pixels off */
 
@@ -820,13 +675,18 @@ rectcopy_gen(long source, long dest, GEN xoff, GEN yoff, long flag)
   long xi, yi;
   if (flag & RECT_CP_RELATIVE) {
     double xd = gtodouble(xoff), yd = gtodouble(yoff);
-
+    if (xd > 1) pari_err_DOMAIN("plotcopy","dx",">",gen_1,xoff);
+    if (xd < 0) pari_err_DOMAIN("plotcopy","dx","<",gen_0,xoff);
+    if (yd > 1) pari_err_DOMAIN("plotcopy","dy",">",gen_1,yoff);
+    if (yd < 0) pari_err_DOMAIN("plotcopy","dy","<",gen_0,yoff);
     PARI_get_plot();
     xi = pari_plot.width - 1;
     yi = pari_plot.height - 1;
     xi = DTOL(xd*xi);
     yi = DTOL(yd*yi);
   } else {
+    if (typ(xoff) != t_INT) pari_err_TYPE("plotcopy",xoff);
+    if (typ(yoff) != t_INT) pari_err_TYPE("plotcopy",yoff);
     xi = itos(xoff);
     yi = itos(yoff);
   }
@@ -850,60 +710,47 @@ rectcopy_gen(long source, long dest, GEN xoff, GEN yoff, long flag)
   rectcopy(source, dest, xi, yi);
 }
 
+static void*
+cp(void* R, size_t t)
+{ void *o = pari_malloc(t); memcpy(o,R,t); return o; }
+
 void
-rectcopy(long source, long dest, long xoff, long yoff)
+rectcopy(long source, long dest, long x, long y)
 {
   PariRect *s = check_rect_init(source), *d = check_rect_init(dest);
-  RectObj *R = RHead(s);
-  RectObj *tail = RTail(d);
-  RectObj *next;
-  int i;
+  RectObj *R, *tail = RTail(d);
+  long i;
 
-  while (R)
+  for (R = RHead(s); R; R = RoNext(R))
   {
+    RectObj *o;
     switch(RoType(R))
     {
       case ROt_PT:
-        next = (RectObj*) pari_malloc(sizeof(RectObj1P));
-        memcpy(next,R,sizeof(RectObj1P));
-        RoPTx(next) += xoff; RoPTy(next) += yoff;
-        RoNext(tail) = next; tail = next;
+        o = (RectObj*)cp(R, sizeof(RectObj1P));
+        RoPTx(o) += x; RoPTy(o) += y;
         break;
       case ROt_LN: case ROt_BX:
-        next = (RectObj*) pari_malloc(sizeof(RectObj2P));
-        memcpy(next,R,sizeof(RectObj2P));
-        RoLNx1(next) += xoff; RoLNy1(next) += yoff;
-        RoLNx2(next) += xoff; RoLNy2(next) += yoff;
-        RoNext(tail) = next; tail = next;
+        o = (RectObj*)cp(R, sizeof(RectObj2P));
+        RoLNx1(o) += x; RoLNy1(o) += y;
+        RoLNx2(o) += x; RoLNy2(o) += y;
         break;
       case ROt_MP: case ROt_ML:
-        next = (RectObj*) pari_malloc(sizeof(RectObjMP));
-        memcpy(next,R,sizeof(RectObjMP));
-        RoMPxs(next) = (double*) pari_malloc(sizeof(double)*RoMPcnt(next));
-        RoMPys(next) = (double*) pari_malloc(sizeof(double)*RoMPcnt(next));
-        memcpy(RoMPxs(next),RoMPxs(R),sizeof(double)*RoMPcnt(next));
-        memcpy(RoMPys(next),RoMPys(R),sizeof(double)*RoMPcnt(next));
-        for (i=0; i<RoMPcnt(next); i++)
-        {
-          RoMPxs(next)[i] += xoff; RoMPys(next)[i] += yoff;
-         }
-        RoNext(tail) = next; tail = next;
+        o = (RectObj*)cp(R, sizeof(RectObjMP));
+        RoMPxs(o) = (double*)cp(RoMPxs(R), sizeof(double)*RoMPcnt(o));
+        RoMPys(o) = (double*)cp(RoMPys(R), sizeof(double)*RoMPcnt(o));
+        for (i=0; i<RoMPcnt(o); i++) { RoMPxs(o)[i] += x; RoMPys(o)[i] += y; }
         break;
       case ROt_ST:
-        next = (RectObj*) pari_malloc(sizeof(RectObjST));
-        memcpy(next,R,sizeof(RectObjST));
-        RoSTs(next) = (char*) pari_malloc(RoSTl(R)+1);
-        memcpy(RoSTs(next),RoSTs(R),RoSTl(R)+1);
-        RoSTx(next) += xoff; RoSTy(next) += yoff;
-        RoNext(tail) = next; tail = next;
+        o = (RectObj*)cp(R, sizeof(RectObjST));
+        RoSTs(o) = (char*)cp(RoSTs(R),RoSTl(R)+1);
+        RoSTx(o) += x; RoSTy(o) += y;
         break;
-      case ROt_PTT: case ROt_LNT: case ROt_PTS:
-        next = (RectObj*) pari_malloc(sizeof(RectObjPN));
-        memcpy(next,R,sizeof(RectObjPN));
-        RoNext(tail) = next; tail = next;
+      default: /* ROt_PTT, ROt_LNT, ROt_PTS */
+        o = (RectObj*)cp(R, sizeof(RectObjPN));
         break;
     }
-    R=RoNext(R);
+    RoNext(tail) = o; tail = o;
   }
   RoNext(tail) = NULL; RTail(d) = tail;
 }
@@ -1686,10 +1533,10 @@ rectplothrawin(long grect, dblPointList *data, long flags)
     flags |= PLOT_PARAMETRIC;
     flags &= (~PLOT_COMPLEX); /* turn COMPLEX to PARAMETRIC*/
   } else i = 1;
-  max_graphcolors = lg(pari_graphcolors)-1;
+  max_graphcolors = lg(GP_DATA->graphcolors)-1;
   for (ltype = 0; ltype < nc; ltype++)
   {
-    current_color[grect] = pari_graphcolors[1+(ltype%max_graphcolors)];
+    current_color[grect] = GP_DATA->graphcolors[1+(ltype%max_graphcolors)];
     if (param) x = data[i++];
 
     y = data[i++]; nbpoints = y.nb;
@@ -1820,7 +1667,7 @@ plot_count(long *w, long lw, col_counter rcolcnt)
   RectObj *O;
   long col, i;
 
-  for (col = 1; col < lg(pari_colormap)-1; col++)
+  for (col = 1; col < lg(GP_DATA->colormap)-1; col++)
     for (i = 0; i < ROt_MAX; i++) rcolcnt[col][i] = 0;
   for (i = 0; i < lw; i++)
   {
@@ -1908,14 +1755,14 @@ rectdraw_flag(GEN list, long flag) { gendraw(list, 0, flag); }
 static void
 ps_sc(void *data, long col)
 {
-  long l = lg(pari_colormap)-1;
+  long l = lg(GP_DATA->colormap)-1;
   int r, g, b;
   if (col >= l)
   {
     pari_warn(warner,"non-existent color: %ld", col);
     col = l-1;
   }
-  color_to_rgb(gel(pari_colormap,col+1), &r, &g, &b);
+  color_to_rgb(gel(GP_DATA->colormap,col+1), &r, &g, &b);
   fprintf((FILE*)data,"%f %f %f setrgbcolor\n", r/255., g/255., b/255.);
 }
 
@@ -2030,7 +1877,7 @@ gen_rectdraw0(struct plot_eng *eng, long *w, long *x, long *y, long lw, double x
   long i, j;
   long hgapsize = eng->pl->hunit, fheight = eng->pl->fheight;
   long vgapsize = eng->pl->vunit,  fwidth = eng->pl->fwidth;
-  long numcolors = lg(pari_colormap)-1;
+  long numcolors = lg(GP_DATA->colormap)-1;
   for(i=0; i<lw; i++)
   {
     PariRect *e = rectgraph[w[i]];
