@@ -86,7 +86,9 @@ typedef long *GEN;
 #define VALPBITS    ((1UL<<VALPnumBITS)-1)
 #define VARNBITS    (MAXVARN<<VARNSHIFT)
 #define MAXVARN     ((1UL<<VARNnumBITS)-1)
-#define NO_VARIABLE (2147483647L) /* > MAXVARN */
+#define NO_VARIABLE (-1)
+#define VARARGBITS  HIGHBIT
+#define ARITYBITS   (~VARARGBITS)
 
 #define HIGHEXPOBIT (1UL<<(EXPOnumBITS-1))
 #define HIGHVALPBIT (1UL<<(VALPnumBITS-1))
@@ -127,27 +129,31 @@ typedef long *GEN;
                       (((ulong*)(x))[0]&(~LGBITS)) | evallg(s))
 #define incrprec(x)   ((x)++)
 
-#define expo(x)       ((long) ((((ulong*)(x))[1] & EXPOBITS) - HIGHEXPOBIT))
+#define expo(x)       ((long) ((((ulong)((x)[1])) & EXPOBITS) - HIGHEXPOBIT))
 #define setexpo(x,s)  (((ulong*)(x))[1]=\
                        (((ulong*)(x))[1]&(~EXPOBITS)) | (ulong)evalexpo(s))
 
-#define valp(x)       ((long) ((((ulong*)(x))[1] & VALPBITS) - HIGHVALPBIT))
+#define valp(x)       ((long) ((((ulong)((x)[1])) & VALPBITS) - HIGHVALPBIT))
 #define setvalp(x,s)  (((ulong*)(x))[1]=\
                        (((ulong*)(x))[1]&(~VALPBITS)) | (ulong)evalvalp(s))
 
-#define precp(x)      ((long) (((ulong*)(x))[1] >> PRECPSHIFT))
+#define precp(x)      ((long) (((ulong)((x)[1])) >> PRECPSHIFT))
 #define setprecp(x,s) (((ulong*)(x))[1]=\
                        (((ulong*)(x))[1]&(~PRECPBITS)) | (ulong)evalprecp(s))
 
-#define varn(x)       ((long)((((ulong*)(x))[1]&VARNBITS) >> VARNSHIFT))
+#define varn(x)       ((long)((((ulong)((x)[1]))&VARNBITS) >> VARNSHIFT))
 #define setvarn(x,s)  (((ulong*)(x))[1]=\
                        (((ulong*)(x))[1]&(~VARNBITS)) | (ulong)evalvarn(s))
 
-#define varncmp(x,y)  ((x)-(y))
-
 /* t_LIST */
-#define list_nmax(x) ((GEN)x)[1]
+
+#define list_typ(x)  ((long)(((ulong)((x)[1])) >> TYPSHIFT))
+#define list_nmax(x) ((long)(((ulong)((x)[1])) & LGBITS))
 #define list_data(x) ((GEN*)x)[2]
+enum {
+  t_LIST_RAW = 0,
+  t_LIST_MAP = 1
+};
 
 /* DO NOT REORDER THESE
  * actual values can be changed. Adapt lontyp in gen2.c */
@@ -173,5 +179,6 @@ enum {
   t_STR    =  21,
   t_VECSMALL= 22,
   t_CLOSURE = 23,
-  t_ERROR   = 24
+  t_ERROR   = 24,
+  t_INFINITY= 25
 };
