@@ -44,6 +44,20 @@ evalprecp(long x)
   return v;
 }
 
+INLINE int
+varncmp(long x, long y)
+{
+  if (varpriority[x] < varpriority[y]) return  1;
+  if (varpriority[x] > varpriority[y]) return -1;
+  return 0;
+}
+INLINE long
+varnmin(long x, long y)
+{ return (varpriority[x] <= varpriority[y])? x: y; }
+INLINE long
+varnmax(long x, long y)
+{ return (varpriority[x] >= varpriority[y])? x: y; }
+
 /* Inhibit some area gerepile-wise: declare it to be a non recursive
  * type, of length l. Thus gerepile won't inspect the zone, just copy it.
  * For the following situation:
@@ -88,17 +102,18 @@ INLINE GEN
 new_chunk(size_t x) /* x is a number of longs */
 {
   GEN z = ((GEN) avma) - x;
-  if (x > (avma-bot) / sizeof(long)) pari_err(e_STACK);
   CHECK_CTRLC
+  if (x > (avma-pari_mainstack->bot) / sizeof(long))
+    new_chunk_resize(x);
   avma = (pari_sp)z;
-
 #ifdef MEMSTEP
-  if (DEBUGMEM && memused != DISABLE_MEMUSED) {
-    long d = (long)memused - (long)z;
-    if (d > 4*MEMSTEP || d < -4*MEMSTEP)
+  if (DEBUGMEM && pari_mainstack->memused != DISABLE_MEMUSED) {
+    long d = (long)pari_mainstack->memused - (long)z;
+    if (labs(d) > 4*MEMSTEP)
     {
-      memused = (pari_sp)z;
-      err_printf("...%4.0lf Mbytes used\n",(top-memused)/1048576.);
+      pari_mainstack->memused = (pari_sp)z;
+      err_printf("...%4.0lf Mbytes used\n",
+                (pari_mainstack->top-pari_mainstack->memused)/1048576.);
     }
   }
 #endif
@@ -339,6 +354,8 @@ real_0_bit(long bitprec) { GEN x=cgetr(2); x[1]=evalexpo(bitprec); return x; }
 INLINE GEN
 real_0(long prec) { return real_0_bit(-prec2nbits(prec)); }
 INLINE GEN
+real_1_bit(long bit) { return real_1(nbits2prec(bit)); }
+INLINE GEN
 real_1(long prec) {
   GEN x = cgetr(prec);
   long i;
@@ -387,7 +404,11 @@ cmpir(GEN x, GEN y)
   GEN z;
 
   if (!signe(x)) return -signe(y);
-  if (!signe(y)) return  signe(x);
+  if (!signe(y))
+  {
+    if (expo(y) >= expi(x)) return 0;
+    return signe(x);
+  }
   av=avma; z = itor(x, realprec(y)); avma=av;
   return cmprr(z,y); /* cmprr does no memory adjustment */
 }
@@ -407,7 +428,7 @@ INLINE int
 cmprs(GEN x, long y) { return -cmpsr(y,x); }
 /* compare x and |y| */
 INLINE int
-cmpui(ulong x, GEN y)
+abscmpui(ulong x, GEN y)
 {
   long l = lgefint(y);
   ulong p;
@@ -419,7 +440,7 @@ cmpui(ulong x, GEN y)
   return p < x ? 1 : -1;
 }
 INLINE int
-cmpiu(GEN x, ulong y) { return -cmpui(y,x); }
+abscmpiu(GEN x, ulong y) { return -abscmpui(y,x); }
 INLINE int
 cmpsi(long x, GEN y)
 {
@@ -465,13 +486,13 @@ equalsi(long x, GEN y)
 }
 /* x == |y| ? */
 INLINE int
-equalui(ulong x, GEN y)
+absequalui(ulong x, GEN y)
 {
   if (!x) return !signe(y);
   return (lgefint(y) == 3 && (ulong)y[2] == x);
 }
 INLINE int
-equaliu(GEN x, ulong y) { return equalui(y,x); }
+absequaliu(GEN x, ulong y) { return absequalui(y,x); }
 INLINE int
 equalis(GEN x, long y) { return equalsi(y,x); }
 
@@ -575,6 +596,24 @@ INLINE long mod4(GEN x)  { return mod2BIL(x) & 3; }
 INLINE long mod2(GEN x)  { return mod2BIL(x) & 1; }
 INLINE int
 mpodd(GEN x) { return signe(x) && mod2(x); }
+/* x mod 2^n, n < BITS_IN_LONG */
+INLINE ulong
+umodi2n(GEN x, long n)
+{
+  long s = signe(x);
+  const ulong _2n = 1UL << n;
+  ulong m;
+  if (!s) return 0;
+  m = *int_LSW(x) & (_2n - 1);
+  if (s < 0 && m) m = _2n - m;
+  return m;
+}
+INLINE ulong Mod64(GEN x){ return umodi2n(x,6); }
+INLINE ulong Mod32(GEN x){ return umodi2n(x,5); }
+INLINE ulong Mod16(GEN x){ return umodi2n(x,4); }
+INLINE ulong Mod8(GEN x) { return umodi2n(x,3); }
+INLINE ulong Mod4(GEN x) { return umodi2n(x,2); }
+INLINE ulong Mod2(GEN x) { return umodi2n(x,1); }
 
 INLINE GEN
 truedivii(GEN a,GEN b) { return truedvmdii(a,b,NULL); }
@@ -600,6 +639,11 @@ smodss(long x, long y)
   long r = x%y;
   return (r >= 0)? r: labs(y) + r;
 }
+INLINE ulong
+umodsu(long x, ulong y)
+{
+  return x>=0 ? x%y: Fl_neg((-x)%y, y);
+}
 
 INLINE long
 sdivss_rem(long x, long y, long *r)
@@ -620,6 +664,13 @@ udivuu_rem(ulong x, ulong y, ulong *r)
   if (!y) pari_err_INV("udivuu_rem",gen_0);
   *r = x % y; return x / y;
 }
+INLINE ulong
+ceildivuu(ulong a, ulong b)
+{
+  ulong c = a/b;
+  return (a%b)? c+1: c;
+}
+
 
 INLINE ulong
 udivui_rem(ulong x, GEN y, ulong *r)
@@ -863,6 +914,66 @@ dvdiiz(GEN x, GEN y, GEN z)
   affii(p1,z); avma=av; return 1;
 }
 
+INLINE ulong
+remlll_pre(ulong u2, ulong u1, ulong u0, ulong n, ulong ninv)
+{
+  u1 = remll_pre(u2, u1, n, ninv);
+  return remll_pre(u1, u0, n, ninv);
+}
+
+INLINE ulong
+Fl_sqr_pre(ulong a, ulong p, ulong pi)
+{
+  register ulong x;
+  LOCAL_HIREMAINDER;
+  x = mulll(a,a);
+  return remll_pre(hiremainder, x, p, pi);
+}
+
+INLINE ulong
+Fl_mul_pre(ulong a, ulong b, ulong p, ulong pi)
+{
+  register ulong x;
+  LOCAL_HIREMAINDER;
+  x = mulll(a,b);
+  return remll_pre(hiremainder, x, p, pi);
+}
+
+INLINE ulong
+Fl_addmul_pre(ulong x0, ulong x1, ulong y0, ulong p, ulong pi)
+{
+  ulong l0, h0;
+  LOCAL_HIREMAINDER;
+  hiremainder = y0;
+  l0 = addmul(x0, x1); h0 = hiremainder;
+  return remll_pre(h0, l0, p, pi);
+}
+
+INLINE ulong
+Fl_addmulmul_pre(ulong x0, ulong y0, ulong x1, ulong y1, ulong p, ulong pi)
+{
+  ulong l0, l1, h0, h1;
+  LOCAL_OVERFLOW;
+  LOCAL_HIREMAINDER;
+  l0 = mulll(x0, y0); h0 = hiremainder;
+  l1 = mulll(x1, y1); h1 = hiremainder;
+  l0 = addll(l0, l1); h0 = addllx(h0, h1);
+  return overflow ? remlll_pre(1, h0, l0, p, pi): remll_pre(h0, l0, p, pi);
+}
+
+INLINE ulong
+Fl_ellj_pre(ulong a4, ulong a6, ulong p, ulong pi)
+{
+  /* a43 = 4 a4^3 */
+  ulong a43 = Fl_double(Fl_double(
+              Fl_mul_pre(a4, Fl_sqr_pre(a4, p, pi), p, pi), p), p);
+  /* a62 = 27 a6^2 */
+  ulong a62 = Fl_mul_pre(Fl_sqr_pre(a6, p, pi), 27 % p, p, pi);
+  ulong z1 = Fl_mul_pre(a43, 1728 % p, p, pi);
+  ulong z2 = Fl_add(a43, a62, p);
+  return Fl_div(z1, z2, p);
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                        MP (INT OR REAL)                         */
@@ -930,6 +1041,15 @@ Fl_triple(ulong a, ulong p)
   res += a;
   return (res >= p || res < a)? res - p: res;
 }
+INLINE ulong
+Fl_halve(ulong a, ulong p)
+{
+  ulong ap, ap2;
+  if ((a&1UL)==0) return a>>1;
+  ap = a + p; ap2 = ap>>1;
+  return ap>=a ? ap2: (ap2|HIGHBIT);
+}
+
 INLINE ulong
 Fl_add(ulong a, ulong b, ulong p)
 {
@@ -1057,7 +1177,7 @@ shift_left(GEN z2, GEN z1, long imin, long imax, ulong f,  ulong sh)
     *te-- = (l << sh) | k;
     k     = l >> m;
   }
-  *te = (*se << sh) | k;
+  *te = (((ulong)*se) << sh) | k;
 }
 /* z2[imin..imax] := f.z1[imin..imax-1] shifted right sh bits
  * (feeding f from the left). Assume sh > 0 */
@@ -1138,7 +1258,7 @@ affsr(long x, GEN y)
     sh = bfffo(x);
     y[1] = evalsigne(1) | _evalexpo((BITS_IN_LONG-1)-sh);
   }
-  y[2] = x<<sh; for (i=3; i<ly; i++) y[i]=0;
+  y[2] = ((ulong)x)<<sh; for (i=3; i<ly; i++) y[i]=0;
 }
 
 INLINE void
