@@ -43,8 +43,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
  }
  * Preprint available from:
  * ftp://ftp.risc.uni-linz.ac.at/pub/techreports/1994/94-64.ps.gz */
-
-/* #define DEBUG_RATLIFT */
 static ulong
 get_vmax(GEN r, long lb, long lbb)
 {
@@ -63,9 +61,6 @@ get_vmax(GEN r, long lb, long lbb)
     else
       vmax = 1UL << (lr-1); /* pessimistic but faster than a division */
   }
-#ifdef DEBUG_RATLIFT
-  err_printf("rl-fs: vmax=%lu\n", vmax);
-#endif
   return vmax;
 }
 
@@ -74,7 +69,7 @@ int
 Fp_ratlift(GEN x, GEN m, GEN amax, GEN bmax, GEN *a, GEN *b)
 {
   GEN d, d1, v, v1, q, r;
-  pari_sp av = avma, av1, lim;
+  pari_sp av = avma, av1;
   long lb, lbb, s, s0;
   ulong vmax;
   ulong xu, xu1, xv, xv1; /* Lehmer stage recurrence matrix */
@@ -97,7 +92,7 @@ Fp_ratlift(GEN x, GEN m, GEN amax, GEN bmax, GEN *a, GEN *b)
   lb = lgefint(bmax);
   lbb = bfffo(*int_MSW(bmax));
   s = 1;
-  av1 = avma; lim = stack_lim(av, 1);
+  av1 = avma;
 
   /* General case: Euclidean division chain starting with m div x, and
    * with bounds on the sequence of convergents' denoms v_j.
@@ -194,9 +189,6 @@ Fp_ratlift(GEN x, GEN m, GEN amax, GEN bmax, GEN *a, GEN *b)
     if (lhmres <= 0 && signe(d1))
     {
       q = dvmdii(d,d1,&r);
-#ifdef DEBUG_LEHMER
-      err_printf("Full division:\n  q = %Ps\n", q);
-#endif
       d = d1; d1 = r;
       r = addii(v, mulii(q,v1));
       v = v1; v1 = r;
@@ -212,7 +204,7 @@ Fp_ratlift(GEN x, GEN m, GEN amax, GEN bmax, GEN *a, GEN *b)
       }
     }
 
-    if (low_stack(lim, stack_lim(av,1)))
+    if (gc_needed(av,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"ratlift");
       gerepileall(av1, 4, &d, &d1, &v, &v1);
@@ -230,19 +222,11 @@ Fp_ratlift(GEN x, GEN m, GEN amax, GEN bmax, GEN *a, GEN *b)
     /* Assertions: lgefint(d)==lgefint(d1)==3.
      * Moreover, we aren't done already, or we would have returned by now.
      * Recompute vmax */
-#ifdef DEBUG_RATLIFT
-    err_printf("rl-fs: d,d1=%Ps,%Ps\n", d, d1);
-    err_printf("rl-fs: v,v1=%Ps,%Ps\n", v, v1);
-#endif
     r = addii(v,v1);
     if (cmpii(r,bmax) > 0) { avma = av; return 0; } /* done, not found */
     vmax = get_vmax(r, lb, lbb);
     /* single-word "Lehmer", discarding the gcd or whatever it returns */
     (void)rgcduu((ulong)*int_MSW(d), (ulong)*int_MSW(d1), vmax, &xu, &xu1, &xv, &xv1, &s0);
-#ifdef DEBUG_RATLIFT
-    err_printf("rl-fs: [%lu,%lu; %lu,%lu] %s\n",
-               xu, xu1, xv, xv1, s0 < 0 ? "-" : "+");
-#endif
     if (xv1 == 1) /* avoid multiplications */
     { /* re-use r = v+v1 computed above */
       v = v1; v1 = r;
