@@ -199,6 +199,83 @@ find:
   return retval;
 }
 
+/********************************************************************/
+/**                                                                **/
+/**                   HASH TABLE MANIPULATIONS                     **/
+/**                                                                **/
+/********************************************************************/
+/* return hashing value for identifier s */
+static ulong
+hashvalue(const char *s)
+{
+  ulong n = 0, c;
+  while ( (c = (ulong)*s++) ) n = (n<<1) ^ c;
+  return n;
+}
+
+static ulong
+hashvalue_raw(const char *s, long len)
+{
+  long n = 0, i;
+  for(i=0; i<len; i++) { n = (n<<1) ^ *s; s++; }
+  return n;
+}
+
+static void
+insertep(entree *ep, entree **table, ulong hash)
+{
+  ep->hash = hash;
+  hash %= functions_tblsz;
+  ep->next = table[hash];
+  table[hash] = ep;
+}
+
+static entree *
+initep(const char *name, long len)
+{
+  const long add = 4*sizeof(long);
+  entree *ep = (entree *) pari_calloc(sizeof(entree) + add + len+1);
+  entree *ep1 = initial_value(ep);
+  char *u = (char *) ep1 + add;
+  ep->name    = u; strncpy(u, name,len); u[len]=0;
+  ep->valence = EpNEW;
+  ep->value   = NULL;
+  ep->menu    = 0;
+  ep->code    = NULL;
+  ep->help    = NULL;
+  ep->pvalue  = NULL;
+  ep->arity   = 0;
+  return ep;
+}
+
+/* Look for s of length len in T; if 'insert', insert if missing */
+static entree *
+findentry(const char *s, long len, entree **T, int insert)
+{
+  ulong hash = hashvalue_raw(s, len);
+  entree *ep;
+  for (ep = T[hash % functions_tblsz]; ep; ep = ep->next)
+    if (ep->hash == hash)
+    {
+      const char *t = ep->name;
+      if (!strncmp(t, s, len) && !t[len]) return ep;
+    }
+  /* not found */
+  if (insert) { ep = initep(s,len); insertep(ep, T, hash); }
+  return ep;
+}
+entree *
+pari_is_default(const char *s)
+{ return findentry(s, strlen(s), defaults_hash, 0); }
+entree *
+is_entry(const char *s)
+{ return findentry(s, strlen(s), functions_hash, 0); }
+entree *
+fetch_entry_raw(const char *s, long len)
+{ return findentry(s, len, functions_hash, 1); }
+entree *
+fetch_entry(const char *s) { return fetch_entry_raw(s, strlen(s)); }
+
 /*******************************************************************/
 /*                                                                 */
 /*                  SYNTACTICAL ANALYZER FOR GP                    */
@@ -208,14 +285,17 @@ GEN
 readseq(char *t)
 {
   pari_sp av = avma;
-  return gerepileupto(av, closure_evalres(pari_compile_str(t,0)));
+  GEN x;
+  if (gp_meta(t,0)) return gnil;
+  x = pari_compile_str(t);
+  return gerepileupto(av, closure_evalres(x));
 }
 
 /* filtered readseq = remove blanks and comments */
 GEN
 gp_read_str(const char *s)
 {
-  char *t = filtre(s, (compatible == OLDALL));
+  char *t = gp_filter(s);
   GEN x = readseq(t);
   pari_free(t); return x;
 }
@@ -223,8 +303,8 @@ gp_read_str(const char *s)
 GEN
 compile_str(const char *s)
 {
-  char *t = filtre(s, (compatible == OLDALL));
-  GEN x = pari_compile_str(t, 1);
+  char *t = gp_filter(s);
+  GEN x = pari_compile_str(t);
   pari_free(t); return x;
 }
 
@@ -233,7 +313,7 @@ check_proto(const char *code)
 {
   long arity = 0;
   const char *s = code, *old;
-  if (*s == 'l' || *s == 'v' || *s == 'i' || *s == 'm') s++;
+  if (*s == 'l' || *s == 'v' || *s == 'i' || *s == 'm' || *s == 'u') s++;
   while (*s && *s != '\n') switch (*s++)
   {
     case '&':
@@ -241,6 +321,7 @@ check_proto(const char *code)
     case 'G':
     case 'I':
     case 'J':
+    case 'U':
     case 'L':
     case 'M':
     case 'P':
@@ -248,8 +329,8 @@ check_proto(const char *code)
     case 'f':
     case 'n':
     case 'p':
+    case 'b':
     case 'r':
-    case 'x':
       arity++;
       break;
     case 'E':
@@ -281,34 +362,24 @@ check_proto(const char *code)
   if (arity > 20) pari_err_IMPL("functions with more than 20 parameters");
   return arity;
 }
-
-static entree *
-installep(const char *name, long len, entree **table)
+static void
+check_name(const char *name)
 {
-  const long add = 4*sizeof(long);
-  entree *ep = (entree *) pari_calloc(sizeof(entree) + add + len+1);
-  entree *ep1 = initial_value(ep);
-  char *u = (char *) ep1 + add;
-  ep->name    = u; strncpy(u, name,len); u[len]=0;
-  ep->valence = EpNEW;
-  ep->value   = NULL;
-  ep->menu    = 0;
-  ep->code    = NULL;
-  ep->help    = NULL;
-  ep->pvalue  = NULL;
-  ep->arity   = 0;
-  ep->next    = *table;
-  return *table = ep;
+  const char *s = name;
+  if (isalpha((int)*s))
+    while (is_keyword_char(*++s)) /* empty */;
+  if (*s) pari_err(e_SYNTAX,"not a valid identifier", s, name);
 }
 
 entree *
 install(void *f, const char *name, const char *code)
 {
-  long hash, arity;
-  entree *ep = is_entry_intern(name, functions_hash, &hash);
+  long arity = check_proto(code);
+  entree *ep;
 
-  arity=check_proto(code);
-  if (ep && ep->valence != EpNEW)
+  check_name(name);
+  ep = fetch_entry(name);
+  if (ep->valence != EpNEW)
   {
     if (ep->valence != EpINSTALL)
       pari_err(e_MISC,"[install] identifier '%s' already in use", name);
@@ -317,34 +388,36 @@ install(void *f, const char *name, const char *code)
   }
   else
   {
-    const char *s = name;
-    if (isalpha((int)*s))
-      while (is_keyword_char(*++s)) /* empty */;
-    if (*s) pari_err(e_SYNTAX,"not a valid identifier", s, name);
-    if (!ep) ep = installep(name, strlen(name), functions_hash + hash);
-    ep->value=f; ep->valence=EpINSTALL;
+    ep->value = f;
+    ep->valence = EpINSTALL;
   }
   ep->code = pari_strdup(code);
-  ep->arity=arity;
-  return ep;
+  ep->arity = arity; return ep;
 }
 
+static void
+killep(entree *ep)
+{
+  GEN p = (GEN)initial_value(ep);
+  freeep(ep);
+  *p = 0; /* otherwise pari_var_create won't regenerate it */
+  ep->valence = EpNEW;
+  ep->value   = NULL;
+  ep->pvalue  = NULL;
+}
 /* Kill ep, i.e free all memory it references, and reset to initial value */
 void
 kill0(const char *e)
 {
   entree *ep = is_entry(e);
   if (!ep || EpSTATIC(ep)) pari_err(e_MISC,"can't kill that");
-  freeep(ep);
-  ep->valence = EpNEW;
-  ep->value   = NULL;
-  ep->pvalue  = NULL;
+  killep(ep);
 }
 
 void
 addhelp(const char *e, char *s)
 {
-  entree *ep = fetch_entry(e, strlen(e));
+  entree *ep = fetch_entry(e);
   if (ep->help && !EpSTATIC(ep)) pari_free((void*)ep->help);
   ep->help = pari_strdup(s);
 }
@@ -363,11 +436,137 @@ type0(GEN x)
 /*******************************************************************/
 
 #ifdef LONG_IS_64BIT
-static const long MAX_DIGITS = 19;
+static const long MAX_DIGITS  = 19;
 #else
-static const long MAX_DIGITS = 9;
+static const long MAX_DIGITS  = 9;
 #endif
 
+static const long MAX_XDIGITS = BITS_IN_LONG>>2;
+static const long MAX_BDIGITS = BITS_IN_LONG;
+
+static int
+ishex(const char **s)
+{
+  if (**s == '0' && ((*s)[1] == 'x' || (*s)[1] == 'X' ))
+  {
+    *s += 2;
+    return 1;
+  }
+  else
+    return 0;
+}
+
+static int
+isbin(const char **s)
+{
+  if (**s == '0' && ((*s)[1] == 'b' || (*s)[1] == 'B' ))
+  {
+    *s += 2;
+    return 1;
+  }
+  else
+    return 0;
+}
+
+static ulong
+bin_number_len(const char *s, long n)
+{
+  ulong m = 0;
+  long i;
+  for (i = 0; i < n; i++,s++)
+    m = 2*m + (*s - '0');
+  return m;
+}
+
+static int
+pari_isbdigit(int c)
+{
+  return c=='0' || c=='1';
+}
+
+static ulong
+hex_number_len(const char *s, long n)
+{
+  ulong m = 0;
+  long i;
+  for(i = 0; i < n; i++, s++)
+  {
+    ulong c;
+    if( *s >= '0' && *s <= '9')
+      c = *s - '0';
+    else if( *s >= 'A' && *s <= 'F')
+      c = *s - 'A' + 10;
+    else
+      c = *s - 'a' + 10;
+    m = 16*m + c;
+  }
+  return m;
+}
+
+static GEN
+strtobin_len(const char *s, long n, long B, ulong num(const char *s, long n))
+{
+  long i, l = (n+B-1)/B;
+  GEN N, Np;
+  N = cgetipos(l+2);
+  Np = int_LSW(N);
+  for (i=1; i<l; i++, Np = int_nextW(Np))
+    uel(Np, 0) = num(s+n-i*B, B);
+  uel(Np, 0) = num(s, n-(i-1)*B);
+  return int_normalize(N, 0);
+}
+
+static GEN
+binary_read(const char **ps, long B, int is(int), ulong num(const char *s, long n))
+{
+  const char *s = *ps;
+  while (is((int)**ps)) (*ps)++;
+  return strtobin_len(s, *ps-s, B, num);
+}
+
+static GEN
+bin_read(const char **ps)
+{
+  return binary_read(ps, MAX_BDIGITS, pari_isbdigit, bin_number_len);
+}
+
+static GEN
+hex_read(const char **ps)
+{
+  return binary_read(ps, MAX_XDIGITS, isxdigit, hex_number_len);
+}
+
+static ulong
+dec_number_len(const char *s, long B)
+{
+  ulong m = 0;
+  long n;
+  for (n = 0; n < B; n++,s++)
+    m = 10*m + (*s - '0');
+  return m;
+}
+
+static GEN
+dec_strtoi_len(const char *s, long n)
+{
+  const long B = MAX_DIGITS;
+  long i, l = (n+B-1)/B;
+  GEN V = cgetg(l+1, t_VECSMALL);
+  for (i=1; i<l; i++)
+    uel(V,i) = dec_number_len(s+n-i*B, B);
+  uel(V, i) = dec_number_len(s, n-(i-1)*B);
+  return fromdigitsu(V, powuu(10, B));
+}
+
+static GEN
+dec_read_more(const char **ps)
+{
+  pari_sp av = avma;
+  const char *s = *ps;
+  while (isdigit((int)**ps)) (*ps)++;
+  return gerepileuptoint(av, dec_strtoi_len(s, *ps-s));
+}
+
 static ulong
 number(int *n, const char **s)
 {
@@ -377,48 +576,25 @@ number(int *n, const char **s)
   return m;
 }
 
-ulong
-u_pow10(int n)
+static GEN
+dec_read(const char **s)
 {
-  const ulong pw10[] = {
-    1UL
-    ,10UL
-    ,100UL
-    ,1000UL
-    ,10000UL
-    ,100000UL
-    ,1000000UL
-    ,10000000UL
-    ,100000000UL
-    ,1000000000UL
-#ifdef LONG_IS_64BIT
-    ,10000000000UL
-    ,100000000000UL
-    ,1000000000000UL
-    ,10000000000000UL
-    ,100000000000000UL
-    ,1000000000000000UL
-    ,10000000000000000UL
-    ,100000000000000000UL
-    ,1000000000000000000UL
-    ,10000000000000000000UL
-#endif
-  };
-  return pw10[n];
+  int nb;
+  ulong y  = number(&nb, s);
+  if (nb < MAX_DIGITS)
+    return utoi(y);
+  *s -= MAX_DIGITS;
+  return dec_read_more(s);
 }
 
 static GEN
-int_read_more(GEN y, const char **ps)
+real_read_more(GEN y, const char **ps)
 {
   pari_sp av = avma;
-  int i = 0, nb;
-  while (isdigit((int)**ps))
-  {
-    ulong m = number(&nb, ps);
-    if (avma != av && ++i > 4) { avma = av; i = 0; } /* HACK gerepile */
-    y = addumului(m, u_pow10(nb), y);
-  }
-  return y;
+  const char *s = *ps;
+  GEN z = dec_read(ps);
+  long e = *ps-s;
+  return gerepileuptoint(av, addmulii(z, powuu(10, e), y));
 }
 
 static long
@@ -461,7 +637,7 @@ real_read(pari_sp av, const char **s, GEN y, long prec)
         }
         --*s; return y; /* member */
       }
-      y = int_read_more(y, s);
+      if (isdigit((int)**s)) y = real_read_more(y, s);
       n = old - *s;
       if (**s != 'E' && **s != 'e')
       {
@@ -489,9 +665,13 @@ real_read(pari_sp av, const char **s, GEN y, long prec)
 static GEN
 int_read(const char **s)
 {
-  int nb;
-  GEN y = utoi(number(&nb, s));
-  if (nb == MAX_DIGITS) y = int_read_more(y, s);
+  GEN y;
+  if (isbin(s))
+    y = bin_read(s);
+  else if (ishex(s))
+    y = hex_read(s);
+  else
+    y = dec_read(s);
   return y;
 }
 
@@ -502,7 +682,7 @@ GEN
 strtor(const char *s, long prec)
 {
   pari_sp av = avma;
-  GEN y = int_read(&s);
+  GEN y = dec_read(&s);
   y = real_read(av, &s, y, prec);
   if (typ(y) == t_REAL) return y;
   return gerepileuptoleaf(av, itor(y, prec));
@@ -615,6 +795,16 @@ pari_lex(union token_value *yylval, struct node_loc *yylloc, char **lex)
     yylloc->end=*lex;
     return '.';
   }
+  if (isbin((const char**)lex))
+  {
+    while (**lex=='0' || **lex=='1') ++*lex;
+    return KINTEGER;
+  }
+  if (ishex((const char**)lex))
+  {
+    while (isxdigit((int)**lex)) ++*lex;
+    return KINTEGER;
+  }
   if (isdigit((int)**lex))
   {
     int token=skipconstante(lex);
@@ -719,106 +909,142 @@ chartoGENstr(char c)
   char *t = GSTR(x);
   t[0] = c; t[1] = 0; return x;
 }
-/********************************************************************/
-/**                                                                **/
-/**                   HASH TABLE MANIPULATIONS                     **/
-/**                                                                **/
-/********************************************************************/
-/* return hashing value for identifier s */
-static ulong
-hashvalue(const char *s)
-{
-  ulong n = 0, c;
-  while ( (c = (ulong)*s++) ) n = (n<<1) ^ c;
-  return n % functions_tblsz;
-}
-
-static ulong
-hashvalue_raw(const char *s, long len, ulong n)
-{
-  long i;
-  for(i=0; i<len; i++) { n = (n<<1) ^ *s; s++; }
-  return n % functions_tblsz;
-}
-
-/* Looking for entry in hashtable. ep1 is the cell's first element */
-static entree *
-findentry(const char *name, long len, entree *ep1)
-{
-  entree *ep;
-  for (ep = ep1; ep; ep = ep->next)
-    if (!strncmp(ep->name, name, len) && !(ep->name)[len]) return ep;
-  return NULL; /* not found */
-}
-
-entree *
-is_entry_intern(const char *s, entree **table, long *pthash)
-{
-  long hash = hashvalue(s);
-  if (pthash) *pthash = hash;
-  return findentry(s, strlen(s), table[hash]);
-}
-
-entree *
-is_entry(const char *s)
-{
-  return is_entry_intern(s,functions_hash,NULL);
-}
-
-entree *
-fetch_entry(const char *s, long len)
-{
-  entree **funhash = functions_hash + hashvalue_raw(s, len, 0);
-  entree *ep = findentry(s, len, *funhash);
-  if (ep) return ep;
-  else return installep(s,len,funhash);
-}
-
-/* Assume s point somewhere in the code text, so s[-1]='.' and s[-2]>0
- * So many kludges, so little time */
-entree *
-fetch_member(const char *s, long len)
-{
-  entree **funhash = functions_hash+hashvalue_raw(s-1, len+1, '_');
-  entree *ep;
-  for (ep = *funhash; ep; ep = ep->next)
-  {
-    if (ep->name[0]!='_' || ep->name[1]!='.') continue;
-    if (!strncmp(ep->name+2, s, len) && !(ep->name)[len+2]) break;
-  }
-  if (ep) return ep;
-  ep=installep(s-2,len+2,funhash);
-  ((char*)ep->name)[0]='_';
-  return ep;
-}
 
 /********************************************************************/
 /*                                                                  */
 /*                Formal variables management                       */
 /*                                                                  */
 /********************************************************************/
-static long max_avail; /* max variable not yet used */
-static long nvar; /* first GP free variable */
+static THREAD long max_priority, min_priority;
+static THREAD long max_avail; /* max variable not yet used */
+static THREAD long nvar; /* first GP free variable */
+static hashtable *h_polvar;
+static struct pari_varstate global_varstate;
+static long *global_varpriority;
 
-void pari_var_init(void) {
+void
+varstate_save(struct pari_varstate *s)
+{
+  s->nvar = nvar;
+  s->max_avail = max_avail;
+  s->max_priority = max_priority;
+  s->min_priority = min_priority;
+}
+
+static void
+varentries_set(long v, entree *ep)
+{
+  hash_insert(h_polvar, (void*)ep->name, (void*)v);
+  varentries[v] = ep;
+}
+static int
+_given_value(void *E, hashentry *e) { return e->val == E; }
+
+static void
+varentries_unset(long v)
+{
+  entree *ep = varentries[v];
+  if (ep)
+  {
+    hashentry *e = hash_remove_select(h_polvar, (void*)ep->name, (void*)v,
+        _given_value);
+    if (!e) pari_err_BUG("varentries_unset [unknown var]");
+    varentries[v] = NULL;
+    pari_free(e);
+    if (v <= nvar && ep == is_entry(ep->name))
+    { /* known to the GP interpreter; entree in functions_hash is permanent */
+      GEN p = (GEN)initial_value(ep);
+      if (ep->value == p) { ep->value = NULL; ep->valence = EpNEW; }
+      *p = 0;
+    }
+    else /* from name_var() or a direct pari_var_create() */
+      pari_free(ep);
+ }
+}
+static void
+varentries_reset(long v, entree *ep)
+{
+  varentries_unset(v);
+  varentries_set(v, ep);
+}
+
+static void
+var_restore(struct pari_varstate *s)
+{
+  nvar = s->nvar;
+  max_avail = s->max_avail;
+  max_priority = s->max_priority;
+  min_priority = s->min_priority;
+}
+
+void
+varstate_restore(struct pari_varstate *s)
+{
+  long i;
+  for (i = nvar; i >= s->nvar; i--)
+  {
+    varentries_unset(i);
+    varpriority[i] = -i;
+  }
+  for (i = max_avail; i < s->max_avail; i++)
+  {
+    varentries_unset(i);
+    varpriority[i] = -i;
+  }
+  var_restore(s);
+}
+
+void
+pari_thread_init_varstate(void)
+{
+  long i;
+  var_restore(&global_varstate);
+  varpriority = (long*)newblock((MAXVARN+2)) + 1;
+  varpriority[-1] = 1-LONG_MAX;
+  for (i = 0; i < max_avail; i++) varpriority[i] = global_varpriority[i];
+}
+
+void
+pari_pthread_init_varstate(void)
+{
+  varstate_save(&global_varstate);
+  global_varpriority = varpriority;
+}
+
+void
+pari_var_close(void)
+{
+  free((void*)varentries);
+  free((void*)(varpriority-1));
+  hash_destroy(h_polvar);
+}
+
+void
+pari_var_init(void)
+{
+  long i;
+  varentries = (entree**) pari_calloc((MAXVARN+1)*sizeof(entree*));
+  varpriority = (long*)pari_malloc((MAXVARN+2)*sizeof(long)) + 1;
+  varpriority[-1] = 1-LONG_MAX;
+  h_polvar = hash_create_str(100, 0);
   nvar = 0; max_avail = MAXVARN;
-  (void)fetch_var();
-  (void)fetch_named_var("x");
+  max_priority = min_priority = 0;
+  (void)fetch_user_var("x");
+  (void)fetch_user_var("y");
+  /* initialize so that people can use pol_x(i) directly */
+  for (i = 2; i <= (long)MAXVARN; i++) varpriority[i] = -i;
+  /* reserve varnum 1..9 for static temps with predictable priority wrt x */
+  nvar = 10;
+  min_priority = -MAXVARN;
 }
 long pari_var_next(void) { return nvar; }
 long pari_var_next_temp(void) { return max_avail; }
-static long
-pari_var_pop(long v)
-{
-  if (v != nvar-1) pari_err(e_MISC,"can't pop user variable %ld", v);
-  return --nvar;
-}
-void
+long
 pari_var_create(entree *ep)
 {
   GEN p = (GEN)initial_value(ep);
   long v;
-  if (*p) return;
+  if (*p) return varn(p);
   if (nvar == max_avail) pari_err(e_MISC,"no more variables available");
   v = nvar++;
   /* set p = pol_x(v) */
@@ -826,74 +1052,105 @@ pari_var_create(entree *ep)
   p[1] = evalsigne(1) | evalvarn(v);
   gel(p,2) = gen_0;
   gel(p,3) = gen_1;
-  varentries[v] = ep;
+  varentries_set(v, ep);
+  varpriority[v]= min_priority--;
+  return v;
 }
 
 long
 delete_var(void)
 { /* user wants to delete one of his/her/its variables */
-  if (max_avail == MAXVARN-1) return 0; /* nothing to delete */
-  max_avail++; return max_avail+1;
+  if (max_avail == MAXVARN) return 0; /* nothing to delete */
+  max_avail++;
+  if      (varpriority[max_avail] == min_priority) min_priority++;
+  else if (varpriority[max_avail] == max_priority) max_priority--;
+  return max_avail+1;
 }
 long
 fetch_var(void)
 {
   if (nvar == max_avail) pari_err(e_MISC,"no more variables available");
+  varpriority[max_avail] = min_priority--;
+  return max_avail--;
+}
+long
+fetch_var_higher(void)
+{
+  if (nvar == max_avail) pari_err(e_MISC,"no more variables available");
+  varpriority[max_avail] = ++max_priority;
   return max_avail--;
 }
 
-/* FIXE: obsolete, kept for backward compatibility */
-long
-manage_var(long n, entree *ep)
-{
-  switch(n) {
-      case manage_var_init: pari_var_init(); return 0;
-      case manage_var_next: return pari_var_next();
-      case manage_var_max_avail: return pari_var_next_temp();
-      case manage_var_pop: return pari_var_pop((long)ep);
-      case manage_var_delete: return delete_var();
-      case manage_var_create:
-        pari_var_create(ep);
-        return varn((GEN)initial_value(ep));
-  }
-  pari_err(e_MISC, "panic");
-  return -1; /* not reached */
-}
+static int
+_higher(void *E, hashentry *e)
+{ long v = (long)e->val; return (varncmp(v, (long)E) < 0); }
+static int
+_lower(void *E, hashentry *e)
+{ long v = (long)e->val; return (varncmp(v, (long)E) > 0); }
 
-entree *
-fetch_named_var(const char *s)
+static GEN
+var_register(long v, const char *s)
 {
-  entree **funhash = functions_hash + hashvalue(s);
-  entree *ep = findentry(s, strlen(s), *funhash);
-  if (!ep) ep = installep(s,strlen(s),funhash);
-  else switch (EpVALENCE(ep))
+  varentries_reset(v, initep(s, strlen(s)));
+  return pol_x(v);
+}
+GEN
+varhigher(const char *s, long w)
+{
+  long v;
+  if (w >= 0)
   {
-    case EpVAR: return ep;
-    case EpNEW: break;
-    default: pari_err(e_MISC, "%s already exists with incompatible valence", s);
+    hashentry *e = hash_select(h_polvar, (void*)s, (void*)w, _higher);
+    if (e) return pol_x((long)e->val);
   }
-  pari_var_create(ep);
-  ep->valence=EpVAR;
-  ep->value=initial_value(ep);
-  return ep;
+  /* no luck: need to create */
+  if (nvar == max_avail) pari_err(e_MISC,"no more variables available");
+  v = nvar++;
+  varpriority[v]= ++max_priority;
+  return var_register(v, s);
+}
+GEN
+varlower(const char *s, long w)
+{
+  long v;
+  if (w >= 0)
+  {
+    hashentry *e = hash_select(h_polvar, (void*)s, (void*)w, _lower);
+    if (e) return pol_x((long)e->val);
+  }
+  /* no luck: need to create */
+  v = fetch_var();
+  return var_register(v, s);
 }
 
 long
 fetch_user_var(const char *s)
 {
-  return varn((GEN)initial_value(fetch_named_var(s)) );
+  entree *ep = fetch_entry(s);
+  long v;
+  switch (EpVALENCE(ep))
+  {
+    case EpVAR: return varn((GEN)initial_value(ep));
+    case EpNEW: break;
+    default: pari_err(e_MISC, "%s already exists with incompatible valence", s);
+  }
+  v = pari_var_create(ep);
+  ep->valence = EpVAR;
+  ep->value = initial_value(ep);
+  return v;
 }
 
 GEN
-fetch_var_value(long vx, GEN t)
+fetch_var_value(long v, GEN t)
 {
-  entree *ep = varentries[vx];
-  long vn;
+  entree *ep = varentries[v];
   if (!ep) return NULL;
-  if (!t)  return (GEN) ep->value;
-  vn=localvars_find(t,ep);
-  if (vn) return get_lex(vn);
-  return (GEN) ep->value;
+  if (t)
+  {
+    long vn = localvars_find(t,ep);
+    if (vn) return get_lex(vn);
+  }
+  return (GEN)ep->value;
 }
 
 void
@@ -912,27 +1169,30 @@ name_var(long n, const char *s)
   ep->valence = EpVAR;
   ep->name = u; strcpy(u,s);
   ep->value = gen_0; /* in case geval is called */
-  if (varentries[n]) pari_free(varentries[n]);
-  varentries[n] = ep;
+  varentries_reset(n, ep);
 }
 
+static int
+cmp_by_var(void *E,GEN x, GEN y)
+{ (void)E; return varncmp((long)x,(long)y); }
+GEN
+vars_sort_inplace(GEN z)
+{ gen_sort_inplace(z,NULL,cmp_by_var,NULL); return z; }
+GEN
+vars_to_RgXV(GEN h)
+{
+  long i, l = lg(h);
+  GEN z = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++) gel(z,i) = pol_x(h[i]);
+  return z;
+}
 GEN
 gpolvar(GEN x)
 {
   long v;
   if (!x) {
-    long k = 1, n = pari_var_next();
-    GEN z = cgetg(n+1, t_VEC);
-    for (v = 0; v < n; v++)
-    {
-      entree *ep = varentries[v];
-      if (ep && ep->name[0] != '_') gel(z,k++) = (GEN)initial_value(ep);
-    }
-    if (k <= n) {
-      setlg(z,k);
-      stackdummy((pari_sp)(z+n), (pari_sp)(z+k));
-    }
-    return z;
+    GEN h = hash_values(h_polvar);
+    return vars_to_RgXV(vars_sort_inplace(h));
   }
   if (typ(x)==t_PADIC) return gcopy( gel(x,2) );
   v = gvar(x);
@@ -943,10 +1203,9 @@ gpolvar(GEN x)
 static void
 fill_hashtable_single(entree **table, entree *ep)
 {
-  long n = hashvalue(ep->name);
   EpSETSTATIC(ep);
-  ep->next = table[n]; table[n] = ep;
-  if (ep->code) ep->arity=check_proto(ep->code);
+  insertep(ep, table,  hashvalue(ep->name));
+  if (ep->code) ep->arity = check_proto(ep->code);
   ep->pvalue = NULL;
 }
 
@@ -983,8 +1242,8 @@ alias0(const char *s, const char *old)
   entree *ep, *e;
   GEN x;
 
-  ep = fetch_entry(old,strlen(old));
-  e  = fetch_entry(s,strlen(s));
+  ep = fetch_entry(old);
+  e  = fetch_entry(s);
   if (EpVALENCE(e) != EpALIAS && EpVALENCE(e) != EpNEW)
     pari_err(e_MISC,"can't replace an existing symbol by an alias");
   freeep(e);
@@ -997,22 +1256,18 @@ GEN
 ifpari(GEN g, GEN a/*closure*/, GEN b/*closure*/)
 {
   if (gequal0(g)) /* false */
-    return b?closure_evalgen(b):gnil;
+    return b? closure_evalgen(b): gnil;
   else /* true */
-    return a?closure_evalgen(a):gnil;
+    return a? closure_evalgen(a): gnil;
 }
 
 void
 ifpari_void(GEN g, GEN a/*closure*/, GEN b/*closure*/)
 {
   if (gequal0(g)) /* false */
-  {
-    if(b) closure_evalvoid(b);
-  }
+  { if (b) closure_evalvoid(b); }
   else /* true */
-  {
-    if(a) closure_evalvoid(a);
-  }
+  { if (a) closure_evalvoid(a); }
 }
 
 GEN
@@ -1053,122 +1308,16 @@ orpari(GEN a, GEN b/*closure*/)
   return gequal0(g)?gen_0:gen_1;
 }
 
-GEN gmule(GEN *x, GEN y)
-{
-  *x=gmul(*x,y);
-  return *x;
-}
-
-GEN gdive(GEN *x, GEN y)
-{
-  *x=gdiv(*x,y);
-  return *x;
-}
-
-GEN gdivente(GEN *x, GEN y)
-{
-  *x=gdivent(*x,y);
-  return *x;
-}
-
-GEN gdivrounde(GEN *x, GEN y)
-{
-  *x=gdivround(*x,y);
-  return *x;
-}
-
-GEN gmode(GEN *x, GEN y)
-{
-  *x=gmod(*x,y);
-  return *x;
-}
-
-GEN gshiftle(GEN *x, long n)
-{
-  *x=gshift(*x,n);
-  return *x;
-}
-
-GEN gshiftre(GEN *x, long n)
-{
-  *x=gshift(*x,-n);
-  return *x;
-}
-
-GEN gadde(GEN *x, GEN y)
-{
-  *x=gadd(*x,y);
-  return *x;
-}
-
-GEN gadd1e(GEN *x)
-{
-  *x=typ(*x)==t_INT?addis(*x,1):gaddgs(*x,1);
-  return *x;
-}
-
-GEN gsube(GEN *x, GEN y)
-{
-  *x=gsub(*x,y);
-  return *x;
-}
-
-GEN gsub1e(GEN *x)
-{
-  *x=typ(*x)==t_INT?subis(*x,1):gsubgs(*x,1);
-  return *x;
-}
+GEN gmule(GEN *x, GEN y) { *x = gmul(*x,y); return *x; }
+GEN gdive(GEN *x, GEN y) { *x = gdiv(*x,y); return *x; }
+GEN gdivente(GEN *x, GEN y) { *x = gdivent(*x,y); return *x; }
+GEN gdivrounde(GEN *x, GEN y) { *x = gdivround(*x,y); return *x; }
+GEN gmode(GEN *x, GEN y) { *x = gmod(*x,y); return *x; }
+GEN gshiftle(GEN *x, long n) { *x = gshift(*x,n); return *x; }
+GEN gshiftre(GEN *x, long n) { *x = gshift(*x,-n); return *x; }
+GEN gadde(GEN *x, GEN y) { *x = gadd(*x,y); return *x; }
+GEN gadd1e(GEN *x) { *x = typ(*x)==t_INT?addis(*x,1):gaddgs(*x,1); return *x; }
+GEN gsube(GEN *x, GEN y) { *x = gsub(*x,y); return *x; }
+GEN gsub1e(GEN *x) { *x = typ(*x)==t_INT?subis(*x,1):gsubgs(*x,1); return *x; }
 
 GEN gshift_right(GEN x, long n) { return gshift(x,-n); }
-
-/********************************************************************/
-/**                                                                **/
-/**            PRINT USER FUNCTION AND MEMBER FUNCTION             **/
-/**                                                                **/
-/********************************************************************/
-static int
-cmp_epname(void *E, GEN e, GEN f)
-{
-  (void)E;
-  return strcmp(((entree*)e)->name, ((entree*)f)->name);
-}
-
-void
-print_all_user_fun(int member)
-{
-  pari_sp av = avma;
-  long iL = 0, lL = 1024;
-  GEN L = cgetg(lL+1, t_VECSMALL);
-  entree *ep;
-  int i;
-  for (i = 0; i < functions_tblsz; i++)
-    for (ep = functions_hash[i]; ep; ep = ep->next)
-    {
-      const char *f;
-      int is_member;
-      if (EpVALENCE(ep) != EpVAR || typ((GEN)ep->value)!=t_CLOSURE) continue;
-      f = ep->name;
-      is_member = (f[0] == '_' && f[1] == '.');
-      if (member != is_member) continue;
-
-      if (iL >= lL)
-      {
-        GEN oL = L;
-        long j;
-        lL *= 2; L = cgetg(lL+1, t_VECSMALL);
-        for (j = 1; j <= iL; j++) gel(L,j) = gel(oL,j);
-      }
-      L[++iL] = (long)ep;
-    }
-  if (iL)
-  {
-    setlg(L, iL+1);
-    L = gen_sort(L, NULL, &cmp_epname);
-    for (i = 1; i <= iL; i++)
-    {
-      ep = (entree*)L[i];
-      pari_printf("%s =\n  %Ps\n\n", ep->name, ep->value);
-    }
-  }
-  avma = av;
-}
