@@ -27,13 +27,11 @@ enum COflags {COsafelex=1, COsafedyn=2};
  **                                                                       **
  ***************************************************************************/
 
-static GEN
-strntoGENexp(const char *str, long len)
+static char *
+translate(const char **src, char *s)
 {
-  GEN z = cgetg(1+nchar2nlong(len-1), t_STR);
-  char *s=GSTR(z);
-  const char *t=str+1;
-  while (t<=str+len)
+  const char *t = *src;
+  while (*t)
   {
     while (*t == '\\')
     {
@@ -42,7 +40,7 @@ strntoGENexp(const char *str, long len)
         case 'e':  *s='\033'; break; /* escape */
         case 'n':  *s='\n'; break;
         case 't':  *s='\t'; break;
-        default:   *s=*t; if (!*t) compile_err("unfinished string",str);
+        default:   *s=*t; if (!*t) { *src=s; return NULL; }
       }
       t++; s++;
     }
@@ -53,7 +51,32 @@ strntoGENexp(const char *str, long len)
     }
     *s++ = *t++;
   }
-  *s = '\0';
+  *s=0; *src=t; return s;
+}
+
+static void
+matchQ(const char *s, char *entry)
+{
+  if (*s != '"')
+    pari_err(e_SYNTAX,"expected character: '\"' instead of",s,entry);
+}
+
+/*  Read a "string" from src. Format then copy it, starting at s. Return
+ *  pointer to char following the end of the input string */
+char *
+pari_translate_string(const char *src, char *s, char *entry)
+{
+  matchQ(src, entry); src++; s = translate(&src, s);
+  if (!s) pari_err(e_SYNTAX,"run-away string",src,entry);
+  matchQ(src, entry); return (char*)src+1;
+}
+
+static GEN
+strntoGENexp(const char *str, long len)
+{
+  GEN z = cgetg(1+nchar2nlong(len-1), t_STR);
+  const char *t = str+1;
+  if (!translate(&t, GSTR(z))) compile_err("run-away string",str);
   return z;
 }
 
@@ -321,7 +344,7 @@ push_frame(GEN C, long lpc, long dummy)
     return;
   }
   if (lg(C)<8) while (j<lfr && frpc[j]==0) j++;
-  for(pc=0; pc<=lpc; pc++)
+  for(pc=0; pc<lpc; pc++) /* do not assume lpc was completed */
   {
     if (pc>0 && (code[pc]==OClocalvar || code[pc]==OClocalvar0))
       var_push((entree*)oper[pc],Llocal);
@@ -449,6 +472,7 @@ parseproto(char const **q, char *c, const char *str)
     break;
   case 'C':
   case 'p':
+  case 'b':
   case 'P':
   case 'f':
     *c=*p;
@@ -499,15 +523,16 @@ get_ret_type(const char **p, long arity, Gtype *t, long *flag)
 {
   *flag = 0;
   if (**p == 'v') { (*p)++; *t=Gvoid; return OCcallvoid; }
-  else if (**p == 'i') { (*p)++; *t=Gsmall; return OCcallint; }
-  else if (**p == 'l') { (*p)++; *t=Gsmall; return OCcalllong; }
+  else if (**p == 'i') { (*p)++; *t=Gsmall;  return OCcallint; }
+  else if (**p == 'l') { (*p)++; *t=Gsmall;  return OCcalllong; }
+  else if (**p == 'u') { (*p)++; *t=Gusmall; return OCcalllong; }
   else if (**p == 'm') { (*p)++; *flag = FLnocopy; }
   *t=Ggen; return arity==2?OCcallgen2:OCcallgen;
 }
 
 /*supported types:
- * type: Gsmall, Ggen, Gvoid, Gvec, Gclosure
- * mode: Gsmall, Ggen, Gvar, Gvoid
+ * type: Gusmall, Gsmall, Ggen, Gvoid, Gvec, Gclosure
+ * mode: Gusmall, Gsmall, Ggen, Gvar, Gvoid
  */
 static void
 compilecast_loc(int type, int mode, const char *loc)
@@ -515,13 +540,21 @@ compilecast_loc(int type, int mode, const char *loc)
   if (type==mode) return;
   switch (mode)
   {
+  case Gusmall:
+    if (type==Ggen)        op_push_loc(OCitou,-1,loc);
+    else if (type==Gvoid)  op_push_loc(OCpushlong,0,loc);
+    else if (type!=Gsmall)
+      compile_err("this should be a small integer >=0",loc);
+    break;
   case Gsmall:
     if (type==Ggen)        op_push_loc(OCitos,-1,loc);
     else if (type==Gvoid)  op_push_loc(OCpushlong,0,loc);
-    else compile_err("this should be a small integer",loc);
+    else if (type!=Gusmall)
+      compile_err("this should be a small integer",loc);
     break;
   case Ggen:
     if (type==Gsmall)      op_push_loc(OCstoi,0,loc);
+    else if (type==Gusmall)op_push_loc(OCutoi,0,loc);
     else if (type==Gvoid)  op_push_loc(OCpushgnil,0,loc);
     break;
   case Gvoid:
@@ -539,14 +572,24 @@ compilecast_loc(int type, int mode, const char *loc)
 static void
 compilecast(long n, int type, int mode) { compilecast_loc(type, mode, tree[n].str); }
 
+static entree *
+fetch_member_raw(const char *s, long len)
+{
+  pari_sp av = avma;
+  char *t = stack_malloc(len+2);
+  entree *ep;
+  t[0] = '_'; strncpy(t+1, s, len); t[++len] = 0; /* prepend '_' */
+  ep = fetch_entry_raw(t, len);
+  avma = av; return ep;
+}
 static entree *
 getfunc(long n)
 {
   long x=tree[n].x;
-  if (tree[x].x==CSTmember)
-    return do_alias(fetch_member(tree[x].str, tree[x].len));
+  if (tree[x].x==CSTmember) /* str-1 points to '.' */
+    return do_alias(fetch_member_raw(tree[x].str - 1, tree[x].len + 1));
   else
-    return do_alias(fetch_entry(tree[x].str, tree[x].len));
+    return do_alias(fetch_entry_raw(tree[x].str, tree[x].len));
 }
 
 static entree *
@@ -673,17 +716,18 @@ matindex_type(long n)
   long fxx = tree[tree[x].x].f, fxy = tree[tree[x].y].f;
   if (y==-1)
   {
-    if (fxy!=Fnoarg) return MAT_range;
-    if (fxx==Fnoarg) compile_err("missing index",tree[n].str);
+    if (fxy!=Fnorange) return MAT_range;
+    if (fxx==Fnorange) compile_err("missing index",tree[n].str);
     return VEC_std;
   }
   else
   {
     long fyx = tree[tree[y].x].f, fyy = tree[tree[y].y].f;
-    if (fxy!=Fnoarg || fyy!=Fnoarg) return MAT_range;
-    if (fxx==Fnoarg && fyx==Fnoarg) compile_err("missing index",tree[n].str);
-    if (fxx==Fnoarg) return MAT_column;
-    if (fyx==Fnoarg) return MAT_line;
+    if (fxy!=Fnorange || fyy!=Fnorange) return MAT_range;
+    if (fxx==Fnorange && fyx==Fnorange)
+      compile_err("missing index",tree[n].str);
+    if (fxx==Fnorange) return MAT_column;
+    if (fyx==Fnorange) return MAT_line;
     return MAT_std;
   }
 }
@@ -816,6 +860,8 @@ compilesmall(long n, long x, long mode)
     op_push(OCpushstoi, x, n);
   else
   {
+    if (mode==Gusmall && x < 0)
+      compile_err("this should be a small integer >=0",tree[n].str);
     op_push(OCpushlong, x, n);
     compilecast(n,Gsmall,mode);
   }
@@ -921,7 +967,7 @@ compilelambda(long n, long y, GEN vep, struct codepos *pos)
 }
 
 static void
-compilecall(long n, int mode)
+compilecall(long n, int mode, entree *ep)
 {
   pari_sp ltop=avma;
   long j;
@@ -929,14 +975,26 @@ compilecall(long n, int mode)
   long y=tree[n].y;
   GEN arg=listtogen(y,Flistarg);
   long nb=lg(arg)-1;
-  long lnc=first_safe_arg(arg, COsafelex);
+  long lnc=first_safe_arg(arg, COsafelex|COsafedyn);
+  long lnl=first_safe_arg(arg, COsafelex);
+  long fl = lnl==0? (lnc==0? FLnocopy: FLnocopylex): 0;
+  if (ep==NULL)
+    compilenode(x, Ggen, fl);
+  else
+  {
+    long vn=getmvar(ep);
+    if (vn)
+      op_push(OCpushlex,vn,n);
+    else
+      op_push(OCpushdyn,(long)ep,n);
+  }
   for (j=1;j<=nb;j++)
   {
     long x = tree[arg[j]].x, f = tree[arg[j]].f;
     if (f==Fseq)
       compile_err("unexpected ';'", tree[x].str+tree[x].len);
     else if (f!=Fnoarg)
-      compilenode(arg[j], Ggen,j>=lnc?FLnocopylex:0);
+      compilenode(arg[j], Ggen,j>=lnl?FLnocopylex:0);
     else
       op_push(OCpushlong,0,n);
   }
@@ -945,18 +1003,6 @@ compilecall(long n, int mode)
   avma=ltop;
 }
 
-static void
-compileuserfunc(entree *ep, long n, int mode)
-{
-  long vn=getmvar(ep);
-  if (tree[n].x<OPnboperator) compile_err("operator unknown",tree[n].str);
-  if (vn)
-    op_push(OCpushlex,vn,n);
-  else
-    op_push(OCpushdyn,(long)ep,n);
-  compilecall(n, mode);
-}
-
 static GEN
 compilefuncinline(long n, long c, long a, long flag, long isif, long lev, long *ev)
 {
@@ -1071,7 +1117,7 @@ compilemy(GEN arg, const char *str, int inl)
         GEN vars = listtogen(tree[x].x,Fmatrixelts);
         long nv = lg(vars)-1;
         compilenode(tree[a].y,Ggen,FLnocopy);
-        op_push(OCdup,nv-1,x);
+        if (nv > 1) op_push(OCdup,nv-1,x);
         for (j=1; j<=nv; j++)
         {
           long v = detag(vars[j]);
@@ -1124,7 +1170,7 @@ compilelocal(GEN arg)
         GEN vars = listtogen(tree[x].x,Fmatrixelts);
         long nv = lg(vars)-1;
         compilenode(tree[a].y,Ggen,FLnocopy);
-        op_push(OCdup,nv-1,x);
+        if (nv > 1) op_push(OCdup,nv-1,x);
         for (j=1; j<=nv; j++)
         {
           long v = detag(vars[j]);
@@ -1254,7 +1300,7 @@ compilefunc(entree *ep, long n, int mode, long flag)
     avma=ltop;
     return;
   }
-  else if (is_func_named(ep,"O") || (compatible==OLDALL && is_func_named(ep,"o")))
+  else if (is_func_named(ep,"O"))
   {
     if (nb!=1)
       compile_err("wrong number of arguments", tree[n].str+tree[n].len-1);
@@ -1318,10 +1364,11 @@ compilefunc(entree *ep, long n, int mode, long flag)
           break;
         case 'W':
           {
-            entree *ep = getlvalue(arg[j]);
+            long a = arg[j];
+            entree *ep = getlvalue(a);
             long vn = getmvar(ep);
-            if (vn) op_push(OCcowvarlex,vn,n);
-            else op_push(OCcowvardyn,(long)ep,n);
+            if (vn) op_push(OCcowvarlex, vn, a);
+            else op_push(OCcowvardyn, (long)ep, a);
             compilenode(arg[j++],Ggen,FLnocopy);
             break;
           }
@@ -1347,9 +1394,12 @@ compilefunc(entree *ep, long n, int mode, long flag)
             }
             break;
           }
-        case 'P': case 'L': /*Fall through*/
+        case 'P': case 'L':
           compilenode(arg[j++],Gsmall,0);
           break;
+        case 'U':
+          compilenode(arg[j++],Gusmall,0);
+          break;
         case 'n':
           compilenode(arg[j++],Gvar,0);
           break;
@@ -1375,9 +1425,9 @@ compilefunc(entree *ep, long n, int mode, long flag)
             }
             else
             {
-              compilenewptr(vn,ep,n);
+              compilenewptr(vn, ep, a);
               compilelvalue(a);
-              op_push(OCpushptr, 0,n);
+              op_push(OCpushptr, 0, a);
             }
             nbpointers++;
             break;
@@ -1456,6 +1506,9 @@ compilefunc(entree *ep, long n, int mode, long flag)
         case 'p':
           op_push(OCprecreal,0,n);
           break;
+        case 'b':
+          op_push(OCbitprecreal,0,n);
+          break;
         case 'P':
           op_push(OCprecdl,0,n);
           break;
@@ -1507,6 +1560,9 @@ compilefunc(entree *ep, long n, int mode, long flag)
         case 'M':
           op_push(OCpushlong,strtol(q+1,NULL,10),n);
           break;
+        case 'U':
+          op_push(OCpushlong,(long)strtoul(q+1,NULL,10),n);
+          break;
         case 'r':
         case 's':
           str_defproto(p, q, tree[n].str);
@@ -1579,15 +1635,30 @@ compilefunc(entree *ep, long n, int mode, long flag)
   avma=ltop;
 }
 
+static void
+genclosurectx(const char *loc, long nbdata)
+{
+  long i;
+  GEN vep = cgetg(nbdata+1,t_VECSMALL);
+  for(i = 1; i <= nbdata; i++)
+  {
+    vep[i] = 0;
+    op_push_loc(OCpushlex,-i,loc);
+  }
+  frame_push(vep);
+}
+
 static GEN
 genclosure(entree *ep, const char *loc, long  nbdata, int check)
 {
+  pari_sp av = avma;
   struct codepos pos;
-  long i, nb=0;
+  long nb=0;
   const char *code=ep->code,*p,*q;
   char c;
+  GEN text;
   long index=ep->arity;
-  long arity=0, maskarg=0, maskarg0=0, stop=0;
+  long arity=0, maskarg=0, maskarg0=0, stop=0, dovararg=0;
   PPproto mod;
   Gtype ret_typ;
   long ret_flag;
@@ -1628,8 +1699,8 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
   dbgstart = loc;
   if (nbdata > arity)
     pari_err(e_MISC,"too many parameters for closure `%s'", ep->name);
-  for(i=1; i<= nbdata; i++)
-    op_push_loc(OCpushgen,data_push(NULL),loc);
+  if (nbdata) genclosurectx(loc, nbdata);
+  text = strtoGENstr(ep->name);
   arity -= nbdata;
   if (maskarg)  op_push_loc(OCcheckargs,maskarg,loc);
   if (maskarg0) op_push_loc(OCcheckargs0,maskarg0,loc);
@@ -1644,6 +1715,9 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
       case 'p':
         op_push_loc(OCprecreal,0,loc);
         break;
+      case 'b':
+        op_push_loc(OCbitprecreal,0,loc);
+        break;
       case 'P':
         op_push_loc(OCprecdl,0,loc);
         break;
@@ -1675,6 +1749,9 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
       case 'L':
         op_push_loc(OCitos,-index,loc);
         break;
+      case 'U':
+        op_push_loc(OCitou,-index,loc);
+        break;
       case 'n':
         op_push_loc(OCvarn,-index,loc);
         break;
@@ -1726,6 +1803,9 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
         op_push_loc(OCpushlong,strtol(q+1,NULL,10),loc);
         op_push_loc(OCdefaultlong,-index,loc);
         break;
+      case 'U':
+        op_push_loc(OCpushlong,(long)strtoul(q+1,NULL,10),loc);
+        op_push_loc(OCdefaultulong,-index,loc);
       case 'r':
       case 's':
         str_defproto(p, q, loc);
@@ -1740,8 +1820,10 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
     case PPstar:
       switch(c)
       {
-      case 'E':
       case 's':
+        dovararg = 1;
+        break;
+      case 'E':
         return NULL;
       default:
         pari_err(e_MISC,"Unknown prototype code `%c*' for `%s'",c,ep->name);
@@ -1756,7 +1838,8 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
   op_push_loc(ret_op, (long) ep, loc);
   if (ret_flag==FLnocopy) op_push_loc(OCcopy,0,loc);
   compilecast_loc(ret_typ, Ggen, loc);
-  return getfunction(&pos,nb+arity,0,strtoGENstr(ep->name),0);
+  if (dovararg) nb|=VARARGBITS;
+  return gerepilecopy(av, getfunction(&pos,nb+arity,nbdata,text,0));
 }
 
 GEN
@@ -1766,7 +1849,7 @@ snm_closure(entree *ep, GEN data)
   long n = data ? lg(data)-1: 0;
   GEN C = genclosure(ep,ep->name,n,0);
   for(i=1; i<=n; i++)
-    gmael(C,4,i) = gel(data,i);
+    gmael(C,7,i) = gel(data,i);
   return C;
 }
 
@@ -1776,7 +1859,7 @@ strtoclosure(const char *s, long n,  ...)
   pari_sp av = avma;
   entree *ep = is_entry(s);
   GEN C;
-  if (!ep) pari_err(e_MISC,"no function named \"%s\"",s);
+  if (!ep) pari_err(e_NOTFUNC, strtoGENstr(s));
   ep = do_alias(ep);
   if ((!EpSTATIC(ep) && EpVALENCE(ep)!=EpINSTALL) || !ep->value)
     pari_err(e_MISC,"not a built-in/install'ed function: \"%s\"",s);
@@ -1788,7 +1871,7 @@ strtoclosure(const char *s, long n,  ...)
     long i;
     va_start(ap,n);
     for(i=1; i<=n; i++)
-      gmael(C,4,i) = va_arg(ap, GEN);
+      gmael(C,7,i) = va_arg(ap, GEN);
     va_end(ap);
   }
   return gerepilecopy(av, C);
@@ -1800,6 +1883,22 @@ strtofunction(const char *s)
   return strtoclosure(s, 0);
 }
 
+GEN
+call0(GEN fun, GEN args)
+{
+  if (!is_vec_t(typ(args))) pari_err_TYPE("call",args);
+  switch(typ(fun))
+  {
+    case t_STR:
+      fun = strtofunction(GSTR(fun));
+    case t_CLOSURE: /* fall through */
+      return closure_callgenvec(fun, args);
+    default:
+      pari_err_TYPE("call", fun);
+      return NULL; /* NOT REACHED */
+  }
+}
+
 static void
 closurefunc(entree *ep, long n, long mode)
 {
@@ -1857,9 +1956,9 @@ compilenode(long n, int mode, long flag)
     if (tree[x].f==Fvec && tree[x].x>=0)
     {
       GEN vars = listtogen(tree[x].x,Fmatrixelts);
-      long i, l = lg(vars)-1;
+      long i, l = lg(vars)-1, d = mode==Gvoid? l-1: l;
       compilenode(y,Ggen,mode==Gvoid?FLnocopy:flag&FLsurvive);
-      op_push(OCdup,mode==Gvoid?l-1:l,x);
+      if (d) op_push(OCdup, d, x);
       for(i=1; i<=l; i++)
       {
         long a = detag(vars[i]);
@@ -1922,8 +2021,8 @@ compilenode(long n, int mode, long flag)
         op_push(OCpushgen,  data_push(strntoGENexp(tree[n].str,tree[n].len)),n);
         break;
       case CSTquote:
-        {
-          entree *ep = fetch_entry(tree[n].str+1,tree[n].len-1);
+        { /* skip ' */
+          entree *ep = fetch_entry_raw(tree[n].str+1,tree[n].len-1);
           if (EpSTATIC(ep)) compile_varerr(tree[n].str+1);
           op_push(OCpushvar, (long)ep,n);
           compilecast(n,Ggen, mode);
@@ -1973,21 +2072,24 @@ compilenode(long n, int mode, long flag)
     {
       entree *ep=getfunc(n);
       if (EpVALENCE(ep)==EpVAR || EpVALENCE(ep)==EpNEW)
-        compileuserfunc(ep,n,mode);
+      {
+        if (tree[n].x<OPnboperator) /* should not happen */
+          compile_err("operator unknown",tree[n].str);
+        compilecall(n,mode,ep);
+      }
       else
         compilefunc(ep,n,mode,flag);
       return;
     }
   case Fcall:
-    compilenode(x,Ggen,0);
-    compilecall(n,mode);
+    compilecall(n,mode,NULL);
     return;
   case Flambda:
     {
       pari_sp ltop=avma;
       struct codepos pos;
       GEN arg=listtogen(x,Flistarg);
-      long nb, lgarg, nbmvar, gap;
+      long nb, lgarg, nbmvar, dovararg=0, gap;
       long strict = GP_DATA->strictargs;
       GEN vep = cgetg_copy(arg, &lgarg);
       GEN text=cgetg(3,t_VEC);
@@ -2004,7 +2106,13 @@ compilenode(long n, int mode, long flag)
         for(i=1;i<=nb;i++)
         {
           long a=arg[i];
-          vep[i]=(long)getvar(tree[a].f==Fassign?tree[a].x:a);
+          if (i==nb && tree[a].f==Fvararg)
+          {
+            dovararg=1;
+            vep[i]=(long)getvar(tree[a].x);
+          }
+          else
+            vep[i]=(long)getvar(tree[a].f==Fassign?tree[a].x:a);
           var_push(NULL,Lmy);
         }
         checkdups(arg,vep);
@@ -2037,6 +2145,7 @@ compilenode(long n, int mode, long flag)
         compilenode(y,Ggen,FLsurvive|FLreturn);
       else
         compilecast(n,Gvoid,Ggen);
+      if (dovararg) nb|=VARARGBITS;
       op_push(OCpushgen, data_push(getfunction(&pos,nb,nbmvar,text,gap)),n);
       if (nbmvar) op_push(OCsaveframe,!!(flag&FLsurvive),n);
       compilecast(n, Gclosure, mode);
@@ -2049,6 +2158,10 @@ compilenode(long n, int mode, long flag)
   case Fnoarg:
     compilecast(n,Gvoid,mode);
     return;
+  case Fnorange:
+    op_push(OCpushlong,LONG_MAX,n);
+    compilecast(n,Gsmall,mode);
+    return;
   default:
     pari_err_BUG("compilenode");
   }
@@ -2072,7 +2185,9 @@ closure_deriv(GEN G)
   struct codepos pos;
   const char *code;
   GEN text;
-  long arity=G[1];
+  long arity=closure_arity(G);
+  if (arity==0 || closure_is_variadic(G))
+    pari_err_TYPE("derivfun",G);
   if (typ(gel(G,6))==t_STR)
   {
     code = GSTR(gel(G,6));
@@ -2198,6 +2313,7 @@ optimizefunc(entree *ep, long n)
         case 'n':
         case 'M':
         case 'L':
+        case 'U':
         case 'P':
           optimizenode(arg[j]);
           fl&=tree[arg[j++]].flags;
@@ -2352,6 +2468,7 @@ optimizenode(long n)
     tree[n].flags=0;
     break;
   case Fnoarg:
+  case Fnorange:
   case Fsmall:
   case Fconst:
   case Fentry:
@@ -2366,6 +2483,9 @@ optimizenode(long n)
   case Frefarg:
     compile_err("unexpected character '&'",tree[n].str);
     return;
+  case Fvararg:
+    compile_err("unexpected characters '..'",tree[n].str);
+    return;
   case Ffunction:
     {
       entree *ep=getfunc(n);
