@@ -13,8 +13,10 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
-/* Simple-minded parsing utilities. These are forbidden to use the GP stack
- * which may not exist at this point [e.g upon GP initialization]  */
+#ifdef _WIN32
+#  include "../systems/mingw/mingw.h"
+#endif
+
 
 /* Return all chars, up to next separator
  * [as strtok but must handle verbatim character string] */
@@ -72,6 +74,9 @@ my_int(char *s)
       case 'k': case 'K': n = safe_mul(n,1000UL);       p++; break;
       case 'm': case 'M': n = safe_mul(n,1000000UL);    p++; break;
       case 'g': case 'G': n = safe_mul(n,1000000000UL); p++; break;
+#ifdef LONG_IS_64BIT
+      case 't': case 'T': n = safe_mul(n,1000000000000UL); p++; break;
+#endif
     }
     if (!n) pari_err(e_SYNTAX,"integer too large",s,s);
   }
@@ -105,6 +110,63 @@ get_uint(const char *s)
   u = my_int(p); avma = av; return u;
 }
 
+#if defined(__EMX__) || defined(_WIN32) || defined(__CYGWIN32__)
+#  define PATH_SEPARATOR ';' /* beware DOSish 'C:' disk drives */
+#else
+#  define PATH_SEPARATOR ':'
+#endif
+
+static const char *
+pari_default_path(void) {
+#if PATH_SEPARATOR == ';'
+  return ".;C:;C:/gp";
+#elif defined(UNIX)
+  return ".:~:~/gp";
+#else
+  return ".";
+#endif
+}
+
+void
+delete_dirs(gp_path *p)
+{
+  char **v = p->dirs, **dirs;
+  if (v)
+  {
+    p->dirs = NULL; /* in case of error */
+    for (dirs = v; *dirs; dirs++) pari_free(*dirs);
+    pari_free(v);
+  }
+}
+
+void
+gp_expand_path(gp_path *p)
+{
+  char **dirs, *s, *v = p->PATH;
+  int i, n = 0;
+
+  delete_dirs(p);
+  v = pari_strdup(v);
+  for (s=v; *s; s++)
+    if (*s == PATH_SEPARATOR) {
+      *s = 0;
+      if (s == v || s[-1] != 0) n++; /* ignore empty path components */
+    }
+  dirs = (char**) pari_malloc((n + 2)*sizeof(char *));
+
+  for (s=v, i=0; i<=n; i++)
+  {
+    char *end, *f;
+    while (!*s) s++; /* skip empty path components */
+    f = end = s + strlen(s);
+    while (f > s && *--f == '/') *f = 0; /* skip trailing '/' */
+    dirs[i] = path_expand(s);
+    s = end + 1; /* next path component */
+  }
+  pari_free((void*)v);
+  dirs[i] = NULL; p->dirs = dirs;
+}
+
 /********************************************************************/
 /*                                                                  */
 /*                            DEFAULTS                              */
@@ -121,7 +183,8 @@ long
 setrealprecision(long n, long *prec)
 {
   GP_DATA->fmt->sigd = n;
-  *prec = precreal = ndec2prec(n);
+  *prec = ndec2prec(n);
+  precreal = prec2nbits(*prec);
   return n;
 }
 
@@ -200,18 +263,19 @@ sd_realprecision(const char *v, long flag)
   pariout_t *fmt = GP_DATA->fmt;
   if (v)
   {
-    ulong newnb = fmt->sigd, prec;
+    ulong newnb = fmt->sigd;
+    long prec;
     sd_ulong_init(v, "realprecision", &newnb, 1, prec2ndec(LGBITS));
     if (fmt->sigd == (long)newnb) return gnil;
     if (fmt->sigd >= 0) fmt->sigd = newnb;
-    prec = (ulong)ndec2prec(newnb);
+    prec = ndec2nbits(newnb);
     if (prec == precreal) return gnil;
     precreal = prec;
   }
-  if (flag == d_RETURN) return stoi(prec2ndec(precreal));
+  if (flag == d_RETURN) return stoi(nbits2ndec(precreal));
   if (flag == d_ACKNOWLEDGE)
   {
-    long n = prec2ndec(precreal);
+    long n = nbits2ndec(precreal);
     pari_printf("   realprecision = %ld significant digits", n);
     if (fmt->sigd < 0)
       pari_puts(" (all digits displayed)");
@@ -222,6 +286,34 @@ sd_realprecision(const char *v, long flag)
   return gnil;
 }
 
+GEN
+sd_realbitprecision(const char *v, long flag)
+{
+  pariout_t *fmt = GP_DATA->fmt;
+  if (v)
+  {
+    ulong newnb = precreal;
+    long n;
+    sd_ulong_init(v, "realbitprecision", &newnb, 1, prec2nbits(LGBITS));
+    if ((long)newnb == precreal) return gnil;
+    n = nbits2ndec(newnb);
+    if (!n) n = 1;
+    if (fmt->sigd >= 0) fmt->sigd = n;
+    precreal = (long) newnb;
+  }
+  if (flag == d_RETURN) return stoi(precreal);
+  if (flag == d_ACKNOWLEDGE)
+  {
+    pari_printf("   realbitprecision = %ld significant bits", precreal);
+    if (fmt->sigd < 0)
+      pari_puts(" (all digits displayed)");
+    else
+      pari_printf(" (%ld decimal digits displayed)", fmt->sigd);
+    pari_putc('\n');
+  }
+  return gnil;
+}
+
 GEN
 sd_seriesprecision(const char *v, long flag)
 {
@@ -280,7 +372,7 @@ sd_colors(const char *v, long flag)
       v = "9, 13, 11, 15, 14, 10, 11";
     if (l <= 6 && strncmp(v, "boldfg", l) == 0)        /* Good for darkbg consoles */
       v = "[1,,1], [5,,1], [3,,1], [7,,1], [6,,1], , [2,,1]";
-    v0 = s = filtre(v, 0);
+    v0 = s = gp_filter(v);
     for (c=c_ERR; c < c_LAST; c++)
       gp_colors[c] = gp_get_color(&s);
     pari_free(v0);
@@ -352,16 +444,12 @@ sd_compatible(const char *v, long flag)
 {
   const char *msg[] = {
     "(no backward compatibility)",
-    "(warn when using obsolete functions)",
-    "(use old functions, don't ignore case)",
-    "(use old functions, ignore case)", NULL
+    "(no backward compatibility)",
+    "(no backward compatibility)",
+    "(no backward compatibility)", NULL
   };
-  ulong old = compatible;
-  GEN r = sd_ulong(v,flag,"compatible",&compatible, 0,3,msg);
-
-  if (old != compatible && flag != d_INITRC && gp_init_functions())
-    pari_warn(warner,"user functions re-initialized");
-  return r;
+  ulong junk = 0;
+  return sd_ulong(v,flag,"compatible",&junk, 0,3,msg);
 }
 
 GEN
@@ -447,9 +535,7 @@ static FILE *
 open_logfile(const char *s) {
   FILE *log = fopen(s, "a");
   if (!log) pari_err_FILE("logfile",s);
-#ifndef WINCE
   setbuf(log,(char *)NULL);
-#endif
   return log;
 }
 
@@ -521,19 +607,43 @@ sd_output(const char *v, long flag)
 }
 
 GEN
-sd_parisize(const char *v, long flag)
+sd_parisizemax(const char *v, long flag)
 {
-  ulong size = top - bot, n = size;
-  GEN r = sd_ulong(v,flag,"parisize",&n, 10000,LONG_MAX,NULL);
+  ulong size = pari_mainstack->vsize, n = size;
+  GEN r = sd_ulong(v,flag,"parisizemax",&n, 0,LONG_MAX,NULL);
   if (n != size) {
     if (flag == d_INITRC)
-      pari_init_stack(n, size);
+      paristack_setsize(pari_mainstack->rsize, n);
     else
-      allocatemem(n);
+      parivstack_resize(n);
   }
   return r;
 }
 
+GEN
+sd_parisize(const char *v, long flag)
+{
+  ulong rsize = pari_mainstack->rsize, n = rsize;
+  GEN r = sd_ulong(v,flag,"parisize",&n, 10000,LONG_MAX,NULL);
+  if (n != rsize) {
+    if (flag == d_INITRC)
+      paristack_setsize(n, pari_mainstack->vsize);
+    else
+      paristack_newrsize(n);
+  }
+  return r;
+}
+
+GEN
+sd_threadsizemax(const char *v, long flag)
+{
+  ulong size = GP_DATA->threadsizemax, n = size;
+  GEN r = sd_ulong(v,flag,"threadsizemax",&n, 0,LONG_MAX,NULL);
+  if (n != size)
+    GP_DATA->threadsizemax = n;
+  return r;
+}
+
 GEN
 sd_threadsize(const char *v, long flag)
 {
@@ -689,23 +799,21 @@ sd_prettyprinter(const char *v, long flag)
   return gnil;
 }
 
-/* compare entrees s1 s2 according to the associated function name */
+/* compare entrees s1 s2 according to the attached function name */
 static int
 compare_name(const void *s1, const void *s2) {
   entree *e1 = *(entree**)s1, *e2 = *(entree**)s2;
   return strcmp(e1->name, e2->name);
 }
-/* return all entries with class '16' */
 static void
 defaults_list(pari_stack *s)
 {
   entree *ep;
   long i;
   for (i = 0; i < functions_tblsz; i++)
-    for (ep = defaults_hash[i]; ep; ep = ep->next)
-      if (ep->menu == 16) pari_stack_pushp(s, ep);
+    for (ep = defaults_hash[i]; ep; ep = ep->next) pari_stack_pushp(s, ep);
 }
-/* ep associated to function f of arity 2. Call f(v,flag) */
+/* ep attached to function f of arity 2. Call f(v,flag) */
 static GEN
 call_f2(entree *ep, const char *v, long flag)
 { return ((GEN (*)(const char*,long))ep->value)(v, flag); }
@@ -725,7 +833,7 @@ setdefault(const char *s, const char *v, long flag)
     pari_stack_delete(&st);
     return gnil;
   }
-  ep = is_entry_intern(s, defaults_hash, NULL);
+  ep = pari_is_default(s);
   if (!ep)
   {
     pari_err(e_MISC,"unknown default: %s",s);
@@ -733,9 +841,6 @@ setdefault(const char *s, const char *v, long flag)
   }
   return call_f2(ep, v, flag);
 }
-int
-pari_is_default(const char *s)
-{ return !!is_entry_intern(s, defaults_hash, NULL); }
 
 GEN
 default0(const char *a, const char *b) { return setdefault(a,b, b? d_SILENT: d_RETURN); }
@@ -774,6 +879,44 @@ init_pp(gp_data *D)
   p->file = NULL;
 }
 
+static char *
+init_help(void)
+{
+  char *h = os_getenv("GPHELP");
+# ifdef GPHELP
+  if (!h) h = (char*)GPHELP;
+# endif
+#ifdef _WIN32
+  win32_set_pdf_viewer();
+#endif
+  if (h) h = pari_strdup(h);
+  return h;
+}
+
+static void
+init_graphs(gp_data *D)
+{
+  const char *cols[] = { "",
+    "white","black","blue","violetred","red","green","grey","gainsboro"
+  };
+  const long N = 8;
+  GEN c = cgetalloc(t_VECSMALL, 3), s;
+  long i;
+  c[1] = 4;
+  c[2] = 5;
+  D->graphcolors = c;
+  c = (GEN)pari_malloc((N+1 + 4*N)*sizeof(long));
+  c[0] = evaltyp(t_VEC)|evallg(N+1);
+  for (i = 1, s = c+N+1; i <= N; i++, s += 4)
+  {
+    GEN lp = s;
+    lp[0] = evaltyp(t_STR)|evallg(4);
+    strcpy(GSTR(lp), cols[i]);
+    gel(c,i) = lp;
+  }
+  D->colormap = c;
+}
+
 gp_data *
 default_gp_data(void)
 {
@@ -809,5 +952,12 @@ default_gp_data(void)
   init_path(D->path, pari_default_path());
   init_path(D->sopath, "");
   init_pp(D);
+  init_graphs(D);
+  D->prompt_comment = (char*)"comment> ";
+  D->prompt = pari_strdup("? ");
+  D->prompt_cont = pari_strdup("");
+  D->help = init_help();
+  D->readline_state = DO_ARGS_COMPLETE;
+  D->histfile = NULL;
   return D;
 }
