@@ -92,15 +92,15 @@ enum {PUSH_VAL = 0, COPY_VAL = 1, DEFAULT_VAL = 2};
 /* ep->args is the stack of old values (INITIAL if initial value, from
  * installep) */
 typedef struct var_cell {
-  struct var_cell *prev; /* cell associated to previous value on stack */
+  struct var_cell *prev; /* cell attached to previous value on stack */
   GEN value; /* last value (not including current one, in ep->value) */
   char flag; /* status of _current_ ep->value: PUSH or COPY ? */
-  long valence; /* valence of entree* associated to 'value', to be restored
+  long valence; /* valence of entree* attached to 'value', to be restored
                     * by pop_val */
 } var_cell;
 #define INITIAL NULL
 
-/* Push x on value stack associated to ep. */
+/* Push x on value stack attached to ep. */
 static void
 new_val_cell(entree *ep, GEN x, char flag)
 {
@@ -124,13 +124,15 @@ static void
 pop_val(entree *ep)
 {
   var_cell *v = (var_cell*) ep->pvalue;
-
-  if (v == INITIAL) return;
-  if (v->flag == COPY_VAL) gunclone_deep((GEN)ep->value);
-  ep->value  = v->value;
-  ep->pvalue = (char*) v->prev;
-  ep->valence=v->valence;
-  pari_free((void*)v);
+  if (v != INITIAL)
+  {
+    GEN old_val = (GEN) ep->value; /* protect against SIGINT */
+    ep->value  = v->value;
+    if (v->flag == COPY_VAL) gunclone_deep(old_val);
+    ep->pvalue = (char*) v->prev;
+    ep->valence=v->valence;
+    pari_free((void*)v);
+  }
 }
 
 void
@@ -213,21 +215,38 @@ copyvalue(entree *ep)
 INLINE void
 err_var(GEN x) { pari_err_TYPE("evaluator [variable name expected]", x); }
 
+enum chk_VALUE { chk_ERROR, chk_NOCREATE, chk_CREATE };
+
 INLINE void
-checkvalue(entree *ep)
+checkvalue(entree *ep, enum chk_VALUE flag)
 {
   if (MT_IS_THREAD)
     pari_err(e_MISC,"mt: global variable not supported: %s",ep->name);
   if (ep->valence==EpNEW)
-  {
-    pari_var_create(ep);
-    ep->valence = EpVAR;
-    ep->value = initial_value(ep);
-  }
+    switch(flag)
+    {
+      case chk_ERROR:
+        /* Do nothing until we can report a meaningful error message
+           The extra variable will be cleaned-up anyway */
+      case chk_CREATE:
+        pari_var_create(ep);
+        ep->valence = EpVAR;
+        ep->value = initial_value(ep);
+        break;
+      case chk_NOCREATE:
+        break;
+    }
   else if (ep->valence!=EpVAR)
     err_var(strtoGENstr(ep->name));
 }
 
+INLINE GEN
+checkvalueptr(entree *ep)
+{
+  checkvalue(ep, chk_NOCREATE);
+  return ep->valence==EpNEW? gen_0: (GEN)ep->value;
+}
+
 /* make GP variables safe for avma = top */
 static void
 lvar_make_safe(void)
@@ -266,7 +285,7 @@ GEN*
 safelistel(GEN x, long l)
 {
   GEN d;
-  if (typ(x)!=t_LIST)
+  if (typ(x)!=t_LIST || list_typ(x)!=t_LIST_RAW)
     pari_err_TYPE("safelistel",x);
   d = list_data(x);
   check_array_index(l, lg(d));
@@ -328,8 +347,9 @@ change_compo(matcomp *c, GEN res)
     if (lg(res) != lg(p)) pari_err_DIM("matrix row assignment");
     for (i=1; i<lg(p); i++)
     {
-      GEN p1 = gcoeff(p,c->full_row,i); if (isclone(p1)) gunclone_deep(p1);
+      GEN p1 = gcoeff(p,c->full_row,i); /* Protect against SIGINT */
       gcoeff(p,c->full_row,i) = gclone(gel(res,i));
+      if (isclone(p1)) gunclone_deep(p1);
     }
     return;
   }
@@ -363,12 +383,12 @@ struct trace
 };
 
 static THREAD long sp, rp, dbg_level;
-static THREAD long *st;
+static THREAD long *st, *precs;
 static THREAD gp_pointer *ptrs;
 static THREAD entree **lvars;
 static THREAD struct var_lex *var;
 static THREAD struct trace *trace;
-static THREAD pari_stack s_st, s_ptrs, s_var, s_lvars, s_trace;
+static THREAD pari_stack s_st, s_ptrs, s_var, s_lvars, s_trace, s_prec;
 
 static void
 changelex(long vn, GEN x)
@@ -400,9 +420,10 @@ pushlex(long vn, GEN x)
 }
 
 INLINE void
-freelex(long vn)
+freelex(void)
 {
-  struct var_lex *v=var+s_var.n+vn;
+  struct var_lex *v=var+s_var.n-1;
+  s_var.n--;
   if (v->flag == COPY_VAL) gunclone_deep(v->value);
 }
 
@@ -411,11 +432,9 @@ restore_vars(long nbmvar, long nblvar)
 {
   long j;
   for(j=1;j<=nbmvar;j++)
-    freelex(-j);
-  s_var.n-=nbmvar;
+    freelex();
   for(j=1;j<=nblvar;j++)
-    pop_val(lvars[s_lvars.n-j]);
-  s_lvars.n-=nblvar;
+    { s_lvars.n--; pop_val(lvars[s_lvars.n]); }
 }
 
 INLINE void
@@ -472,8 +491,7 @@ pop_lex(long n)
 {
   long j;
   for(j=1; j<=n; j++)
-    freelex(-j);
-  s_var.n-=n;
+    freelex();
   s_trace.n--;
 }
 
@@ -496,6 +514,7 @@ pari_init_evaluator(void)
   pari_stack_init(&s_trace,sizeof(*trace),(void**)&trace);
   br_res = NULL;
   pari_stack_init(&s_relocs,sizeof(*relocs),(void**)&relocs);
+  pari_stack_init(&s_prec,sizeof(*precs),(void**)&precs);
 }
 void
 pari_close_evaluator(void)
@@ -506,6 +525,7 @@ pari_close_evaluator(void)
   pari_stack_delete(&s_lvars);
   pari_stack_delete(&s_trace);
   pari_stack_delete(&s_relocs);
+  pari_stack_delete(&s_prec);
 }
 
 static gp_pointer *
@@ -526,10 +546,61 @@ new_ptr(void)
   return &ptrs[rp++];
 }
 
+void
+push_localprec(long p)
+{
+  long n = pari_stack_new(&s_prec);
+  precs[n] = prec2nbits(p);
+}
+
+void
+push_localbitprec(long p)
+{
+  long n = pari_stack_new(&s_prec);
+  precs[n] = p;
+}
+
+void
+pop_localprec(void)
+{
+  s_prec.n--;
+}
+
+long
+get_localbitprec(void)
+{
+  return s_prec.n? precs[s_prec.n-1]: precreal;
+}
+
+long
+get_localprec(void)
+{
+  return nbits2prec(get_localbitprec());
+}
+
+void
+localprec(long p)
+{
+  long pmax = prec2ndec(LGBITS);
+  if (p < 1) pari_err_DOMAIN("localprec", "p", "<", gen_1, stoi(p));
+  if (p > pmax)
+    pari_err_DOMAIN("localprec", "p", ">", utoi(pmax), stoi(p));
+  push_localprec(ndec2prec(p));
+}
+
+void
+localbitprec(long p)
+{
+  if (p < 1) pari_err_DOMAIN("localprec", "p", "<", gen_1, stoi(p));
+  if (p > (long)LGBITS)
+    pari_err_DOMAIN("localbitprec", "p", ">", utoi(LGBITS), stoi(p));
+  push_localbitprec(p);
+}
+
 INLINE GEN
 copyupto(GEN z, GEN t)
 {
-  if (is_universal_constant(z) || (z>(GEN)bot && z<=t))
+  if (is_universal_constant(z) || (z>(GEN)pari_mainstack->bot && z<=t))
     return z;
   else
     return gcopy(z);
@@ -587,6 +658,9 @@ closure_castgen(GEN z, long mode)
   case Gsmall:
     st[sp++]=gtos(z);
     break;
+  case Gusmall:
+    st[sp++]=gtou(z);
+    break;
   case Gvar:
     st[sp++]=closure_varn(z);
     break;
@@ -704,6 +778,14 @@ closure_err(long level)
   }
 }
 
+GEN
+pari_self(void)
+{
+  long fun = s_trace.n - 1;
+  if (fun > 0) while (lg(trace[fun].closure)==6) fun--;
+  return fun >= 0 ? trace[fun].closure: NULL;
+}
+
 long
 closure_context(long start, long level)
 {
@@ -749,7 +831,7 @@ closure_eval(GEN C)
   GEN data=closure_get_data(C);
   long loper=lg(oper);
   long saved_sp=sp-closure_arity(C);
-  long saved_rp=rp;
+  long saved_rp=rp, saved_prec=s_prec.n;
   long j, nbmvar=0, nblvar=0;
   long pc, t;
   clone_lock(C);
@@ -788,7 +870,7 @@ closure_eval(GEN C)
       gel(st,sp++)=gel(data,operand);
       break;
     case OCpushreal:
-      gel(st,sp++)=strtor(GSTR(data[operand]),precreal);
+      gel(st,sp++)=strtor(GSTR(data[operand]),get_localprec());
       break;
     case OCpushstoi:
       gel(st,sp++)=stoi(operand);
@@ -796,14 +878,13 @@ closure_eval(GEN C)
     case OCpushvar:
       {
         entree *ep = (entree *)operand;
-        pari_var_create(ep);
-        gel(st,sp++)=(GEN)initial_value(ep);
+        gel(st,sp++)=pol_x(pari_var_create(ep));
         break;
       }
     case OCpushdyn:
       {
         entree *ep = (entree *)operand;
-        checkvalue(ep);
+        checkvalue(ep, chk_CREATE);
         gel(st,sp++)=(GEN)ep->value;
         break;
       }
@@ -815,8 +896,7 @@ closure_eval(GEN C)
         gp_pointer *g = new_ptr();
         g->vn=0;
         g->ep = (entree*) operand;
-        checkvalue(g->ep);
-        g->x = (GEN) g->ep->value;
+        g->x = checkvalueptr(g->ep);
         g->ox = g->x; clone_lock(g->ox);
         g->sp = sp;
         gel(st,sp++) = (GEN)&(g->x);
@@ -838,7 +918,7 @@ closure_eval(GEN C)
         entree *ep = (entree *)operand;
         gp_pointer *g = new_ptr();
         matcomp *C;
-        checkvalue(ep);
+        checkvalue(ep, chk_ERROR);
         g->sp = -1;
         g->x = copyvalue(ep);
         g->ox = g->x; clone_lock(g->ox);
@@ -890,7 +970,7 @@ closure_eval(GEN C)
     case OCstoredyn:
       {
         entree *ep = (entree *)operand;
-        checkvalue(ep);
+        checkvalue(ep, chk_NOCREATE);
         changevalue(ep, gel(st,--sp));
         break;
       }
@@ -913,7 +993,10 @@ closure_eval(GEN C)
         break;
       }
     case OCprecreal:
-      st[sp++]=precreal;
+      st[sp++]=get_localprec();
+      break;
+    case OCbitprecreal:
+      st[sp++]=get_localbitprec();
       break;
     case OCprecdl:
       st[sp++]=precdl;
@@ -924,7 +1007,7 @@ closure_eval(GEN C)
     case OCcowvardyn:
       {
         entree *ep = (entree *)operand;
-        checkvalue(ep);
+        checkvalue(ep, chk_ERROR);
         (void)copyvalue(ep);
         break;
       }
@@ -934,9 +1017,15 @@ closure_eval(GEN C)
     case OCstoi:
       gel(st,sp-1)=stoi(st[sp-1]);
       break;
+    case OCutoi:
+      gel(st,sp-1)=utoi(st[sp-1]);
+      break;
     case OCitos:
       st[sp+operand]=gtos(gel(st,sp+operand));
       break;
+    case OCitou:
+      st[sp+operand]=gtou(gel(st,sp+operand));
+      break;
     case OCtostr:
       {
         GEN z = gel(st,sp+operand);
@@ -987,6 +1076,8 @@ closure_eval(GEN C)
         case t_LIST:
           {
             long lx;
+            if (list_typ(p)!=t_LIST_RAW)
+              pari_err_TYPE("_[_] OCcompo1 [not a vector]", p);
             p = list_data(p); lx = p? lg(p): 1;
             check_array_index(c, lx);
             closure_castgen(gel(p,c),operand);
@@ -1023,6 +1114,8 @@ closure_eval(GEN C)
           g->x = stoi(p[c]);
           break;
         case t_LIST:
+          if (list_typ(p)!=t_LIST_RAW)
+            pari_err_TYPE("&_[_] OCcompo1 [not a vector]", p);
           p = list_data(p); lx = p? lg(p): 1;
           check_array_index(c,lx);
           C->ptcell = (GEN *) p+c;
@@ -1130,7 +1223,7 @@ closure_eval(GEN C)
       {
         long n = pari_stack_new(&s_lvars);
         entree *ep = (entree *)operand;
-        checkvalue(ep);
+        checkvalue(ep, chk_NOCREATE);
         lvars[n] = ep;
         nblvar++;
         pushvalue(ep,gel(st,--sp));
@@ -1140,7 +1233,7 @@ closure_eval(GEN C)
       {
         long n = pari_stack_new(&s_lvars);
         entree *ep = (entree *)operand;
-        checkvalue(ep);
+        checkvalue(ep, chk_NOCREATE);
         lvars[n] = ep;
         nblvar++;
         zerovalue(ep);
@@ -1228,17 +1321,29 @@ closure_eval(GEN C)
       {
         long n=operand;
         GEN fun = gel(st,sp-1-n);
-        long arity;
+        long arity, isvar;
         GEN z;
         if (typ(fun)!=t_CLOSURE) pari_err(e_NOTFUNC, fun);
+        isvar = closure_is_variadic(fun);
         arity = closure_arity(fun);
-        if (n!=arity)
+        if (!isvar || n < arity)
         {
+          st_alloc(arity-n);
           if (n>arity)
             pari_err(e_MISC,"too many parameters in user-defined function call");
-          st_alloc(arity-n);
           for (j=n+1;j<=arity;j++)
             gel(st,sp++)=0;
+          if (isvar) gel(st,sp-1) = cgetg(1,t_VEC);
+        }
+        else
+        {
+          GEN v;
+          long j, m = n-arity+1;
+          v = cgetg(m+1,t_VEC);
+          sp-=m;
+          for (j=1; j<=m; j++)
+            gel(v,j) = gel(st,sp+j-1)? gcopy(gel(st,sp+j-1)): gen_0;
+          gel(st,sp++)=v;
         }
 #ifdef STACK_CHECK
         if (PARI_stack_limit && (void*) &z <= PARI_stack_limit)
@@ -1313,6 +1418,13 @@ closure_eval(GEN C)
       else
         st[sp+operand]=st[sp];
       break;
+    case OCdefaultulong:
+      sp--;
+      if (st[sp+operand])
+        st[sp+operand]=gtou(gel(st,sp+operand));
+      else
+        st[sp+operand]=st[sp];
+      break;
     case OCdefaultgen:
       sp--;
       if (!st[sp+operand])
@@ -1360,6 +1472,7 @@ endeval:
       clone_unlock(g->ox);
     }
   }
+  s_prec.n = saved_prec;
   s_trace.n--;
   restore_vars(nbmvar, nblvar);
   clone_unlock(C);
@@ -1380,6 +1493,7 @@ evalstate_save(struct pari_evalstate *state)
   state->avma = avma;
   state->sp   = sp;
   state->rp   = rp;
+  state->prec = s_prec.n;
   state->var  = s_var.n;
   state->lvars= s_lvars.n;
   state->trace= s_trace.n;
@@ -1394,6 +1508,7 @@ evalstate_restore(struct pari_evalstate *state)
   mtstate_restore(&state->pending_threads);
   sp = state->sp;
   rp = state->rp;
+  s_prec.n = state->prec;
   restore_vars(s_var.n-state->var,s_lvars.n-state->lvars);
   restore_trace(s_trace.n-state->trace);
   reset_break();
@@ -1420,7 +1535,7 @@ evalstate_reset(void)
   reset_break();
   compilestate_reset();
   parsestate_reset();
-  avma = top;
+  avma = pari_mainstack->top;
 }
 
 void
@@ -1478,24 +1593,6 @@ closure_returnupto(GEN C)
   return copyupto(closure_return(C),(GEN)av);
 }
 
-void
-closure_callvoid1(GEN C, GEN x)
-{
-  long i, ar = closure_arity(C);
-  gel(st,sp++) = x;
-  for(i=2; i <= ar; i++) gel(st,sp++) = NULL;
-  closure_evalvoid(C);
-}
-
-GEN
-closure_callgen1(GEN C, GEN x)
-{
-  long i, ar = closure_arity(C);
-  gel(st,sp++) = x;
-  for(i=2; i<= ar; i++) gel(st,sp++) = NULL;
-  return closure_returnupto(C);
-}
-
 GEN
 pareval_worker(GEN C)
 {
@@ -1562,7 +1659,7 @@ parvector(long n, GEN code)
 GEN
 parsum(GEN a, GEN b, GEN code, GEN x)
 {
-  pari_sp av = avma, av2, lim;
+  pari_sp av = avma, av2;
   long pending = 0;
   GEN worker = snm_closure(is_entry("_parvector_worker"), mkvec(code));
   GEN done;
@@ -1574,7 +1671,7 @@ parsum(GEN a, GEN b, GEN code, GEN x)
   mt_queue_start(&pt, worker);
   b = gfloor(b);
   a = mkvec(setloop(a));
-  av2=avma; lim = stack_lim(av2,1);
+  av2=avma;
   for (; cmpii(gel(a,1),b) <= 0 || pending; gel(a,1) = incloop(gel(a,1)))
   {
     mt_queue_submit(&pt, 0, cmpii(gel(a,1),b) <= 0? a: NULL);
@@ -1582,7 +1679,7 @@ parsum(GEN a, GEN b, GEN code, GEN x)
     if (done)
     {
       x = gadd(x, done);
-      if (low_stack(lim, stack_lim(av2,1)))
+      if (gc_needed(av2,1))
       {
         if (DEBUGMEM>1) pari_warn(warnmem,"sum");
         x = gerepileupto(av2,x);
@@ -1594,7 +1691,7 @@ parsum(GEN a, GEN b, GEN code, GEN x)
 }
 
 void
-parfor(GEN a, GEN b, GEN code, GEN code2)
+parfor(GEN a, GEN b, GEN code, void *E, long call(void*, GEN, GEN))
 {
   pari_sp av = avma, av2;
   long running, pending = 0;
@@ -1603,29 +1700,31 @@ parfor(GEN a, GEN b, GEN code, GEN code2)
   GEN done, stop = NULL;
   struct pari_mt pt;
   if (typ(a) != t_INT) pari_err_TYPE("parfor",a);
-  if (b && gcmp(b,a) < 0) return;
-
+  if (b)
+  {
+    if (gcmp(b,a) < 0) return;
+    if (typ(b) == t_INFINITY)
+    {
+      if (inf_get_sign(b) < 0) return;
+      b = NULL;
+    }
+    else
+      b = gfloor(b);
+  }
   mt_queue_start(&pt, worker);
-  b = b ? gfloor(b): NULL;
   a = mkvec(setloop(a));
   av2 = avma;
   while ((running = (!stop && (!b || cmpii(gel(a,1),b) <= 0))) || pending)
   {
     mt_queue_submit(&pt, 0, running ? a: NULL);
     done = mt_queue_get(&pt, NULL, &pending);
-    if (code2 && done && (!stop || cmpii(gel(done,1),stop) < 0))
-    {
-      push_lex(gel(done,1), code2);
-      push_lex(gel(done,2), NULL);
-      closure_evalvoid(code2);
-      pop_lex(2);
-      if (loop_break())
+    if (call && done && (!stop || cmpii(gel(done,1),stop) < 0))
+      if (call(E, gel(done,1), gel(done,2)))
       {
         status = br_status;
         br_status = br_NONE;
         stop = gerepileuptoint(av2, gel(done,1));
       }
-    }
     gel(a,1) = incloop(gel(a,1));
     if (!stop) avma = av2;
   }
@@ -1635,8 +1734,66 @@ parfor(GEN a, GEN b, GEN code, GEN code2)
   avma = av;
 }
 
+static long
+gp_evalvoid2(void *E, GEN x, GEN y)
+{
+  GEN code =(GEN) E;
+  push_lex(x, code);
+  push_lex(y, NULL);
+  closure_evalvoid(code);
+  pop_lex(2);
+  return loop_break();
+}
+
 void
-parforprime(GEN a, GEN b, GEN code, GEN code2)
+parfor0(GEN a, GEN b, GEN code, GEN code2)
+{
+  parfor(a, b, code, (void*)code2, code2 ? gp_evalvoid2: NULL);
+}
+
+void
+parforprime(GEN a, GEN b, GEN code, void *E, long call(void*, GEN, GEN))
+{
+  pari_sp av = avma, av2;
+  long running, pending = 0;
+  long status = br_NONE;
+  GEN worker = snm_closure(is_entry("_parfor_worker"), mkvec(code));
+  GEN v, done, stop = NULL;
+  struct pari_mt pt;
+  forprime_t T;
+
+  if (!forprime_init(&T, a,b)) { avma = av; return; }
+  mt_queue_start(&pt, worker);
+  v = mkvec(gen_0);
+  av2 = avma;
+  while ((running = (!stop && forprime_next(&T))) || pending)
+  {
+    gel(v, 1) = T.pp;
+    mt_queue_submit(&pt, 0, running ? v: NULL);
+    done = mt_queue_get(&pt, NULL, &pending);
+    if (call && done && (!stop || cmpii(gel(done,1),stop) < 0))
+      if (call(E, gel(done,1), gel(done,2)))
+      {
+        status = br_status;
+        br_status = br_NONE;
+        stop = gerepileuptoint(av2, gel(done,1));
+      }
+    if (!stop) avma = av2;
+  }
+  avma = av2;
+  mt_queue_end(&pt);
+  br_status = status;
+  avma = av;
+}
+
+void
+parforprime0(GEN a, GEN b, GEN code, GEN code2)
+{
+  parforprime(a, b, code, (void*)code2, code2? gp_evalvoid2: NULL);
+}
+
+void
+parforvec(GEN x, GEN code, long flag, void *E, long call(void*, GEN, GEN))
 {
   pari_sp av = avma, av2;
   long running, pending = 0;
@@ -1644,28 +1801,25 @@ parforprime(GEN a, GEN b, GEN code, GEN code2)
   GEN worker = snm_closure(is_entry("_parfor_worker"), mkvec(code));
   GEN done, stop = NULL;
   struct pari_mt pt;
-  forprime_t T;
+  forvec_t T;
+  GEN a, v = gen_0;
 
-  if (!forprime_init(&T, a,b)) { avma = av; return; }
+  if (!forvec_init(&T, x, flag)) { avma = av; return; }
   mt_queue_start(&pt, worker);
+  a = mkvec(gen_0);
   av2 = avma;
-  while ((running = (!stop && forprime_next(&T))) || pending)
+  while ((running = (!stop && v && (v = forvec_next(&T)))) || pending)
   {
-    mt_queue_submit(&pt, 0, running ? mkvec(T.pp): NULL);
+    gel(a, 1) = v;
+    mt_queue_submit(&pt, 0, running ? a: NULL);
     done = mt_queue_get(&pt, NULL, &pending);
-    if (code2 && done && (!stop || cmpii(gel(done,1),stop) < 0))
-    {
-      push_lex(gel(done,1), code2);
-      push_lex(gel(done,2), NULL);
-      closure_evalvoid(code2);
-      pop_lex(2);
-      if (loop_break())
+    if (call && done && (!stop || lexcmp(gel(done,1),stop) < 0))
+      if (call(E, gel(done,1), gel(done,2)))
       {
         status = br_status;
         br_status = br_NONE;
-        stop = gerepileuptoint(av2, gel(done,1));
+        stop = gerepilecopy(av2, gel(done,1));
       }
-    }
     if (!stop) avma = av2;
   }
   avma = av2;
@@ -1674,6 +1828,43 @@ parforprime(GEN a, GEN b, GEN code, GEN code2)
   avma = av;
 }
 
+void
+parforvec0(GEN x, GEN code, GEN code2, long flag)
+{
+  parforvec(x, code, flag, (void*)code2, code2? gp_evalvoid2: NULL);
+}
+
+void
+closure_callvoid1(GEN C, GEN x)
+{
+  long i, ar = closure_arity(C);
+  gel(st,sp++) = x;
+  for(i=2; i <= ar; i++) gel(st,sp++) = NULL;
+  closure_evalvoid(C);
+}
+
+GEN
+closure_callgen1(GEN C, GEN x)
+{
+  long i, ar = closure_arity(C);
+  gel(st,sp++) = x;
+  for(i=2; i<= ar; i++) gel(st,sp++) = NULL;
+  return closure_returnupto(C);
+}
+
+GEN
+closure_callgen1prec(GEN C, GEN x, long prec)
+{
+  GEN z;
+  long i, ar = closure_arity(C);
+  gel(st,sp++) = x;
+  for(i=2; i<= ar; i++) gel(st,sp++) = NULL;
+  push_localprec(prec);
+  z = closure_returnupto(C);
+  pop_localprec();
+  return z;
+}
+
 GEN
 closure_callgen2(GEN C, GEN x, GEN y)
 {
@@ -1688,19 +1879,35 @@ closure_callgen2(GEN C, GEN x, GEN y)
 GEN
 closure_callgenvec(GEN C, GEN args)
 {
-  long i, l = lg(args), ar = closure_arity(C);
+  long i, l = lg(args)-1, ar = closure_arity(C);
   st_alloc(ar);
-  for (i = 1; i < l;   i++) gel(st,sp++) = gel(args,i);
+  if (l > ar)
+    pari_err(e_MISC,"too many parameters in user-defined function call");
+  if (closure_is_variadic(C) && l==ar && typ(gel(args,l))!=t_VEC)
+    pari_err_TYPE("call", gel(args,l));
+  for (i = 1; i <= l;  i++) gel(st,sp++) = gel(args,i);
   for(      ; i <= ar; i++) gel(st,sp++) = NULL;
   return closure_returnupto(C);
 }
 
+GEN
+closure_callgenvecprec(GEN C, GEN args, long prec)
+{
+  GEN z;
+  push_localprec(prec);
+  z = closure_callgenvec(C, args);
+  pop_localprec();
+  return z;
+}
+
 GEN
 closure_callgenall(GEN C, long n, ...)
 {
   va_list ap;
   long i, ar = closure_arity(C);
   va_start(ap,n);
+  if (n > ar)
+    pari_err(e_MISC,"too many parameters in user-defined function call");
   st_alloc(ar);
   for (i = 1; i <=n;  i++) gel(st,sp++) = va_arg(ap, GEN);
   for(      ; i <=ar; i++) gel(st,sp++) = NULL;
@@ -1723,6 +1930,16 @@ gp_evalupto(void *E, GEN x)
   return copyupto(gp_eval(E,x), (GEN)av);
 }
 
+GEN
+gp_evalprec(void *E, GEN x, long prec)
+{
+  GEN z;
+  push_localprec(prec);
+  z = gp_eval(E, x);
+  pop_localprec();
+  return z;
+}
+
 long
 gp_evalbool(void *E, GEN x)
 {
@@ -1747,6 +1964,20 @@ gp_call(void *E, GEN x)
   return closure_callgen1(code, x);
 }
 
+GEN
+gp_callprec(void *E, GEN x, long prec)
+{
+  GEN code = (GEN)E;
+  return closure_callgen1prec(code, x, prec);
+}
+
+GEN
+gp_call2(void *E, GEN x, GEN y)
+{
+  GEN code = (GEN)E;
+  return closure_callgen2(code, x, y);
+}
+
 long
 gp_callbool(void *E, GEN x)
 {
@@ -1870,15 +2101,24 @@ closure_disassemble(GEN C)
     case OCprecreal:
       pari_printf("precreal\n");
       break;
+    case OCbitprecreal:
+      pari_printf("bitprecreal\n");
+      break;
     case OCprecdl:
       pari_printf("precdl\n");
       break;
     case OCstoi:
       pari_printf("stoi\n");
       break;
+    case OCutoi:
+      pari_printf("utoi\n");
+      break;
     case OCitos:
       pari_printf("itos\t\t%ld\n",operand);
       break;
+    case OCitou:
+      pari_printf("itou\t\t%ld\n",operand);
+      break;
     case OCtostr:
       pari_printf("tostr\t\t%ld\n",operand);
       break;
@@ -1927,6 +2167,9 @@ closure_disassemble(GEN C)
     case OCdefaultlong:
       pari_printf("defaultlong\t%ld\n",operand);
       break;
+    case OCdefaultulong:
+      pari_printf("defaultulong\t%ld\n",operand);
+      break;
     case OCdefaultgen:
       pari_printf("defaultgen\t%ld\n",operand);
       break;
@@ -2040,9 +2283,12 @@ opcode_need_relink(op_code opcode)
   case OCstackgen:
   case OCendptr:
   case OCprecreal:
+  case OCbitprecreal:
   case OCprecdl:
   case OCstoi:
+  case OCutoi:
   case OCitos:
+  case OCitou:
   case OCtostr:
   case OCvarn:
   case OCcopy:
@@ -2062,6 +2308,7 @@ opcode_need_relink(op_code opcode)
   case OCdefaultarg:
   case OCdefaultgen:
   case OCdefaultlong:
+  case OCdefaultulong:
   case OCcalluser:
   case OCvec:
   case OCcol:
@@ -2205,11 +2452,6 @@ copybin_unlink(GEN C)
   return res;
 }
 
-static ulong
-hash_id(void *x) { return (ulong)x; }
-static int
-eq_id(void *x, void *y) { return x == y; }
-
 /* e = t_VECSMALL of entree *ep [ addresses ],
  * names = t_VEC of strtoGENstr(ep.names),
  * Return hashtable : ep => is_entry(ep.name) */
@@ -2217,12 +2459,12 @@ hashtable *
 hash_from_link(GEN e, GEN names, int use_stack)
 {
   long i, l = lg(e);
-  hashtable *h = hash_create(l-1, &hash_id, &eq_id, use_stack);
+  hashtable *h = hash_create_ulong(l-1, use_stack);
   if (lg(names) != l) pari_err_DIM("hash_from_link");
   for (i = 1; i < l; i++)
   {
     char *s = GSTR(gel(names,i));
-    hash_insert(h, (void*)e[i], (void*)fetch_entry(s, strlen(s)));
+    hash_insert(h, (void*)e[i], (void*)fetch_entry(s));
   }
   return h;
 }
