@@ -32,7 +32,7 @@ typedef struct Red {
   GEN (*red)(GEN x, struct Red*);
 } Red;
 
-typedef struct Cache { /* data associated to p^k */
+typedef struct Cache { /* data attached to p^k */
   GEN aall, tall;
   GEN cyc;
   GEN E;
@@ -246,11 +246,11 @@ _powpolmod(Cache *C, GEN jac, Red *R, GEN (*_sqr)(GEN, Red *))
   const long efin = lg(taba)-1, lv = R->lv;
   GEN L, res = jac, pol2 = _sqr(res, R);
   long f;
-  pari_sp av0 = avma, av, lim;
+  pari_sp av0 = avma, av;
 
   L = cgetg(lv+1, t_VEC); gel(L,1) = res;
   for (f=2; f<=lv; f++) gel(L,f) = _mul(gel(L,f-1), pol2, R);
-  av = avma; lim = stack_lim(av, 1);
+  av = avma;
   for (f = efin; f >= 1; f--)
   {
     GEN t = gel(L, taba[f]);
@@ -258,7 +258,7 @@ _powpolmod(Cache *C, GEN jac, Red *R, GEN (*_sqr)(GEN, Red *))
     res = (f==efin)? t: _mul(t, res, R);
     while (tf--) {
       res = _sqr(res, R);
-      if (low_stack(lim, stack_lim(av,1))) {
+      if (gc_needed(av,1)) {
         res = gerepilecopy(av, res);
         if(DEBUGMEM>1) pari_warn(warnmem,"powpolmod: f = %ld",f);
       }
@@ -612,7 +612,7 @@ filltabs(Cache *C, Cache *Cp, Red *R, long p, long pk, long ltab)
 
     for (i=1; i<=pk; i++)
     {
-      GEN z = FpX_rem(monomial(gen_1, i-1, 0), C->cyc, R->N);
+      GEN z = FpX_rem(pol_xn(i-1, 0), C->cyc, R->N);
       gel(eta,i) = FpX_center(z, R->N, R->N2);
     }
     C->eta = eta;
@@ -965,7 +965,7 @@ aprcl(GEN N)
   for (i=2; i<lfat; i++)
   {
     ulong p = fat[i];
-    flaglp[i] = equaliu(Fp_powu(N, p-1, sqru(p)), 1);
+    flaglp[i] = absequaliu(Fp_powu(N, p-1, sqru(p)), 1);
   }
   vecsmall_sort(faet);
 
