@@ -19,9 +19,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /*               degree of a p-adic field                   */
 /* Xavier Roblot                                            */
 /************************************************************/
-
-#undef CHECK_EXTENSIONS
-
 /* cf. Math. Comp, vol. 70, No. 236, pp. 1641-1659 for more details.
    Note that n is now e (since the e from the paper is always = 1) and l
    is now f */
@@ -34,10 +31,9 @@ typedef struct {
   long r;     /* pr = p^r */
   GEN pr;     /* p-adic precision for poly. reduction */
   GEN q, qm1; /* p^f, q - 1 */
-  GEN upl;    /* cyclotomic polynomial (in v) generating K^ur (mod pr) */
-  GEN mv;     /* -v mod upl (mod pr) */
-  GEN uplr;   /* upl reduced mod p */
-  GEN frob;   /* Frobenius acting of the root of upl (mod pr) */
+  GEN T;      /* polynomial defining K^ur */
+  GEN frob;   /* Frobenius acting of the root of T (mod pr) */
+  GEN u;      /* suitable root of unity (expressed in terms of the root of T) */
   GEN nbext;  /* number of extensions */
   GEN roottable; /* table of roots of polynomials over the residue field */
   GEN pk;     /* powers of p: [p^1, p^2, ..., p^c] */
@@ -56,13 +52,7 @@ typedef struct {
   long cj;  /* number of conjugate fields */
 } FAD_t;
 
-static long
-ceildiv(ulong a, ulong b)
-{
-  long c = a/b;
-  if (a%b) c++;
-  return c;
-}
+#undef CHECK
 
 /* Eval P(x) assuming P has positive coefficients and the result is a long */
 static ulong
@@ -210,19 +200,19 @@ NumberExtensions(KRASNER_t *data)
 static GEN
 get_topx(KRASNER_t *data, GEN eis)
 {
-  GEN p1, p2, rpl;
+  GEN p1, p2, rpl, c;
+  pari_sp av;
   long j;
-  pari_sp av, lim;
-  /* top poly. is the minimal polynomial of root(pol) + root(upl) */
-  rpl = FqX_translate(FqX_red(eis, data->upl, data->pr),
-                      data->mv, data->upl, data->pr);
-  p1 = p2 = rpl;
-  av = avma; lim = stack_lim(av, 1);
+  /* top poly. is the minimal polynomial of root(pol) + root(T) */
+  if (data->f == 1) return eis;
+  c = FpX_neg(pol_x(data->v),data->pr);
+  rpl = FqX_translate(eis, c, data->T, data->pr);
+  p1 = p2 = rpl; av = avma;
   for (j = 1; j < data->f; j++)
   {
-    p1 = FqX_FpXQ_eval(p1, data->frob, data->upl, data->pr);
-    p2 = FqX_mul(p2, p1, data->upl, data->pr);
-    if (low_stack(lim, stack_lim(av, 1))) gerepileall(av, 2, &p1, &p2);
+    p1 = FqX_FpXQ_eval(p1, data->frob, data->T, data->pr);
+    p2 = FqX_mul(p2, p1, data->T, data->pr);
+    if (gc_needed(av,2)) gerepileall(av, 2, &p1,&p2);
   }
   return simplify_shallow(p2); /* ZX */
 }
@@ -242,7 +232,7 @@ FieldData(KRASNER_t *data, FAD_t *fdata, GEN eis, GEN topx)
   fdata->topr = FpX_red(t, data->pr);
 
   zq  = pol_x(data->v);
-  /* FIXME: do as in CycloPol (not so easy) */
+
   for(;;)
   {
     GEN zq2 = zq;
@@ -333,33 +323,6 @@ GetSharp(FAD_t *fdata, GEN pp, GEN ppp, GEN pol, GEN beta, long *pl)
   return gerepilecopy(av, normalizepol(p1));
 }
 
-#ifdef CHECK_EXTENSIONS
-static void
-PrintValuations(GEN pol, GEN mod, GEN p)
-{
-  long i, d = degpol(pol);
-  for (i = 0; i <= d; i++)
-    err_printf("%d ", Z_pval(RgXQ_norm(gel(pol, i+2), mod), p));
-}
-
-/* Return the degree of pol mod the prime ideal of top */
-static long
-DegreeMod(FAD_t *fdata, GEN pp, GEN ppp, GEN pol)
-{
-  long d = degpol(pol); /* should be > 0 */
-  pari_sp av = avma;
-
-  do
-  {
-    GEN c = gel(pol, d+2);
-    if (!gequal0(c) && !DivideByPi(fdata, pp, ppp, c))
-      return d;
-  }
-  while (--d >= 1);
-  avma = av; return 0;
-}
-#endif
-
 /* Compute roots of pol in the residue field. Use table look-up if possible */
 static GEN
 Quick_FqX_roots(KRASNER_t *data, GEN pol)
@@ -367,15 +330,16 @@ Quick_FqX_roots(KRASNER_t *data, GEN pol)
   GEN rts;
   ulong ind = 0;
 
-  pol = FpXQX_red(pol, data->uplr, data->p);
-
+  if (data->f == 1)
+    pol = FpXY_evalx(pol, gen_0, data->p);
+  else
+    pol = FqX_red(pol, data->T, data->p);
   if (data->roottable)
   {
     ind = ZXY_z_eval(pol, data->q[2], data->p[2]);
     if (gel(data->roottable, ind)) return gel(data->roottable, ind);
   }
-  rts = FqX_roots(pol, data->uplr, data->p);
-
+  rts = FqX_roots(pol, data->T, data->p);
   if (ind) gel(data->roottable, ind) = gclone(rts);
   return rts;
 }
@@ -405,10 +369,6 @@ RootCongruents(KRASNER_t *data, FAD_t *fdata, GEN pol, GEN alpha, GEN pp, GEN pp
     long l;
     pol = GetSharp(fdata, pp, ppp, pol, alpha, &l);
     if (l <= 1) return l;
-#ifdef CHECK_EXTENSIONS
-    if (l != DegreeMod(fdata, pp, ppp, pol))
-      pari_err_BUG("RootCongruents [degree mismatch in RCA]");
-#endif
     /* decrease precision if sl gets bigger than a multiple of e */
     sl += l;
     if (sl >= data->e)
@@ -469,7 +429,7 @@ IsIsomorphic(KRASNER_t *data, FAD_t *fdata, GEN pol)
     nb = RootCountingAlgorithm(data, fdata, p1, 1);
     if (nb) { avma = av; return nb; }
     if (j < data->f)
-      pol = FqX_FpXQ_eval(pol, data->frob, data->upl, data->pr);
+      pol = FqX_FpXQ_eval(pol, data->frob, data->T, data->pr);
   }
   avma = av; return 0;
 }
@@ -492,8 +452,8 @@ NbConjugateFields(KRASNER_t *data, FAD_t *fdata)
   {
     GEN p1 = FqX_FpXQ_eval(pol, fdata->zq, fdata->top, data->pr);
     nb += RootCountingAlgorithm(data, fdata, p1, 0);
-    if (j < data->f)
-      pol = FqX_FpXQ_eval(pol, data->frob, data->upl, data->pr);
+    if (j < data->f) /* FIXME */
+      pol = FqX_FpXQ_eval(pol, data->frob, data->T, data->pr);
   }
   avma = av;
   fdata->cj = data->e * data->f / nb;
@@ -506,81 +466,34 @@ NbConjugateFields(KRASNER_t *data, FAD_t *fdata)
 static GEN
 TamelyRamifiedCase(KRASNER_t *data)
 {
-  long av = avma, g;
-  GEN rep, p2, topx, m, eis, Xe = gpowgs(pol_x(0), data->e);
+  long av = avma;
+  long g = ugcd(data->e, umodiu(data->qm1, data->e)); /* (e, q-1) */
+  GEN rep, eis, Xe = gpowgs(pol_x(0), data->e), m = stoi(data->e / g);
 
-#ifdef CHECK_EXTENSIONS
-  FAD_t fdata;
-  long cnt = 0, nb, j;
-  GEN vpl;
-  err_printf("Number of extensions: %ld\n", itos(data->nbext));
-#endif
-
-  g   = ugcd(data->e, umodiu(data->qm1, data->e)); /* (e, q-1) */
-  m   = stoi(data->e/g);
   rep = zerovec(g);
-
   eis = gadd(Xe, data->p);
-  topx = get_topx(data, eis);
-  p2 = mkvec2(topx, m);
-  gel(rep, 1) = p2;
-#ifdef CHECK_EXTENSIONS
-  vpl = zerovec(g);
-  gel(vpl, 1) = eis;
-  if (data->e == 1)
-    nb = 1;
-  else
-  {
-    FieldData(data, &fdata, eis, topx);
-    NbConjugateFields(data, &fdata);
-    nb = fdata.cj;
-  }
-  err_printf("Found %ld field(s)\n", nb);
-  cnt += nb;
-#endif
-
+  gel(rep, 1) = mkvec2(get_topx(data,eis), m);
   if (g > 1)
   {
     ulong pmodg = umodiu(data->p, g);
     long r = 1, ct = 1;
     GEN sv = InitSieve(g-1);
+    /* we compute the action of the Frobenius to get a minimal set of polynomials over Q_p */
     while (r)
     {
       long gr;
-      GEN p1 = FpXQ_powu(pol_x(data->v), r, data->uplr, data->p);
-      eis = gadd(Xe, ZX_Z_mul(p1, data->p)); /* Adding a ZX and a ZY (cste coefficient) */
+      GEN p1 = (typ(data->u) == t_INT)?
+        mulii(Fp_powu(data->u, r, data->p), data->p):
+        ZX_Z_mul(FpXQ_powu(data->u, r, data->T, data->p), data->p);
+      eis = gadd(Xe, p1); /* add cst coef */
       ct++;
-      topx = get_topx(data, eis);
-      p2 = mkvec2(topx, m);
-      gel(rep, ct) = p2;
-#ifdef CHECK_EXTENSIONS
-      gel(vpl, ct) = eis;
-      FieldData(data, &fdata, eis, topx);
-      for (j = 1; j < ct; j++)
-        if (IsIsomorphic(data, &fdata, gel(vpl, j)))
-          pari_err_BUG("TamelyRamifiedCase [isomorphic fields]");
-      NbConjugateFields(data, &fdata);
-      nb = fdata.cj;
-      err_printf("Found %ld field(s)\n", nb);
-      cnt += nb;
-#endif
+      gel(rep, ct) = mkvec2(get_topx(data,eis), m);
       gr = r;
-      do
-      {
-        SetSieveValue(sv, gr);
-        gr = Fl_mul(gr, pmodg, g);
-      }
-      while (gr != r);
+      do { SetSieveValue(sv, gr); gr = Fl_mul(gr, pmodg, g); } while (gr != r);
       r  = NextZeroValue(sv, r);
     }
     setlg(rep, ct+1);
   }
-
-#ifdef CHECK_EXTENSIONS
-  if (!equaliu(data->nbext, cnt))
-    pari_err_BUG("TamelyRamifiedCase [incorrect #fields]");
-#endif
-
   return gerepilecopy(av, rep);
 }
 
@@ -654,7 +567,13 @@ RandomPol(KRASNER_t *data, GEN Omega)
     for (j = 1; j <= end-st; j++)
       c = ZX_add(c, ZX_Z_mul(RandomFF(data), gel(data->pk, j)));
     c = ZX_Z_mul(c, gel(data->pk, st));
-    gel(pol, i+1) = FpX_red(c, data->pr);
+    c = FpX_red(c, data->pr);
+    switch(degpol(c))
+    {
+      case -1: c = gen_0; break;
+      case  0: c = gel(c,2); break;
+    }
+    gel(pol, i+1) = c;
   }
   gel(pol, i+1) = gen_1; /* monic */
   return pol;
@@ -715,16 +634,6 @@ WildlyRamifiedCase(KRASNER_t *data)
   { /* Jump randomly among the polynomials : seems best... */
     rpl = RandomPol(data, Omega);
     if (DEBUGLEVEL>3) err_printf("considering polynomial %Ps\n", rpl);
-#ifdef CHECK_EXTENSIONS
-    {
-      GEN disc = poldisc0(rpl, 0);
-      long e = data->e, f = data->f, j = data->j;
-      disc = RgXQ_norm(disc, data->upl);
-      if (Z_pval(disc, data->p) != f*(e+j-1))
-        pari_err_BUG("WildlyRamifiedCase [incorrect discriminant]");
-    }
-#endif
-
     for (j = 0; j < ct; j++)
     {
       nb = IsIsomorphic(data, vfd[j], rpl);
@@ -733,11 +642,11 @@ WildlyRamifiedCase(KRASNER_t *data)
     if (!nb)
     {
       GEN topx = get_topx(data, rpl);
-      FAD_t *fdata = (FAD_t*)vfd[ct];
-      FieldData(data, fdata, rpl, topx);
-      CloneFieldData(fdata);
-      NbConjugateFields(data, fdata);
-      nb = fdata->cj;
+      FAD_t *f = (FAD_t*)vfd[ct];
+      FieldData(data, f, rpl, topx);
+      CloneFieldData(f);
+      NbConjugateFields(data, f);
+      nb = f->cj;
       fd += nb;
       ct++;
       if (DEBUGLEVEL>1)
@@ -750,29 +659,37 @@ WildlyRamifiedCase(KRASNER_t *data)
   rep = cgetg(ct+1, t_VEC);
   for (j = 0; j < ct; j++)
   {
-    GEN topx = ZX_copy(((FAD_t*)vfd[j])->top);
-    GEN p1;
+    FAD_t *f = (FAD_t*)vfd[j];
+    GEN topx = ZX_copy(f->top);
     setvarn(topx, 0);
-    p1 = mkvec2(topx, stoi(((FAD_t*)vfd[j])->cj));
-    gel(rep, j+1) = p1;
-    FreeFieldData((FAD_t*)vfd[j]);
+    gel(rep, j+1) = mkvec2(topx, stoi(f->cj));
+    FreeFieldData(f);
   }
   FreeRootTable(data->roottable);
   return gerepileupto(av, rep);
 }
 
-/* return the minimal polynomial (mod pr) of a generator of (F_p^f)^x with variable v */
-static GEN
-CycloPol(KRASNER_t *d)
+/* return the minimal polynomial of a generator of K^ur and the expression (mod pr)
+ * in terms of this generator of a root of unity nu such that nu is l-maximal
+ * for all primes l dividing g = (e,q-1). */
+static void
+setUnramData(KRASNER_t *d)
 {
-  GEN T, z;
-  /* v - primroot(p) */
   if (d->f == 1)
-    return deg1pol_shallow(gen_1, Fp_neg(pgener_Fp(d->p), d->pr), d->v);
-  T = init_Fq(d->p, d->f, d->v);
-  z = gener_FpXQ(T, d->p, NULL);
-  z = ZpXQ_sqrtnlift(scalarpol(gen_1,varn(T)), d->qm1, z, T, d->p, d->r);
-  return FpX_red(ZXQ_charpoly(z, T, d->v), d->pr);
+  {
+    d->T = NULL;
+    d->u = pgener_Fp(d->p);
+    d->frob = NULL;
+  }
+  else
+  {
+    GEN L, z, T, p = d->p;
+    d->T = T = init_Fq(p, d->f, d->v);
+    L = gel(factoru(ugcd(d->e, umodiu(d->qm1, d->e))), 1);
+    z = gener_FpXQ_local(T, p, zv_to_ZV(L));
+    d->u = ZpXQ_sqrtnlift(pol_1(d->v), d->qm1, z, T, p, d->r);
+    d->frob = ZpX_ZpXQ_liftroot(T, FpXQ_pow(pol_x(d->v), p, T, p), T, p, d->r);
+  }
 }
 
 /* return [ p^1, p^2, ..., p^c ] */
@@ -811,19 +728,15 @@ GetRamifiedPol(GEN p, GEN efj, long flag)
   data.q   = powiu(p, f);
   data.qm1 = subis(data.q, 1);
   data.v   = v;
-  data.r   = 1 + ceildiv(2*j + 3, e); /* enough precision */
+  data.r   = 1 + (long)ceildivuu(2*j + 3, e); /* enough precision */
   data.pr  = powiu(p, data.r);
   data.nbext = NumberExtensions(&data);
 
   if (flag == 2) return data.nbext;
 
-  data.upl   = CycloPol(&data);
-  /* mv = -v mod upl. If f = 1, then upl = v + c, hence mv = c */
-  data.mv    = f == 1? gel(data.upl, 2)
-                     : FpX_neg(pol_x(v), data.pr);
-  data.uplr  = FpX_red(data.upl, data.p);
-  data.frob  = FpXQ_pow(pol_x(v), p, data.upl, data.pr);
-  if (DEBUGLEVEL>1) err_printf("  Unramified part %Ps\n", data.upl);
+  setUnramData(&data);
+  if (DEBUGLEVEL>1)
+    err_printf("  Unramified part %Ps\n", data.T);
   data.roottable = NULL;
   if (j)
   {
@@ -903,13 +816,76 @@ possible_efj(GEN p, long m)
   setlg(L, nb); return L;
 }
 
+#ifdef CHECK
+static void
+checkpols(GEN p, GEN EFJ, GEN pols)
+{
+  GEN pol, p1, p2;
+  long c1, c2, e, f, j, i, l = lg(pols);
+
+  if (typ(pols) == t_INT) return;
+
+  e = EFJ[1];
+  f = EFJ[2];
+  j = EFJ[3];
+
+  for (i = 1; i < l; i++)
+  {
+    pol = gel(pols, i);
+    if (typ(pol) == t_VEC) pol = gel(pol, 1);
+    if (!isirreducible(pol)) pari_err_BUG("Polynomial is reducible");
+    p1 = poldisc0(pol, -1);
+    if (gvaluation(p1, p) != f*(e+j-1)) pari_err_BUG("Discriminant is incorrect");
+    /* only compute a p-maximal order */
+    p1 = nfinitall(mkvec2(pol, mkvec(p)), 0, DEFAULTPREC);
+    p2 = idealprimedec(p1, p);
+    if(lg(p2) > 2) pari_err_BUG("Prime p is split");
+    p2 = gel(p2, 1);
+    if (cmpis(gel(p2, 3), e)) pari_err_BUG("Ramification index is incorrect");
+    if (cmpis(gel(p2, 4), f)) pari_err_BUG("inertia degree is incorrect");
+  }
+
+  if (l == 2) return;
+  if (e*f > 20) return;
+
+  /* TODO: check that (random) distinct polynomials give non-isomorphic extensions */
+  for (i = 1; i < 3*l; i++)
+  {
+    c1 = random_Fl(l-1)+1;
+    c2 = random_Fl(l-1)+1;
+    if (c1 == c2) continue;
+    p1 = gel(pols, c1);
+    if (typ(p1) == t_VEC) p1 = gel(p1, 1);
+    p2 = gel(pols, c2);
+    if (typ(p2) == t_VEC) p2 = gel(p2, 1);
+    pol = polcompositum0(p1, p2, 0);
+    pol = gel(pol, 1);
+    if (poldegree(pol, -1) > 100) continue;
+    p1 = factorpadic(pol, p, 2);
+    p1 = gmael(p1, 1, 1);
+    if (poldegree(p1, -1) == e*f) pari_err_BUG("fields are isomorphic");
+    /*
+      p1 = nfinitall(mkvec2(pol, mkvec(p)), 0, DEFAULTPREC);
+      p2 = gel(idealprimedec(p1, p), 1);
+      if (!cmpis(mulii(gel(p2, 3), gel(p2, 4)), e*f)) pari_err_BUG("fields are isomorphic");
+    */
+  }
+}
+#endif
+
 static GEN
 pols_from_efj(pari_sp av, GEN EFJ, GEN p, long flag)
 {
   long i, l;
   GEN L = cgetg_copy(EFJ, &l);
   if (l == 1) { avma = av; return flag == 2? gen_0: cgetg(1, t_VEC); }
-  for (i = 1; i < l; i++) gel(L,i) = GetRamifiedPol(p, gel(EFJ,i), flag);
+  for (i = 1; i < l; i++)
+  {
+    gel(L,i) = GetRamifiedPol(p, gel(EFJ,i), flag);
+#ifdef CHECK
+    checkpols(p, gel(EFJ, i), gel(L, i));
+#endif
+  }
   if (flag == 2) return gerepileuptoint(av, ZV_sum(L));
   return gerepilecopy(av, shallowconcat1(L));
 }
