@@ -14,15 +14,19 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 #include "mt.h"
+#if defined(_WIN32)
+#  include "../systems/mingw/mingw.h"
+#endif
 
 struct mt_queue
 {
   long no;
+  pari_sp avma;
   GEN input, output;
   GEN worker;
   long workid;
-  pthread_cond_t cond, cond1, cond2;
-  pthread_mutex_t mut, mut1, mut2;
+  pthread_cond_t cond;
+  pthread_mutex_t mut;
   pthread_cond_t *pcond;
   pthread_mutex_t *pmut;
 };
@@ -66,9 +70,10 @@ mt_err_recover(long er)
   {
     struct mt_pstate *mt = pari_mt;
     struct mt_queue *mq = mt->mq+mt_thread_no;
+    GEN err = bin_copy(copy_bin(pari_err_last()));
     LOCK(mq->pmut)
     {
-      mq->output = pari_err_last();
+      mq->output = err;
       pthread_cond_signal(mq->pcond);
     } UNLOCK(mq->pmut);
     pthread_exit((void*)1);
@@ -99,7 +104,9 @@ void pari_mt_init(void)
 {
   pari_mt = NULL;
 #ifdef _SC_NPROCESSORS_CONF
-  pari_mt_nbthreads = sysconf(_SC_NPROCESSORS_CONF);
+  if (!pari_mt_nbthreads) pari_mt_nbthreads = sysconf(_SC_NPROCESSORS_CONF);
+#elif defined(_WIN32)
+  if (!pari_mt_nbthreads) pari_mt_nbthreads = win32_nbthreads();
 #else
   pari_mt_nbthreads = 1;
 #endif
@@ -107,6 +114,13 @@ void pari_mt_init(void)
 
 void pari_mt_close(void) { }
 
+static void
+mt_queue_cleanup(void *arg)
+{
+  (void) arg;
+  pari_thread_close();
+}
+
 static void*
 mt_queue_run(void *arg)
 {
@@ -114,6 +128,12 @@ mt_queue_run(void *arg)
   pari_sp av = avma;
   struct mt_queue *mq = (struct mt_queue *) args;
   mt_thread_no = mq->no;
+  pthread_cleanup_push(mt_queue_cleanup,NULL);
+  LOCK(mq->pmut)
+  {
+    mq->avma = av;
+    pthread_cond_signal(mq->pcond);
+  } UNLOCK(mq->pmut);
   for(;;)
   {
     GEN work, done;
@@ -122,28 +142,23 @@ mt_queue_run(void *arg)
       while(!mq->input)
         pthread_cond_wait(&mq->cond, &mq->mut);
     } UNLOCK(&mq->mut);
-    work = gcopy(mq->input);
-    LOCK(&mq->mut1)
-    {
-      mq->input = NULL;
-      pthread_cond_signal(&mq->cond1);
-    } UNLOCK(&mq->mut1);
+    avma = mq->avma;
+    work = mq->input;
     pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);
     done = closure_callgenvec(mq->worker,work);
     pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED,NULL);
     LOCK(mq->pmut)
     {
+      mq->avma = av;
+      mq->input = NULL;
       mq->output = done;
       pthread_cond_signal(mq->pcond);
     } UNLOCK(mq->pmut);
-    LOCK(&mq->mut2)
-    {
-      while(mq->output)
-        pthread_cond_wait(&mq->cond2, &mq->mut2);
-    } UNLOCK(&mq->mut2);
-    avma = av;
   }
-  return NULL;
+  pthread_cleanup_pop(1);
+#ifdef __GNUC__
+  return NULL; /* NOT REACHED */
+#endif
 }
 
 static long
@@ -166,7 +181,12 @@ mtpthread_queue_get(struct mt_state *junk, long *workid, long *pending)
   GEN done = NULL;
   long last;
   (void) junk;
-  if (mt->nbint<mt->n) { mt->last = mt->nbint; *pending = mt->pending; return NULL; }
+  if (mt->nbint<mt->n)
+  {
+    mt->last = mt->nbint;
+    *pending = mt->pending;
+    return NULL;
+  }
   BLOCK_SIGINT_START
   LOCK(&mt->pmut)
   {
@@ -187,19 +207,13 @@ mtpthread_queue_get(struct mt_state *junk, long *workid, long *pending)
   } UNLOCK(&mt->pmut);
   BLOCK_SIGINT_END
   mq = mt->mq+last;
-  if (mq->output==err_e_STACK) pari_err(e_STACK);
+  if (mq->output==err_e_STACK) pari_err(e_STACKTHREAD);
   done = gcopy(mq->output);
+  mq->output = NULL;
   if (workid) *workid = mq->workid;
   if (typ(done)==t_ERROR) pari_err(0,done);
-  BLOCK_SIGINT_START
-  LOCK(&mq->mut2)
-  {
-    mq->output = NULL;
-    pthread_cond_signal(&mq->cond2);
-  } UNLOCK(&mq->mut2);
   mt->last = last;
   mt->pending--;
-  BLOCK_SIGINT_END
   *pending = mt->pending;
   return done;
 }
@@ -212,19 +226,23 @@ mtpthread_queue_submit(struct mt_state *junk, long workid, GEN work)
   (void) junk;
   if (!work) { mt->nbint=mt->n; return; }
   BLOCK_SIGINT_START
-  if (mt->nbint<mt->n) mt->nbint++;
+  if (mt->nbint<mt->n)
+  {
+    mt->nbint++;
+    LOCK(mq->pmut)
+    {
+      while(!mq->avma)
+        pthread_cond_wait(mq->pcond, mq->pmut);
+    } UNLOCK(mq->pmut);
+  }
   LOCK(&mq->mut)
   {
-    mq->input = work;
+    mq->output = NULL;
     mq->workid = workid;
+    mq->input = gcopy_avma(work, &mq->avma);
     pthread_cond_signal(&mq->cond);
   } UNLOCK(&mq->mut);
   mt->pending++;
-  LOCK(&mq->mut1)
-  {
-    while (mq->input)
-      pthread_cond_wait(&mq->cond1, &mq->mut1);
-  } UNLOCK(&mq->mut1);
   BLOCK_SIGINT_END
 }
 
@@ -233,21 +251,19 @@ mt_queue_reset(void)
 {
   struct mt_pstate *mt = pari_mt;
   long i;
+  BLOCK_SIGINT_START
   for (i=0; i<mt->n; i++)
     pthread_cancel(mt->th[i]);
   for (i=0; i<mt->n; i++)
     pthread_join(mt->th[i],NULL);
-  if (DEBUGLEVEL) pari_warn(warner,"stop threads");
   pari_mt = NULL;
+  BLOCK_SIGINT_END
+  if (DEBUGLEVEL) pari_warn(warner,"stop threads");
   for (i=0;i<mt->n;i++)
   {
     struct mt_queue *mq = mt->mq+i;
     pthread_cond_destroy(&mq->cond);
-    pthread_cond_destroy(&mq->cond1);
-    pthread_cond_destroy(&mq->cond2);
     pthread_mutex_destroy(&mq->mut);
-    pthread_mutex_destroy(&mq->mut1);
-    pthread_mutex_destroy(&mq->mut2);
     pari_thread_free(&mt->pth[i]);
   }
   pari_free(mt->mq);
@@ -266,7 +282,8 @@ mt_queue_start(struct pari_mt *pt, GEN worker)
     long NBT = pari_mt_nbthreads;
     struct mt_pstate *mt =
            (struct mt_pstate*) pari_malloc(sizeof(struct mt_pstate));
-    long mtparisize = GP_DATA->threadsize? GP_DATA->threadsize: top-bot;
+    long mtparisize = GP_DATA->threadsize? GP_DATA->threadsize: pari_mainstack->rsize;
+    long mtparisizemax = GP_DATA->threadsizemax;
     long i;
     mt->mq  = (struct mt_queue *) pari_malloc(sizeof(*mt->mq)*NBT);
     mt->th  = (pthread_t *) pari_malloc(sizeof(*mt->th)*NBT);
@@ -275,29 +292,32 @@ mt_queue_start(struct pari_mt *pt, GEN worker)
     mt->n = NBT;
     mt->nbint = 0;
     mt->last = 0;
-    pari_mt = mt;
     pthread_cond_init(&mt->pcond,NULL);
     pthread_mutex_init(&mt->pmut,NULL);
+    pari_thread_sync();
     for (i=0;i<NBT;i++)
     {
       struct mt_queue *mq = mt->mq+i;
       mq->no     = i;
+      mq->avma   = 0;
       mq->worker = worker;
       mq->input  = NULL;
       mq->output = NULL;
       mq->pcond  = &mt->pcond;
       mq->pmut   = &mt->pmut;
       pthread_cond_init(&mq->cond,NULL);
-      pthread_cond_init(&mq->cond1,NULL);
-      pthread_cond_init(&mq->cond2,NULL);
       pthread_mutex_init(&mq->mut,NULL);
-      pthread_mutex_init(&mq->mut1,NULL);
-      pthread_mutex_init(&mq->mut2,NULL);
-      pari_thread_alloc(&mt->pth[i],mtparisize,(GEN)mq);
+      if (mtparisizemax)
+        pari_thread_valloc(&mt->pth[i],mtparisize,mtparisizemax,(GEN)mq);
+      else
+        pari_thread_alloc(&mt->pth[i],mtparisize,(GEN)mq);
     }
     if (DEBUGLEVEL) pari_warn(warner,"start threads");
+    BLOCK_SIGINT_START
     for (i=0;i<NBT;i++)
       pthread_create(&mt->th[i],NULL, &mt_queue_run, (void*)&mt->pth[i]);
+    pari_mt = mt;
+    BLOCK_SIGINT_END
     pt->get=&mtpthread_queue_get;
     pt->submit=&mtpthread_queue_submit;
     pt->end=&mt_queue_reset;
